Package: RGraphSpace
Type: Package
Version: 1.3.0
Title: A Lightweight Interface Between 'igraph' and 'ggplot2' Graphics
Description: An interface to integrate 'igraph' and 'ggplot2' graphics 
    within a normalized coordinate system. 'RGraphSpace' implements 
    geometric objects based on 'ggplot2' prototypes, optimized for 
    the representation of large networks. The package provides three 
    specialized 'geoms' to translate graph data into geometric layers, 
    supporting customization of aesthetics and visual styles. These 
    'geoms' use a dual-anchor normalization approach to align layers, 
    required for analyses where network elements must be referenced 
    to a spatial map. 'RGraphSpace' aims to facilitate side-by-side 
    visualization of multiple graphs spatially aligned with reference 
    maps and images.
Authors@R: 
    c(person("Sysbiolab Team", role = "aut"),
      person("Flávio", "Kessler", role = "ctb"),
      person("Jonathan", "Back", role = "ctb"),
      person("Lana", "Querne", role = "ctb"),
      person("Victor", "Apolonio", role = c("ctb")),
      person("Vinicius", "Chagas", role = "ctb"),
      person("Mauro", "Castro", role = c("cre"),
              email = "mauro.a.castro@gmail.com",
              comment = c(ORCID = "0000-0003-4942-8131")))
Depends: R(>= 4.5), methods, ggplot2
Imports: grDevices, grid, igraph, tidygraph, scales, rlang, ggrastr,
        lifecycle
Suggests: knitr, rmarkdown, testthat, ggraph, ggnewscale, patchwork
Enhances: RedeR
License: Artistic-2.0
VignetteBuilder: knitr
URL: https://github.com/sysbiolab/RGraphSpace,
        https://sysbiolab.github.io/RGraphSpace/
BugReports: https://github.com/sysbiolab/RGraphSpace/issues
Collate: 'geom-edgespace.R' 'geom-graphspace.R' 'geom-nodespace.R'
        'gspace-checks.R' 'gspace-classes.R' 'gspace-constructor.R'
        'gspace-generics.R' 'gspace-methods.R' 'gspace-misc.R'
        'gspace-normalize.R' 'gspace-supplements.R' 'gspace-themes.R'
        'gspace-validation.R' 'inject-nodespace.R'
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-21 15:30:26 UTC; maac
Author: Sysbiolab Team [aut],
  Flávio Kessler [ctb],
  Jonathan Back [ctb],
  Lana Querne [ctb],
  Victor Apolonio [ctb],
  Vinicius Chagas [ctb],
  Mauro Castro [cre] (ORCID: <https://orcid.org/0000-0003-4942-8131>)
Maintainer: Mauro Castro <mauro.a.castro@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-22 06:30:02 UTC
Built: R 4.6.0; ; 2026-05-25 04:25:31 UTC; unix

Package: RGraphSpace
Type: Package
Version: 1.4.0
Title: A Lightweight Interface Between 'igraph' and 'ggplot2' Graphics
Description: An interface to integrate 'igraph' and 'ggplot2' graphics within 
    a normalized coordinate system. 'RGraphSpace' extends 'ggplot2' with 
    graph-aware geometries optimized for large networks. The 'GraphSpace' class 
    integrates directly with 'ggplot2' through specialized 'geoms' and lazy 
    resolution of node attributes, supporting customization of aesthetics and 
    visual styles. These 'geoms' use a dual-anchor normalization approach to 
    align layers, particularly useful for analyses in which network elements   
    must be spatially aligned with reference maps and images. 
Authors@R: 
    c(person("Sysbiolab Team", role = "aut"),
      person("Flávio", "Kessler", role = "ctb"),
      person("Jonathan", "Back", role = "ctb"),
      person("Lana", "Querne", role = "ctb"),
      person("Victor", "Apolonio", role = c("ctb")),
      person("Vinicius", "Chagas", role = "ctb"),
      person("Mauro", "Castro", role = c("cre"),
              email = "mauro.a.castro@gmail.com",
              comment = c(ORCID = "0000-0003-4942-8131")))
Depends: R(>= 4.5), methods, ggplot2 (>= 4.0)
Imports: grDevices, grid, igraph, tidygraph, scales, rlang, ggrastr,
        Matrix, lifecycle
Suggests: knitr, rmarkdown, testthat, ggraph, ggnewscale, patchwork,
        Seurat, SeuratObject
Enhances: RedeR
License: Artistic-2.0
VignetteBuilder: knitr
URL: https://github.com/sysbiolab/RGraphSpace,
        https://sysbiolab.github.io/RGraphSpace/
BugReports: https://github.com/sysbiolab/RGraphSpace/issues
Collate: 'geom-edgespace.R' 'geom-graphspace.R' 'geom-nodespace.R'
        'gspace-checks.R' 'gspace-classes.R' 'gspace-constructor.R'
        'gspace-generics.R' 'gspace-methods.R' 'gspace-misc.R'
        'gspace-normalize.R' 'gspace-supplements.R' 'gspace-themes.R'
        'gspace-validation.R' 'gspace-ggplot-constructor.R'
        'annotation-gspace.R' 'gspace-coercion.R' 'gspace-features.R'
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-12 21:32:39 UTC; maac
Author: Sysbiolab Team [aut],
  Flávio Kessler [ctb],
  Jonathan Back [ctb],
  Lana Querne [ctb],
  Victor Apolonio [ctb],
  Vinicius Chagas [ctb],
  Mauro Castro [cre] (ORCID: <https://orcid.org/0000-0003-4942-8131>)
Maintainer: Mauro Castro <mauro.a.castro@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-13 05:10:23 UTC

# Generated by roxygen2: do not edit by hand

S3method(.DollarNames,GraphSpace)
S3method(as.igraph,GraphSpace)
S3method(fortify,GraphSpace)
S3method(ggplot_add,inject_nodespace)
S3method(plot,GraphSpace)
export(GeomEdgeSpace)
export(GeomGraphSpace)
export(GeomNodeSpace)
export(StatEdgeSpace)
export(StatNodeSpace)
export(as_colorraster)
export(edgespace_handler)
export(geom_edgespace)
export(geom_graphspace)
export(geom_nodespace)
export(inject_nodespace)
export(nodespace_handler)
export(theme_gspace_coords)
export(theme_gspace_legend)
export(theme_gspace_th0)
export(theme_gspace_th1)
export(theme_gspace_th2)
export(theme_gspace_th3)
exportClasses(GraphSpace)
exportMethods("$")
exportMethods("$<-")
exportMethods("gs_edge_attr<-")
exportMethods("gs_vertex_attr<-")
exportMethods("names<-")
exportMethods(GraphSpace)
exportMethods(cropGraphSpace)
exportMethods(getGraphSpace)
exportMethods(gs_ecount)
exportMethods(gs_edge_attr)
exportMethods(gs_edges)
exportMethods(gs_graph)
exportMethods(gs_nodes)
exportMethods(gs_vcount)
exportMethods(gs_vertex_attr)
exportMethods(names)
exportMethods(normalizeGraphSpace)
exportMethods(plotGraphSpace)
import(igraph)
import(methods)
importFrom(ggplot2,"%+replace%")
importFrom(ggplot2,.pt)
importFrom(ggplot2,Geom)
importFrom(ggplot2,aes)
importFrom(ggplot2,annotation_raster)
importFrom(ggplot2,coord_fixed)
importFrom(ggplot2,draw_key_point)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,element_line)
importFrom(ggplot2,element_rect)
importFrom(ggplot2,element_text)
importFrom(ggplot2,expansion)
importFrom(ggplot2,fortify)
importFrom(ggplot2,from_theme)
importFrom(ggplot2,geom_blank)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_segment)
importFrom(ggplot2,geom_text)
importFrom(ggplot2,gg_par)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,ggplot_add)
importFrom(ggplot2,ggproto)
importFrom(ggplot2,guide_legend)
importFrom(ggplot2,guides)
importFrom(ggplot2,is_waiver)
importFrom(ggplot2,labs)
importFrom(ggplot2,layer)
importFrom(ggplot2,margin)
importFrom(ggplot2,remove_missing)
importFrom(ggplot2,scale_colour_identity)
importFrom(ggplot2,scale_linetype_manual)
importFrom(ggplot2,scale_x_continuous)
importFrom(ggplot2,scale_y_continuous)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_bw)
importFrom(ggplot2,theme_gray)
importFrom(ggplot2,translate_shape_string)
importFrom(ggplot2,unit)
importFrom(ggplot2,zeroGrob)
importFrom(ggrastr,rasterize)
importFrom(grDevices,as.raster)
importFrom(grDevices,col2rgb)
importFrom(grDevices,hcl.colors)
importFrom(grDevices,is.raster)
importFrom(graphics,plot)
importFrom(grid,arrow)
importFrom(grid,gList)
importFrom(grid,gpar)
importFrom(grid,grobName)
importFrom(grid,grobTree)
importFrom(grid,pointsGrob)
importFrom(grid,segmentsGrob)
importFrom(grid,unit)
importFrom(igraph,'E<-')
importFrom(igraph,'V<-')
importFrom(igraph,'edge_attr<-')
importFrom(igraph,E)
importFrom(igraph,V)
importFrom(igraph,any_loop)
importFrom(igraph,any_multiple)
importFrom(igraph,as.igraph)
importFrom(igraph,as_adjacency_matrix)
importFrom(igraph,as_edgelist)
importFrom(igraph,as_undirected)
importFrom(igraph,degree)
importFrom(igraph,delete_edge_attr)
importFrom(igraph,delete_graph_attr)
importFrom(igraph,delete_vertex_attr)
importFrom(igraph,ecount)
importFrom(igraph,edge_attr)
importFrom(igraph,edge_attr_names)
importFrom(igraph,empty_graph)
importFrom(igraph,ends)
importFrom(igraph,graph_attr_names)
importFrom(igraph,is_directed)
importFrom(igraph,is_simple)
importFrom(igraph,layout_nicely)
importFrom(igraph,simplify)
importFrom(igraph,vcount)
importFrom(igraph,vertex_attr)
importFrom(igraph,vertex_attr_names)
importFrom(igraph,which_mutual)
importFrom(lifecycle,deprecate_soft)
importFrom(lifecycle,deprecated)
importFrom(lifecycle,is_present)
importFrom(methods,setClass)
importFrom(methods,setOldClass)
importFrom(rlang,'%||%')
importFrom(rlang,exec)
importFrom(rlang,list2)
importFrom(rlang,warn)
importFrom(scales,alpha)
importFrom(scales,rescale)
importFrom(scales,squish)
importFrom(tidygraph,as.igraph)
importFrom(tidygraph,tbl_graph)
importFrom(utils,.DollarNames)
importFrom(utils,modifyList)

## ----Load packages for quick start, eval=TRUE, message=FALSE------------------
#--- Load required packages
library("igraph")
library("ggplot2")
library("RGraphSpace")

## ----Toy igraph - 1, eval=TRUE, message=FALSE, results=FALSE------------------
# Make a 'toy' igraph with 5 nodes and 4 edges;
# ..either a directed or undirected graph
gtoy1 <- make_star(5, mode="out")

# Check whether the graph is directed or not
is_directed(gtoy1)
## [1] TRUE

# Check graph size
vcount(gtoy1)
## [1] 5
ecount(gtoy1)
## [1] 4

# Assign 'x' and 'y' coordinates to each vertex;
# ..this can be an arbitrary unit in (-Inf, +Inf)
V(gtoy1)$x <- c(0, 2, -2, -4, -8)
V(gtoy1)$y <- c(0, 0,  2, -4,  0)

# Assign a name to each vertex
V(gtoy1)$name <- paste0("n", 1:5)

## ----Toy igraph - 2, eval=TRUE, message=FALSE, out.width="100%"---------------
# Plot the 'gtoy1' using standard R graphics
plot(gtoy1)

## ----Toy igraph - 3, eval=TRUE, message=FALSE, out.width="80%"----------------
# Plot the 'gtoy1' using RGraphSpace
plotGraphSpace(gtoy1, add.labels = TRUE)

## ----Node attributes, eval=TRUE, message=FALSE--------------------------------
# Node size (numeric in [0, 100], as '%' of the plot space)
V(gtoy1)$nodeSize <- c(8, 5, 5, 10, 5)

# Node shape (integer code between 0 and 25; see 'help(points)')
V(gtoy1)$nodeShape <- c(21, 22, 23, 24, 25)

# Node color (Hexadecimal or color name)
V(gtoy1)$nodeColor <- c("red", "#00ad39", "grey80", "lightblue", "cyan")

# Node line width (as in 'lwd' standard graphics; see 'help(gpar)')
V(gtoy1)$nodeLineWidth <- 1

# Node line color (Hexadecimal or color name)
V(gtoy1)$nodeLineColor <- "grey20"

# Node labels ('NA' will omit labels)
V(gtoy1)$nodeLabel <- c("V1", "V2", "V3", "V4", NA)

# Node label size (in pts)
V(gtoy1)$nodeLabelSize <- 8

# Node label color (Hexadecimal or color name)
V(gtoy1)$nodeLabelColor <- "black"

# Node transparency (in [0,1])
V(gtoy1)$nodeAlpha <- 1

## ----Edge attributes - 1, eval=TRUE, message=FALSE----------------------------
# Edge width (as in 'lwd' standard graphics; see 'help(gpar)')
E(gtoy1)$edgeLineWidth <- 0.8

# Edge color (Hexadecimal or color name)
E(gtoy1)$edgeLineColor <- c("red","green","blue","black")

# Edge type (as in 'lty' standard graphics; see 'help(gpar)')
E(gtoy1)$edgeLineType <- c("solid", "11", "dashed", "2124")

# Edge transparency (in [0,1])
E(gtoy1)$edgeAlpha <- 1

## ----Edge attributes - 2, eval=TRUE, message=FALSE----------------------------
# Arrowhead types in directed graphs
## Integer or character code:
## 0 = "---", 1 = "-->", -1 = "--|"
E(gtoy1)$arrowType <- 1

## ----Edge attributes - 3, eval=TRUE, message=FALSE----------------------------
# Arrowhead types in undirected graphs
## Integer or character code:
##  0 = "---"
##  1 = "-->",  2 = "<--",  3 = "<->",  4 = "|->",
## -1 = "--|", -2 = "|--", -3 = "|-|", -4 = "<-|", 
E(gtoy1)$arrowType <- 1
# Note: in undirected graphs, this attribute overrides 
# the edge's orientation in the edge list

## ----A shortcut for RGraphSpace, eval=TRUE, message=FALSE, out.width="80%"----
# Plot the updated 'gtoy1' using RGraphSpace
plotGraphSpace(gtoy1, add.labels = TRUE)

## ----Using geoms, eval=TRUE, message=FALSE, out.width="80%"-------------------
gs <- GraphSpace(gtoy1)
gs <- normalizeGraphSpace(gs)
ggplot(gs) + 
  geom_edgespace() +
  geom_nodespace() +
  theme_gspace_coords(is_norm = TRUE)

## ----label='Session information', eval=TRUE, echo=FALSE-----------------------
sessionInfo()

################################################################################
### Package documentation
################################################################################
#' @aliases RGraphSpace-package
#'
#' @section Index:
#' \tabular{ll}{
#' \link{GraphSpace}:
#' \tab Constructor of GraphSpace-class objects.\cr
#' \link{plotGraphSpace}:
#' \tab Wrapper function to plot GraphSpace objects in ggplot2.
#' }
#' Further information is available in the vignettes by typing 
#' \code{vignette('RGraphSpace')}. Documented topics are also available in
#' HTML by typing \code{help.start()} and selecting the RGraphSpace package
#' from the menu.
#'
#' @references
#' `r paste(format(citation("RGraphSpace"), style = "text"), collapse = "\n\n")`
#' 
#' @importFrom lifecycle deprecated deprecate_soft is_present
#' @keywords internal
"_PACKAGE"


################################################################################
### Documentation for some 'toy' datasets
################################################################################
#' @title Toy 'igraph' objects
#'
#' @description Small 'igraph' objects used for workflow demonstrations.
#' All graphs include 'x', 'y', and 'name' vertex attributes.
#'
#' @format igraph
#'
#' @usage data(gtoy1)
#'
#' @source This package.
#'
#' @docType data
#' @keywords gtoys
#' @name gtoys
#' @aliases gtoy1 gtoy2
#' @return A pre-processed igraph object.
#' @examples
#' data(gtoy1)
#' data(gtoy2)
NULL

#-------------------------------------------------------------------------------
#' Map numeric values to a color raster
#'
#' Helper function that converts numeric values to colors and 
#' returns a raster image. Useful for visualizing numeric matrices 
#' as color backgrounds.
#'
#' @param x A numeric vector or matrix containing values to be mapped to colors.
#' @param palette A vector of colors used as the palette. By default,
#'   \code{hcl.colors(30)} is used.
#' @param na.color Color used for \code{NA} values. Defaults to \code{NA}.
#'
#' @details
#' Values in \code{x} are rescaled to the range of the palette using
#' \code{scales::rescale()}, and each value is mapped to a corresponding
#' color. If \code{x} is a matrix, the resulting raster preserves the same
#' dimensions.
#'
#' @return A raster object as produced by \code{as.raster()}.
#'
#' @examples
#' # Convert the volcano matrix to a color raster
#' img <- as_colorraster(volcano)
#' plot(img)
#'
#' @importFrom grDevices hcl.colors
#' @export
as_colorraster <- function(x, palette = hcl.colors(30), na.color = "white") {
  
  if(!is.numeric(x)){
    stop("'x' must be a numeric vector or matrix.", call. = FALSE)
  }
  .validate_gs_colors("allColors", "palette", palette)
  .validate_gs_colors("singleColor", "na.color", na.color)
  
  z <- scales::rescale(x, to = c(1, length(palette)))
  z <- pmin(pmax(round(z), 1), length(palette))
  
  m <- palette[z]
  m[is.na(x)] <- na.color
  
  if (is.matrix(x)) {
    dim(m) <- dim(x)
  }
  
  as.raster(m)

}

#-------------------------------------------------------------------------------
#' @title Draw node elements in a 2D graph layout
#' 
#' @description
#' 
#' Constructor for \link{GeomNodeSpace} ggproto objects.
#' 
#' A wrapper around \link[ggplot2]{geom_point} that enables direct use of
#' node attributes stored in \link{GraphSpace} objects as aesthetics.
#'
#' This geom is designed to map node-level attributes (e.g., \code{fill},
#' \code{size}) or any aesthetics supported by \link[ggplot2]{GeomPoint}.
#'
#' @param mapping Set of aesthetic mappings created by [ggplot2::aes()].
#' These mappings override global aesthetics and are not inherited 
#' from the top-level plot.
#'
#' @param data A \link{GraphSpace} object.
#'
#' @param stat The statistical transformation to use on the data.
#' Defaults to \code{identity}.
#'
#' @param position Position adjustment, either as a string or
#' the result of a call to a position adjustment function.
#'
#' @param ... Additional parameters passed to the underlying
#' drawing function in \link{GeomNodeSpace}.
#'
#' @param na.rm Logical. Should missing values be removed?
#' Defaults to \code{FALSE}.
#' 
#' @param show.legend Logical or a named logical vector indicating
#' whether this layer should be included in legends.
#'
#' @param inherit.aes Logical. If \code{FALSE} (default), the layer will use 
#' aesthetics defined in \code{mapping}.
#' 
#' @return A ggplot2 layer that renders node glyphs defined by
#' \link{GeomNodeSpace}.
#'
#' @section Aesthetics:
#'
#' \code{geom_nodespace()} understands \link[ggplot2]{geom_point} aesthetics.
#' 
#' If these aesthetics are not explicitly provided in \code{aes()}, they 
#' are automatically retrieved from the \link{GraphSpace} object.
#'
#' \tabular{ll}{
#'   \strong{\code{x}, \code{y}} \tab Required (automatically supplied).\cr
#'   \code{fill} \tab Node interior colour (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
#'   \code{colour} \tab Node border colour (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
#'   \code{alpha} \tab Transparency (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
#'   \code{shape} \tab Node shape (see \link{points} and \link[ggplot2]{aes_linetype_size_shape}).\cr
#'   \code{size} \tab Node size (see *drawing* section and \link[ggplot2]{aes_linetype_size_shape}).\cr
#'   \code{stroke} \tab Node line width (see \link[ggplot2]{gg_par} and \link[ggplot2]{aes_linetype_size_shape}).
#' }
#' 
#' Required aesthetics \code{x} and \code{y} are supplied from the 
#' \link{GraphSpace} object and do not need to be manually mapped.
#' 
#' Additional parameters can be passed to control fixed values for the layer.
#' For example: `fill = "red"`, `stroke = 3`, `alpha = 0.5`, or `shape = 21`.
#' 
#' @section Drawing:
#' 
#' The interpretation of \strong{\code{size}} depends on how it is provided:
#' \itemize{
#'   \item **As an aesthetic**: When mapped within \code{aes()}, \code{size}
#'   follows the behavior of \link[ggplot2]{geom_point}, using absolute
#'   units to ensure consistency with the plot legends.
#'   \item **As a parameter**: When set outside \code{aes()}, \code{size} is 
#'   treated as a percentage of the viewport (\code{[0, 100]}), scaling 
#'   in \code{npc} units. This allows nodes to resize dynamically with 
#'   viewport changes.
#' }
#' 
#' @seealso
#' \link{GraphSpace}, \link{geom_edgespace}, \link{geom_graphspace}, 
#' \link[ggplot2]{geom_point}
#'
#' @examples
#' 
#' # Make a demo igraph
#' library(igraph)
#' gtoy1 <- make_star(15, mode="out")
#' 
#' # Set some node attributes
#' V(gtoy1)$nodeSize <- runif(vcount(gtoy1), 1, 20)
#' V(gtoy1)$nodeColor <- rainbow(vcount(gtoy1))
#' 
#' # Set some variables
#' V(gtoy1)$user_var1 <- runif(vcount(gtoy1), 1, 3)^3
#' V(gtoy1)$user_var2 <-  rep(c(1, 2, 3), each = 5)
#' 
#' # Create a GraphSpace object
#' gs <- GraphSpace(gtoy1, layout = layout_in_circle(gtoy1))
#' 
#' \dontrun{
#' 
#' # Example 1: Nodes scaling with the legend
#' # When 'size' is mapped inside aes(), it follows
#' # ggplot2 default behavior: sizes are translated 
#' # to absolute units (mm) via 'scale_size()'.
#' 
#' ggplot() + 
#' geom_edgespace(data = gs, arrow_offset = 0.01) +
#'   geom_nodespace(mapping = aes(size = nodeSize, fill = user_var2), 
#'   data = gs) + 
#'   scale_size(range = c(1, 12)) + 
#'   theme(aspect.ratio = 1)
#'   
#' # Example 2: Nodes scaling with the viewport
#' # When 'size' is passed as a node attribute, 
#' # as inherited from the igraph object, it is
#' # interpreted as a percentage of the plot area
#' # and translated to NPC units.
#' 
#' ggplot() + 
#' geom_edgespace(data = gs, arrow_offset = 0.01) +
#' geom_nodespace(mapping = aes(fill = user_var2), data = gs) +
#' theme(aspect.ratio = 1)
#'   
#' }
#' 
#' @export
geom_nodespace <- function(mapping = NULL, data = NULL, 
  stat = "identity", position = "identity", ...,
  na.rm = FALSE, show.legend = NA, inherit.aes = FALSE) {
  
  if (inherits(data, "GraphSpace")) {
    .geom_check_slots(data)
    nodes <- gs_nodes(data)
  } else {
    stop("'data' must be a 'GraphSpace' object.")
  }
  
  mapping <- .mapping_nodespace(mapping)
  
  params <- rlang::list2(na.rm = na.rm, ...)
  
  params <- .params_nodespace(params, mapping, nodes)
  
  ggplot2::layer(
    geom = GeomNodeSpace,
    mapping = mapping,
    data = nodes,
    stat = stat,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = params
  )
  
}

#-------------------------------------------------------------------------------
.geom_check_slots <- function(data){
  
  nodes <- gs_nodes(data)
  edges <- gs_edges(data)
  
  essential_natt <- c("x", "y", "vertex")
  essential_eatt <- c("vertex1", "vertex2", "arrowType")
  
  missing_n <- essential_natt[!(essential_natt %in% names(nodes))]
  if (length(missing_n) > 0) {
    stop("Required columns missing in slot 'data@nodes': ", 
      paste(missing_n, collapse = ", "), call. = FALSE)
  }
  
  missing_e <- essential_eatt[!(essential_eatt %in% names(edges))]
  if (length(missing_e) > 0) {
    stop("Required columns missing in slot 'data@edges': ", 
      paste(missing_e, collapse = ", "), call. = FALSE)
  }
  
  if (any(is.na(nodes[, essential_natt, drop = FALSE]))) {
    stop("Slot 'data@nodes' contains NAs in essential geometry columns: ", 
      paste(essential_natt, collapse = ", "), call. = FALSE)
  }
  
  if (any(is.na(edges[, essential_eatt, drop = FALSE]))) {
    stop("Slot 'data@edges' contains NAs in essential geometry columns: ", 
      paste(essential_eatt, collapse = ", "), call. = FALSE)
  }
  
}

#-------------------------------------------------------------------------------
.mapping_nodespace <- function(mapping) {
  x <- y <- NULL
  default_mapping <- ggplot2::aes(x = x, y = y)
  if (is.null(mapping)) {
    mapping <- default_mapping
  } else {
    mapping <- utils::modifyList(default_mapping, mapping)
  }
  return(mapping)
}

#-------------------------------------------------------------------------------
.params_nodespace <- function(params, mapping, nodes){
  
  params$.size_unit <- if("size" %in% names(mapping)) "mm" else "npc"
  
  if(is.null(params[["size"]]) && is.null(mapping[["size"]])){
    if("nodeSize" %in% names(nodes) ){
      params[["size"]] <- nodes[["nodeSize"]]
    }
  }
  
  if(is.null(params[["stroke"]]) && is.null(mapping[["stroke"]])){
    if("nodeLineWidth" %in% names(nodes) ){
      params[["stroke"]] <- nodes[["nodeLineWidth"]]
    }
  }
  
  if(is.null(params[["shape"]]) && is.null(mapping[["shape"]])){
    if("nodeShape" %in% names(nodes) ){
      params[["shape"]] <- nodes[["nodeShape"]]
    }
  }
  
  if(is.null(params[["fill"]]) && is.null(mapping[["fill"]])){
    if("nodeColor" %in% names(nodes) ){
      params[["fill"]] <- nodes[["nodeColor"]]
    }
  }
  
  if(is.null(params[["colour"]]) && is.null(mapping[["colour"]])){
    if("nodeLineColor" %in% names(nodes) ){
      params[["colour"]] <- nodes[["nodeLineColor"]]
    }
  }
  
  if(is.null(params[["alpha"]]) && is.null(mapping[["alpha"]])){
    if("nodeAlpha" %in% names(nodes) ){
      params[["alpha"]] <- nodes[["nodeAlpha"]]
    }
  }
  
  params
}


#-------------------------------------------------------------------------------
#' @title GeomNodeSpace: a ggplot2 prototype for GraphSpace-class methods
#'
#' @description
#' 
#' \code{GeomNodeSpace} is the underlying \link[ggplot2]{ggproto} object 
#' used by \link{geom_nodespace} to draw node elements in a graph layout.
#'
#' This geom is designed for network diagrams, where graph attributes 
#' are often already in their final form (e.g., hex colors).
#'
#' @section Aesthetics:
#'
#' \code{GeomNodeSpace} understands ggplot2's conventions for point-like geoms.
#' 
#' @seealso
#' \link{geom_nodespace}, \link[ggplot2]{geom_point}
#'
#' @importFrom ggplot2 scale_colour_identity draw_key_point from_theme zeroGrob
#' @importFrom ggplot2 geom_point geom_segment aes Geom .pt geom_text gg_par
#' @importFrom ggplot2 element_rect margin element_blank layer theme_bw
#' @importFrom ggplot2 element_line element_text ggproto theme theme_gray
#' @importFrom ggplot2 scale_linetype_manual annotation_raster coord_fixed
#' @importFrom ggplot2 scale_x_continuous scale_y_continuous expansion labs
#' @importFrom ggplot2 expansion translate_shape_string is_waiver 
#' @importFrom ggplot2 remove_missing
#' @importFrom grid gpar arrow unit pointsGrob segmentsGrob
#' @importFrom grid grobTree gList grobName
#' @importFrom scales alpha squish
#' @importFrom utils modifyList
#' @importFrom rlang list2 warn exec '%||%'
#' @export
GeomNodeSpace <- ggproto(
  
  "GeomNodeSpace", ggplot2::GeomPoint, 
  
  required_aes = c("x", "y"),
  
  non_missing_aes = c("size", "stroke", "shape", "colour"),
  
  default_aes = aes(
    size = 5,
    shape = 21,
    colour = "grey20",
    fill = "grey80",
    stroke = 0.5,
    alpha = NA
  ),
  
  draw_panel = function(self, data, panel_params, coord, 
    na.rm = FALSE, .size_unit = "mm") {
    
    data$shape <- translate_shape_string(data$shape)
    
    data <- .geom_check_node_size(data, .size_unit)
    
    coords <- coord$transform(data, panel_params)
    
    # Create node grobs
    grobs <- .get_node_grobs(coords, .size_unit)
    grobs$name  <- grobName(grobs, "geom_nodespace")
    grobs
    
  },
  
  draw_key = draw_key_point
  
)

#-------------------------------------------------------------------------------
.geom_check_node_size <- function(nodes, size_unit = "mm", 
  range = NULL, verbose = TRUE){
  # 1. Passed as a aesthetic, 'size' follows 'geom_point' behavior, in mm;
  # 2. Passed as a parameter, 'size' scales with the viewport (%) with
  # range expected in [0, 100], then converted to 'npc'.
  if (size_unit == "mm") {
    if(!is.null(range)){
      nodes$size <- scales::rescale(sqrt(nodes$size), to = range, 
        from = c(0, max(sqrt(nodes$size), na.rm = T)))
    }
  } else {
    if(verbose) .check_node_size(nodes[["size"]])
    nodes$size <- scales::squish(nodes[["size"]], range = c(0, 100))
  }
  return(nodes)
}

.check_node_size <- function(size) {
  if ( any(size < 0 | size > 100, na.rm = TRUE) ){
    rlang::warn(
      message = c(
        "Node 'size' outside the expected range of [0, 100].",
        "i" = "Passed as a parameter, 'size' scales with the viewport (%).",
        "*" = "Use aes(size = ...) for data-driven scaling."
      ),
      .frequency = "always",
      .frequency_id = "graphspace_node_size_range"
    )
  }
}

#-------------------------------------------------------------------------------
# pointsGrob
.get_node_grobs <- function(coords, size_unit){
  
  if(size_unit == "npc"){
    # For 'npc', 'size' is in [0, 100], here transformed by a 0.01 factor
    # For 'pch' in 0:25, 'size' is about 75% of the character height
    # (see 'points()' graphics); here rectified by a 1/0.75 factor
    coords$size <- coords$size * 0.01 * 1/0.75
  }
  
  grid::pointsGrob(
    x = coords$x,
    y = coords$y,
    pch = coords$shape,
    size = grid::unit(coords$size, size_unit),
    gp = ggplot2::gg_par(
      fill = scales::alpha(coords$fill, coords$alpha),
      col = scales::alpha(coords$colour, coords$alpha),
      stroke = coords$stroke
    )
  )
  
}

################################################################################
### Validate igraph for RGraphSpace
################################################################################
.validate_igraph <- function(g, layout = NULL, verbose = FALSE) {
    if (verbose) message("Validating the 'igraph' object...")
    if (!is(g, "igraph")) {
        stop("'g' should be an 'igraph' object.", call. = FALSE)
    }
    if (!is.null(layout)) {
        if (nrow(layout) != vcount(g)) {
            msg <- paste("'layout' must have xy-coordinates",
                "for the exact number of nodes in 'g'")
            stop(msg, call. = FALSE)
        } else {
            igraph::V(g)$x <- layout[, 1]
            igraph::V(g)$y <- layout[, 2]
        }
    } else if (is.null(igraph::V(g)$x) || is.null(igraph::V(g)$y)) {
        layout <- igraph::layout_nicely(g)
        igraph::V(g)$x <- layout[, 1]
        igraph::V(g)$y <- layout[, 2]
        msg <- paste0("Vertex attributes 'x' and 'y' missing; ",
            "computing layout...")
        if (verbose) message(msg)
    }
    if (is.null(igraph::V(g)$name)) {
        msg <- "Vertex attribute 'name' missing; assigning names... "
        if (verbose) message(msg)
        igraph::V(g)$name <- paste0("n", seq_len(igraph::vcount(g)))
    } else {
        if(is.vector(igraph::V(g)$name) || !is.list(igraph::V(g)$name)){
            if(any(is.na(igraph::V(g)$name))){
                msg <- "NA values found in vertex attribute 'name'."
                stop(msg, call. = FALSE)
            }
            if(!.all_characterValues(igraph::V(g)$name)){
                warning("vertex attribute 'name' converted to character.", 
                    call. = FALSE)
                igraph::V(g)$name <- as.character(igraph::V(g)$name)
            }
        } else {
            msg <- "vertex attribute 'name' should be a character vector."
            stop(msg, call. = FALSE) 
        }
        if (anyDuplicated(igraph::V(g)$name) > 0){
            stop("vertex names must be unique.", call. = FALSE)
        }
    }
    if (!igraph::is_simple(g)) {
        if (verbose && igraph::any_loop(g)) message("Removing loops...")
        if (verbose && igraph::any_multiple(g))
            message("Merging duplicated edges...")
        g <- igraph::simplify(g, remove.loops = TRUE, remove.multiple = TRUE,
          edge.attr.comb = list(weight = "max", "first"))
    }
    if (is.null(igraph::V(g)$nodeLabel)){
        igraph::V(g)$nodeLabel <- as.character(igraph::V(g)$name)
    }
    if (is.null(igraph::V(g)$nodeSize)){
        igraph::V(g)$nodeSize <- .get_default_vatt()[["nodeSize"]]
    }
    if (is.null(igraph::E(g)$arrowType)){
        if (is_directed(g)) {
            igraph::E(g)$arrowType <- 1
        } else {
            igraph::E(g)$arrowType <- 0
        }
    }
    g <- .validate_attributes(g)
    return(g)
    
}

################################################################################
### Validate graph attributes
################################################################################
.validate_attributes <- function(g){
    g <- .validate_nodes(g)
    g <- .validate_edges(g)
    g <- .validate_graph(g)
    return(g)
}

#-------------------------------------------------------------------------------
.validate_nodes <- function(g) {
    
    # get default attributes
    atts <- c(.get_required_vatt(), .get_default_vatt())
    a_names <- names(atts)
    # check default attributes
    b_names <- a_names[a_names %in% igraph::vertex_attr_names(g)]
    if(length(b_names)>0){
        if (vcount(g) > 0) {
            .validate_vatt(igraph::vertex_attr(g)[b_names])
        }
    }
    
    # put default attributes 1st
    d_names <- igraph::vertex_attr_names(g)
    a_names <- a_names[a_names %in% d_names]
    a_names <- c(a_names, d_names[ ! d_names %in% a_names ])
    igraph::vertex_attr(g) <- igraph::vertex_attr(g)[a_names]
    
    # attributes that require transformation
    g <- .validate_nodeshape(g)
    
    return(g)
}

#-------------------------------------------------------------------------------
.validate_edges <- function(g) {
    
    g <- .remove_hidden_eatt(g)
    
    # get default attributes
    atts <- .get_default_eatt(igraph::is_directed(g))
    a_names <- names(atts)
    # check default attributes
    b_names <- a_names[a_names %in% igraph::edge_attr_names(g)]
    if(length(b_names)>0){
        if (igraph::ecount(g) > 0) {
            .validate_eatt(igraph::edge_attr(g)[b_names])
        }
    }
    
    # put default attributes 1st
    d_names <- igraph::edge_attr_names(g)
    a_names <- a_names[a_names %in% d_names]
    a_names <- c(a_names, d_names[ ! d_names %in% a_names ])
    igraph::edge_attr(g) <- igraph::edge_attr(g)[a_names]
    
    # attributes that require transformation
    g <- .validate_arrowtype(g)
    g <- .validate_linetype(g)
    return(g)
}

#-------------------------------------------------------------------------------
.validate_graph <- function(g) {
    d_names <- igraph::graph_attr_names(g)
    if (length(d_names) > 0) {
        for (at in d_names) {
            g <- igraph::delete_graph_attr(g, name = at)
        }
    }
    return(g)
}

################################################################################
### Default RGraphSpace attributes
################################################################################
.get_required_vatt <- function() {
    atts <- list("x" = NA, "y" = NA, "name" = NA)
    return(atts)
}
.get_default_vatt <- function() {
    atts <- list(
        "nodeLabel" = NA, "nodeLabelSize" = 8, "nodeLabelColor" = "grey40",
        "nodeShape" = 21, "nodeSize" = 5, "nodeColor" = "grey80",
        "nodeLineWidth" = 0.5, "nodeLineColor" = "grey20")
    return(atts)
}
.get_default_eatt <- function(is.directed = FALSE) {
    atts <- list("edgeLineType" = "solid", "edgeLineColor" = "grey80",
        "edgeLineWidth" = 0.5)
    if (is.directed) {
        atts$arrowType <- 1
    } else {
        atts$arrowType <- 0
    }
    atts$weight <- 1
    return(atts)
}
# remove internally used hidden attributes
.remove_hidden_eatt <- function(g){
    atts <- names(.get_default_eatt(igraph::is_directed(g)))
    hidden <- setdiff(names(.get_empty_edgedf()), atts)
    hidden <- hidden[hidden %in% igraph::edge_attr_names(g)]
    if (length(hidden) > 0) {
        for (at in hidden) {
            g <- igraph::delete_edge_attr(g, name = at)
        }
    }
    g
}

################################################################################
### Validate attribute values
################################################################################
.validate_vatt <- function(atts) {
    if (!is.null(atts$x)) {
        .validate_gs_args("numeric_vec", "x", atts$x)
    }
    if (!is.null(atts$y)) {
        .validate_gs_args("numeric_vec", "y", atts$y)
    }
    if (!is.null(atts$name)) {
        .validate_gs_args("allCharacter", "name", atts$name)
    }
    if (!is.null(atts$nodeLabel)) {
        .validate_gs_args("allCharacterOrNa", "nodeLabel", atts$nodeLabel)
    }
    if (!is.null(atts$nodeLabelSize)) {
        .validate_gs_args("numeric_vec", "nodeLabelSize", atts$nodeLabelSize)
        if (min(atts$nodeLabelSize) <= 0) {
            stop("'nodeLabelSize' should be a vector of numeric values >0", 
              call. = FALSE)
        }
    }
    if (!is.null(atts$nodeLabelColor)) {
        .validate_gs_colors("allColors", "nodeLabelColor", atts$nodeLabelColor)
    }
    if (!is.null(atts$nodeSize)) {
        .validate_gs_args("numeric_vec", "nodeSize", atts$nodeSize)
        if (max(atts$nodeSize) > 100 || min(atts$nodeSize) < 0) {
            stop("'nodeSize' should be a vector of numeric values in [0, 100]", 
              call. = FALSE)
        }
    }
    if (!is.null(atts$nodeShape)) {
        .validate_gs_args("allCharacterOrInteger", "nodeShape", atts$nodeShape)
    }
    if (!is.null(atts$nodeColor)) {
        .validate_gs_colors("allColors", "nodeColor", atts$nodeColor)
    }
    if (!is.null(atts$nodeLineWidth)) {
        .validate_gs_args("numeric_vec", "nodeLineWidth", atts$nodeLineWidth)
        if (min(atts$nodeLineWidth) < 0) {
            stop("'nodeLineWidth' should be a vector of numeric values >=0", 
              call. = FALSE)
        }
    }
    if (!is.null(atts$nodeLineColor)) {
        .validate_gs_colors("allColors", "nodeLineColor", atts$nodeLineColor)
    }
}
#-------------------------------------------------------------------------------
.validate_eatt <- function(atts) {
    if (!is.null(atts$edgeLineType)) {
        .validate_gs_args("allCharacterOrInteger", "edgeLineType",
            atts$edgeLineType)
    }
    if (!is.null(atts$edgeLineWidth)) {
        .validate_gs_args("numeric_vec", "edgeLineWidth", atts$edgeLineWidth)
        if (min(atts$edgeLineWidth) <= 0) {
            stop("'edgeLineWidth' should be a vector of numeric values >0", 
                call. = FALSE)
        }
    }
    if (!is.null(atts$edgeLineColor)) {
        .validate_gs_colors("allColors", "edgeLineColor", atts$edgeLineColor)
    }
    if (!is.null(atts$arrowType)) {
        .validate_gs_args("allCharacterOrInteger", "arrowType", atts$arrowType)
    }
    if (!is.null(atts$weight)) {
        .validate_gs_args("numeric_vec", "weight", atts$weight)
    }
}

################################################################################
### Transform attribute types
################################################################################

#-------------------------------------------------------------------------------
.validate_nodeshape <- function(g) {
    if (vcount(g) > 0 && "nodeShape" %in% names(vertex_attr(g))) {
        V(g)$nodeShape  <- .transform_nodeshape(V(g)$nodeShape)
    }
    return(g)
}
.transform_nodeshape <- function(vshapes) {
    if (.all_integerValues(vshapes)) {
        vshapes[vshapes > 25] <- 21
        vshapes[vshapes < 0] <- 1
    } else {
        vshapes <- tolower(vshapes)
        pch <- rep(21, length(vshapes))
        pch[grep("circle", vshapes)] <- 21
        pch[grep("ellipse", vshapes)] <- 21
        pch[grep("square", vshapes)] <- 22
        pch[grep("diamond", vshapes)] <- 23
        pch[grep("triangle", vshapes)] <- 24
        pch[grep("rectangle", vshapes)] <- 22
        vshapes <- pch
    }
    return(vshapes)
}

#-------------------------------------------------------------------------------
.validate_arrowtype <- function(g) {
    if (ecount(g) > 0 && "arrowType" %in% names(edge_attr(g))) {
        E(g)$arrowType  <- .transform_arrowtype(E(g)$arrowType, is_directed(g))
    }
    return(g)
}
.transform_arrowtype <- function(eatt, is_dir = FALSE) {
        aty <- .arrowtypes(is_dir)
        if (.all_integerValues(eatt)) {
            idx <- !eatt %in% aty
            if (any(idx)) {
                eatt[idx] <- ifelse(is_dir, 1, 0)
                .arrowtypes_warning(is_dir)
            }
        } else {
            idx <- eatt %in% as.character(aty)
            if (any(idx)) {
                eatt[idx] <- names(aty)[match(eatt[idx], as.character(aty))]
            }
            idx <- !eatt %in% names(aty)
            if (any(idx)) {
                eatt[idx] <- "-->"
                .arrowtypes_warning(is_dir)
            }
            eatt <- aty[eatt]
        }
    return(eatt)
}
.arrowtypes <- function(is.dir = FALSE, unique = FALSE) {
    atp1 <- c(
        "---" = 0, "--" = 0, "-" = 0,
        "-->" = 1, "->" = 1, ">" = 1,
        "<--" = 2, "<-" = 2, "<" = 2,
        "<->" = 3, "<>" = 3,
        "|->" = 4, "|>" = 4)
    atp2 <- c(
        "--|" = -1, "-|" = -1, "|" = -1,
        "|--" = -2, "|-" = -2, "|" = -2,
        "|-|" = -3, "||" = -3,
        "<-|" = -4, "<|" = -4)
    atypes <- c(atp1, atp2)
    if (is.dir) {
        atypes <- atypes[atypes %in% c(-1, 0, 1)]
        if(unique){
            atypes <- atypes[match(unique(atypes), atypes)]
        }
    } else {
        if(unique){
            atp1 <- atp1[match(unique(atp1), atp1)]
            atp2 <- atp2[match(unique(atp2), atp2)]
            atypes <- c(atp1, atp2)
        }
    }
    return(atypes)
}
.arrowtypes_warning <- function(is.dir = FALSE){
    
    atypes <- .arrowtypes(is.dir, unique = TRUE)
    
    if (is.dir) {
        msg1 <- paste("'arrow type' for directed graphs must be one of:\n")
        msg2 <- paste(paste0("'", names(atypes), "'"), atypes, sep = " or ")
        msg2 <- paste(msg2, collapse = ", ")
    } else {
        msg1 <- paste("'arrow type' for undirected graphs must be one of:\n")
        idx <- atypes >= 0
        atp1 <- paste(paste0("'", names(atypes)[idx], "'"), atypes[idx], sep = " or ")
        atp2 <- paste(paste0("'", names(atypes)[!idx], "'"), atypes[!idx], sep = " or ")
        atp1 <- paste0(paste(atp1, collapse = ", "), "\n")
        atp2 <- paste(atp2, collapse = ", ")
        msg2 <- paste0(atp1, atp2)
    }
    msg3 <- c("\n...using default values.")
    warning(msg1, msg2, msg3, call. = FALSE)
    
}

#-------------------------------------------------------------------------------
.validate_linetype <- function(g) {
    if (ecount(g) > 0 && "edgeLineType" %in% names(edge_attr(g))) {
        E(g)$edgeLineType  <- .transform_linetype(E(g)$edgeLineType)
    }
    return(g)
}
.transform_linetype <- function(lty) {
    if (.all_integerValues(lty)) {
        ltypes <- .linetypes()
        lty[!lty %in% ltypes] <- 1
        lty <- ltypes[match(lty, ltypes)]
        lty <- names(lty)
    } else {
        lty <- tolower(lty)
        lty[grep("solid", lty)] <- "solid"
        lty[grep("dotted", lty)] <- "dotted"
        lty[grep("dashed", lty)] <- "dashed"
        lty[grep("long", lty)] <- "longdash"
        lty[grep("two", lty)] <- "twodash"
    }
    return(lty)
}
.linetypes <- function() {
    c('blank' = 0, 'solid' = 1, 'dashed' = 2, 'dotted' = 3,
        'dotdash' = 4, 'longdash' = 5, 'twodash' = 6)
}

#-------------------------------------------------------------------------------
.validate_gs_args <- function(check, name, para) {
  if (check == "numeric_vec") {
    msg <- paste0("'", name, "' should be a numeric vector.")
    if (!is.vector(para) || !.all_numericValues(para)) stop(msg, call. = FALSE)
  } else if (check == "character_vec") {
    msg <- paste0("'", name, "' should be a character vector.")
    if (!is.vector(para) || !.all_characterValues(para)) 
      stop(msg, call. = FALSE)
  } else if (check == "integer_vec") {
    msg <- paste0("'", name, "' should be an integer vector.")
    if (!is.vector(para) || !.all_integerValues(para)) 
      stop(msg, call. = FALSE)
  } else if (check == "numeric_mtx") {
    msg <- paste0("'", name, "' should be a numeric matrix")
    if (!is.numeric(para) || !is.matrix(para)) stop(msg, call. = FALSE)
  } else if (check == "image_mtx") {
    msg1 <- paste0("Invalid '", name, "' input. Expected a raster object\n")
    msg2 <- c("or a numeric matrix with values in the range [0, 1]")
    if (!is.raster(para) || !is.matrix(para) || !is.array(para))
      stop(msg1, msg2, call. = FALSE)
    if(!is.raster(para)){
      if(!is.numeric(para)){
        stop(msg1, msg2, call. = FALSE)
      } else {
        rg <- range(para, na.rm = TRUE)
        if(rg[1]<0 || rg[2]>1){
          stop(msg1, msg2, call. = FALSE)
        }
      }
    }
  } else if (check == "allCharacter") {
    msg <- paste0("'", name, "' should be a vector of strings.")
    if (!.all_characterValues(para)) stop(msg, call. = FALSE)
  } else if (check == "allCharacterOrInteger") {
    msg <- paste("'", name, "'should be a vector of strings of integers.")
    if (! (.all_characterValues(para) | .all_integerValues(para) ) ) 
      stop(msg, call. = FALSE)
  } else if (check == "allCharacterOrNa") {
    msg <- paste0("'", name, "' should be a vector of strings.")
    if (!.all_characterValues(para, notNA=FALSE)) stop(msg, call. = FALSE)
  } else if (check == "allBinary") {
    msg <- paste0("'", name, "' should be a vector of binary values.")
    if (!.all_binaryValues(para)) stop(msg, call. = FALSE)
  } else if (check == "allInteger") {
    msg <- paste0("'", name, "' should be a vector of integer values.")
    if (!.all_integerValues(para)) stop(msg, call. = FALSE)
  } else if (check == "singleString") {
    msg <- paste0("'", name, "' should be a single string.")
    if (!.is_singleString(para)) stop(msg, call. = FALSE)
  } else if (check == "singleInteger") {
    msg <- paste0("'", name, "' should be a single integer value.")
    if (!.is_singleInteger(para)) stop(msg, call. = FALSE)
  } else if (check == "singleNumber") {
    msg <- paste0("'", name, "' should be a single numeric value.")
    if (!.is_singleNumber(para)) stop(msg, call. = FALSE)
  } else if (check == "function") {
    msg <- paste0("'", name, "' should be a function.")
    if (!is.function(para)) stop(msg, call. = FALSE)
  } else if (check == "singleLogical") {
    msg <- paste0("'", name, "' should be a single logical value.")
    if (!.is_singleLogical(para)) stop(msg, call. = FALSE)
  } else {
    warning("Skipped arg validation.", call. = FALSE)
  }
}

#-------------------------------------------------------------------------------
.validate_gs_colors <- function(check, name, para) {
  if (check == "singleColor") {
    if (!.is_singleColor(para)) {
      msg <- paste0("'", name, "' should be a single color.")
      stop(msg, call. = FALSE)
    }
  } else if (check == "allColors") {
    if (!.is_color(para)) {
      msg <- paste0("'", name, "' should be a vector with colors.")
      stop(msg, call. = FALSE)
    }
  } else {
    warning("Skipped color validation.", call. = FALSE)
  }
}

#-------------------------------------------------------------------
.is_singleNumber <- function(para) {
  (is.integer(para) || is.numeric(para)) &&
    length(para) == 1L && !is.na(para)
}
.is_singleInteger <- function(para) {
  lg <- (is.integer(para) || is.numeric(para)) &&
    length(para) == 1L && !is.na(para)
  if (lg) {
    para <- abs(para)
    lg <- abs(para - round(para)) <= para
  }
  return(lg)
}
.is_singleString <- function(para) {
  is.character(para) && length(para) == 1L && !is.na(para)
}
.is_singleLogical <- function(para) {
  is.logical(para) && length(para) == 1L && !is.na(para)
}
.all_binaryValues <- function(para) {
  all(para %in% c(0, 1, NA))
}
.all_integerValues <- function(para, notNA = TRUE) {
  lg <- is.integer(para) || is.numeric(para) || all(is.na(para))
  if (lg) {
    para <- abs(para)
    lg <- all(abs(para - round(para)) <= para, na.rm=TRUE)
  }
  if(lg && notNA) lg <- !any(is.na(para))
  return(lg)
}
.all_numericValues <- function(para, notNA = TRUE) {
  lg <- is.numeric(para) || all(is.na(para))
  if(lg && notNA) lg <- !any(is.na(para))
  return(lg)
}
.all_characterValues <- function(para, notNA = TRUE) {
  lg <- is.character(para) || all(is.na(para))
  if(lg && notNA) lg <- !any(is.na(para))
  return(lg)
}
.is_numericVector <- function(para){
  is.vector(para) && .all_numericValues(para)
}
.is_integerVector <- function(para){
  is.vector(para) && .all_integerValues(para)
}
.is_characterVector <- function(para){
  is.vector(para) && .all_characterValues(para)
}
.is_color <- function(x) {
  if (is.numeric(x)) {
    if (any(x < 1 | x %% 1 != 0, na.rm = TRUE)) {
      return(FALSE)
    }
  }
  res <- try(col2rgb(x), silent = TRUE)
  return( !inherits(res, "try-error") )
}
.is_singleColor <- function(para) {
  .is_color(para) && length(para) == 1L && !is.na(para)
}

#-------------------------------------------------------------------------------
#' @title Draw node and edge elements in a 2D graph layout
#' 
#' @description
#' 
#' Constructor for \link{GeomGraphSpace} ggproto objects.
#' 
#' A wrapper around \link[ggplot2]{geom_point} that enables direct use of
#' node attributes stored in \link{GraphSpace} objects as aesthetics.
#'
#' This geom is designed to map node-level attributes (e.g., \code{fill},
#' \code{size}) or any aesthetics supported by \link[ggplot2]{GeomPoint}.
#'
#' @param mapping Set of aesthetic mappings created by [ggplot2::aes()].
#' These mappings override global aesthetics and are not inherited 
#' from the top-level plot.
#'
#' @param data A \link{GraphSpace} object.
#'
#' @param stat The statistical transformation to use on the data.
#' Defaults to \code{identity}.
#'
#' @param position Position adjustment, either as a string or
#' the result of a call to a position adjustment function.
#'
#' @param ... Additional parameters passed to the underlying
#' drawing function in \link{GeomGraphSpace}.
#'
#' @param na.rm Logical. Should missing values be removed?
#' Defaults to \code{FALSE}.
#' 
#' @param show.legend Logical or a named logical vector indicating
#' whether this layer should be included in legends.
#'
#' @param inherit.aes Logical. If \code{FALSE} (default), the layer will use 
#' aesthetics defined in \code{mapping}.
#' 
#' @param arrow_size Numeric scaling factor controlling arrowhead 
#' geometry (see 'drawing' section).
#' 
#' @param arrow_offset Numeric value controlling the base offset of arrows  
#' at edge endpoints (see 'drawing' section).
#' 
#' @return A ggplot2 layer that renders node glyphs defined by
#' \link{GeomGraphSpace}.
#'
#' @section Aesthetics for node drawing:
#' 
#' Nodes are drawn in the main layer of \code{geom_graphspace()}, which 
#' understands \link[ggplot2]{geom_point} aesthetics.
#' 
#' If these aesthetics are not explicitly provided in \code{aes()}, they 
#' are automatically retrieved from the \link{GraphSpace} object.
#' 
#' \tabular{ll}{
#'   \strong{\code{x}, \code{y}, \code{vertex}} \tab  
#'   Required (automatically supplied).\cr
#'   \code{fill} \tab Node interior colour (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
#'   \code{colour} \tab Node border colour (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
#'   \code{alpha} \tab Transparency (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
#'   \code{shape} \tab Node shape (see \link{points} and \link[ggplot2]{aes_linetype_size_shape}).\cr
#'   \code{size} \tab Node size (see *drawing* section and \link[ggplot2]{aes_linetype_size_shape}).\cr
#'   \code{stroke} \tab Node line width (see \link[ggplot2]{gg_par} and \link[ggplot2]{aes_linetype_size_shape}).
#' }
#' 
#' Required aesthetics \code{x}, \code{y}, and \code{vertex} are supplied from
#' the \link{GraphSpace} object and do not need to be manually mapped.
#' 
#' Additional parameters can be passed to control fixed values for the layer.
#' For example: `fill = "red"`, `stroke = 3`, `alpha = 0.5`, or `shape = 21`.
#' 
#' The interpretation of \strong{\code{size}} depends on how it is provided:
#' \itemize{
#'   \item **As an aesthetic**: When mapped within \code{aes()}, \code{size}
#'   follows the behavior of \link[ggplot2]{geom_point}, using absolute
#'   units to ensure consistency with the plot legends.
#'   \item **As a parameter**: When set outside \code{aes()}, \code{size} is 
#'   treated as a percentage of the viewport (\code{[0, 100]}), scaling 
#'   in \code{npc} units. This allows nodes to resize dynamically with 
#'   viewport changes.
#' }
#' 
#' @section Edge context-aware parameters:
#' 
#' These parameters control the edge appearance. If not explicitly provided, 
#' they are automatically retrieved from the \link{GraphSpace} object. 
#' They can be a single value or a vector matching the number of edges:
#' 
#' \tabular{ll}{
#'   \code{edge_colour} \tab Node border colour.\cr
#'   \code{edge_linetype} \tab Edge line type.\cr
#'   \code{edge_linewidth} \tab Edge line width.\cr
#'    \code{edge_alpha} \tab Edge transparency.
#' }
#' 
#' @section Edge global parameters:
#' 
#' These parameters apply globally to all edges in the layer:
#' 
#' \tabular{ll}{
#'    \code{arrow_size} \tab Arrow scaling factor (default = 1).\cr
#'    \code{arrow_offset} \tab Arrow offset from nodes (default = 0.01).\cr
#'    \code{arrow_lineend} \tab Line end style (see \link[grid]{gpar}).\cr
#'    \code{arrow_linejoin} \tab Line join style (see \link[grid]{gpar}).
#' }
#' 
#' **arrow_size** is a numeric scaling factor controlling arrowhead geometry. 
#' The value is interpreted in the same numeric space as line width (\code{lwd}), 
#' ensuring consistent scaling between edge strokes and arrowheads.
#' 
#' **arrow_offset** is an additive term that offsets arrow endpoints 
#' uniformly in graph space and is bounded by the edge length, in NPC units.
#' 
#' Arrowhead types are specified in the \link{GraphSpace} constructor.
#' 
#' @seealso
#' \link{GraphSpace}, \link{geom_nodespace}, \link{geom_edgespace}, 
#' \link[ggplot2]{geom_point}
#'
#' @examples
#' 
#' # Make a demo igraph
#' library(igraph)
#' gtoy1 <- make_star(15, mode="out")
#' 
#' # Set some node attributes
#' V(gtoy1)$nodeSize <- runif(vcount(gtoy1), 1, 20)
#' V(gtoy1)$nodeColor <- rainbow(vcount(gtoy1))
#' 
#' # Set some variables
#' V(gtoy1)$user_var1 <- runif(vcount(gtoy1), 1, 3)^3
#' V(gtoy1)$user_var2 <-  rep(c(1, 2, 3), each = 5)
#' 
#' # Create a GraphSpace object
#' gs <- GraphSpace(gtoy1, layout = layout_in_circle(gtoy1))
#' 
#' \dontrun{
#' 
#' # Example 1: Nodes scaling with the legend
#' # When 'size' is mapped inside aes(), it follows
#' # ggplot2 default behavior: sizes are translated 
#' # to absolute units (mm) via 'scale_size()'.
#' 
#' ggplot() + 
#'   geom_graphspace(
#'   mapping = aes(size = nodeSize, fill = user_var2), 
#'   data = gs, arrow_offset = 0.01) + 
#'   scale_size(range = c(1, 12)) + 
#'   theme(aspect.ratio = 1)
#'   
#' # Example 2: Nodes scaling with the viewport
#' # When 'size' is passed as a node attribute, 
#' # as inherited from the igraph object, it is
#' # interpreted as a percentage of the plot area
#' # and translated to NPC units.
#' 
#' ggplot() +
#'   geom_graphspace(mapping = aes(fill = user_var2), 
#'   data = gs, arrow_offset = 0.01) + 
#'   theme(aspect.ratio = 1)
#'   
#' }
#' 
#' @export
geom_graphspace <- function(mapping = NULL, data = NULL, 
  stat = "identity", position = "identity", ...,
  na.rm = FALSE, show.legend = NA, inherit.aes = FALSE,
  arrow_size = 1, arrow_offset = 0.01) {
  
  if (inherits(data, "GraphSpace")) {
    .geom_check_slots(data)
    nodes <- gs_nodes(data)
    edges <- gs_edges(data)
  } else {
    stop("'data' must be a 'GraphSpace' object.")
  }
  
  params <- rlang::list2(
    na.rm = na.rm,
    edges = edges,
    arrow_size = arrow_size,
    arrow_offset = arrow_offset,
    ...)
  
  mapping <- .mapping_graphspace(mapping)
  
  params <- .params_graphspace(params, mapping, nodes, edges)
  
  ggplot2::layer(
    geom = GeomGraphSpace,
    mapping = mapping,
    data = nodes,
    stat = stat,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = params
  )
  
}

#-------------------------------------------------------------------------------
.mapping_graphspace <- function(mapping) {
  x <- y <- vertex <- NULL
  default_mapping <- ggplot2::aes(x = x, y = y, vertex = vertex)
  if (is.null(mapping)) {
    mapping <- default_mapping
  } else {
    mapping <- utils::modifyList(default_mapping, mapping)
  }
  return(mapping)
}

#-------------------------------------------------------------------------------
.params_graphspace <- function(params, mapping, nodes, edges){
  
  params$.size_unit <- if("size" %in% names(mapping)) "mm" else "npc"
  
  #--- nodes
  
  if(is.null(params[["size"]]) && is.null(mapping[["size"]])){
    if("nodeSize" %in% names(nodes) ){
      params[["size"]] <- nodes[["nodeSize"]]
    }
  }
  
  if(is.null(params[["stroke"]]) && is.null(mapping[["stroke"]])){
    if("nodeLineWidth" %in% names(nodes) ){
      params[["stroke"]] <- nodes[["nodeLineWidth"]]
    }
  }
  
  if(is.null(params[["shape"]]) && is.null(mapping[["shape"]])){
    if("nodeShape" %in% names(nodes) ){
      params[["shape"]] <- nodes[["nodeShape"]]
    }
  }
  
  if(is.null(params[["fill"]]) && is.null(mapping[["fill"]])){
    if("nodeColor" %in% names(nodes) ){
      params[["fill"]] <- nodes[["nodeColor"]]
    }
  }
  
  if(is.null(params[["colour"]]) && is.null(mapping[["colour"]])){
    if("nodeLineColor" %in% names(nodes) ){
      params[["colour"]] <- nodes[["nodeLineColor"]]
    }
  }
  
  if(is.null(params[["alpha"]]) && is.null(mapping[["alpha"]])){
    if("nodeAlpha" %in% names(nodes) ){
      params[["alpha"]] <- nodes[["nodeAlpha"]]
    }
  }
  
  #--- edges
  
  if(is.null(params[["edge_colour"]])){
    if("edgeLineColor" %in% names(edges) ){
      params[["edge_colour"]] <- edges[["edgeLineColor"]]
    }
  }
  
  if(is.null(params[["edge_linewidth"]])){
    if("edgeLineWidth" %in% names(edges) ){
      params[["edge_linewidth"]] <- edges[["edgeLineWidth"]]
    }
  }
  
  if(is.null(params[["edge_linetype"]])){
    if("edgeLineType" %in% names(edges) ){
      params[["edge_linetype"]] <- edges[["edgeLineType"]]
    }
  }
  
  if(is.null(params[["edge_alpha"]])){
    if("edgeAlpha" %in% names(edges) ){
      params[["edge_alpha"]] <- edges[["edgeAlpha"]]
    }
  }
  
  params
}


#-------------------------------------------------------------------------------
#' @title GeomGraphSpace: a ggplot2 prototype for GraphSpace-class methods
#'
#' @description
#' 
#' \code{GeomGraphSpace} is the underlying \link[ggplot2]{ggproto} object 
#' used by \link{geom_graphspace} to draw node and edge elements in a 
#' graph layout. 
#'
#' This geom is designed for network diagrams, where graph attributes 
#' are often already in their final form (e.g., hex colors).
#'
#' @section Aesthetics:
#'
#' \code{GeomGraphSpace} understands ggplot2's conventions for point-like geoms.
#' 
#' @seealso
#' \link{geom_graphspace}, \link[ggplot2]{geom_point}
#'
#' @export
GeomGraphSpace <- ggproto(
  
  "GeomGraphSpace", ggplot2::Geom, 
  
  required_aes = c("x", "y", "vertex"),
  
  non_missing_aes = c("size", "shape", "colour"),
  
  default_aes = aes(
    size = 5,
    shape = 21,
    colour = "grey20",
    fill = "#E5E5E5B3",
    stroke = 0.5,
    alpha = NA
  ),
  
  draw_panel = function(self, data, panel_params, coord, edges, 
    edge_colour = "grey80", edge_alpha = NA,
    edge_linewidth = 0.5, edge_linetype = "solid", 
    arrow_size = 1, arrow_offset = 0.01, 
    arrow_lineend = "butt", arrow_linejoin = "mitre", 
    na.rm = FALSE, .size_unit = "mm") {
    
    data$shape <- translate_shape_string(data$shape)
    
    data <- .geom_check_node_size(data, size_unit = .size_unit)
    
    coords <- coord$transform(data, panel_params)
    
    # Create node grobs
    node_grobs <- .get_node_grobs(coords, size_unit = .size_unit)
    node_grobs$name <- grobName(node_grobs, "nodes")
    
    if(.empty(edges)){
      
      edge_grobs <- zeroGrob()
      
    } else {
      
      edges$colour <- edge_colour %||% "grey80"
      edges$linewidth <- edge_linewidth %||% 0.5
      edges$linetype <- edge_linetype %||% "solid"
      
      edges$alpha <- edge_alpha %||% NA
      edges$arrow_size <- (arrow_size %||% 1)
      edges$arrow_offset <- arrow_offset %||% 0
      
      edges <- remove_missing(edges, na.rm = na.rm,
        vars = c("vertex1", "vertex2", "arrowType"), 
        name = "geom_graphspace-edges")
      
      edges <- .geom_remap_edge_coords(edges = edges, nodes = coords)
      
      edges <- .geom_get_edge_offsets(edges = edges, nodes = coords,
        size_unit = .size_unit)
      
      edges <- .geom_adj_arrow_offsets(edges)
      
      edges <- .geom_adj_arrow_size(edges, size_unit = .size_unit)
      
      edges <- remove_missing(edges, na.rm = na.rm,
        vars = c("x", "y", "xend", "yend"), 
        name = "geom_graphspace-coords")
      
      # Create edge grobs
      edge_grobs <- .get_edge_grobs(edges, lineend = arrow_lineend, 
        linejoin = arrow_linejoin, size_unit = .size_unit)
      
    }
    
    grid::gTree(
      children = grid::gList(edge_grobs, node_grobs),
      name = grid::grobName(prefix = "geom_graphspace")
    )
    
  },
  
  draw_key = draw_key_point
  
)

#-------------------------------------------------------------------------------
# Remap x, y, xend, and yend to updated coords
.geom_remap_edge_coords <- function(edges, nodes){
  
  if(.empty(edges)){
    return( edges )
  }
  
  # remap segments
  idx <- match(edges[["vertex1"]], nodes[["vertex"]])
  edges[["x"]] <- nodes[["x"]][idx]
  edges[["y"]] <- nodes[["y"]][idx]
  idx <- match(edges[["vertex2"]], nodes[["vertex"]])
  edges[["xend"]] <- nodes[["x"]][idx]
  edges[["yend"]] <- nodes[["y"]][idx]
  
  return(edges)
  
}

################################################################################
### Main constructor of GraphSpace-class objects
################################################################################
.buildGraphSpace <- function(g, layout = NULL, verbose = TRUE) {
    
    gg <- .validate_igraph(g, layout, verbose)
    nodes <- .get_nodes(gg)
    edges <- .get_edges(gg)
    
    if(verbose) message("Creating a 'GraphSpace' object...")
    pars <- list(is.directed = is_directed(gg), 
        is.normalized = FALSE, image.layer = FALSE)
    gs <- new(Class = "GraphSpace", 
        nodes = nodes, 
        edges = edges, 
        graph = gg, 
        image = as.raster(matrix()), 
        pars = pars, 
        misc = list(g = g))
    
    return(gs)
    
}

################################################################################
### Get nodes and edges in a df object
################################################################################
.get_nodes <- function(gg){
    lt <- vertex_attr(gg)
    nodes <- data.frame(row.names = seq_along(lt[[1]]))
    for(nm in names(lt)){
        nodes[[nm]] <- lt[[nm]]
    }
    vertex <- seq_len(igraph::vcount(gg))
    nodes <- cbind(vertex = vertex, nodes)
    rownames(nodes) <- nodes$name
    return(nodes)
}
.get_edges <- function(gg){
    if (igraph::is_directed(gg)) {
        edges <- .get_directed_edges(gg)
    } else {
        edges <- .get_undirected_edges(gg)
    }
    return(edges)
}

################################################################################
### Get edges in a df object
################################################################################
.get_undirected_edges <- function(g){
    if(ecount(g)>0){
        vertex <- igraph::V(g)$name
        edges <- igraph::as_edgelist(g, names = FALSE)
        rownames(edges) <- colnames(edges) <- NULL
        edges <- as.data.frame(edges)
        colnames(edges) <- c("vertex1", "vertex2")
        edges$name1 <- vertex[edges$vertex1]
        edges$name2 <- vertex[edges$vertex2]
        atts <- .get_eatt(g)
        if(!all(atts[,c(1,2)]==edges[,c(1,2)])){
            stop("unexpected indexing during edge attribute combination.", 
                call. = FALSE)
        }
        edges <- cbind(edges, atts[,-c(1,2), drop = FALSE])
        edges <- edges[order(edges$vertex1,edges$vertex2), ]
        idx <- colnames(edges) %in% names(.get_empty_edgedf())
        edges <- edges[, c(which(idx), which(!idx))]
        rownames(edges) <- NULL
    } else {
        edges <- .get_empty_edgedf()
    }
    return(edges)
}
.get_eatt <- function(g){
    lt <- igraph::edge_attr(g)
    atts <- data.frame(row.names = seq_along(lt[[1]]))
    for(nm in names(lt)){
        atts[[nm]] <- lt[[nm]]
    }
    e <- igraph::as_edgelist(g, names = FALSE)
    colnames(e) <- c("vertex1", "vertex2")
    atts <- cbind(e, atts)
    return(atts)
}
.get_empty_edgedf <- function(){
    n <- numeric(); c <- character()
    edges <- data.frame(n, n, c, c, c, c, n, n, n)
    colnames(edges) <- c("vertex1","vertex2", "name1", "name2", 
        "edgeLineType", "edgeLineColor", "edgeLineWidth",
        "arrowType", "weight")
    return(edges)
}

################################################################################
### Get undirected edges in a df object
################################################################################
.get_directed_edges <- function(g) {
    if (ecount(g) > 0) {
        atts <- .extract_directed_att(g)
        vertex <- igraph::V(g)$name
        E(g)$emode <- 1
        E(g)$emode[igraph::which_mutual(g)] <- 3
        e <- emode <- .adjacency(g, attr = "emode")
        bl <- lower.tri(emode) & emode == 3
        emode[bl] <- 0
        edges <- arrayInd(seq_len(prod(dim(emode))), dim(emode), 
            useNames = TRUE)
        edges <- as.data.frame(edges)
        colnames(edges) <- c("vertex1", "vertex2")
        edges$emode <- as.numeric(emode)
        edges$name1 <- vertex[edges$vertex1]
        edges$name2 <- vertex[edges$vertex2]
        edges$e <- as.numeric(e > 0)
        eid <- e; eid[,] <- 0
        ut <- upper.tri(eid)
        eid[ut] <- seq_len(sum(ut))
        eid <- t(eid)
        eid[ut] <- seq_len(sum(ut))
        edges$eid <- as.numeric(eid)
        edges$ut <- as.numeric(upper.tri(e))
        edges$lt <- as.numeric(lower.tri(e))
        if (!all(atts[, c(1, 2)] == edges[, c(1, 2)])) {
            stop("unexpected indexing during edge attribute combination.", 
                call. = FALSE)
        }
        edges <- cbind(edges, atts[, -c(1, 2), drop=FALSE])
        eid <- unique(edges$eid[edges$e > 0])
        edges <- edges[edges$eid %in% eid, ]
        edges <- edges[order(edges$eid), ]
        rownames(edges) <- NULL
        edges <- .set_arrowtype_dir(edges)
        idx <- colnames(edges) %in% names(.get_empty_edgedf())
        edges <- edges[, c(which(idx), which(!idx))]
        rownames(edges) <- NULL
    } else {
        edges <- .get_empty_edgedf()
    }
    return(edges)
}
.set_arrowtype_dir <- function(edges) {
    # Flip ut/lt from single-edge arrows; this
    # for collecting arrows from the same mtx side
    idx <- which(edges$emode == 1 & edges$lt == 1)
    if (length(idx) > 0) {
        for (i in idx) {
            ii <- which(edges$eid == edges$eid[i])
            edges[ii, c("ut", "lt")] <- edges[ii, c("lt", "ut")]
        }
    }
    # collect left-side arrows
    arrow1 <- edges[edges$lt == 1, "arrowType"]
    arrow1[is.na(arrow1)] <- 0
    # collect right-side arrows
    arrow2 <- edges[edges$ut == 1, "arrowType"]
    arrow2[is.na(arrow2)] <- 0
    # get single-edge assignments
    edges <- edges[edges$e == 1, ]
    eid <- sort(unique(edges$eid))
    edges <- edges[order(-edges$ut, edges$eid), ]
    edges <- edges[match(eid, edges$eid), ]
    # add arrows and remove intermediate columns
    edges <- .merge_arrowtypes_dir(edges, arrow1, arrow2)
    edges <- edges[, -which(colnames(edges) %in% 
            c("e", "eid", "ut", "lt","emode"))]
    return(edges)
}
.merge_arrowtypes_dir <- function(edges, arrow1, arrow2) {
    ##  0 = "---", 1 = "-->",  2 = "<--",  3 = "<->",  4 = "|->",
    ## -1 = "--|", -2 = "|--", -3 = "|-|", -4 = "<-|",
    atypes <- c(0, 1, 2, 3, 4, -1, -2, -3, -4)
    names(atypes) <- c("00","01","10","11","-11","0-1","-10","-1-1","1-1")
    arrowType <- paste0(format(arrow1, digits = 1, trim = TRUE),
        format(arrow2, digits = 1, trim = TRUE))
    edges$arrowType <- as.numeric(atypes[arrowType])
    return(edges)
}
.extract_directed_att <- function(g) {
    # e <- igraph::as_adjacency_matrix(g, sparse = FALSE)
    e <- .adjacency(g)
    atts <- arrayInd(seq_len(prod(dim(e))), dim(e), useNames = TRUE)
    atts <- as.data.frame(atts)
    colnames(atts) <- c("vertex1", "vertex2")
    atts$e <- as.numeric(e)
    # a_names <- names(.get.default.eatt())
    a_names <- igraph::edge_attr_names(g)
    ne <- e == 0
    for (at in a_names) {
        x <- .adjacency(g, attr = at)
        x[ne] <- NA
        if (is.data.frame(x)){
            atts[[at]] <- I(unlist(x, recursive=FALSE))
        } else {
            if (is.numeric(x)) {
                atts[[at]] <- as.numeric(x)
            } else if (is.character(x)) {
                atts[[at]] <- as.character(x)
            }
        }
    }
    rownames(atts) <- NULL
    atts <- atts[, c("vertex1", "vertex2", a_names)]
    return(atts)
}
# ..this is a fix for 'as_adjacency_matrix', when 'attr' is character
.adjacency <- function(g, attr = NULL) {
    if(is.null(attr)){
        exattr <- rep(1, ecount(g))
        x <- matrix(0, nrow = vcount(g), ncol = vcount(g))
    } else {
        exattr <- edge_attr(g, as.character(attr))
        x <- matrix(NA, nrow = vcount(g), ncol = vcount(g))
        if(is.list(exattr)) x <- as.data.frame(x)
    }
    e <- igraph::ends(g, seq_len(ecount(g)), names = FALSE)
    x[e] <- exattr
    if (!is_directed(g)) x[e[,c(2,1)]] <- exattr
    colnames(x) <- rownames(x) <- V(g)$name
    return(x)
}

################################################################################
### Other functions
################################################################################

.get_gs_edges <- function(gs){
    nodes <- gs@nodes
    edges <- gs@edges
    coord <- data.frame(
        x = nodes[edges$vertex1, "x"],
        y = nodes[edges$vertex1, "y"],
        xend = nodes[edges$vertex2, "x"],
        yend = nodes[edges$vertex2, "y"]
        )
    edges <- cbind(coord, edges)
    return(edges)
}

.get_emode <- function(arrow_type){
    emode <- abs(arrow_type)
    emode[emode>3] <- 3
    return(emode)
}

#-------------------------------------------------------------------------------
#' @title Create a GraphSpace object
#' 
#' @description \code{GraphSpace} is the main constructor for 
#' \linkS4class{GraphSpace} objects, designed to store graph data and 
#' metadata for optimized rendering in RGraphSpace.
#'
#' @param g An \link[igraph]{igraph} object or a \code{data.frame} used to 
#' initialize a \code{GraphSpace} object. If an \code{igraph} is provided, 
#' it should include vertex coordinates in \code{x} and \code{y} attributes, 
#' and vertex labels in the \code{name} attribute. If a \code{data.frame} is 
#' provided, it must contain at least \code{x} and \code{y} columns 
#' representing the node coordinates; additional columns will be treated 
#' as vertex attributes. For graphs requiring edge definitions, use the 
#' \code{igraph} initialization.
#' @param layout An optional numeric matrix with two columns for \code{x} and
#' \code{y} vertex coordinates. If provided, it overrides coordinates in \code{g}.
#' @param verbose A logical value. If \code{TRUE}, displays detailed messages.
#' @param ... Additional arguments passed to the \code{GraphSpace} constructor.
#' @param mar `r lifecycle::badge("deprecated")` Deprecated since
#' RGraphSpace 1.1.1; use \link{normalizeGraphSpace} instead.
#' @param image `r lifecycle::badge("deprecated")` Deprecated since
#' RGraphSpace 1.1.1; use \link{normalizeGraphSpace} instead.
#' 
#' @return A \linkS4class{GraphSpace} class object.
#' 
#' @details
#' \code{GraphSpace} objects are designed to bridge the gap between network 
#' analysis (via \code{igraph}) and high-quality visualization (via \code{ggplot2}). 
#' The constructor ensures that all necessary aesthetics for 
#' \code{\link{geom_graphspace}} are pre-processed and validated.
#' 
#' \strong{Coordinate System and Normalization:}
#' By default, the constructor expects coordinates in the \code{x} and \code{y} 
#' vertex attributes, along with unique IDs in the \code{name} vertex 
#' attribute. If these are not provided, the constructor will generate 
#' sequential IDs and assign a layout using the 
#' \code{\link[igraph]{layout_nicely}} function. These coordinates define the 
#' relative positioning of nodes. For optimal rendering, it is recommended 
#' to pass the object through \code{\link{normalizeGraphSpace}} after 
#' construction. This converts vertex positions to Normalized Parent Coordinates 
#' (NPC), ensuring the graph remains centered and scaled relative to the 
#' plotting area.
#' 
#' \strong{Data Structure:}
#' The resulting object stores nodes and edges in separate internal slots, 
#' preserving metadata such as \code{nodeSize} and \code{edgeLineColor}. 
#' If an \code{igraph} object is provided without specific styling attributes, 
#' \code{GraphSpace} will assign the default values defined in the 
#' \code{\link{geom_graphspace}} aesthetics. Users can also specify custom 
#' variables in the input graph to be used as aesthetics within the 
#' \code{ggplot2} grammar.
#' 
#' \strong{Arrowhead Mapping:}
#' The \code{arrowType} attribute (see \emph{Arrowhead types} section) 
#' allows for a mapping between symbolic aliases (such as \code{"-->"}) 
#' and internal integer codes. This is useful for assigning interaction 
#' types in directed or undirected graphs (e.g., activation vs. inhibition).
#'  
#' @section Vertex attributes:
#' The following attributes in \code{g} are evaluated by the constructor:
#' 
#' \tabular{ll}{
#'   \code{nodeSize} \tab Numeric \code{[0, 100]}, representing % of the plotting space. \cr
#'   \code{nodeShape} \tab Integer code \code{[0-25]}; see \link[graphics]{points}. \cr
#'   \code{nodeColor} \tab A valid color name or hexadecimal code. \cr
#'   \code{nodeLineWidth} \tab Border thickness; see \link[grid]{gpar}. \cr
#'   \code{nodeLineColor} \tab A valid color name or hexadecimal code. \cr
#'   \code{nodeLabel} \tab Character string (\code{NA} will omit labels). \cr
#'   \code{nodeLabelSize} \tab Font size in \code{pts}; see \link[grid]{gpar}. \cr
#'   \code{nodeLabelColor} \tab A valid color name or hexadecimal code.
#' }
#' 
#' @section Edge attributes:
#' The following attributes in \code{g} are evaluated by the constructor:
#' 
#' \tabular{ll}{
#'   \code{edgeLineWidth} \tab Edge thickness; see \code{\link[grid]{gpar}}. \cr
#'   \code{edgeLineColor} \tab A valid color name or hexadecimal code. \cr
#'   \code{edgeLineType}  \tab Line style (e.g., "solid", "dashed"); see \code{\link[grid]{gpar}}. \cr
#'   \code{arrowType}     \tab Arrowhead style (see \emph{Arrowhead types} section).
#' }
#' 
#' @section Arrowhead types:
#' 
#' Arrowheads are controlled via the \code{arrowType} attribute using 
#' integer or character codes (see examples in the \emph{RGraphSpace} vignette).
#' 
#' In directed graphs, arrows follow the edge list orientation by default, 
#' representing forward directions (\emph{e.g.}, \code{A -> B}). 
#' While undirected graphs do not show arrows by default, specific styles 
#' can be manually assigned for detailed visualization, including forward, 
#' backward, or bidirectional arrowheads.
#' 
#' \subsection{Directed graphs (A -> B):}{
#' \tabular{lll}{
#'   \strong{Code} \tab \strong{Alias} \tab \strong{Description} \cr
#'   \code{0} \tab \code{"---"} \tab No arrow \cr
#'   \code{1} \tab \code{"-->"} \tab Forward arrow \cr
#'   \code{-1} \tab \code{"--|"} \tab Forward bar
#' }
#' }
#' 
#' \subsection{Undirected graphs (A -- B):}{
#' \tabular{lll}{
#'   \strong{Code} \tab \strong{Alias} \tab \strong{Description} \cr
#'   \code{0} \tab \code{"---"} \tab No arrow \cr
#'   \code{1} \tab \code{"-->"} \tab Forward arrow \cr
#'   \code{2} \tab \code{"<--"} \tab Backward arrow \cr
#'   \code{3} \tab \code{"<->"} \tab Bidirectional arrow \cr
#'   \code{4} \tab \code{"|->"} \tab Forward arrow / backward bar \cr
#'   \code{-1} \tab \code{"--|"} \tab Forward bar \cr
#'   \code{-2} \tab \code{"|--"} \tab Backward bar \cr
#'   \code{-3} \tab \code{"|-|"} \tab Bidirectional bar \cr
#'   \code{-4} \tab \code{"<-|"} \tab Backward arrow / forward bar
#' }
#' }
#' 
#' @author Sysbiolab.
#' 
#' @seealso \code{\link{geom_graphspace}}, \code{\link{plotGraphSpace}}
#' 
#' @examples
#' library(igraph)
#' 
#' # Create a star graph
#' gtoy1 <- make_full_graph(15)
#' 
#' # Custom attributes
#' V(gtoy1)$nodeSize <- 5
#' E(gtoy1)$edgeLineColor <- "red"
#' E(gtoy1)$arrowType <- "-->"
#' 
#' # Create a GraphSpace
#' gs <- GraphSpace(gtoy1)
#' 
#' @importFrom igraph degree vcount ecount which_mutual any_loop any_multiple
#' @importFrom igraph as_edgelist as_adjacency_matrix is_simple 
#' @importFrom igraph graph_attr_names delete_graph_attr empty_graph
#' @importFrom igraph simplify V E 'V<-' 'E<-' is_directed vertex_attr
#' @importFrom igraph layout_nicely as_undirected delete_edge_attr
#' @importFrom igraph vertex_attr_names edge_attr edge_attr_names
#' @importFrom igraph ends delete_vertex_attr 'edge_attr<-'
#' @importFrom scales rescale
#' @importFrom grDevices is.raster as.raster
#' @aliases GraphSpace
#' @rdname GraphSpace-methods
#' @export
setMethod("GraphSpace", signature(g = "igraph"),
  function(g, layout = NULL, verbose = TRUE,  
    mar = deprecated(), image = deprecated(), ...) {
    ### deprecate
    if (lifecycle::is_present(mar)) {
      deprecate_soft("1.1.1", "GraphSpace(mar)", 
        "normalizeGraphSpace(mar)")
    }
    if (lifecycle::is_present(image)) {
      deprecate_soft("1.1.1", "GraphSpace(image)", 
        "normalizeGraphSpace(image)")
    }
    .validate_gs_args("singleLogical", "verbose", verbose)
    #--- validate argument values
    if(!is.null(layout)){
      .validate_gs_args("numeric_mtx", "layout", layout)
      if (ncol(layout) != 2) {
        stop("'layout' matrix should have two columns.", call. = FALSE)
      } 
    }
    #--- validate igraph and build a gs object
    gs <- .buildGraphSpace(g, layout, verbose)
    
    return(gs)
  }
)

#' @rdname GraphSpace-methods
#' @export
setMethod("GraphSpace", signature(g = "data.frame"),
  function(g, verbose = TRUE, ...){
    
    if(!all(c("x", "y") %in% colnames(g))){
      ms_i <- c("i" = "GraphSpace requires x/y columns for 'data.frame' initialization.")
      rlang::abort(
        message = "Input 'g' is missing 'x' and 'y' coordinates.",
        body = ms_i
      )
    }
    
    g <- .graphFromCoordinates(g)
    
    #--- build GraphSpace-class
    gs <- GraphSpace(g = g, verbose = verbose, ...)
    
    return(gs)
    
  }
)

#-------------------------------------------------------------------------------
.graphFromCoordinates <- function(coord){
  
  # Check attributes
  attr <- unique(colnames(coord))
  attr <- attr[!is.na(attr)]
  attr <- attr[!attr%in%c("x","y")]
  
  # Initialize a graph using 'coord' as vertices, with no edges
  g <- make_empty_graph(n = nrow(coord), directed = FALSE)
  if(!is.null(rownames(coord))){
    V(g)$name <- rownames(coord)
  }
  
  # Add coordinates
  V(g)$x <- coord$x
  V(g)$y <- coord$y
  
  if(length(attr)>0){
    for(name in attr){
      igraph::vertex_attr(g, name) <- coord[[name]]
    }
  }
  
  return(g)
  
}


#-------------------------------------------------------------------------------
#' @title Wrapper function to plot GraphSpace objects in ggplot2
#'
#' @description 
#' \code{plotGraphSpace()} is a High-level plotting interface that translates 
#' \code{igraph} and \code{GraphSpace} data objects into \code{ggplot2} layers.
#'
#' @param gs Either an \code{igraph} or \linkS4class{GraphSpace} class object.
#' If \code{gs} is an \code{igraph}, then it must include \code{x}, \code{y}, 
#' and \code{name}  vertex attributes (see \code{\link{GraphSpace}}).
#' @param theme Name of a custom RGraphSpace theme. These themes 
#' (from 'th0' to 'th3') consist of preconfigured ggplot settings, 
#' which can subsequently refine using \code{\link[ggplot2]{ggplot2}}.
#' @param xlab The title for the 'x' axis of a 2D-image space.
#' @param ylab The title for the 'y' axis of a 2D-image space.
#' @param font.size A single numeric value passed to ggplot themes.
#' @param bg.color A single color for background.
#' @param add.labels A logical value indicating whether to plot vertex labels.
#' @param node.labels A vector of vertex names to be highlighted in the graph
#' space. This argument overrides 'add.labels'.
#' @param label.size A size argument passed to \code{\link[ggplot2]{geom_text}}.
#' @param label.color A color passed to \code{\link[ggplot2]{geom_text}}.
#' @param add.image A logical value indicating whether to add a background 
#' image, when one is available (see \code{\link{GraphSpace}}).
#' @param raster A logical value indicating whether to rasterize the main plot.
#' See \code{\link[ggrastr]{rasterise}} for further specifications.
#' @param dpi Raster resolution, in dots per inch.
#' @param dev Device used in the \code{\link[ggrastr]{rasterise}} call.
#' @return A ggplot-class object.
#' @author Sysbiolab.
#' @seealso \code{\link{GraphSpace}}
#' @examples
#' library(igraph)
#' 
#' # Load a demo igraph
#' data('gtoy1', package = 'RGraphSpace')
#' 
#' # Generate a ggplot for gtoy1
#' plotGraphSpace(gtoy1)
#' 
#' # Create a star graph
#' gtoy_star <- make_full_graph(15)
#' 
#' # Example of setting node and edge attributes
#' V(gtoy_star)$nodeSize <- 5
#' E(gtoy_star)$edgeLineColor <- "red"
#' E(gtoy_star)$arrowType <- "<->"
#' 
#' # Create a GraphSpace object
#' gs_star <- GraphSpace(gtoy_star)
#' 
#' # Normalize graph coordinates
#' gs_star <- normalizeGraphSpace(gs_star)
#' 
#' # Generate a ggplot for gs_star
#' plotGraphSpace(gs_star)
#' 
#' @import methods
#' @importFrom ggplot2 ggplot
#' @importFrom grDevices col2rgb
#' @importFrom ggrastr rasterize
#' @docType methods
#' @rdname plotGraphSpace-methods
#' @aliases plotGraphSpace
#' @export
setMethod("plotGraphSpace", "GraphSpace", 
    function(gs, theme = "th0", xlab = "Graph coordinates 1", 
        ylab = "Graph coordinates 2", font.size = 1,
        bg.color = "grey95", add.labels = FALSE,
        node.labels = NULL, label.size = 3, 
        label.color = "grey20", add.image = FALSE, 
        raster = FALSE, dpi = 300, dev = "cairo_png") {
        #--- validate the gs object and args
        .validate_gs_args("singleString", "xlab", xlab)
        .validate_gs_args("singleString", "ylab", ylab)
        .validate_gs_args("singleNumber", "font.size", font.size)
        .validate_gs_colors("singleColor", "bg.color", bg.color)
        .validate_gs_args("singleLogical", "add.labels", add.labels)
        .validate_gs_args("singleNumber", "label.size", label.size)
        .validate_gs_colors("singleColor", "label.color", label.color)
        .validate_gs_args("singleLogical", "add.image", add.image)
        .validate_gs_args("singleLogical", "raster", raster)
        .validate_gs_args("singleInteger", "dpi", dpi)
        .validate_gs_args("singleString", "dev", dev)
        if (!is.null(node.labels)) {
            .validate_gs_args("allCharacter", "node.labels", node.labels)
        }
        
        theme <- match.arg(theme, choices = c("th0", "th1", "th2", "th3"))
        
        #--- get gs slots
        nodes <- gs_nodes(gs)
        pars <- getGraphSpace(gs, "pars")
        
        #--- initialize a ggplot object
        ggp <- ggplot()
        
        #--- add image
        if(pars$image.layer){
            img <- getGraphSpace(gs, "image")
            if(add.image){
                ggp <- .add_image(ggp, img)
            } else {
                ggi <- .add_image(ggp, img)
                ggi <- ggi + theme_gspace_coords(theme = theme, 
                  is_norm = pars$is.normalized, xlab = xlab, ylab = ylab, 
                  txt_size = font.size, leg_size = font.size, 
                  bg_color = bg.color)
            }
        }

        #--- add graph
        ggp <- ggp + geom_graphspace(data = gs)
        
        #--- add node labels
        if (!is.null(node.labels)){
            ggp <- .add_labels1(ggp, nodes, node.labels, 
                label.size, label.color)
        } else if(add.labels){
            ggp <- .add_labels2(ggp, nodes)
        }
        
        #--- apply custom theme
        ggp <- ggp + theme_gspace_coords(theme = theme, 
          is_norm = pars$is.normalized, xlab = xlab, ylab = ylab, 
          txt_size = font.size, leg_size = font.size,
          bg_color = bg.color)
        
        if(raster){
          ggp <- ggrastr::rasterize(ggp, layers = "GraphSpace", 
              dpi = dpi, dev = dev)
        }
        
        if(pars$image.layer && !add.image){
            ggl <- list(image = ggi, graph = ggp)
            return(ggl)
        } else {
            return(ggp)
        }
        
    }
)

#-------------------------------------------------------------------------------
#' @param ... Additional arguments passed to the 
#' \code{\link{plotGraphSpace}} function.
#' @import methods
#' @docType methods
#' @rdname plotGraphSpace-methods
#' @aliases plotGraphSpace
#' @export
setMethod("plotGraphSpace", "igraph", 
    function(gs, ...) {
        gs <- GraphSpace(gs, verbose=FALSE)
        gs <- normalizeGraphSpace(gs)
        plotGraphSpace(gs = gs, ...)
    }
)

#' Plot GraphSpace objects
#' 
#' @param x A \linkS4class{GraphSpace} class object.
#' @param ... Additional arguments passed to the 
#' \code{\link{plotGraphSpace}} function.
#' @seealso \code{\link{plotGraphSpace}}
#' 
#' @importFrom graphics plot
#' @export
plot.GraphSpace <- function(x, ...) {
    plotGraphSpace(x, ...)
}

#-------------------------------------------------------------------------------
#' @title Accessors for fetching slots from a GraphSpace object
#'
#' @description \code{getGraphSpace} retrives information from
#' individual slots available in a GraphSpace object.
#'
#' @param gs A preprocessed \linkS4class{GraphSpace} class object
#' @param what A single character value specifying which information should 
#' be retrieved from the slots.
#' Options: 'graph','gxy','gxyz','pars','misc','status','summits',
#' 'summit_mask', and 'summit_contour'.
#' @return Content from slots in the \linkS4class{GraphSpace} object.
#' @examples
#' # Load a demo igraph
#' data('gtoy1', package = 'RGraphSpace')
#'
#' # Create a new GraphSpace object
#' gs <- GraphSpace(gtoy1)
#'
#' # Get the 'summary' slot in gs
#' getGraphSpace(gs, what = 'graph')
#'
#' @import methods
#' @docType methods
#' @rdname getGraphSpace-methods
#' @aliases getGraphSpace
#' @export
setMethod("getGraphSpace", "GraphSpace", function(gs, what = "graph") {
    opts <- c("nodes", "edges", "graph","pars", "misc", "image")
    if (!what %in% opts) {
        opts <- paste0(opts, collapse = ", ")
        stop("'what' must be one of:\n", opts, call. = FALSE)
    }
    if (what == "nodes") {
        obj <- gs@nodes
    } else if (what == "edges") {
        obj <- gs@edges
    } else if (what == "graph") {
        obj <- gs@graph
    } else if (what == "pars") {
        obj <- gs@pars
    } else if (what == "misc") {
        obj <- gs@misc
    } else if (what == "image") {
        obj <- gs@image
    } else {
        obj <- gs@graph
    }
    return(obj)
})

#-------------------------------------------------------------------------------
# show summary information on screen
setMethod("show", "GraphSpace", function(object) {
    message("A GraphSpace-class object for:")
    obj <- getGraphSpace(object, what = "graph")
    summary(obj)
})

#-------------------------------------------------------------------------------
#' @title Accessors for applying essential igraph methods to modify
#' attributes of GraphSpace objects.
#' 
#' @description Access and modify individual slots of a GraphSpace 
#' object. Selected 'igraph' methods are applied to the 'graph' slot and 
#' propagated to downstream components.
#' 
#' @param x A \linkS4class{GraphSpace} class object
#' @param name Name of the attribute.
#' @param value The new value of the attribute.
#' @param ... Additional arguments passed to igraph methods.
#' @return Updated \linkS4class{GraphSpace} object.
#' @seealso \code{\link[igraph]{vertex_attr}}, \code{\link[igraph]{edge_attr}}
#' @examples
#' # Load a demo igraph
#' data('gtoy1', package = 'RGraphSpace')
#' 
#' # Create a new GraphSpace object
#' gs <- GraphSpace(gtoy1)
#' 
#' #--- Usage of GraphSpace attribute accessors:
#'
#' # Get a data frame with nodes for plotting methods
#' gs_nodes(gs)
#' 
#' # Get a data frame with edges for plotting methods
#' gs_edges(gs)
#' 
#' # Get vertex count
#' gs_vcount(gs)
#' 
#' # Get edge count
#' gs_ecount(gs)
#' 
#' # Get vertex names
#' names(gs)
#' 
#' # Access all vertex attributes
#' gs_vertex_attr(gs)
#' 
#' # Access a specific vertex attribute
#' gs_vertex_attr(gs, "nodeLabel")
#' 
#' # Modify a single value within a vertex attribute
#' gs_vertex_attr(gs, "nodeSize")["n1"] <- 10
#' 
#' # Replace an entire vertex attribute
#' gs_vertex_attr(gs, "nodeSize") <- 10
#' 
#' # Alternative syntax using `$` accessor
#' gs_vertex_attr(gs)$nodeSize <- 10
#' 
#' # Access a specific edge attribute
#' gs_edge_attr(gs, "edgeLineColor")
#' 
#' # Replace an entire edge attribute
#' gs_edge_attr(gs, "edgeLineWidth") <- 1
#' 
#'  # Alternative syntax using `$` for edge attributes
#' gs_edge_attr(gs)$edgeLineWidth <- 3
#' 
#' @aliases names
#' @aliases names<-
#' @aliases gs_nodes
#' @aliases gs_edges
#' @aliases gs_vcount
#' @aliases gs_ecount
#' @aliases gs_vertex_attr
#' @aliases gs_edge_attr
#' @aliases gs_vertex_attr<-
#' @aliases gs_edge_attr<-
#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_nodes", "GraphSpace", function(x) {
    x@nodes
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_edges", "GraphSpace", function(x) {
    .get_gs_edges(x)
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_vcount", "GraphSpace", function(x) {
    igraph::vcount(x@graph)
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_ecount", "GraphSpace", function(x) {
    igraph::ecount(x@graph)
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("names", "GraphSpace", function(x) {
    x@nodes$name
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("names<-", "GraphSpace", function(x, value) {
    gs_vertex_attr(x, "name") <- value
    return(x)
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_vertex_attr", "GraphSpace", function(x, name, ...) {
    g <- x@graph
    if(missing(name)){
        att <- igraph::vertex_attr(graph = g, ...=...)
    } else {
        if(name %in% igraph::vertex_attr_names(g)){
            att <- igraph::vertex_attr(graph = g, name = name, ...=...)
            if(name!="name") names(att) <- V(g)$name
        } else {
            att <- NULL
        }
    }
    return(att)
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_vertex_attr<-", "GraphSpace", function(x, name, ..., value) {
    g <- x@graph
    if (missing(name) && is.list(value)){
        #workaround for analogy to igraph's "syntactic sugar" $<-
        len1 <- unlist(lapply(value, length))==1
        if(any(len1)){
            for(i in which(len1)){
                vl <- value[[i]]
                vl <- ifelse(.is_replicable(vl), vl, list(vl))
                value[[i]] <- rep(vl, igraph::vcount(g))
            }
        }
        igraph::vertex_attr(graph = g) <- value
    } else {
        if(length(value)==1){
            value <- ifelse(.is_replicable(value), value, list(value))
        }
        igraph::vertex_attr(graph = g, name = name, ...=...) <- value  
    }
    x <- .updateNodeSpace(x, g)
    return(x)
})
.is_replicable <- function(x) {
    tryCatch({
        rep(x, 2)
        TRUE
    }, error = function(e) FALSE)
}

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_edge_attr", "GraphSpace", function(x, name, ...) {
    g <- x@graph
    att <- igraph::edge_attr(graph = g, name = name, ...=...)
    return(att)
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_edge_attr<-", "GraphSpace", function(x, name, ..., value) {
    g <- x@graph
    if (missing(name) && is.list(value)){
        #workaround for analogy to igraph's "syntactic sugar" $<-
        len1 <- unlist(lapply(value, length))==1
        if(any(len1)){
            for(i in which(len1)){
                vl <- value[[i]]
                vl <- ifelse(.is_replicable(vl), vl, list(vl))
                value[[i]] <- rep(vl, igraph::ecount(g))
            }
        }
        igraph::edge_attr(graph = g) <- value
    } else {
        if(length(value)==1){
            value <- ifelse(.is_replicable(value), value, list(value))
        }
        igraph::edge_attr(graph = g, name = name, ...=...) <- value  
    }
    x <- .updateEdgeSpace(x, g)
    return(x)
})
.updateEdgeSpace <- function(x, g){
  x@graph <- .validate_igraph(g)
  x@edges <- .get_edges(x@graph)
  return(x)
}
.updateNodeSpace <- function(x, g) {
    x@graph <- .validate_igraph(g)
    pars <- x@pars
    if(pars$is.normalized){
        if(pars$image.layer){
            x <- normalizeGraphSpace(x, 
                image = x@misc$image,
                mar = pars$mar,
                flip.x = pars$flip.x %||% FALSE, 
                flip.y = pars$flip.y %||% FALSE, 
                rotate.xy = pars$rotate.xy %||% FALSE, 
                flip.v = pars$flip.v %||% TRUE, 
                flip.h = pars$flip.h %||% FALSE,
                verbose = FALSE)
        } else {
            x <- normalizeGraphSpace(x, 
                mar = pars$mar,
                flip.x = pars$flip.x %||% FALSE, 
                flip.y = pars$flip.y %||% FALSE, 
                rotate.xy = pars$rotate.xy %||% FALSE,
                verbose = FALSE)
        }
    } else {
      x@nodes <- .get_nodes(x@graph)
    }
    return(x)
}

# .recenter_nodes <- function(nodes, image, mar){
#     if(is.null(image)){
#         temp_list <- .center_graph_nodes(nodes, mar)
#     } else {
#         temp_list <- .adjust_image_nodes(nodes, image, mar)
#     }
#     return(temp_list)
# }

################################################################################
### Plotting adjusts for GraphSpace-class methods
################################################################################

#-------------------------------------------------------------------------------
.add_labels1 <- function(ggp, nodes, node.labels, label.size, label.color){
  node.labels <- node.labels[!duplicated(node.labels)]
  if (is.null(names(node.labels))) names(node.labels) <- node.labels
  names(node.labels) <- ifelse(is.na(names(node.labels)), 
    node.labels, names(node.labels))
  names(node.labels) <- ifelse(names(node.labels) == "",
    node.labels, names(node.labels))
  idx_df <- .label_idx(node.labels, nodes)
  if(any(is.na(idx_df$idx))){
    stop("All 'node.labels' should be annotated in the 'PathwaySpace' object.",
      call. = FALSE)
  }
  nodes_ft <- nodes[idx_df$idx, , drop = FALSE]
  nodes_ft$ID <- idx_df$label
  x <- y <- ID <- NULL
  ggp <- ggp + ggplot2::geom_text(
    mapping = aes(x = x, y = y, label = ID), 
    data = nodes_ft, fontface = "bold", size.unit = "mm",
    size = label.size, colour = label.color, 
    vjust="center", hjust="center")
  return(ggp)
}
.label_idx <- function(node.labels, nodes){
  idx_df <- data.frame(name=node.labels, label=names(node.labels))
  idx_df$name1 <- match(node.labels, nodes$name)
  idx_df$name2 <- match(names(node.labels), nodes$name)
  idx_df$labl1  <- match(node.labels, nodes$nodeLabel)
  idx_df$labl2  <- match(names(node.labels), nodes$nodeLabel)
  idx_df$idx <- sapply(seq_len(nrow(idx_df)), function(i){
    idx <- idx_df[i,seq.int(3,6)]
    idx[!is.na(idx)][1]
  })
  idx_df <- idx_df[,c(1,2,7)]
  return(idx_df)
}

#-------------------------------------------------------------------------------
.add_labels2 <- function(ggp, nodes){
  vatt <- .get_default_vatt()
  nodes_ft <- nodes[!is.na(nodes$nodeLabel), ]
  x <- y <- nodeLabel <- NULL
  ggp <- ggp + geom_text(
    mapping = aes(x = x, y = y, label = nodeLabel), 
    data = nodes_ft, fontface = "bold",
    size = nodes_ft$nodeLabelSize %||% vatt$nodeLabelSize, 
    colour = nodes_ft$nodeLabelColor %||% vatt$nodeLabelColor, 
    size.unit = "pt", vjust="center", hjust="center")
  return(ggp)
}

#-------------------------------------------------------------------------------
.add_image <- function(ggp, image){
  ggp <- ggp + annotation_raster(image, xmin = 0, xmax = 1, ymin = 0, ymax = 1)
  return(ggp)
}

################################################################################
### Package documentation
################################################################################
#' @aliases RGraphSpace-package
#'
#' @section Index:
#' \tabular{ll}{
#' \link{GraphSpace}:
#' \tab Constructor of GraphSpace-class objects.\cr
#' \link{plotGraphSpace}:
#' \tab Wrapper function to plot GraphSpace objects in ggplot2.
#' }
#' Further information is available in the vignettes by typing 
#' \code{vignette('RGraphSpace')}. Documented topics are also available in
#' HTML by typing \code{help.start()} and selecting the RGraphSpace package
#' from the menu.
#'
#' @references
#' `r paste(format(citation("RGraphSpace"), style = "text"), collapse = "\n\n")`
#' 
#' @importFrom lifecycle deprecated deprecate_soft is_present
#' @keywords internal
"_PACKAGE"


################################################################################
### Documentation for some 'toy' datasets
################################################################################
#' @title Toy 'igraph' objects
#'
#' @description Small 'igraph' objects used for workflow demonstrations.
#' All graphs include 'x', 'y', and 'name' vertex attributes.
#'
#' @format igraph
#'
#' @usage data(gtoy1)
#'
#' @source This package.
#'
#' @docType data
#' @keywords gtoys
#' @name gtoys
#' @aliases gtoy1 gtoy2
#' @return A pre-processed igraph object.
#' @examples
#' library(RGraphSpace)
#' data(gtoy1)
#' data(gtoy2)
NULL

#-------------------------------------------------------------------------------
#' Map numeric values to a color raster
#'
#' Helper function that converts numeric values to colors and 
#' returns a raster image. Useful for visualizing numeric matrices 
#' as color backgrounds.
#'
#' @param x A numeric vector or matrix containing values to be mapped to colors.
#' @param palette A vector of colors used as the palette. By default,
#'   \code{hcl.colors(30)} is used.
#' @param na.color Color used for \code{NA} values. Defaults to \code{NA}.
#'
#' @details
#' Values in \code{x} are rescaled to the range of the palette using
#' \code{scales::rescale()}, and each value is mapped to a corresponding
#' color. If \code{x} is a matrix, the resulting raster preserves the same
#' dimensions.
#'
#' @return A raster object as produced by \code{as.raster()}.
#'
#' @examples
#' library(RGraphSpace)
#' 
#' # Convert the volcano matrix to a color raster
#' img <- as_colorraster(volcano)
#' plot(img)
#'
#' @importFrom grDevices hcl.colors
#' @export
as_colorraster <- function(x, palette = hcl.colors(30), na.color = "white") {
  
  if(!is.numeric(x)){
    stop("'x' must be a numeric vector or matrix.", call. = FALSE)
  }
  .validate_gs_colors("allColors", "palette", palette)
  .validate_gs_colors("singleColor", "na.color", na.color)
  
  z <- scales::rescale(x, to = c(1, length(palette)))
  z <- pmin(pmax(round(z), 1), length(palette))
  
  m <- palette[z]
  m[is.na(x)] <- na.color
  
  if (is.matrix(x)) {
    dim(m) <- dim(x)
  }
  
  as.raster(m)

}

#-------------------------------------------------------------------------------
#' @title Draw node elements in a 2D graph layout
#' 
#' @description
#' 
#' Constructor for \link{GeomNodeSpace} ggproto objects.
#' 
#' A wrapper around \link[ggplot2]{geom_point} that enables direct use of
#' node attributes stored in \link{GraphSpace} objects as aesthetics.
#'
#' This geom is designed to map node-level attributes (e.g., \code{fill},
#' \code{size}) or any aesthetics supported by \link[ggplot2]{GeomPoint}.
#'
#' @param mapping Set of aesthetic mappings created by [ggplot2::aes()].
#' These mappings override global aesthetics and are not inherited 
#' from the top-level plot.
#' 
#' @param data The data to be displayed in this layer. It can be a 
#' \link{GraphSpace} object, an \link[igraph]{igraph} object, or the 
#' \code{nodespace_handler()} closure. When \code{NULL} (default),
#' a handler is created internally from the \code{mapping} argument.
#' 
#' @param stat The statistical transformation to use on the data.
#' Defaults to \code{identity}.
#'
#' @param position Position adjustment, either as a string or
#' the result of a call to a position adjustment function.
#'
#' @param ... Additional parameters passed to the underlying
#' drawing function in \link{GeomNodeSpace}.
#'
#' @param na.rm Logical. Should missing values be removed?
#' Defaults to \code{FALSE}.
#' 
#' @param show.legend Logical or a named logical vector indicating
#' whether this layer should be included in legends.
#'
#' @param inherit.aes Logical. If \code{FALSE} (default), the layer will use 
#' aesthetics defined in \code{mapping}.
#' 
#' @param raster Logical. Should node glyphs be rasterized? 
#' Rasterization support is based on \code{\link[ggrastr]{rasterise}}.
#' 
#' @param dpi Numeric. Rasterization resolution.
#' 
#' @param dev Character. Rasterization backend. One of `"cairo"`,
#' `"ragg"`, `"ragg_png"`, or `"cairo_png"`.
#' 
#' @param scale Numeric. Rasterization scaling factor
#' (see \code{\link[ggrastr]{rasterise}}).
#' 
#' @return A ggplot2 layer that renders node glyphs defined by
#' \link{GeomNodeSpace}.
#'
#' @section Aesthetics:
#'
#' \code{geom_nodespace()} understands \link[ggplot2]{geom_point} aesthetics.
#' 
#' If these aesthetics are not explicitly provided in \code{aes()}, they 
#' are automatically retrieved from the \link{GraphSpace} object.
#'
#' \tabular{ll}{
#'   \strong{\code{x}, \code{y}} \tab Required (automatically supplied).\cr
#'   \code{fill} \tab Node interior colour (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
#'   \code{colour} \tab Node border colour (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
#'   \code{alpha} \tab Transparency (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
#'   \code{shape} \tab Node shape (see \link{points} and \link[ggplot2]{aes_linetype_size_shape}).\cr
#'   \code{size} \tab Node size (see *drawing* section and \link[ggplot2]{aes_linetype_size_shape}).\cr
#'   \code{stroke} \tab Node line width (see \link[ggplot2]{gg_par} and \link[ggplot2]{aes_linetype_size_shape}).
#' }
#' 
#' Required aesthetics \code{x} and \code{y} are supplied from the 
#' \link{GraphSpace} object and do not need to be manually mapped.
#' 
#' Additional parameters can be passed to control fixed values for the layer.
#' For example: `fill = "red"`, `stroke = 3`, `alpha = 0.5`, or `shape = 21`.
#' 
#' @section Integration with ggraph:
#' 
#' \code{geom_nodespace} is compatible with the \code{ggraph} methods.
#' When used within a \code{ggraph()} call, the default \code{nodespace_handler()} 
#' automatically:
#' \itemize{
#'   \item Identifies the current \code{layout_ggraph}.
#'   \item Extracts the \code{x} and \code{y} coordinates calculated by \code{ggraph}.
#'   \item Reconstructs a temporary \code{GraphSpace} object to inject spatial 
#'   metadata and user-chosen \code{ggraph} layout.
#' }
#' 
#' @details
#' 
#' The interpretation of \strong{\code{size}} depends on how it is provided:
#' \itemize{
#'   \item **As an aesthetic**: When mapped within \code{aes()}, \code{size}
#'   follows the behavior of \link[ggplot2]{geom_point}, using absolute
#'   units to ensure consistency with the plot legends.
#'   \item **As a parameter**: When set outside \code{aes()}, \code{size} is 
#'   treated as a percentage of the viewport (\code{[0, 100]}), scaling 
#'   in \code{npc} units. This allows nodes to resize dynamically with 
#'   viewport changes.
#' }
#' 
#' @seealso
#' \link{GraphSpace}, \link{geom_edgespace}, \link{geom_graphspace}, 
#' \link[ggplot2]{geom_point}
#'
#' @examples
#' library(RGraphSpace)
#' library(igraph)
#' library(ggplot2)
#' 
#' # Make a demo igraph
#' gtoy1 <- make_star(15, mode="out")
#' 
#' # Set some node attributes
#' V(gtoy1)$nodeSize <- runif(vcount(gtoy1), 1, 20)
#' V(gtoy1)$nodeColor <- rainbow(vcount(gtoy1))
#' 
#' # Set some variables
#' V(gtoy1)$user_var1 <- runif(vcount(gtoy1), 1, 3)^3
#' V(gtoy1)$user_var2 <-  rep(c(1, 2, 3), each = 5)
#' 
#' # Create a GraphSpace object
#' gs <- GraphSpace(gtoy1, layout = layout_in_circle(gtoy1))
#' 
#' \dontrun{
#' 
#' # Example 1: Nodes scaling with the legend
#' # When 'size' is mapped inside aes(), it follows
#' # ggplot2 default behavior: size is translated 
#' # to absolute units (mm) via 'scale_size()'.
#' 
#' ggplot() + 
#' geom_edgespace(data = gs, arrow_offset = 0.01) +
#'   geom_nodespace(mapping = aes(size = nodeSize, fill = user_var2), 
#'   data = gs) + 
#'   scale_size(range = c(1, 12)) + 
#'   theme(aspect.ratio = 1)
#'   
#' # Example 2: Nodes scaling with the viewport
#' # When 'size' is passed as a node attribute, 
#' # inherited from the igraph object, it is 
#' # interpreted as a percentage of the plotting 
#' # area and translated to NPC units.
#' 
#' ggplot() + 
#' geom_edgespace(data = gs, arrow_offset = 0.01) +
#' geom_nodespace(mapping = aes(fill = user_var2), data = gs) +
#' theme(aspect.ratio = 1)
#'   
#' }
#' 
#' @export
geom_nodespace <- function(mapping = NULL, data = NULL, 
  stat = StatNodeSpace, position = "identity", ...,
  na.rm = FALSE, show.legend = NA, inherit.aes = FALSE,  
  raster = FALSE, dpi = NULL, dev = "cairo", scale = 1) {
  
  if (is.null(data)) {
    data <- nodespace_handler(mapping)
  } else if (!inherits(data, "nodespace_handler")){
    if (is.function(data)){
      rlang::abort(
        message = c(
          "x" = "Invalid handler function provided to `data`.",
          "*" = "Use `nodespace_handler()` to create a compatible handler."
        )
      )
    }
    data <- nodespace_handler(mapping)(data)
  }
  
  mapping <- .mapping_nodespace(mapping)
  
  params <- rlang::list2(
    na.rm = na.rm, 
    raster = raster, 
    dpi = dpi, 
    dev = dev, 
    scale = scale, 
    ...)
  
  params$.size_unit <- if("size" %in% names(mapping)) "mm" else "npc"
  
  ggplot2::layer(
    geom = GeomNodeSpace,
    mapping = mapping,
    data = data,
    stat = stat,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = params
  )
  
}

#-------------------------------------------------------------------------------
#' Attribute Processing for GeomNodeSpace
#'
#' Manage visual attribute precedence (color, size, shape) for `GeomNodeSpace` 
#' objects.
#'
#' @section Attribute Priority:
#' 1. Explicit `aes()` mappings.
#' 2. Fixed `geom_nodespace()` arguments.
#' 3. Original graph attributes (via `optional_aes`).
#' 
#' During the `setup_data` stage, the Stat invokes internal functions 
#' to resolve value priority:
#' \enumerate{
#'   \item **Explicit Mapping**: Values defined by the user inside `aes()`.
#'   \item **Fixed Parameters**: Constant values passed as arguments in the `geom_nodespace()` call.
#'   \item **Graph Attributes**: Original attributes stored within the GraphSpace 
#'   object, retrieved from the data columns.
#' }
#'
#' @format A \code{ggproto} object.
#' @seealso \code{\link{geom_nodespace}}
#' @export
StatNodeSpace <- ggproto(
  "StatNodeSpace", ggplot2::Stat,
  optional_aes = c("nodeSize", "nodeShape", "nodeLineWidth", 
    "nodeColor", "nodeLineColor", "nodeAlpha"),
  setup_data = function(data, params) {
    data <- .params_nodespace(params, data)
    return(data)
  },
  compute_panel = function(data, scales){
    return(data)
  }
)

#-------------------------------------------------------------------------------
#' @rdname geom_nodespace
#' @export
nodespace_handler <- function(mapping = NULL) {
  
  fn <- function(data) {
    if (is_waiver(data)) return(NULL)
    vars <- .detect_mapping_vars(mapping)
    if ( inherits(data, c("igraph", "layout_ggraph")) ) {
      data <- gs_nodes(GraphSpace(data, verbose = FALSE))
    } else if (inherits(data, "GraphSpace")){
      data <- gs_nodes(data, vars = vars)
    } else if (inherits(data, "gs_nodes")){
      graph <- attr(data, ".gs_graph")
      if(inherits(graph, "GraphSpace")){
        data <- gs_nodes(graph, vars = vars)
      }
    } else {
      rlang::abort(
        message = c(
          "x" = "`nodespace_handler()` received an unsupported object type.",
          "i" = "Input must be a 'GraphSpace', 'gs_nodes', 'igraph', 'tidygraph', or 'ggraph' layout."
        )
      )
    }
    
    return(data)
  }
  attr(fn, "gs_handler_type") <- "node"
  class(fn) <- c("nodespace_handler", class(fn))
  return(fn)
}

#-------------------------------------------------------------------------------
.detect_mapping_vars <- function(mapping){
  if(is.null(mapping)){
    return(mapping)
  }
  vars <- unique(unlist(
    lapply(mapping, function(x) {
      tryCatch(all.vars(x), error = function(e) character())
    })
  ))
  return(vars)
}

#-------------------------------------------------------------------------------
.mapping_nodespace <- function(mapping) {
  
  x <- y <- NULL
  default_mapping <- ggplot2::aes(x = x, y = y)
  
  nodeColor <- nodeSize <- nodeShape <- nodeLineWidth <- 
    nodeLineColor <- nodeAlpha <- NULL
  optional_mapping <- ggplot2::aes(
    nodeColor = nodeColor, 
    nodeSize = nodeSize,
    nodeShape = nodeShape,
    nodeLineWidth = nodeLineWidth,
    nodeLineColor = nodeLineColor,
    nodeAlpha = nodeAlpha)
  
  if (is.null(mapping)) {
    mapping <- utils::modifyList(
      default_mapping, optional_mapping)
  } else {
    mapping <- utils::modifyList(utils::modifyList(
      default_mapping, optional_mapping), mapping)
  }
  
  return(mapping)
  
}

#-------------------------------------------------------------------------------
.params_nodespace <- function(params, nodes, mapping){
  
  # Note: 'mapping' is read from colnames(nodes) because, at this
  # stage, ggplot2 has already evaluated aes() and pruned unmapped
  # columns. Any standard aesthetic name present here (e.g. "size",
  # "fill") was explicitly mapped by the user.
  if(missing(mapping)){
    mapping <- colnames(nodes)
  } else {
    mapping <- names(mapping)
  }
  
  if(is.null(params[["size"]]) && !"size" %in% mapping){
    if("nodeSize" %in% names(nodes) ){
      nodes[["size"]] <- nodes[["nodeSize"]]
    }
  }
  
  if(is.null(params[["stroke"]]) && !"stroke" %in% mapping ){
    if("nodeLineWidth" %in% names(nodes) ){
      nodes[["stroke"]] <- nodes[["nodeLineWidth"]]
    }
  }
  
  if(is.null(params[["shape"]]) && !"shape" %in% mapping ){
    if("nodeShape" %in% names(nodes) ){
      nodes[["shape"]] <- nodes[["nodeShape"]]
    }
  }
  
  if(is.null(params[["fill"]]) && !"fill" %in% mapping ){
    if("nodeColor" %in% names(nodes) ){
      nodes[["fill"]] <- nodes[["nodeColor"]]
    }
  }
  
  if(is.null(params[["colour"]]) && !"colour" %in% mapping ){
    if("nodeLineColor" %in% names(nodes) ){
      nodes[["colour"]] <- nodes[["nodeLineColor"]]
    }
  }
  
  if(is.null(params[["alpha"]]) && !"alpha" %in% mapping ){
    if("nodeAlpha" %in% names(nodes) ){
      nodes[["alpha"]] <- nodes[["nodeAlpha"]]
    }
  }
  
  return(nodes)
  
}

#-------------------------------------------------------------------------------
#' @title GeomNodeSpace: a ggplot2 prototype for GraphSpace-class methods
#'
#' @description
#' 
#' \code{GeomNodeSpace} is the underlying \link[ggplot2]{ggproto} object 
#' used by \link{geom_nodespace} to draw node elements in a graph layout.
#'
#' This geom is designed for network diagrams, where graph attributes 
#' are often already in their final form (e.g., hex colors).
#'
#' @section Aesthetics:
#'
#' \code{GeomNodeSpace} understands ggplot2's conventions for point-like geoms.
#' 
#' @seealso
#' \link{geom_nodespace}, \link[ggplot2]{geom_point}
#'
#' @importFrom ggplot2 scale_colour_identity zeroGrob
#' @importFrom ggplot2 geom_point geom_segment aes Geom .pt geom_text gg_par
#' @importFrom ggplot2 element_rect margin element_blank layer theme_bw
#' @importFrom ggplot2 element_line element_text ggproto theme theme_gray
#' @importFrom ggplot2 scale_linetype_manual annotation_raster coord_fixed
#' @importFrom ggplot2 scale_x_continuous scale_y_continuous expansion labs
#' @importFrom ggplot2 expansion translate_shape_string is_waiver 
#' @importFrom ggplot2 remove_missing geom_blank draw_key_point
#' @importFrom grid gpar arrow unit pointsGrob segmentsGrob
#' @importFrom grid grobTree gList grobName
#' @importFrom scales alpha squish
#' @importFrom utils modifyList
#' @importFrom rlang list2 warn exec '%||%'
#' @export
GeomNodeSpace <- ggproto(
  
  "GeomNodeSpace", ggplot2::GeomPoint, 
  
  required_aes = c("x", "y"),
  
  non_missing_aes = c("size", "stroke", "shape", "colour"),
  
  default_aes = aes(
    size = 5,
    shape = 21,
    colour = "grey20",
    fill = "grey80",
    stroke = 0.5,
    alpha = NA
  ),
  
  draw_panel = function(self, data, panel_params, coord, 
    na.rm = FALSE, .size_unit = "mm", raster = FALSE, 
    dpi = NULL, dev = "cairo", scale = 1) {
    
    data$shape <- translate_shape_string(data$shape)
    
    data <- .geom_check_node_size(data, .size_unit)
    
    coords <- coord$transform(data, panel_params)
    
    # Create node grobs
    grobs <- .get_node_grobs(coords, .size_unit)
    grobs$name  <- grobName(grobs, "geom_nodespace")
    
    if(raster){
      grobs <- .as_rasteriser(grobs, dpi = dpi, 
        dev = dev, scale = scale)
    }
    
    grobs
    
  },
  
  draw_key = draw_key_point
  
)

#-------------------------------------------------------------------------------
.geom_check_node_size <- function(nodes, size_unit = "npc"){
  # 1. Passed as a aesthetic, 'size' follows 'geom_point' behavior, in mm;
  # 2. Passed as a parameter, 'size' scales with the viewport (%) with
  # range expected in [0, 100], then later converted to 'npc'.
  if (size_unit == "npc") {
    .check_node_size(nodes[["size"]])
    nodes$size <- scales::squish(nodes[["size"]], range = c(0, 100))
  }
  return(nodes)
}

.check_node_size <- function(size) {
  if ( any(size < 0 | size > 100, na.rm = TRUE) ){
    rlang::warn(
      message = c(
        "Node 'size' outside the expected range of [0, 100].",
        "i" = "Passed as a parameter, 'size' scales with the viewport (%).",
        "*" = "Use aes(size = ...) for data-driven scaling."
      )
    )
  }
}

#-------------------------------------------------------------------------------
# pointsGrob
.get_node_grobs <- function(coords, size_unit){
  
  if(size_unit == "npc"){
    coords$size <- coords$size * .gs_pch_to_npc()
    grob <- grid::pointsGrob(
      x = coords$x,
      y = coords$y,
      pch = coords$shape,
      size = grid::unit(coords$size, size_unit),
      gp = ggplot2::gg_par(
        fill = scales::alpha(coords$fill, coords$alpha),
        col = scales::alpha(coords$colour, coords$alpha),
        stroke = coords$stroke
      )
    )
  } else {
    grob <- grid::pointsGrob(
      x = coords$x,
      y = coords$y,
      pch = coords$shape,
      gp = ggplot2::gg_par(
        fill = scales::alpha(coords$fill, coords$alpha),
        col = scales::alpha(coords$colour, coords$alpha),
        pointsize = coords$size, stroke = coords$stroke
      )
    )
  }
  
  grob

}

#-------------------------------------------------------------------------------
# For gspace, node 'size' is defined on a [0, 100] scale and converted
# to NPC units using a 0.01 factor.
# For plotting symbols ('pch') in 0:25, the effective symbol diameter is
# approximately 75% of the character height (see graphics::points()); 
# a 1/0.75 correction is applied to recover the intended size.
.gs_pch_to_npc <- function(){
  .gs_nsz_to_npc() * (1/0.75)
}
.gs_nsz_to_npc <- function(){
  0.01
}

#-------------------------------------------------------------------------------
# Internal rasterization adapter
# Currently delegates to the ggrastr rasteriser mechanism.
# Isolated here so GraphSpace can switch to a native implementation
# without requiring changes to geoms.
.as_rasteriser <- function(grob, dpi = NULL, dev = "cairo", scale = 1){
  class(grob) <- unique( c("rasteriser", class(grob)) )
  grob$dpi <- dpi
  grob$dev <- dev
  grob$scale <- scale
  grob
}

#-------------------------------------------------------------------------------
#' RGraphSpace ggplot2 themes
#'
#' A set of \pkg{ggplot2} themes used by \pkg{RGraphSpace} plots.
#' 
#' @param txt_size Numeric value to scale plot- and axis-related text elements.
#' @param leg_size Numeric value to scale legend-related elements.
#' @param bg_color A color name or hex code specifying the panel background.
#' @param key_fill Logical; if TRUE, treats the fill legend as discrete 
#' to adjust key size.
#' @param key_colour Logical; if TRUE, treats the colour legend as discrete 
#' to adjust key size.
#' @param ... Additional arguments passed to \code{theme_gspace_th*} and
#' \link[ggplot2]{ggtheme}.
#' @details
#' \code{theme_gspace_th0()} is a minimal wrapper around 
#' \link[ggplot2]{theme_gray} that simplifies axis and legend scaling.
#' The \code{txt_size} and \code{leg_size} arguments aggregate related 
#' \link[ggplot2]{theme} parameters for quick thematic overrides.
#' 
#' \code{theme_gspace_th1()} builds on \code{theme_gspace_th0()} and 
#' modifies grid lines, axis appearance, and panel borders.
#'
#' \code{theme_gspace_th2()} is similar to \code{theme_gspace_th1()} with
#' simplified grid elements and a customizable panel background.
#'
#' \code{theme_gspace_th3()} is similar to \code{theme_gspace_th2()} but with
#' slightly adjusted margins, tick appearance, and legend formatting.
#'
#' The \code{theme_gspace_coords()} is a helper function that also adds axes
#' scales for normalized coordinates. It configures axis breaks, limits, 
#' and expansion for graph layouts. Plot coordinates are ideally normalized 
#' to the interval \code{[0, 1]}.
#' 
#' @return
#' \code{theme_gspace_th*()} return a \code{ggplot2} theme object.
#' 
#' \code{theme_gspace_coords()} returns a list containing scale and theme
#' components that can be added to a \pkg{ggplot2} plot.
#'
#' @seealso \link[ggplot2]{ggtheme}, \link[ggplot2]{theme}
#' 
#' @examples
#' library(ggplot2)
#' 
#' ggplot(mtcars, aes(wt, mpg)) +
#'   geom_point() +
#'   theme_gspace_th0()
#'   
#' ggplot(mtcars, 
#'   aes(scales::rescale(wt), 
#'       scales::rescale(mpg))) +
#'   geom_point() +
#'   theme_gspace_coords("th2", is_norm = TRUE)
#'   
#' @importFrom ggplot2 "%+replace%"
#' @name theme_gspace
#' @export
theme_gspace_th0 <- function(txt_size = 1, leg_size = 1, 
  bg_color = "grey95", key_fill = FALSE,
  key_colour = FALSE, ...) {
  
  et0 <- ggplot2::element_text(size = 14 * txt_size)
  et1 <- ggplot2::element_text(size = 12 * txt_size)
  et2 <- ggplot2::element_text(size = 11 * txt_size)
  et3 <- ggplot2::element_text(size = 9 * leg_size)
  et4 <- ggplot2::element_text(size = 8 * leg_size)
  et1s <- ggplot2::element_text(size = 12 * txt_size, hjust = 0)
  
  l_th <- list()
  l_th[[1]] <- ggplot2::theme_gray(...) %+replace%
  ggplot2::theme(
    plot.title = et0, plot.subtitle = et1s,
    axis.title = et1, axis.text = et2,
    
    legend.title = et3, legend.text = et4,
    legend.key.spacing = unit(leg_size, "mm"),
    legend.spacing = unit(leg_size, "mm"),
    legend.key.size = unit(4 * leg_size, "mm"),
    legend.margin = margin(leg_size, leg_size, leg_size, leg_size),
    legend.background = element_blank(),
    legend.box.background = element_blank(),
    
    panel.background = element_rect(fill = bg_color, colour = NA),
    aspect.ratio = 1, complete = TRUE)
  
  l_args <- .guide_args(leg_size, key_fill, key_colour)
  if (length(l_args) > 0) {
    l_th[[2]] <- do.call(guides, l_args)
  }
  
  return(l_th)
  
}

#-------------------------------------------------------------------------------
#' @rdname theme_gspace
#' @export
theme_gspace_th1 <- function(txt_size = 1, leg_size = 1, 
  bg_color = "grey95", key_fill = FALSE,
  key_colour = FALSE, ...) {
  
  et0 <- ggplot2::element_text(size = 14 * txt_size)
  et1 <- ggplot2::element_text(size = 12 * txt_size)
  et2 <- ggplot2::element_text(size = 11 * txt_size)
  et3 <- ggplot2::element_text(size = 9 * leg_size)
  et4 <- ggplot2::element_text(size = 8 * leg_size)
  et1s <- ggplot2::element_text(size = 12 * txt_size, hjust = 0)
  
  l_th <- list()
  l_th[[1]] <- ggplot2::theme_gray(...) %+replace%
    ggplot2::theme(
      plot.title = et0, plot.subtitle = et1s,
      axis.title = et1, axis.text = et2,
      
      legend.title = et3, legend.text = et4,
      legend.key.spacing = unit(leg_size, "mm"),
      legend.spacing = unit(leg_size, "mm"),
      legend.key.size = unit(4 * leg_size, "mm"),
      legend.margin = margin(leg_size, leg_size, leg_size, leg_size),
      legend.background = element_blank(),
      legend.box.background = element_blank(),
      
      panel.background = element_rect(fill = bg_color, colour = NA),
      panel.grid.minor = element_line(linewidth = 0.7),
      panel.grid.major = element_line(linewidth = 0.7),
      panel.border = element_rect(linewidth = 1.2),
      
      axis.ticks = element_line(linewidth = 0.7),
      axis.line = element_blank(),
      plot.margin = margin(1, 1, 1, 1), 
      aspect.ratio = 1, complete = TRUE)
  
  l_args <- .guide_args(leg_size, key_fill, key_colour)
  if (length(l_args) > 0) {
    l_th[[2]] <- do.call(guides, l_args)
  }
  return(l_th)
}

#-------------------------------------------------------------------------------
#' @rdname theme_gspace
#' @export
theme_gspace_th2 <- function(txt_size = 1, leg_size = 1, 
  bg_color = "grey95", key_fill = FALSE, 
  key_colour = FALSE, ...) {
  
  et0 <- ggplot2::element_text(size = 14 * txt_size)
  et1 <- ggplot2::element_text(size = 12 * txt_size)
  et2 <- ggplot2::element_text(size = 11 * txt_size)
  et3 <- ggplot2::element_text(size = 9 * leg_size)
  et4 <- ggplot2::element_text(size = 8 * leg_size)
  et1s <- ggplot2::element_text(size = 12 * txt_size, hjust = 0)
  
  l_th <- list()
  l_th[[1]] <- ggplot2::theme_gray(...) %+replace%
    ggplot2::theme(
      plot.title = et0, plot.subtitle = et1s,
      axis.title = et1, axis.text = et2,
      
      legend.title = et3, legend.text = et4,
      legend.key.spacing = unit(leg_size, "mm"),
      legend.spacing = unit(leg_size, "mm"),
      legend.key.size = unit(4 * leg_size, "mm"),
      legend.margin = margin(leg_size, leg_size, leg_size, leg_size),
      legend.background = element_blank(),
      legend.box.background = element_blank(),
      
      panel.background = element_rect(fill = bg_color, colour = NA),
      panel.grid.minor = element_blank(),
      panel.grid.major = element_line(linewidth = 0.75),
      axis.ticks = element_line(linewidth = 0.7),
      
      axis.line = element_blank(), panel.border = element_blank(),
      plot.margin = margin(5, 10, 0, 10), 
      aspect.ratio = 1, complete = TRUE)
  
  l_args <- .guide_args(leg_size, key_fill, key_colour)
  if (length(l_args) > 0) {
    l_th[[2]] <- do.call(guides, l_args)
  }
  return(l_th)
}

#-------------------------------------------------------------------------------
#' @rdname theme_gspace
#' @export
theme_gspace_th3 <- function(txt_size = 1, leg_size = 1, 
  bg_color = "grey95", key_fill = FALSE, 
  key_colour = FALSE, ...) {
  
  et0 <- ggplot2::element_text(size = 14 * txt_size)
  et1 <- ggplot2::element_text(size = 12 * txt_size)
  et2 <- ggplot2::element_text(size = 11 * txt_size)
  et3 <- ggplot2::element_text(size = 9 * leg_size)
  et4 <- ggplot2::element_text(size = 8 * leg_size)
  et1s <- ggplot2::element_text(size = 12 * txt_size, hjust = 0)
  
  l_th <- list()
  l_th[[1]] <- ggplot2::theme_gray(...) %+replace%
    ggplot2::theme(
      plot.title = et0, plot.subtitle = et1s,
      axis.title = et1, axis.text = et2,
      
      legend.title = element_text(size = 9 * leg_size, vjust = 1), 
      legend.text = element_text(size = 8 * leg_size, vjust = 1),
      legend.key.spacing = unit(leg_size, "mm"), 
      legend.spacing = unit(leg_size, "mm"),
      legend.key.size = unit(4 * leg_size, "mm"),
      legend.margin = margin(leg_size, leg_size, leg_size, leg_size),
      legend.background = element_blank(),
      legend.box.background = element_blank(),
      legend.box.margin = margin(0, 0, 0, 0),
      
      panel.background = element_rect(fill = bg_color, colour = NA),
      panel.grid.minor = element_blank(),
      panel.grid.major = element_line(linewidth = 0.75),
      
      axis.ticks = element_line(linewidth = 0.5),
      axis.line = element_blank(), panel.border = element_blank(),
      plot.margin = margin(5, 5, 5, 5),
      aspect.ratio = 1, complete = TRUE)
  
  l_args <- .guide_args(leg_size, key_fill, key_colour)
  if (length(l_args) > 0) {
    l_th[[2]] <- do.call(guides, l_args)
  }
  return(l_th)
}

#' @param theme Character string specifying the GraphSpace theme variant. 
#' Options: `th0`, `th1`, `th2`, and `th3`.
#' @param is_norm Logical; if TRUE, assumes plot coordinates are 
#' already normalized in \code{[0, 1]}.
#' @param xlab The title for the 'x' axis.
#' @param ylab The title for the 'y' axis.
#' @param expand A range expansion factor applied to both the lower and 
#' upper limits of the 'x' and 'y' scales.
#' @rdname theme_gspace
#' @export
theme_gspace_coords <- function(theme = "th0", is_norm = FALSE,
  xlab = "Graph coordinates 1", ylab = "Graph coordinates 2", 
  expand = NULL, ...) {
  
  theme <- match.arg(theme, choices = c("th0", "th1", "th2", "th3"))
  bks <- .set_theme_bks(theme)
  bks$expand <- expand %||% bks$expand
  th <- list()
  if(is_norm){
    th[[length(th)+1]] <- scale_x_continuous(breaks = bks$axis.ticks,
      labels = format(bks$axis.ticks), position = bks$x.position,
      limits = bks$xylim, expand = ggplot2::expansion(mult = bks$expand))
    th[[length(th)+1]] <- scale_y_continuous(breaks = bks$axis.ticks,
      labels = format(bks$axis.ticks), limits = bks$xylim,
      expand = ggplot2::expansion(mult = bks$expand))
  } else {
    th[[length(th)+1]] <- scale_x_continuous(position = bks$x.position)
  }
  ftheme <- .get_gspace_theme(theme)
  th[[length(th)+1]] <- ftheme(...)
  th[[length(th)+1]] <- labs(x = xlab, y = ylab)
  attr(th, "gspace_pars") <- bks
  return(th)
}
.get_gspace_theme <- function(theme) {
  if (theme == "th3") {
    ftheme <- theme_gspace_th3
  } else if (theme == "th2") {
    ftheme <- theme_gspace_th2
  } else if (theme == "th1") {
    ftheme <- theme_gspace_th1 
  } else {
    ftheme <- theme_gspace_th0
  }
  return(ftheme)
}
.set_theme_bks <- function(theme){
  bks <- list()
  if (theme %in% c("th3")) {
    bks$axis.ticks <- c(0.25, 0.5, 0.75)
    bks$xylim <- c(0, 1)
    bks$expand <- 0.01
    bks$x.position <- "top"
    bks$justify <- "centre"
    bks$leg.position <- "bottom"
  } else if (theme %in% c("th2")) {
    bks$axis.ticks <- seq(0.1, 0.9, 0.2)
    bks$xylim <- c(0, 1)
    bks$expand <- 0.02
    bks$x.position <- "bottom"
    bks$justify <- "right"
    bks$leg.position <- "right"
  } else {
    bks$axis.ticks <- seq(0, 1, 0.2)
    bks$xylim <- c(0, 1)
    bks$expand <- 0.05
    bks$x.position <- "bottom"
    bks$justify <- "right"
    bks$leg.position <- "right"
  }
  return(bks)
}

#-------------------------------------------------------------------------------
#' @details
#' \code{theme_gspace_legend()} is helper function that adjusts legend text,
#' title, and key sizes by a single scaling factor.
#' 
#' @return
#' \code{theme_gspace_legend()} returns a list of theme and guide components.
#' 
#' @examples
#' # Small scale legends
#' ggplot(mtcars, aes(wt, mpg, fill = factor(cyl))) + 
#'   geom_point(shape = 21) + 
#'   theme_gspace_legend(0.8)
#'
#' @importFrom ggplot2 theme element_text unit guides guide_legend
#' @rdname theme_gspace
#' @export
theme_gspace_legend <- function(leg_size = 1, 
  key_fill = FALSE, key_colour = FALSE, ...) {
  
  et1 <- ggplot2::element_text(size = 9 * leg_size)
  et2 <- ggplot2::element_text(size = 8 * leg_size)
  
  l_th <- list(
    theme(
      legend.title = et1, legend.text = et2,
      legend.key.spacing = unit(2*leg_size, "mm"),
      legend.spacing = unit(leg_size, "mm"),
      legend.key.size = unit(4 * leg_size, "mm"),
      legend.margin = margin(2*leg_size, 2*leg_size, 2*leg_size, 2*leg_size), 
      legend.background = element_blank(),
      legend.box.background = element_blank(),
      legend.box.margin = margin(0, 0, 0, 0), 
    )
  )
  
  l_args <- .guide_args(leg_size, key_fill, key_colour)
  if (length(l_args) > 0) {
    l_th[[length(l_th)+1]] <- do.call(guides, l_args)
  }
  return(l_th)
  
}

#-------------------------------------------------------------------------------
.guide_args <- function(leg_size = 1, key_fill = FALSE, key_colour = FALSE) {
  l_args <- list()
  l_args$shape <- guide_legend(override.aes = list(size = 3 * leg_size))
  l_args$linetype <- guide_legend(
    theme = theme(legend.key.width = unit(10 * leg_size, "mm")),
    override.aes = list(size = 3 * leg_size))
  if (key_fill) {
    l_args$fill <- guide_legend(override.aes = list(size = 3 * leg_size))
  }
  if (key_colour) {
    l_args$colour <- guide_legend(override.aes = list(size = 3 * leg_size))
  }
  l_args
}

#-------------------------------------------------------------------------------
#' Annotate a GraphSpace Plot
#'
#' @description
#' \code{annotation_gspace()} is a generic dispatcher for adding annotation
#' layers to a \code{ggplot}-based \code{GraphSpace} plot. The \code{type}
#' argument selects the annotation type; additional arguments are forwarded
#' to the corresponding handler.
#' 
#' @param type A string specifying the annotation type. Currently available: 
#' \code{"image"} (default), dispatched to \code{annotation_gspace_image()}.
#' @param ... Arguments forwarded to the annotation handler selected by
#'   \code{type}.
#' @param raster An image to be displayed. Accepted types:
#'   \itemize{
#'     \item A \code{\link{GraphSpace}} object — the image is extracted via
#'       \code{\link{gs_image}}.
#'     \item A \code{raster} object
#'       (see \code{\link[grDevices]{as.raster}}).
#'     \item A \code{matrix} or 3D array (RGB/RGBA), coerced to
#'       \code{raster} automatically.
#'   }
#' @param interpolate A logical value indicating whether to apply linear
#'   interpolation when the image is rendered at a different resolution than
#'   its native size. Defaults to \code{FALSE}.
#' @param opacity A numeric value in \code{[0, 1]} controlling the
#'   transparency of the image. \code{1} is fully opaque (default);
#'   \code{0} is fully transparent.
#' @param flip.v A logical value; if \code{TRUE}, the image is flipped
#'   vertically (top-to-bottom). Defaults to \code{FALSE}.
#' @param flip.h A logical value; if \code{TRUE}, the image is flipped
#'   horizontally (left-to-right). Defaults to \code{FALSE}.
#'
#' @return A ggplot2 layer object that can be added to a \code{ggplot()}
#'   call with \code{+}, or \code{invisible(NULL)} with a warning if the
#'   image could not be resolved.
#'
#' @seealso
#' \code{\link[ggplot2]{annotation_raster}},
#' \code{\link{gs_image}},
#' \code{\link{geom_nodespace}},
#' \code{\link{geom_edgespace}}
#'
#' @examples
#' 
#' # Assuming 'gs' is a GraphSpace object with 
#' # an image stored in gs_image(gs)
#' 
#' \dontrun{
#' # Pass a GraphSpace object directly
#' ggplot(gs) +
#'   annotation_gspace(gs) +
#'   geom_edgespace() +
#'   geom_nodespace()
#'
#' # Extract the image explicitly
#' ggplot(gs) +
#'   annotation_gspace(gs_image(gs)) +
#'   geom_edgespace() +
#'   geom_nodespace()
#'
#' # Dim the background and flip vertically
#' ggplot(gs) +
#'   annotation_gspace(gs, opacity = 0.5, flip.v = TRUE) +
#'   geom_edgespace() +
#'   geom_nodespace()
#'   
#' }
#' @rdname annotation_gspace
#' @export
annotation_gspace <- function(..., type = "image") {
  fn <- switch(type,
    image = annotation_gspace_image,
    rlang::abort(c(
      "x" = sprintf("Unknown annotation type '%s'.", type),
      "i" = "Available types: 'image'."
    ))
  )
  fn(...)
}

#' @importFrom ggplot2 annotation_raster
#' @importFrom grDevices col2rgb rgb
#' @rdname annotation_gspace
#' @export
annotation_gspace_image <- function(raster, interpolate = FALSE, opacity = 1,
  flip.v = FALSE, flip.h = FALSE) {

  if (missing(raster)) {
    rlang::abort("Argument 'raster' is missing, with no default.")
  }

  if (inherits(raster, "GraphSpace")) {
    raster <- gs_image(raster)
    if (is.null(raster)) {
      rlang::warn("The 'GraphSpace' object contains no image.")
      return(invisible(NULL))
    }
  }

  if (!inherits(raster, "raster")) {
    raster <- tryCatch({
      grDevices::as.raster(raster)
    }, error = function(e) {
      rlang::warn(c(
        "x" = "Failed to convert 'raster' to a valid raster object.",
        "i" = "Accepted types: matrix, array (RGB/RGBA), or raster."
      ))
      NULL
    })
    if (is.null(raster)) return(invisible(NULL))
  }

  if (opacity != 1) {
    alpha_channel <- as.integer(max(0, min(1, opacity)) * 255)
    img <- grDevices::col2rgb(as.character(raster), alpha = TRUE)
    img[4, ] <- alpha_channel
    raster <- as.raster(t(matrix(
      grDevices::rgb(img[1,], img[2,], img[3,], img[4,], maxColorValue = 255),
      nrow = nrow(raster), ncol = ncol(raster)
    )))
  }

  if (flip.v) raster <- raster[rev(seq_len(nrow(raster))), , drop = FALSE]
  if (flip.h) raster <- raster[, rev(seq_len(ncol(raster))), drop = FALSE]

  ggplot2::annotation_raster(
    raster      = raster,
    xmin        = 0,
    xmax        = 1,
    ymin        = 0,
    ymax        = 1,
    interpolate = interpolate
  )
}

#-------------------------------------------------------------------------------
#' Dynamic Scale Injection for Edge Clipping
#'
#' @description
#' Utility function for **RGraphSpace** that enables edge layers to scan 
#' adjacent nodes and determine their dimensions. This information is used 
#' to compute arrow clipping offsets, preventing edge geometry from 
#' overlapping node symbols.
#' 
#' @param ... Additional parameters passed to other methods (currently ignored).
#'
#' @details
#' This function operates in two stages within the `ggplot2` workflow:
#' 
#' 1. **Capture:** It scans the `plot` layers for a \link{GeomNodeSpace} 
#' to extract both mapping variables (from `aes()`) and static parameters 
#' (specifically `size` and `stroke`).
#' 
#' 2. **Injection:** It locates \link{GeomEdgeSpace} layers and injects
#' scale rules, captured mappings, and fixed parameters into the geometry 
#' parameters.
#'    
#' This "lazy injection" calculates edge clipping based on the actual scales 
#' used by the nodes, even if scales are defined after the layers.
#' 
#' **Note:** `inject_nodespace()` must be called last in the `ggplot` chain to 
#' allow the function to correctly scan all previously added layers and scales.
#'
#' @return An object of class `inject_nodespace`, which interacts with the 
#' ggplot2 `+` operator.
#' 
#' @seealso
#' \link{geom_edgespace}, \link{geom_nodespace}
#' 
#' @examples
#' 
#' library(igraph)
#' library(ggplot2)
#' 
#' # Generate a toy star graph
#' gtoy1 <- make_star(15, mode="out")
#'
#' # Set node and edge attributes
#' V(gtoy1)$my_node_var <- runif(vcount(gtoy1), 1, 20)
#' E(gtoy1)$my_edge_var <-  runif(ecount(gtoy1), 1, 20)
#'
#' # Create a GraphSpace object with a circular layout
#' gs <- GraphSpace(gtoy1, layout = layout_in_circle(gtoy1))
#'
#' \dontrun{
#' # Build the plot
#' # Note that inject_nodespace() is called at the end to
#' # synchronize node sizes with edge clipping.
#' ggplot() +
#'   geom_edgespace(aes(colour = my_edge_var), data = gs) +
#'   geom_nodespace(aes(size = my_node_var), data = gs) +
#'   scale_size(range = c(2, 15)) +
#'   inject_nodespace()
#' }
#' 
#' @importFrom ggplot2 ggplot_add
#' @rdname inject_nodespace
#' @export
inject_nodespace <- function(...) {
  structure(list(...), class = "inject_nodespace")
}

#' @export
#' @method ggplot_add inject_nodespace
ggplot_add.inject_nodespace <- function(object, plot, ...) {
  
  object$size_par <- list()
  object$size_aes <- NULL
  object$size_rule <- NULL
  object$size_unit <- NULL
  
  ##--- ATTEMPT AUTO-DETECTION OF MAPPINGS
  has_nodes <- FALSE
  has_edges <- FALSE
  is_compatible <- FALSE
  p_layers <- list()
  for (i in seq_along(plot@layers)){
    layer <- plot@layers[[i]]
    if (inherits(layer$geom, "GeomNodeSpace") && !has_nodes){
      has_nodes <- TRUE
      p_layers[["nodes"]] <- layer
    }
    if (inherits(layer$geom, "GeomEdgeSpace") && !has_edges){
      has_edges <- TRUE
      p_layers[["edges"]] <- layer
    }
    if(has_nodes && has_edges){
      atts <- c("vertex", "x", "y", "name")
      is_compatible <- identical(p_layers[["edges"]]$geom_params$.nodes[,atts],
        p_layers[["nodes"]]$data[,atts])
      break
    }
  }
  
  if(!is_compatible){
    warning("`inject_nodespace()` failed to detect a compatible layers.",
      call. = FALSE)
    return(plot)
  }
  
  has_aes <- FALSE
  layer <- p_layers[["nodes"]]
  if (inherits(layer$mapping, "uneval")) {
    # Extract aes for clipping calculations
    m <- layer$mapping[names(layer$mapping) %in% c("size", "stroke")]
    if (length(m) > 0) {
      has_aes <- TRUE
      object$size_aes <- m
    }
  }
  if(is.list(layer$aes_params)){
    p <- layer$aes_params[names(layer$aes_params) %in% c("size", "stroke")]
    if (length(p) > 0) object$size_par <- p
  }
  if(is.list(layer$geom_params)){
    u <- layer$geom_params[[".size_unit"]]
    if (!is.null(u)) object$size_unit <- u
  }

  ##--- PREPARE SCALES
  
  has_scales <- FALSE
  size_scales <- Filter(function(s) "size" %in% s$aesthetics, 
    plot$scales$scales)

  if (length(size_scales) > 0){
    sz <- size_scales[[1]] 
    if (inherits(sz, "Scale") && !is.null(sz$palette)){
      has_scales <- TRUE
      object$size_rule <- sz
    }
  }
  
  if(has_aes && !has_scales){
    warning("`inject_nodespace()` could not find a size scale; ",
      "clipping might be inaccurate.",
      call. = FALSE)
  }
  
  ##--- INJECTION INTO EDGE LAYERS
  
  for (i in seq_along(plot@layers)) {

    if (inherits(plot@layers[[i]]$geom, "GeomEdgeSpace")) {
      
      # Inject the scale rule if found and valid
      if (!is.null(object$size_rule)) {
        plot@layers[[i]]$geom_params$.size_rule <- object$size_rule
      }
      
      # Inject the node mappings (captured or manual)
      if (!is.null(object$size_aes)) {
        plot@layers[[i]]$geom_params$size_aes <- object$size_aes
      }
      
      # Inject the user-defined params if present
      if (!is.null(object$size_par$size)) {
          plot@layers[[i]]$geom_params$.nodes[["size"]] <- object$size_par$size
      }
      
      if (!is.null(object$size_par$stroke)) {
          plot@layers[[i]]$geom_params$.nodes[["stroke"]] <- object$size_par$stroke
      }
      
      if (!is.null(object$size_unit)) {
        plot@layers[[i]]$geom_params$.size_unit <- object$size_unit
      }
    }
  }
  
  return(plot)
  
}

setOldClass("igraph")
setOldClass("raster")

#' @title GraphSpace: An S4 class for igraph objects
#'
#' @slot nodes A data frame with xy-vertex coordinates.
#' @slot edges  A data frame with edges.
#' @slot graph An igraph object.
#' @slot image A raster background image matrix.
#' @slot pars A list with parameters.
#' @slot misc A list with intermediate objects for downstream methods.
#'
#' @method plotGraphSpace \code{\link{plotGraphSpace}}
#' @method getGraphSpace \code{\link{getGraphSpace}}
#' @aliases GraphSpace-class
#' @return An S4 class object.
#' @section Constructor:
#' see \code{\link{GraphSpace}} constructor.
#' @import igraph
#' @exportClass GraphSpace
#'
## Class GraphSpace
setClass("GraphSpace",
    slot = c(
        nodes = "data.frame",
        edges = "data.frame",
        graph = "igraph",
        image = "raster",
        pars = "list",
        misc = "list"
    ),
    prototype = list(
        nodes = data.frame(),
        edges = data.frame(),
        graph = igraph::empty_graph(),
        image = as.raster(matrix()),
        pars = list(),
        misc = list()
    )
)
setValidity("GraphSpace", function(object) {
  errors <- character()
  
  # Check if node and edge slots are data.frames
  if (!is.data.frame(object@nodes)) {
    errors <- c(errors, "'nodes' must be a data.frame.")
  }
  if (!is.data.frame(object@edges)) {
    errors <- c(errors, "'edges' must be a data.frame.")
  }
  
  # Check if graph is an igraph object
  if (!inherits(object@graph, "igraph")) {
    errors <- c(errors, "'graph' must be an igraph object.")
  }
  
  # If node names exist, check consistency with graph vertices
  if (nrow(object@nodes) > 0 && igraph::vcount(object@graph) > 0) {
    g_vertex_names <- igraph::V(object@graph)$name
    node_row_names <- rownames(object@nodes)
    
    if (!is.null(node_row_names)) {
      if (!setequal(g_vertex_names, node_row_names)) {
        errors <- c(errors, "Vertex names in 'graph' must match row names in 'nodes'.")
      }
    }
  }
  
  if (length(errors) == 0) TRUE else errors
})

################################################################################
### Validate igraph for RGraphSpace
################################################################################
.validate_igraph <- function(g, layout = NULL, verbose = FALSE) {
    if (verbose) message("Validating the 'igraph' object...")
    if (!inherits(g, "igraph")) {
        stop("'g' should be an 'igraph' object.", call. = FALSE)
    }
    if (inherits(g, "tbl_graph")) {
        g <- tidygraph::as.igraph(g)
    }
    if (!is.null(layout)) {
        if (nrow(layout) != vcount(g)) {
            msg <- paste("'layout' must have xy-coordinates",
                "for the exact number of nodes in 'g'")
            stop(msg, call. = FALSE)
        } else {
            igraph::V(g)$x <- layout[, 1]
            igraph::V(g)$y <- layout[, 2]
        }
    } else if (is.null(igraph::V(g)$x) || is.null(igraph::V(g)$y)) {
        layout <- igraph::layout_nicely(g)
        igraph::V(g)$x <- layout[, 1]
        igraph::V(g)$y <- layout[, 2]
        msg <- paste0("Vertex attributes 'x' and 'y' missing; ",
            "computing layout...")
        if (verbose) message(msg)
    }
    if (is.null(igraph::V(g)$name)) {
        msg <- "Vertex attribute 'name' missing; assigning names... "
        if (verbose) message(msg)
        igraph::V(g)$name <- paste0("n", seq_len(igraph::vcount(g)))
    } else {
        if(is.vector(igraph::V(g)$name) || !is.list(igraph::V(g)$name)){
            if(any(is.na(igraph::V(g)$name))){
                msg <- "NA values found in vertex attribute 'name'."
                stop(msg, call. = FALSE)
            }
            if(!.all_characterValues(igraph::V(g)$name)){
                warning("vertex attribute 'name' converted to character.", 
                    call. = FALSE)
                igraph::V(g)$name <- as.character(igraph::V(g)$name)
            }
        } else {
            msg <- "vertex attribute 'name' should be a character vector."
            stop(msg, call. = FALSE) 
        }
        if (anyDuplicated(igraph::V(g)$name) > 0){
            stop("vertex names must be unique.", call. = FALSE)
        }
    }
    if (!igraph::is_simple(g)) {
        if (verbose && igraph::any_loop(g)) 
            message("Removing loops...")
        if (verbose && igraph::any_multiple(g))
            message("Merging duplicated edges...")
        g <- igraph::simplify(g, remove.loops = TRUE, remove.multiple = TRUE,
          edge.attr.comb = list(weight = "max", "first"))
    }
    if (is.null(igraph::V(g)$nodeLabel)){
        igraph::V(g)$nodeLabel <- as.character(igraph::V(g)$name)
    }
    if (is.null(igraph::V(g)$nodeSize)){
        igraph::V(g)$nodeSize <- .get_default_vatt()[["nodeSize"]]
    }
    if (is.null(igraph::E(g)$arrowType)){
        if (is_directed(g)) {
            igraph::E(g)$arrowType <- 1
        } else {
            igraph::E(g)$arrowType <- 0
        }
    }
    if(verbose && any(which_mutual(g))){
        message("Mutual edges detected: Simplified for data frame...")
        message("Arrows recoded to bidirectional display")
        message("Edge attributes retained from the first occurrence")
    }
    g <- .validate_attributes(g)
    return(g)
    
}

################################################################################
### Validate graph attributes
################################################################################
.validate_attributes <- function(g){
    g <- .validate_nodes(g)
    g <- .validate_edges(g)
    g <- .validate_graph(g)
    return(g)
}

#-------------------------------------------------------------------------------
.validate_nodes <- function(g) {
    
    # get default attributes
    atts <- c(.get_required_vatt(), .get_default_vatt())
    a_names <- names(atts)
    # check default attributes
    b_names <- a_names[a_names %in% igraph::vertex_attr_names(g)]
    if(length(b_names)>0){
        if (vcount(g) > 0) {
            .validate_vatt(igraph::vertex_attr(g)[b_names])
        }
    }
    
    # put default attributes 1st
    d_names <- igraph::vertex_attr_names(g)
    a_names <- a_names[a_names %in% d_names]
    a_names <- c(a_names, d_names[ ! d_names %in% a_names ])
    igraph::vertex_attr(g) <- igraph::vertex_attr(g)[a_names]
    
    # attributes that require transformation
    g <- .validate_nodeshape(g)
    
    return(g)
}

#-------------------------------------------------------------------------------
.validate_edges <- function(g) {
    
    g <- .remove_hidden_eatt(g)
    
    # get default attributes
    atts <- .get_default_eatt(igraph::is_directed(g))
    a_names <- names(atts)
    # check default attributes
    b_names <- a_names[a_names %in% igraph::edge_attr_names(g)]
    if(length(b_names)>0){
        if (igraph::ecount(g) > 0) {
            .validate_eatt(igraph::edge_attr(g)[b_names])
        }
    }
    
    # put default attributes 1st
    d_names <- igraph::edge_attr_names(g)
    a_names <- a_names[a_names %in% d_names]
    a_names <- c(a_names, d_names[ ! d_names %in% a_names ])
    igraph::edge_attr(g) <- igraph::edge_attr(g)[a_names]
    
    # attributes that require transformation
    g <- .validate_arrowtype(g)
    g <- .validate_linetype(g)
    return(g)
}

#-------------------------------------------------------------------------------
.validate_graph <- function(g) {
    d_names <- igraph::graph_attr_names(g)
    if (length(d_names) > 0) {
        for (at in d_names) {
            g <- igraph::delete_graph_attr(g, name = at)
        }
    }
    return(g)
}

################################################################################
### Default RGraphSpace attributes
################################################################################
.get_required_vatt <- function() {
    atts <- list("x" = NA, "y" = NA, "name" = NA)
    return(atts)
}
.get_default_vatt <- function() {
    atts <- list(
        "nodeLabel" = NA, "nodeLabelSize" = 8, "nodeLabelColor" = "grey40",
        "nodeShape" = 21, "nodeSize" = 5, "nodeColor" = "grey80",
        "nodeLineWidth" = 0.5, "nodeLineColor" = "grey20")
    return(atts)
}
.get_default_eatt <- function(is.directed = FALSE) {
    atts <- list("edgeLineType" = "solid", "edgeLineColor" = "grey80",
        "edgeLineWidth" = 0.5)
    if (is.directed) {
        atts$arrowType <- 1
    } else {
        atts$arrowType <- 0
    }
    atts$weight <- 1
    return(atts)
}
# remove internally used hidden attributes
.remove_hidden_eatt <- function(g){
    atts <- names(.get_default_eatt(igraph::is_directed(g)))
    hidden <- setdiff(names(.get_empty_edgedf()), atts)
    hidden <- hidden[hidden %in% igraph::edge_attr_names(g)]
    if (length(hidden) > 0) {
        for (at in hidden) {
            g <- igraph::delete_edge_attr(g, name = at)
        }
    }
    g
}

################################################################################
### Validate attribute values
################################################################################
.validate_vatt <- function(atts) {
    if (!is.null(atts$x)) {
        .validate_gs_args("numeric_vec", "x", atts$x)
    }
    if (!is.null(atts$y)) {
        .validate_gs_args("numeric_vec", "y", atts$y)
    }
    if (!is.null(atts$name)) {
        .validate_gs_args("allCharacter", "name", atts$name)
    }
    if (!is.null(atts$nodeLabel)) {
        .validate_gs_args("allCharacterOrNa", "nodeLabel", atts$nodeLabel)
    }
    if (!is.null(atts$nodeLabelSize)) {
        .validate_gs_args("numeric_vec", "nodeLabelSize", atts$nodeLabelSize)
        if (min(atts$nodeLabelSize) <= 0) {
            stop("'nodeLabelSize' should be a vector of numeric values >0", 
              call. = FALSE)
        }
    }
    if (!is.null(atts$nodeLabelColor)) {
        .validate_gs_colors("allColors", "nodeLabelColor", atts$nodeLabelColor)
    }
    if (!is.null(atts$nodeSize)) {
        .validate_gs_args("numeric_vec", "nodeSize", atts$nodeSize)
        if (max(atts$nodeSize) > 100 || min(atts$nodeSize) < 0) {
            stop("'nodeSize' should be a vector of numeric values in [0, 100]", 
              call. = FALSE)
        }
    }
    if (!is.null(atts$nodeShape)) {
        .validate_gs_args("allCharacterOrInteger", "nodeShape", atts$nodeShape)
    }
    if (!is.null(atts$nodeColor)) {
        .validate_gs_colors("allColors", "nodeColor", atts$nodeColor)
    }
    if (!is.null(atts$nodeLineWidth)) {
        .validate_gs_args("numeric_vec", "nodeLineWidth", atts$nodeLineWidth)
        if (min(atts$nodeLineWidth) < 0) {
            stop("'nodeLineWidth' should be a vector of numeric values >=0", 
              call. = FALSE)
        }
    }
    if (!is.null(atts$nodeLineColor)) {
        .validate_gs_colors("allColors", "nodeLineColor", atts$nodeLineColor)
    }
}
#-------------------------------------------------------------------------------
.validate_eatt <- function(atts) {
    if (!is.null(atts$edgeLineType)) {
        .validate_gs_args("allCharacterOrInteger", "edgeLineType",
            atts$edgeLineType)
    }
    if (!is.null(atts$edgeLineWidth)) {
        .validate_gs_args("numeric_vec", "edgeLineWidth", atts$edgeLineWidth)
        if (min(atts$edgeLineWidth) <= 0) {
            stop("'edgeLineWidth' should be a vector of numeric values >0", 
                call. = FALSE)
        }
    }
    if (!is.null(atts$edgeLineColor)) {
        .validate_gs_colors("allColors", "edgeLineColor", atts$edgeLineColor)
    }
    if (!is.null(atts$arrowType)) {
        .validate_gs_args("allCharacterOrInteger", "arrowType", atts$arrowType)
    }
    if (!is.null(atts$weight)) {
        .validate_gs_args("numeric_vec", "weight", atts$weight)
    }
}

################################################################################
### Transform attribute types
################################################################################

#-------------------------------------------------------------------------------
.validate_nodeshape <- function(g) {
    if (vcount(g) > 0 && "nodeShape" %in% names(vertex_attr(g))) {
        V(g)$nodeShape  <- .transform_nodeshape(V(g)$nodeShape)
    }
    return(g)
}
.transform_nodeshape <- function(vshapes) {
    if (.all_integerValues(vshapes)) {
        vshapes[vshapes > 25] <- 21
        vshapes[vshapes < 0] <- 1
    } else {
        vshapes <- tolower(vshapes)
        pch <- rep(21, length(vshapes))
        pch[grep("circle", vshapes)] <- 21
        pch[grep("ellipse", vshapes)] <- 21
        pch[grep("square", vshapes)] <- 22
        pch[grep("diamond", vshapes)] <- 23
        pch[grep("triangle", vshapes)] <- 24
        pch[grep("rectangle", vshapes)] <- 22
        vshapes <- pch
    }
    return(vshapes)
}

#-------------------------------------------------------------------------------
.validate_arrowtype <- function(g) {
    if (ecount(g) > 0 && "arrowType" %in% names(edge_attr(g))) {
        E(g)$arrowType  <- .transform_arrowtype(E(g)$arrowType, is_directed(g))
    }
    return(g)
}
.transform_arrowtype <- function(eatt, is_dir = FALSE) {
        aty <- .arrowtypes(is_dir)
        if (.all_integerValues(eatt)) {
            idx <- !eatt %in% aty
            if (any(idx)) {
                eatt[idx] <- ifelse(is_dir, 1, 0)
                .arrowtypes_warning(is_dir)
            }
        } else {
            idx <- eatt %in% as.character(aty)
            if (any(idx)) {
                eatt[idx] <- names(aty)[match(eatt[idx], as.character(aty))]
            }
            idx <- !eatt %in% names(aty)
            if (any(idx)) {
                eatt[idx] <- "-->"
                .arrowtypes_warning(is_dir)
            }
            eatt <- aty[eatt]
        }
    return(eatt)
}
.arrowtypes <- function(is.dir = FALSE, unique = FALSE) {
    atp1 <- c(
        "---" = 0, "--" = 0, "-" = 0,
        "-->" = 1, "->" = 1, ">" = 1,
        "<--" = 2, "<-" = 2, "<" = 2,
        "<->" = 3, "<>" = 3,
        "|->" = 4, "|>" = 4)
    atp2 <- c(
        "--|" = -1, "-|" = -1, "|" = -1,
        "|--" = -2, "|-" = -2, "|" = -2,
        "|-|" = -3, "||" = -3,
        "<-|" = -4, "<|" = -4)
    atypes <- c(atp1, atp2)
    if (is.dir) {
        atypes <- atypes[atypes %in% c(-1, 0, 1)]
        if(unique){
            atypes <- atypes[match(unique(atypes), atypes)]
        }
    } else {
        if(unique){
            atp1 <- atp1[match(unique(atp1), atp1)]
            atp2 <- atp2[match(unique(atp2), atp2)]
            atypes <- c(atp1, atp2)
        }
    }
    return(atypes)
}
.arrowtypes_warning <- function(is.dir = FALSE){
    
    atypes <- .arrowtypes(is.dir, unique = TRUE)
    
    if (is.dir) {
        msg1 <- paste("'arrow type' for directed graphs must be one of:\n")
        msg2 <- paste(paste0("'", names(atypes), "'"), atypes, sep = " or ")
        msg2 <- paste(msg2, collapse = ", ")
    } else {
        msg1 <- paste("'arrow type' for undirected graphs must be one of:\n")
        idx <- atypes >= 0
        atp1 <- paste(paste0("'", names(atypes)[idx], "'"), atypes[idx], sep = " or ")
        atp2 <- paste(paste0("'", names(atypes)[!idx], "'"), atypes[!idx], sep = " or ")
        atp1 <- paste0(paste(atp1, collapse = ", "), "\n")
        atp2 <- paste(atp2, collapse = ", ")
        msg2 <- paste0(atp1, atp2)
    }
    msg3 <- c("\n...using default values.")
    warning(msg1, msg2, msg3, call. = FALSE)
    
}

#-------------------------------------------------------------------------------
.validate_linetype <- function(g) {
    if (ecount(g) > 0 && "edgeLineType" %in% names(edge_attr(g))) {
        E(g)$edgeLineType  <- .transform_linetype(E(g)$edgeLineType)
    }
    return(g)
}
.transform_linetype <- function(lty) {
    if (.all_integerValues(lty)) {
        ltypes <- .linetypes()
        lty[!lty %in% ltypes] <- 1
        lty <- ltypes[match(lty, ltypes)]
        lty <- names(lty)
    } else {
        lty <- tolower(lty)
        lty[grep("solid", lty)] <- "solid"
        lty[grep("dotted", lty)] <- "dotted"
        lty[grep("dashed", lty)] <- "dashed"
        lty[grep("long", lty)] <- "longdash"
        lty[grep("two", lty)] <- "twodash"
    }
    return(lty)
}
.linetypes <- function() {
    c('blank' = 0, 'solid' = 1, 'dashed' = 2, 'dotted' = 3,
        'dotdash' = 4, 'longdash' = 5, 'twodash' = 6)
}

#-------------------------------------------------------------------------------
#' Normalize node coordinates to graph and image spaces
#' 
#' Accessory functions to normalize node coordinates in GraphSpace, 
#' either by centering them within the graph boundaries 
#' or by mapping them to pixel coordinates of a background image.
#' 
#' @param gs A \code{GraphSpace} object to be normalized.
#' @param image An optional background image. When provided, \code{x} and 
#' \code{y} coordinates must represent pixel positions in the image matrix.
#' @param ... Additional arguments passed to specific normalization workflows.
#' @param mar A single numeric value in \code{[0, 0.5]} controlling the size of
#' the outer margins around the graph. Without an image, \code{mar} specifies
#' symmetric margins as a fraction of the graph space. With an image,
#' \code{mar} is interpreted as a fraction of the available image margins
#' surrounding the graph.
#' @param flip.y Logical; whether to flip the node coordinates along the y-axis. 
#' Useful for aligning nodes with image backgrounds, which often use an 
#' inverted coordinate system.
#' @param flip.x Logical; whether to flip the node coordinates along the x-axis.
#' @param rotate.xy Logical; whether to rotate x-y coordinates.
#' @param flip.v Logical; whether to vertically flip the background image  
#' matrix (top-to-bottom) to align with the graph coordinate system.
#' @param flip.h Logical; whether to horizontally flip the background image  
#' matrix (left-to-right) to align with the graph coordinate system.
#' @param crop.coord An optional numeric vector of length four specifying a  
#' cropping region (xmin, xmax, ymin, ymax), with values in normalized 
#' coordinates \code{[0,1]}.
#' @param verbose A single logical value specifying to display detailed 
#' messages (when \code{verbose=TRUE}) or not (when \code{verbose=FALSE}).
#' 
#' @usage
#' normalizeGraphSpace(gs, image = NULL, ...)
#' 
#' @details 
#' These functions provide different strategies for coordinate transformation:
#' \itemize{
#'   \item \bold{normalizeGraphSpace}: Re-scales node coordinates to a 
#'   \code{[0, 1]} unit square based on the graph's bounding box (when 
#'   \code{image} is missing) or maps them to pixel coordinates (when 
#'   \code{image} is provided). It handles image-to-graph alignment via 
#'   \code{flip.} and \code{rotate.} arguments, used to adjust the graph 
#'   origin with the image matrix layout. Users should be aware of the 
#'   potential discrepancy between image matrix orientation (top-down) and
#'   graph coordinates (bottom-up). The function attempts to automatically 
#'   adjust the y-axis to align the graph's bottom-up coordinates with the 
#'   image's top-down layout, but further manual adjustments might be required.
#'   \item \bold{cropGraphSpace}: Subsets the normalized graph space into a 
#'   specific region defined by \code{crop.coord}. It recalculates node positions 
#'   and background image boundaries to maintain spatial consistency after cropping. 
#'   This function requires a previously normalized \code{GraphSpace} object.
#' }
#' 
#' @return A \code{GraphSpace} object with updated \code{nodes} 
#' and \code{image} slots.
#' 
#' @note This is an accessory function typically called during 
#' the preprocessing of \code{GraphSpace} objects before rendering.
#'
#' @examples
#' library(igraph)
#' 
#' # Create a star graph
#' gtoy1 <- make_full_graph(15)
#' 
#' # Create a GraphSpace
#' gs <- GraphSpace(gtoy1)
#' 
#' gs <- normalizeGraphSpace(gs)
#' 
#' @aliases normalizeGraphSpace
#' @rdname normalizeGraphSpace-methods
#' @export
setMethod("normalizeGraphSpace",
  signature = c(gs = "GraphSpace", image = "missing"),
  function(gs, image, ...,  mar = 0.1, flip.x = FALSE, flip.y = FALSE, 
    rotate.xy = FALSE, crop_method = "", verbose = TRUE){
    
    if(gs@pars$image.layer) {
      gs@pars$image.layer <- FALSE
    }
    
    extra_args <- list(...)
    if(length(extra_args) > 0 && verbose) {
      warning("Arguments ignored: ", paste(names(extra_args), collapse = ", "))
    }
    
    .validate_gs_args("singleLogical", "flip.x", flip.x)
    .validate_gs_args("singleLogical", "flip.y", flip.y)
    .validate_gs_args("singleLogical", "rotate.xy", rotate.xy)
    .validate_gs_args("singleLogical", "verbose", verbose)
    .validate_gs_args("singleNumber", "mar", mar)
    
    if (mar < 0 || mar > 0.5) {
      warning("'mar' should be in [0, 0.5]", call. = FALSE)
      mar <- max(0, min(mar, 0.5))
    }
    
    nodes <- .get_nodes(gs@graph)
    if(nrow(nodes)>0){
      nodes <- .setCoordToGraph(nodes, flip.x, flip.y, rotate.xy, verbose)
      if(verbose) message("Normalizing node coordinates to graph space...")
      gs@nodes <- .fit_graph_space(nodes, mar)
      gs@pars$is.normalized <- TRUE
      gs@pars$flip.x <- flip.x
      gs@pars$flip.y <- flip.y
      gs@pars$rotate.xy <- rotate.xy
      gs@pars$mar <- mar
    }
    
  return(gs)
    
})

#-------------------------------------------------------------------------------
#' @aliases normalizeGraphSpace
#' @rdname normalizeGraphSpace-methods
#' @export
setMethod("normalizeGraphSpace", 
  signature = c(gs = "GraphSpace", image = "ANY"),
  function(gs, image, ..., mar = 0.1, flip.x = FALSE, flip.y = FALSE, 
    rotate.xy = FALSE, flip.v = FALSE, flip.h = FALSE, verbose = TRUE){
    
    if(is.null(image)) {
      return(normalizeGraphSpace(gs = gs, mar = mar, 
        flip.x = flip.x, flip.y = flip.y, 
        rotate.xy = rotate.xy, verbose = verbose))
    }
    
    extra_args <- list(...)
    if(length(extra_args) > 0 && verbose) {
      warning("Arguments ignored: ", paste(names(extra_args), collapse = ", "))
    }
    
    .validate_gs_args("image_mtx", "image", image)
    .validate_gs_args("singleLogical", "flip.v", flip.v)
    .validate_gs_args("singleLogical", "flip.h", flip.h)
    .validate_gs_args("singleLogical", "flip.x", flip.x)
    .validate_gs_args("singleLogical", "flip.y", flip.y)
    .validate_gs_args("singleLogical", "rotate.xy", rotate.xy)
    .validate_gs_args("singleLogical", "verbose", verbose)
    .validate_gs_args("singleNumber", "mar", mar)
    
    if (mar < 0 || mar > 0.5) {
      warning("'mar' should be in [0, 0.5]", call. = FALSE)
      mar <- max(0, min(mar, 0.5))
    }
    
    nodes <- .get_nodes(gs@graph)
    if(nrow(nodes)>0){
      if(verbose) message("Normalizing node coordinates to image space...")
      image_adj <- image
      if(!is.raster(image_adj)) image_adj <- as.raster(image_adj)
      if(flip.v){
        if(verbose) message("Flipping image top-to-bottom...")  
        image_adj <- image_adj[rev(seq_len(nrow(image_adj))), , drop = FALSE]
      } 
      if(flip.h){
        if(verbose) message("Flipping image left-to-right...")  
        image_adj <- image_adj[, rev(seq_len(ncol(image_adj))), drop = FALSE]
      } 
      
      nodes <- .setCoordToImage(nodes, image_adj, flip.x, flip.y, rotate.xy, verbose)
      
      l_temp <- .fitImageNodes(nodes, image_adj, mar)
      
      gs@image <- l_temp$image
      gs@nodes <- l_temp$nodes
      gs@pars$is.normalized <- TRUE
      gs@pars$image.layer <- TRUE
      gs@pars$flip.v <- flip.v
      gs@pars$flip.x <- flip.x
      gs@pars$flip.y <- flip.y
      gs@pars$rotate.xy <- rotate.xy
      gs@pars$mar <- mar
      gs@misc$image <- image
    }
    
    return(gs)
    
  })

#-------------------------------------------------------------------------------
#' @aliases cropGraphSpace
#' @rdname normalizeGraphSpace-methods
#' @export
setMethod("cropGraphSpace", "GraphSpace", 
  function(gs, crop.coord = c(0, 1, 0, 1), verbose = TRUE){
    .validate_gs_args("numeric_vec", "crop.coord", crop.coord)
    if(length(crop.coord)!=4){
      stop("'crop.coord' should be a numeric vector of length = 4.", 
        call. = FALSE)
    }
    if(any(crop.coord < 0) || any(crop.coord > 1)){
      stop("'crop.coord' should be in [0,1].", call. = FALSE)
    }
    if(!gs@pars$is.normalized){
      stop("'GraphSpace' must be normalized.", call. = FALSE)
    }
    
    gs <- .crop_gspace(gs, crop.coord)
    
    return(gs)
    
  })

################################################################################
### Graph adjusts
################################################################################
.setCoordToGraph <- function(nodes, flip.x = FALSE, flip.y = FALSE, 
  rotate.xy= FALSE, verbose = TRUE){
  
  # Rotated coordinates
  coord <- nodes[,c("x","y")]
  if(rotate.xy){
    if(verbose) message("Rotating xy-coordinates...")
    coord$x2 <- coord$y
    coord$y2 <- coord$x
  } else {
    coord$x2 <- coord$x
    coord$y2 <- coord$y
  }
  
  # Flip y-coordinates
  if(flip.y){
    if(verbose) message("Flipping y-coordinates...")
    y <- coord$y2
    coord$y2 <- -(y - max(y)) - max(y) + 1
  }
  
  # Flip x-coordinates
  if(flip.x){
    if(verbose) message("Flipping x-coordinates...")
    x <- coord$x2
    coord$x2 <- -(x - max(x)) - max(x) + 1
  }
  # Update coordinates
  nodes$x <- coord$x2
  nodes$y <- coord$y2
  
  return(nodes)
}

################################################################################
### Adjust node coordinates
################################################################################
# Fit graph in a [0, 1] space with focus on adjusting margins
.fit_graph_space <- function(nodes, mar = 0.1){
  
  mar <- max(0, min(mar, 0.49))
  
  nds <- nodes
  
  if(nrow(nds)>0){
    
    nds$x <- nds$x - mean(range(nds$x))
    nds$y <- nds$y - mean(range(nds$y))
    
    from <- range(c(nds$x, nds$y))
    to <- c(mar, 1 - mar)
    
    nds$x <- scales::rescale(nds$x, from = from, to=to)
    nds$y <- scales::rescale(nds$y, from = from, to=to)
    
  }
  
  return(nds)
}

################################################################################
### Graph-to-image adjusts
################################################################################
.setCoordToImage <- function(nodes, image, 
  flip.x = FALSE, flip.y = FALSE, rotate.xy = FALSE, 
  verbose = TRUE){
  
  # Rotated coordinates
  coord <- nodes[,c("x","y")]
  if(rotate.xy){
    if(verbose) message("Rotating xy-coordinates...")
    coord$x2 <- coord$y
    coord$y2 <- coord$x
  } else {
    coord$x2 <- coord$x
    coord$y2 <- coord$y
  }
  
  if(flip.y){
    if(verbose) message("Flipping y-coordinates...")
  } else {
    # Flip 'y' by default if an image is provided
    y <- coord$y2
    y <- -(y - max(y)) + nrow(image) - max(y) + 1
    coord$y2 <- y
  }
  
  # Flip x-coordinates over image axis
  if(flip.x){
    if(verbose) message("Flipping x-coordinates...")
    x <- coord$x2
    x <- -(x - max(x)) + ncol(image) - max(x) + 1
    coord$x2 <- x
  }
  
  # Update coordinates
  nodes$x <- coord$x2
  nodes$y <- coord$y2
  
  .check_final_coords(nodes, image)
  
  return(nodes)
  
}

#-------------------------------------------------------------------------------
.check_final_coords <- function(nodes, image){
  
  d <- dim(image)
  xr <- range(nodes$x, na.rm = TRUE)
  yr <- range(nodes$y, na.rm = TRUE)
  
  xr_int <- c(floor(xr[1]), ceiling(xr[2]))
  yr_int <- c(floor(yr[1]), ceiling(yr[2]))
  
  out_x <- (xr_int[1] < 1) || (xr_int[2] > d[2])
  out_y <- (yr_int[1] < 1) || (yr_int[2] > d[1])
  
  if( out_x || out_y ){
    msg <- "Graph coordinates outside the image boundaries."
    
    ms_i <- c("i" = "Note: node 'x' or 'y' coordinates are treated as indices of the image matrix.")
    ms_x <- c("x" = sprintf(
      "Observed ranges: x[%s, %s], y[%s, %s]. Image dimensions: %s x %s (yx).",
      xr_int[1], xr_int[2], yr_int[1], yr_int[2], d[1], d[2]
    ))
    ms_a <- c("*" = "Try adjusting 'flip' and 'rotate' options in `normalizeGraphSpace()`.")
    ms_f <- "See `vignette('RGraphSpace')` for more information on coordinate normalization."
    
    rlang::abort(message = msg, 
      body = c(ms_i, ms_x, ms_a), footer = ms_f, 
      call = rlang::caller_env())
  }
  
  invisible(TRUE)
  
}

################################################################################
### Adjust image to node coordinates
################################################################################

#-------------------------------------------------------------------------------
.fitImageNodes <- function(nodes, image, mar = 0.1){
  
  l_temp <- .fit_image_nodes(nodes, image, mar)
  l_temp <- .adjust_aspect_ratio(l_temp)
  l_temp <- .normalize_image_nodes(l_temp)
  
  return(l_temp)
}

#-------------------------------------------------------------------------------
# Fit image to nodes with focus on adjusting graph margins
.fit_image_nodes <- function(nodes, image, mar = 0.1) {
  
  nds <- nodes
  img <- image
  d <- dim(img)
  mar <- max(0, min(mar, 0.49))
  
  # bounding box around nodes
  xl_nds <- range(nds$x)
  yl_nds <- range(nds$y)
  center_x <- mean(xl_nds)
  center_y <- mean(yl_nds)
  
  # target dimension centered on nodes;
  # side_length is calculated so that max_d is 
  # exactly (1 - 2*mar) of the total width
  max_d <- max(diff(xl_nds), diff(yl_nds))
  side_length <- max_d / (1 - 2 * mar)
  half_side <- side_length / 2
  
  # initial crop coordinates
  x_start <- center_x - half_side
  x_end   <- x_start + side_length
  y_start <- center_y - half_side
  y_end   <- y_start + side_length
  
  # shift crop coordinates to the image boundaries
  if (x_start < 1) { 
    shift <- 1 - x_start
    x_start <- 1
    x_end <- min(d[2], x_end + shift)
  }
  if (x_end > d[2]) { 
    shift <- x_end - d[2]
    x_end <- d[2]
    x_start <- max(1, x_start - shift)
  }
  if (y_start < 1) { 
    shift <- 1 - y_start
    y_start <- 1
    y_end <- min(d[1], y_end + shift)
  }
  if (y_end > d[1]) { 
    shift <- y_end - d[1]
    y_end <- d[1]
    y_start <- max(1, y_start - shift)
  }
  
  # force the limits to include the node bounding box
  x_start <- max(1, min(x_start, xl_nds[1]))
  x_end   <- min(d[2], max(x_end, xl_nds[2]))
  y_start <- max(1, min(y_start, yl_nds[1]))
  y_end   <- min(d[1], max(y_end, yl_nds[2]))
  
  # convert to indices
  x_s_idx <- floor(x_start)
  x_e_idx   <- ceiling(x_end)
  y_s_idx <- floor(y_start)
  y_e_idx   <- ceiling(y_end)
  
  # final validity check
  x_s_idx <- max(1L, x_s_idx)
  x_e_idx   <- min(d[2], x_e_idx)
  y_s_idx <- max(1L, y_s_idx)
  y_e_idx   <- min(d[1], y_e_idx)
  
  # execute crop on flipped image
  img_res <- img[seq.int(d[1], 1), ]
  img_res <- img_res[seq.int(y_s_idx, y_e_idx), seq.int(x_s_idx, x_e_idx)]
  img_res <- img_res[seq.int(nrow(img_res), 1), ]
  
  # update node coordinates
  nds$x <- nds$x - x_s_idx + 1
  nds$y <- nds$y - y_s_idx + 1
  
  # calculate final side_length
  x_side_length <- x_end - x_start
  y_side_length <- y_end - y_start
  x_side_length <- x_side_length + (x_start - x_s_idx)
  y_side_length <- y_side_length + (y_start - y_s_idx)
  side_length <- max(x_side_length, y_side_length)
  
  return(list(nodes = nds, image = img_res, 
    side_length = side_length))
}

#-------------------------------------------------------------------------------
.adjust_aspect_ratio <- function(l_temp){
  d <- dim(l_temp$image)
  if(d[1] > d[2]){
    n <- ceiling( (d[1] - d[2]) )/2
    img_d <- matrix(NA, nrow = d[1], ncol = d[1])
    img_d[ , seq(n + 1, n + d[2])] <- as.matrix(l_temp$image)
    l_temp$nodes$x <- l_temp$nodes$x + n
    l_temp$image  <- as.raster(img_d)
  } else if(d[1] < d[2]){
    n <- ceiling( (d[2] - d[1])/2 )
    img_d <- matrix(NA, nrow = d[2], ncol = d[2])
    img_d[seq(n + 1, n + d[1]), ] <- as.matrix(l_temp$image)
    l_temp$nodes$y <- l_temp$nodes$y + n
    l_temp$image  <- as.raster(img_d)
  }
  return(l_temp)
}

#-------------------------------------------------------------------------------
.normalize_image_nodes <- function(l_temp){
  d <- dim(l_temp$image)
  l_temp$nodes$x <- .rescale_direct(l_temp$nodes$x, d[2], 0.5 / d[2])
  l_temp$nodes$y <- .rescale_direct(l_temp$nodes$y, d[1], 0.5 / d[1])
  return(l_temp)
}
.rescale_direct <- function(x, n, half_pixel) {
  ((x - 1) / (n - 1)) * (1 - 2 * half_pixel) + half_pixel
}

################################################################################
### Crop graph and image
################################################################################
.crop_gspace <- function(gs, crop.coord){
  
  # remove nodes
  nodes <- gs@nodes
  cx <- nodes$x >= crop.coord[1] & nodes$x <= crop.coord[2]
  cy <- nodes$y >= crop.coord[3] & nodes$y <= crop.coord[4]
  nodes <- nodes[ which(cx & cy), ]
  
  # remove edges
  idx <- (gs@edges$name1 %in% gs@nodes$name) & 
    (gs@edges$name2 %in% gs@nodes$name) 
  gs@edges <- gs@edges[idx,]
  
  # center nodes
  nodes$x <- nodes$x - mean(range(nodes$x))
  nodes$y <- nodes$y - mean(range(nodes$y))
  from <- range(c(nodes$x, nodes$y))
  nodes$x <- scales::rescale(nodes$x, from = from, to=c(0,1))
  nodes$y <- scales::rescale(nodes$y, from = from, to=c(0,1))
  
  # update graph
  idx <- V(gs@graph)$name %in% rownames(nodes)
  gs@graph <- igraph::delete_vertices(gs@graph, which(!idx))
  idx <- match(rownames(nodes), V(gs@graph)$name)
  V(gs@graph)$x[idx] <- nodes$x
  V(gs@graph)$y[idx] <- nodes$y
  gs@nodes <- nodes
  
  # crop image
  if(gs@pars$image.layer){
    nrow_mat <- nrow(gs@image)
    ncol_mat <- ncol(gs@image)
    xmin <- max(1, floor(crop.coord[1] * ncol_mat) + 1)
    xmax <- min(ncol_mat, ceiling(crop.coord[2] * ncol_mat))
    ymin <- max(1, floor(crop.coord[3] * nrow_mat) + 1)
    ymax <- min(nrow_mat, ceiling(crop.coord[4] * nrow_mat))
    gs@image <- gs@image[ymin:ymax, xmin:xmax, drop = FALSE]
  }
  
  return(gs)
}

#-------------------------------------------------------------------------------
.validate_gs_args <- function(check, name, para) {
  if (check == "numeric_vec") {
    msg <- paste0("'", name, "' should be a numeric vector.")
    if (!is.vector(para) || !.all_numericValues(para)) stop(msg, call. = FALSE)
  } else if (check == "character_vec") {
    msg <- paste0("'", name, "' should be a character vector.")
    if (!is.vector(para) || !.all_characterValues(para)) 
      stop(msg, call. = FALSE)
  } else if (check == "integer_vec") {
    msg <- paste0("'", name, "' should be an integer vector.")
    if (!is.vector(para) || !.all_integerValues(para)) 
      stop(msg, call. = FALSE)
  } else if (check == "numeric_mtx") {
    msg <- paste0("'", name, "' should be a numeric matrix")
    if (!is.numeric(para) || !is.matrix(para)) stop(msg, call. = FALSE)
  } else if (check == "image_mtx") {
    msg1 <- paste0("Invalid '", name, "' input. Expected a raster object\n")
    msg2 <- c("or a numeric matrix with values in the range [0, 1]")
    if (!is.raster(para)){
      if (!is.matrix(para) || !is.numeric(para)) {
        stop(msg1, msg2, call. = FALSE)
      }
      rg <- range(para, na.rm = TRUE)
      if (rg[1] < 0 || rg[2] > 1) {
        stop(msg1, msg2, call. = FALSE)
      }
    }
  } else if (check == "allCharacter") {
    msg <- paste0("'", name, "' should be a vector of strings.")
    if (!.all_characterValues(para)) stop(msg, call. = FALSE)
  } else if (check == "allCharacterOrInteger") {
    msg <- paste("'", name, "'should be a vector of strings of integers.")
    if (! (.all_characterValues(para) | .all_integerValues(para) ) ) 
      stop(msg, call. = FALSE)
  } else if (check == "allCharacterOrNa") {
    msg <- paste0("'", name, "' should be a vector of strings.")
    if (!.all_characterValues(para, notNA=FALSE)) stop(msg, call. = FALSE)
  } else if (check == "allBinary") {
    msg <- paste0("'", name, "' should be a vector of binary values.")
    if (!.all_binaryValues(para)) stop(msg, call. = FALSE)
  } else if (check == "allInteger") {
    msg <- paste0("'", name, "' should be a vector of integer values.")
    if (!.all_integerValues(para)) stop(msg, call. = FALSE)
  } else if (check == "singleString") {
    msg <- paste0("'", name, "' should be a single string.")
    if (!.is_singleString(para)) stop(msg, call. = FALSE)
  } else if (check == "singleInteger") {
    msg <- paste0("'", name, "' should be a single integer value.")
    if (!.is_singleInteger(para)) stop(msg, call. = FALSE)
  } else if (check == "singleNumber") {
    msg <- paste0("'", name, "' should be a single numeric value.")
    if (!.is_singleNumber(para)) stop(msg, call. = FALSE)
  } else if (check == "function") {
    msg <- paste0("'", name, "' should be a function.")
    if (!is.function(para)) stop(msg, call. = FALSE)
  } else if (check == "singleLogical") {
    msg <- paste0("'", name, "' should be a single logical value.")
    if (!.is_singleLogical(para)) stop(msg, call. = FALSE)
  } else {
    warning("Skipped arg validation.", call. = FALSE)
  }
}

#-------------------------------------------------------------------------------
.validate_gs_colors <- function(check, name, para) {
  if (check == "singleColor") {
    if (!.is_singleColor(para)) {
      msg <- paste0("'", name, "' should be a single color.")
      stop(msg, call. = FALSE)
    }
  } else if (check == "allColors") {
    if (!.is_color(para)) {
      msg <- paste0("'", name, "' should be a vector with colors.")
      stop(msg, call. = FALSE)
    }
  } else {
    warning("Skipped color validation.", call. = FALSE)
  }
}

#-------------------------------------------------------------------
.is_singleNumber <- function(para) {
  (is.integer(para) || is.numeric(para)) &&
    length(para) == 1L && !is.na(para)
}
.is_singleInteger <- function(para) {
  lg <- (is.integer(para) || is.numeric(para)) &&
    length(para) == 1L && !is.na(para)
  if (lg) {
    para <- abs(para)
    lg <- abs(para - round(para)) <= para
  }
  return(lg)
}
.is_singleString <- function(para) {
  is.character(para) && length(para) == 1L && !is.na(para)
}
.is_singleLogical <- function(para) {
  is.logical(para) && length(para) == 1L && !is.na(para)
}
.all_binaryValues <- function(para) {
  all(para %in% c(0, 1, NA))
}
.all_integerValues <- function(para, notNA = TRUE) {
  lg <- is.integer(para) || is.numeric(para) || all(is.na(para))
  if (lg) {
    para <- abs(para)
    lg <- all(abs(para - round(para)) <= para, na.rm=TRUE)
  }
  if(lg && notNA) lg <- !any(is.na(para))
  return(lg)
}
.all_numericValues <- function(para, notNA = TRUE) {
  lg <- is.numeric(para) || all(is.na(para))
  if(lg && notNA) lg <- !any(is.na(para))
  return(lg)
}
.all_characterValues <- function(para, notNA = TRUE) {
  lg <- is.character(para) || all(is.na(para))
  if(lg && notNA) lg <- !any(is.na(para))
  return(lg)
}
.is_numericVector <- function(para){
  is.vector(para) && .all_numericValues(para)
}
.is_integerVector <- function(para){
  is.vector(para) && .all_integerValues(para)
}
.is_characterVector <- function(para){
  is.vector(para) && .all_characterValues(para)
}
.is_color <- function(x) {
  if (is.numeric(x)) {
    if (any(x < 1 | x %% 1 != 0, na.rm = TRUE)) {
      return(FALSE)
    }
  }
  res <- try(col2rgb(x), silent = TRUE)
  return( !inherits(res, "try-error") )
}
.is_singleColor <- function(para) {
  .is_color(para) && length(para) == 1L && !is.na(para)
}

setOldClass("igraph")

setGeneric("GraphSpace", function(g, ...) standardGeneric("GraphSpace"))

setGeneric("plotGraphSpace", function(gs, ...)
  standardGeneric("plotGraphSpace"),
  package = "RGraphSpace"
)

setGeneric("getGraphSpace", function(gs, what = "summary") 
  standardGeneric("getGraphSpace"), 
  package = "RGraphSpace"
)

setGeneric("normalizeGraphSpace", function(gs, image = NULL, ...) 
  standardGeneric("normalizeGraphSpace"),
  signature = c("gs", "image"),
  package = "RGraphSpace"
)

setGeneric("cropGraphSpace", function(gs, ...) 
  standardGeneric("cropGraphSpace"),
  package = "RGraphSpace"
)

setGeneric("gs_nodes", function(x) 
  standardGeneric("gs_nodes"),
  package = "RGraphSpace"
)

setGeneric("gs_edges", function(x) 
  standardGeneric("gs_edges"),
  package = "RGraphSpace"
)

setGeneric("gs_vcount", function(x) 
  standardGeneric("gs_vcount"),
  package = "RGraphSpace"
)

setGeneric("gs_ecount", function(x) 
  standardGeneric("gs_ecount"),
  package = "RGraphSpace"
)

setGeneric("gs_vertex_attr", function(x, ...) 
  standardGeneric("gs_vertex_attr"),
  package = "RGraphSpace"
)

setGeneric("gs_vertex_attr<-", function(x, ..., value) 
  standardGeneric("gs_vertex_attr<-"),
  package = "RGraphSpace"
)

setGeneric("gs_edge_attr", function(x, ...) 
  standardGeneric("gs_edge_attr"),
  package = "RGraphSpace"
)

setGeneric("gs_edge_attr<-", function(x, ..., value) 
  standardGeneric("gs_edge_attr<-"),
  package = "RGraphSpace"
)

#-------------------------------------------------------------------------------
#' @title Draw node and edge elements in a 2D graph layout
#' 
#' @description
#' 
#' Constructor for \link{GeomGraphSpace} ggproto objects.
#' 
#' A wrapper around \link[ggplot2]{geom_point} that enables direct use of
#' node attributes stored in \link{GraphSpace} objects as aesthetics.
#'
#' This geom is designed to map node-level attributes (e.g., \code{fill},
#' \code{size}) or any aesthetics supported by \link[ggplot2]{GeomPoint}.
#'
#' @param mapping Set of aesthetic mappings created by [ggplot2::aes()].
#' These mappings override global aesthetics and are not inherited 
#' from the top-level plot.
#'
#' @param data A \link{GraphSpace} object.
#'
#' @param stat The statistical transformation to use on the data.
#' Defaults to \code{identity}.
#'
#' @param position Position adjustment, either as a string or
#' the result of a call to a position adjustment function.
#'
#' @param ... Additional parameters passed to the underlying
#' drawing function in \link{GeomGraphSpace}.
#'
#' @param na.rm Logical. Should missing values be removed?
#' Defaults to \code{FALSE}.
#' 
#' @param show.legend Logical or a named logical vector indicating
#' whether this layer should be included in legends.
#'
#' @param inherit.aes Logical. If \code{FALSE} (default), the layer will use 
#' aesthetics defined in \code{mapping}.
#' 
#' @param arrow_size Numeric scaling factor controlling arrowhead 
#' geometry (see 'drawing' section).
#' 
#' @param arrow_offset Numeric value controlling the base offset of arrows  
#' at edge endpoints (see 'drawing' section).
#' 
#' @param raster Logical. Should node glyphs be rasterized? 
#' Rasterization support is based on \code{\link[ggrastr]{rasterise}}.
#' 
#' @param dpi Numeric. Rasterization resolution.
#' 
#' @param dev Character. Rasterization backend. One of `"cairo"`,
#' `"ragg"`, `"ragg_png"`, or `"cairo_png"`.
#' 
#' @param scale Numeric. Rasterization scaling factor
#' (see \code{\link[ggrastr]{rasterise}}).
#' 
#' @return A ggplot2 layer that renders node glyphs defined by
#' \link{GeomGraphSpace}.
#'
#' @section Aesthetics for node drawing:
#' 
#' Nodes are drawn in the main layer of \code{geom_graphspace()}, which 
#' understands \link[ggplot2]{geom_point} aesthetics.
#' 
#' If these aesthetics are not explicitly provided in \code{aes()}, they 
#' are automatically retrieved from the \link{GraphSpace} object.
#' 
#' \tabular{ll}{
#'   \strong{\code{x}, \code{y}, \code{vertex}} \tab  
#'   Required (automatically supplied).\cr
#'   \code{fill} \tab Node interior colour (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
#'   \code{colour} \tab Node border colour (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
#'   \code{alpha} \tab Transparency (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
#'   \code{shape} \tab Node shape (see \link{points} and \link[ggplot2]{aes_linetype_size_shape}).\cr
#'   \code{size} \tab Node size (see *drawing* section and \link[ggplot2]{aes_linetype_size_shape}).\cr
#'   \code{stroke} \tab Node line width (see \link[ggplot2]{gg_par} and \link[ggplot2]{aes_linetype_size_shape}).
#' }
#' 
#' Required aesthetics \code{x}, \code{y}, and \code{vertex} are supplied from
#' the \link{GraphSpace} object and do not need to be manually mapped.
#' 
#' Additional parameters can be passed to control fixed values for the layer.
#' For example: `fill = "red"`, `stroke = 3`, `alpha = 0.5`, or `shape = 21`.
#' 
#' The interpretation of \strong{\code{size}} depends on how it is provided:
#' \itemize{
#'   \item **As an aesthetic**: When mapped within \code{aes()}, \code{size}
#'   follows the behavior of \link[ggplot2]{geom_point}, using absolute
#'   units to ensure consistency with the plot legends.
#'   \item **As a parameter**: When set outside \code{aes()}, \code{size} is 
#'   treated as a percentage of the viewport (\code{[0, 100]}), scaling 
#'   in \code{npc} units. This allows nodes to resize dynamically with 
#'   viewport changes.
#' }
#' 
#' @section Edge context-aware parameters:
#' 
#' These parameters control the edge appearance. If not explicitly provided, 
#' they are automatically retrieved from the \link{GraphSpace} object. 
#' They can be a single value or a vector matching the number of edges:
#' 
#' \tabular{ll}{
#'   \code{edge_colour} \tab Node border colour.\cr
#'   \code{edge_linetype} \tab Edge line type.\cr
#'   \code{edge_linewidth} \tab Edge line width.\cr
#'    \code{edge_alpha} \tab Edge transparency.
#' }
#' 
#' @section Edge global parameters:
#' 
#' These parameters apply globally to all edges in the layer:
#' 
#' \tabular{ll}{
#'    \code{arrow_size} \tab Arrow scaling factor (default = 1).\cr
#'    \code{arrow_offset} \tab Arrow offset from nodes (default = 0.01).\cr
#'    \code{arrow_lineend} \tab Line end style (see \link[grid]{gpar}).\cr
#'    \code{arrow_linejoin} \tab Line join style (see \link[grid]{gpar}).
#' }
#' 
#' **arrow_size** is a numeric scaling factor controlling arrowhead geometry. 
#' The value is interpreted in the same numeric space as line width (\code{lwd}), 
#' ensuring consistent scaling between edge strokes and arrowheads.
#' 
#' **arrow_offset** is an additive term that offsets arrow endpoints 
#' uniformly in graph space and is bounded by the edge length, in NPC units.
#' 
#' Arrowhead types are specified in the \link{GraphSpace} constructor.
#' 
#' @seealso
#' \link{GraphSpace}, \link{geom_nodespace}, \link{geom_edgespace}, 
#' \link[ggplot2]{geom_point}
#'
#' @examples
#' library(RGraphSpace)
#' library(igraph)
#' library(ggplot2)
#' 
#' # Make a demo igraph
#' gtoy1 <- make_star(15, mode="out")
#' 
#' # Set some node attributes
#' V(gtoy1)$nodeSize <- runif(vcount(gtoy1), 1, 20)
#' V(gtoy1)$nodeColor <- rainbow(vcount(gtoy1))
#' 
#' # Set some variables
#' V(gtoy1)$user_var1 <- runif(vcount(gtoy1), 1, 3)^3
#' V(gtoy1)$user_var2 <-  rep(c(1, 2, 3), each = 5)
#' 
#' # Create a GraphSpace object
#' gs <- GraphSpace(gtoy1, layout = layout_in_circle(gtoy1))
#' 
#' \dontrun{
#' 
#' # Example 1: Nodes scaling with the legend
#' # When 'size' is mapped inside aes(), it follows
#' # ggplot2 default behavior: size is translated 
#' # to absolute units (mm) via 'scale_size()'.
#' 
#' ggplot() + 
#'   geom_graphspace(
#'   mapping = aes(size = nodeSize, fill = user_var2), 
#'   data = gs, arrow_offset = 0.01) + 
#'   scale_size(range = c(1, 12)) + 
#'   theme(aspect.ratio = 1)
#'   
#' # Example 2: Nodes scaling with the viewport
#' # When 'size' is passed as a node attribute, 
#' # inherited from the igraph object, it is
#' # interpreted as a percentage of the plotting 
#' # area and translated to NPC units.
#' 
#' ggplot() +
#'   geom_graphspace(mapping = aes(fill = user_var2), 
#'   data = gs, arrow_offset = 0.01) + 
#'   theme(aspect.ratio = 1)
#'   
#' }
#' 
#' @export
geom_graphspace <- function(mapping = NULL, data, 
  stat = "identity", position = "identity", ...,
  na.rm = FALSE, show.legend = NA, inherit.aes = FALSE,
  arrow_size = 1, arrow_offset = 0.01,
  raster = FALSE, dpi = NULL, dev = "cairo", scale = 1) {
  
  if (missing(data) || is.null(data)){
    rlang::warn(
      message = c(
        "!" = "`geom_graphspace()` ignored: explicit `data` object is required.",
        "i" = "Accepted: 'GraphSpace', 'igraph', 'tidygraph', or 'ggraph' layout.",
        "*" = "For inherited data, use `geom_nodespace()` and `geom_edgespace()` instead."
      )
    )
    return(ggplot2::geom_blank())
  }
  
  mapping <- .mapping_graphspace(mapping)
  
  params <- rlang::list2(
    na.rm = na.rm,
    arrow_size = arrow_size,
    arrow_offset = arrow_offset,
    raster = raster, 
    dpi = dpi, 
    dev = dev, 
    scale = scale,
    ...)
  
  data <- .graphspace_handler(data)
  edges <- gs_edges(data)
  data <- gs_nodes(data, vars = .detect_mapping_vars(mapping))
  params <- .params_graphspace(params, mapping, data, edges)
  
  ggplot2::layer(
    geom = GeomGraphSpace,
    mapping = mapping,
    data = data,
    stat = stat,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = params
  )
  
}

#-------------------------------------------------------------------------------
.graphspace_handler <- function(data) {
  if ( inherits(data, c("igraph", "layout_ggraph")) ) {
    data <- GraphSpace(data, verbose = FALSE)
  } else if (!inherits(data, "GraphSpace")) {
    rlang::abort(
      message = c(
        "x" = "Unsupported `data` type in `geom_graphspace()`.",
        "i" = "Accepted: 'GraphSpace', 'igraph', 'tidygraph', or 'ggraph' layout.",
        "*" = "For inherited data, use `geom_nodespace()` and `geom_edgespace()` instead."
      )
    )
  }
  return(data)
}

#-------------------------------------------------------------------------------
.mapping_graphspace <- function(mapping) {
  x <- y <- vertex <- NULL
  default_mapping <- ggplot2::aes(x = x, y = y, vertex = vertex)
  if (is.null(mapping)) {
    mapping <- default_mapping
  } else {
    mapping <- utils::modifyList(default_mapping, mapping)
  }
  return(mapping)
}

#-------------------------------------------------------------------------------
.params_graphspace <- function(params, mapping, nodes, edges){
  
  params$.size_unit <- if("size" %in% names(mapping)) "mm" else "npc"
  params$.edges <- edges
  
  #--- nodes
  
  if(is.null(params[["size"]]) && is.null(mapping[["size"]])){
    if("nodeSize" %in% names(nodes) ){
      params[["size"]] <- nodes[["nodeSize"]]
    }
  }
  
  if(is.null(params[["stroke"]]) && is.null(mapping[["stroke"]])){
    if("nodeLineWidth" %in% names(nodes) ){
      params[["stroke"]] <- nodes[["nodeLineWidth"]]
    }
  }
  
  if(is.null(params[["shape"]]) && is.null(mapping[["shape"]])){
    if("nodeShape" %in% names(nodes) ){
      params[["shape"]] <- nodes[["nodeShape"]]
    }
  }
  
  if(is.null(params[["fill"]]) && is.null(mapping[["fill"]])){
    if("nodeColor" %in% names(nodes) ){
      params[["fill"]] <- nodes[["nodeColor"]]
    }
  }
  
  if(is.null(params[["colour"]]) && is.null(mapping[["colour"]])){
    if("nodeLineColor" %in% names(nodes) ){
      params[["colour"]] <- nodes[["nodeLineColor"]]
    }
  }
  
  if(is.null(params[["alpha"]]) && is.null(mapping[["alpha"]])){
    if("nodeAlpha" %in% names(nodes) ){
      params[["alpha"]] <- nodes[["nodeAlpha"]]
    }
  }
  
  #--- edges
  
  if(is.null(params[["edge_colour"]])){
    if("edgeLineColor" %in% names(edges) ){
      params[["edge_colour"]] <- edges[["edgeLineColor"]]
    }
  }
  
  if(is.null(params[["edge_linewidth"]])){
    if("edgeLineWidth" %in% names(edges) ){
      params[["edge_linewidth"]] <- edges[["edgeLineWidth"]]
    }
  }
  
  if(is.null(params[["edge_linetype"]])){
    if("edgeLineType" %in% names(edges) ){
      params[["edge_linetype"]] <- edges[["edgeLineType"]]
    }
  }
  
  if(is.null(params[["edge_alpha"]])){
    if("edgeAlpha" %in% names(edges) ){
      params[["edge_alpha"]] <- edges[["edgeAlpha"]]
    }
  }
  
  params
}

#-------------------------------------------------------------------------------
#' @title GeomGraphSpace: a ggplot2 prototype for GraphSpace-class methods
#'
#' @description
#' 
#' \code{GeomGraphSpace} is the underlying \link[ggplot2]{ggproto} object 
#' used by \link{geom_graphspace} to draw node and edge elements in a 
#' graph layout. 
#'
#' This geom is designed for network diagrams, where graph attributes 
#' are often already in their final form (e.g., hex colors).
#'
#' @section Aesthetics:
#'
#' \code{GeomGraphSpace} understands ggplot2's conventions for point-like geoms.
#' 
#' @seealso
#' \link{geom_graphspace}, \link[ggplot2]{geom_point}
#'
#' @export
GeomGraphSpace <- ggproto(
  
  "GeomGraphSpace", ggplot2::Geom, 
  
  required_aes = c("x", "y", "vertex"),
    
  non_missing_aes = c("size", "stroke", "shape", "colour"),
  
  default_aes = aes(
    size = 5,
    stroke = 0.5,
    shape = 21,
    colour = "grey20",
    fill = "#E5E5E5B3",
    alpha = NA
  ),
  
  draw_panel = function(self, data, panel_params, coord, 
    edge_colour = "grey80", edge_alpha = NA, edge_linewidth = 0.5, 
    edge_linetype = "solid", arrow_size = 1, arrow_offset = 0.01, 
    arrow_lineend = "butt", arrow_linejoin = "mitre", na.rm = FALSE, 
    raster = FALSE, dpi = NULL, dev = "cairo", scale = 1, 
    .size_unit = "mm", .edges = NULL) {
    
    data$shape <- translate_shape_string(data$shape)
    
    data <- .geom_check_node_size(data, size_unit = .size_unit)
    
    coords <- coord$transform(data, panel_params)
    
    # Create node grobs
    node_grobs <- .get_node_grobs(coords, size_unit = .size_unit)
    node_grobs$name <- grobName(node_grobs, "nodes")
    
    if(.empty(.edges)){
      
      edge_grobs <- zeroGrob()
      
    } else {
      
      .edges$colour <- edge_colour %||% "grey80"
      .edges$linewidth <- edge_linewidth %||% 0.5
      .edges$linetype <- edge_linetype %||% "solid"
      
      .edges$alpha <- edge_alpha %||% NA
      .edges$arrow_size <- (arrow_size %||% 1)
      .edges$arrow_offset <- arrow_offset %||% 0
      
      .edges <- remove_missing(.edges, na.rm = na.rm,
        vars = c("vertex1", "vertex2", "arrowType"), 
        name = "geom_graphspace-edges")
      
      .edges <- .geom_remap_edge_coords(edges = .edges, nodes = coords)
      
      .edges <- .geom_remap_edge_offsets(edges = .edges, nodes = coords,
        size_unit = .size_unit)
      
      .edges <- .geom_adj_arrow_offsets(.edges)
      
      .edges <- .geom_adj_arrow_size(.edges, size_unit = .size_unit)
      
      .edges <- remove_missing(.edges, na.rm = na.rm,
        vars = c("x", "y", "xend", "yend"), 
        name = "geom_graphspace-coords")
      
      # Create edge grobs
      edge_grobs <- .get_edge_grobs(.edges, lineend = arrow_lineend, 
        linejoin = arrow_linejoin, size_unit = .size_unit)
      
    }
    
    graph_grob <- grid::gTree(
      children = grid::gList(edge_grobs, node_grobs),
      name = grid::grobName(prefix = "geom_graphspace")
    )
    
    if (raster) {
      graph_grob <- .as_rasteriser(graph_grob, 
        dpi = dpi, dev = dev, scale = scale)
    }
    
    graph_grob
    
  },
  
  draw_key = draw_key_point
  
)

#-------------------------------------------------------------------------------
# Remap x, y, xend, and yend to updated coords
.geom_remap_edge_coords <- function(edges, nodes){
  
  if(.empty(edges)){
    return( edges )
  }
  
  # remap segments
  idx <- match(edges[["vertex1"]], nodes[["vertex"]])
  edges[["x"]] <- nodes[["x"]][idx]
  edges[["y"]] <- nodes[["y"]][idx]
  idx <- match(edges[["vertex2"]], nodes[["vertex"]])
  edges[["xend"]] <- nodes[["x"]][idx]
  edges[["yend"]] <- nodes[["y"]][idx]
  
  return(edges)
  
}

#-------------------------------------------------------------------------------
#' @title Draw edge elements in a 2D graph layout
#' 
#' @description
#' 
#' Constructor for \link{GeomEdgeSpace} ggproto objects.
#' 
#' A wrapper around \link[ggplot2]{geom_segment} that enables direct use of
#' edge attributes stored in \link{GraphSpace} objects as aesthetics.
#'
#' This \code{geom} is designed to create edge-level aesthetics such as
#' \code{colour} and \code{linewidth}, or any custom aesthetics defined 
#' in \link{GeomEdgeSpace}.
#'
#' @param mapping Set of aesthetic mappings created by [ggplot2::aes()].
#' These mappings override global aesthetics and are not inherited 
#' from the top-level plot.
#'
#' @param data A \link{GraphSpace} object.
#'
#' @param stat The statistical transformation to use on the data.
#' Defaults to \code{identity}.
#'
#' @param position Position adjustment, either as a string or
#' the result of a call to a position adjustment function.
#'
#' @param ... Additional parameters passed to the underlying
#' drawing function in \link{GeomEdgeSpace}.
#'
#' @param na.rm Logical. Should missing values be removed?
#' Defaults to \code{FALSE}.
#' 
#' @param show.legend Logical or a named logical vector indicating
#' whether this layer should be included in legends.
#'
#' @param inherit.aes Logical. If \code{FALSE} (default), the layer will use 
#' aesthetics defined in \code{mapping}.
#' 
#' @param size_aes Optional node size mapping, created with [ggplot2::aes()]. 
#' Can be used to synchronize \link{geom_nodespace} and \link{geom_edgespace} 
#' to ensure directed edges and arrows are offset from node boundaries. 
#' This prevents overlaps when \link{geom_nodespace} includes a dynamic 
#' `size` or `stroke` mappings. For automated scale detection, use
#' \link{inject_nodespace} instead.
#' 
#' @param range Optional numeric vector of length 2 (default `c(1, 6)`) 
#' specifying the min/max node diameters in mm. Used with `size_aes` and 
#' should match the `range` in [ggplot2::scale_size_continuous()] for 
#' accurate edge clipping. For automated scale detection, use
#' \link{inject_nodespace} instead.
#' 
#' @param arrow_size Numeric scaling factor controlling arrowhead 
#' geometry (see 'drawing' section).
#' 
#' @param arrow_offset Numeric value controlling the base offset of arrows  
#' at edge endpoints (see 'drawing' section).
#'
#' @param lineend Line end style (round, butt, square). Supplied for 
#' compatibility with \link[ggplot2]{geom_segment}.
#' 
#' @param linejoin Line join style (round, mitre, bevel). Supplied for 
#' compatibility with \link[ggplot2]{geom_segment}.
#' 
#' @return A ggplot2 layer that renders edge segments defined by
#' \link{GeomEdgeSpace}.
#' 
#' @section Aesthetics:
#' 
#' \code{geom_edgespace()} understands \link[ggplot2]{geom_segment} aesthetics.
#' 
#' If these aesthetics are not explicitly provided in \code{aes()}, they 
#' are automatically retrieved from the \link{GraphSpace} object.
#'
#' \tabular{ll}{
#'   \strong{\code{x}, \code{y}, \code{xend}, \code{yend}} \tab Required (automatically supplied).\cr
#'   \code{colour} \tab Node border colour (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
#'   \code{alpha} \tab Transparency (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
#'   \code{linetype} \tab Edge line type (see \link[ggplot2]{aes_linetype_size_shape}).\cr
#'   \code{linewidth} \tab Edge line width (see \link[ggplot2]{aes_linetype_size_shape}).
#' }
#' 
#' Required aesthetics (\code{x}, \code{y}, \code{xend}, \code{yend}, ...)  
#' are supplied from the \link{GraphSpace} object and do not need to be 
#' manually mapped.
#' 
#' Additional parameters can be passed to control fixed values for the layer.
#' For example: `colour = "grey"`, `linetype = 2`, `linewidth = 1`.
#' 
#' Arrows can be further adjusted by \code{arrow_size} and \code{arrow_offset} 
#' arguments (see *details*).
#' 
#' @details
#' 
#' **arrow_size** is a numeric scaling factor controlling arrowhead geometry. 
#' The value is interpreted in the same numeric space as line width (`lwd`), 
#' ensuring consistent scaling between edge strokes and arrowheads.
#' 
#' **arrow_offset** is an additive term that offsets arrow endpoints 
#' uniformly in graph space and is bounded by the edge length, in NPC units.
#' 
#' Arrowhead types are specified in the \link{GraphSpace} constructor.
#' 
#' @seealso
#' \link{GraphSpace}, \link{geom_nodespace}, \link{geom_graphspace}, 
#' \link[ggplot2]{geom_segment}
#'
#' @examples
#' 
#' # Load a demo igraph
#' data('gtoy1', package = 'RGraphSpace')
#' 
#' # Create a GraphSpace object
#' gs <- GraphSpace(gtoy1)
#' 
#' \dontrun{
#' 
#' ggplot() +
#'   geom_edgespace(data = gs) +
#'   geom_nodespace(data = gs) +
#'   theme(aspect.ratio = 1)
#' 
#' }
#' 
#' @export
geom_edgespace <- function(mapping = NULL, data = NULL,
  stat = "identity", position = "identity", ..., 
  na.rm = FALSE, show.legend = NA, inherit.aes = FALSE,
  size_aes = NULL, range = c(1, 6), 
  arrow_size = 1, arrow_offset = 0.01,
  lineend = "butt", linejoin = "mitre") {
  
  if (inherits(data, "GraphSpace")) {
    .geom_check_slots(data)
    nodes <- gs_nodes(data)
    edges <- gs_edges(data)
  } else {
    stop("'data' must be a 'GraphSpace' object.")
  }
  
  # Check custom params
  .validate_gs_args("singleLogical", "na.rm", na.rm)
  .validate_gs_args("singleNumber", "arrow_size", arrow_size)
  .validate_gs_args("singleNumber", "arrow_offset", arrow_offset)
  .validate_gs_args("singleString", "lineend", lineend)
  .validate_gs_args("singleString", "linejoin", linejoin)
  if (!is.null(size_aes) && !inherits(size_aes, "uneval")) {
    stop("Argument 'size_aes' must be a mapping created with aes().", 
      call. = FALSE)
  }
  if (!is.numeric(range) || length(range) != 2) {
    stop("Argument 'range' must be a numeric vector of length 2.", 
      call. = FALSE)
  }
  
  params <- list2(
    na.rm = na.rm, 
    arrow_size = arrow_size,
    arrow_offset = arrow_offset,
    lineend = lineend,
    linejoin = linejoin,
    size_aes = size_aes,
    range = range,
    .nodes = nodes,
    .size_unit = "mm",
    .size_rule = NULL,
    ...)
  
  mapping <- .mapping_edgespace(mapping)
  
  params <- .params_edgespace(params, mapping, edges)
  
  ggplot2::layer(
    geom = GeomEdgeSpace,
    stat = stat,
    mapping = mapping,
    data = edges,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = params
  )
  
}

#-------------------------------------------------------------------------------
.mapping_edgespace <- function(mapping) {
  x <- y <- xend <- yend <- arrowType <- NULL
  vertex1 <- vertex2 <- NULL
  default_mapping <- ggplot2::aes(
    x = x, y = y, xend = xend, yend = yend,
    arrowType = arrowType, 
    vertex1 = vertex1, vertex2 = vertex2
  )
  if (is.null(mapping)) {
    mapping <- default_mapping
  } else {
    mapping <- utils::modifyList(default_mapping, mapping)
  }
  return(mapping)
}

#-------------------------------------------------------------------------------
.params_edgespace <- function(params, mapping, edges){
  
  if(is.null(params[["colour"]]) && is.null(mapping[["colour"]])){
    if("edgeLineColor" %in% names(edges) ){
      params[["colour"]] <- edges[["edgeLineColor"]]
    }
  }
  
  if(is.null(params[["linewidth"]]) && is.null(mapping[["linewidth"]])){
    if("edgeLineWidth" %in% names(edges) ){
      params[["linewidth"]] <- edges[["edgeLineWidth"]]
    }
  }
  
  if(is.null(params[["linetype"]]) && is.null(mapping[["linetype"]])){
    if("edgeLineType" %in% names(edges) ){
      params[["linetype"]] <- edges[["edgeLineType"]]
    }
  }
  
  if(is.null(params[["alpha"]]) && is.null(mapping[["alpha"]])){
    if("edgeAlpha" %in% names(edges) ){
      params[["alpha"]] <- edges[["edgeAlpha"]]
    }
  }
  
  params
}


#-------------------------------------------------------------------------------
#' @title GeomEdgeSpace: a ggplot2 prototype for GraphSpace-class methods
#'
#' @description
#' 
#' \code{GeomEdgeSpace} is the underlying \link[ggplot2]{ggproto} object 
#' used by \link{geom_edgespace} to draw edge elements in a graph layout. 
#'
#' This geom is designed for network diagrams, where graph attributes 
#' are often already in their final form (e.g., hex colors).
#' 
#' @section Aesthetics:
#'
#' \code{GeomEdgeSpace} understands ggplot2's conventions for segment-like geoms.
#' 
#' @seealso
#' \link{geom_edgespace}, \link[ggplot2]{geom_segment}
#'
#' @export
GeomEdgeSpace <- ggproto(
  
  "GeomEdgeSpace", ggplot2::GeomSegment,
  
  required_aes = c("x", "y", "xend", "yend", 
    "vertex1", "vertex2", "arrowType"),
  
  non_missing_aes = c("linewidth", "linetype", "colour"),
  
  setup_params = function(data, params){
    
    nodes <- .geom_map_size_aes(params$.nodes, size_aes = params$size_aes)
    
    if(is.null(params$.size_rule)){
      size_range <- params$range
    } else {
      size_range <- params$.size_rule$palette(c(0, 1))
    }
    
    params$.nodes <- .geom_check_node_size(nodes, 
      size_unit = params$.size_unit, range = size_range)
    
    params
  },
  
  default_aes = ggplot2::aes(
    linewidth = 0.5,
    linetype = "solid",
    colour = "grey80",
    alpha = NA,
    offset_start = 0,
    offset_end = 0
  ),
  
  draw_panel = function(self, data, panel_params, coord, .nodes,   
    arrow_size = 1, arrow_offset = 0.01, lineend = "butt", 
    linejoin = "mitre", na.rm = FALSE, size_aes = NULL, 
    range = c(1, 6), .size_unit = "mm", .size_rule = NULL) {
    
    data <- .geom_get_edge_offsets(data, .nodes, size_unit = .size_unit)
    
    data$arrow_size <- (arrow_size %||% 1)
    data$arrow_offset <- arrow_offset %||% 0
    
    coords <- coord$transform(data, panel_params)
    
    coords <- .geom_adj_arrow_offsets(coords)
    
    coords <- .geom_adj_arrow_size(coords, size_unit = .size_unit)
    
    # Create edge grobs
    grobs <- .get_edge_grobs(coords, lineend = lineend, 
      linejoin = linejoin, size_unit = .size_unit)
    
    grid::gTree(children = grobs,
      name = grid::grobName(prefix = "geom_edgespace")
    )
    
  },
  draw_key = draw_key_path
)

#-------------------------------------------------------------------------------
.geom_map_size_aes <- function(nodes, size_aes){
  
  required_att <- c("x", "y", "vertex", "size", "stroke")
  eval_att <- c(size = "nodeSize", stroke = "nodeLineWidth")
  default_att <- .get_default_vatt()
  
  if (!is.null(size_aes) && !inherits(size_aes, "uneval")) {
    size_aes <- NULL
    warning("Argument 'size_aes' must be a mapping created with aes().", 
      call. = FALSE)
  }
  
  for(att in names(eval_att)){
    if (.is_valid_aes(att, size_aes, nodes)) {
      nodes[[att]] <- rlang::eval_tidy(size_aes[[att]], nodes)
    } else {
      if( !(att %in% colnames(nodes)) ){
        nodes[[att]] <- nodes[[eval_att[att]]] %||% default_att[[eval_att[att]]]
      }
      if(!is.numeric(nodes[[att]])){
        nodes[[att]] <- default_att[[eval_att[att]]]
      }
    }
  }
  nodes <- nodes[, intersect(required_att, colnames(nodes)), drop = FALSE]
  
  return(nodes)
  
}

#-------------------------------------------------------------------------------
.is_valid_aes <- function(att, mapping, data_df) {
  if (is.null(mapping)) return(FALSE)
  expr <- mapping[[att]]
  if (is.null(expr)) return(FALSE)
  vars <- all.vars(expr)
  if (length(vars) != 1) return(FALSE)
  vars %in% names(data_df)
}

#-------------------------------------------------------------------------------
.geom_get_edge_offsets <- function(edges, nodes, size_unit){
  
  # --- Calculate Edge Offset ---
  # The offset is the effective node radius: (size/2) + (stroke/2).
  # 1. 'size' is the node diameter in points (mm * .pt). We use half for the radius.
  # 2. 'stroke' is approx. 0.75 mm per unit (see 'aes_linetype_size_shape');
  #    it's pre-processed by gg_par() as (stroke * .stroke / 2).
  # 3. Since the border sits half-in/half-out of the node's edge, we rectify 
  #    'stroke' to find the actual external thickness added to the radius.
  # 4. Final sum is converted to 'npc' for grid coordinate alignment.
  mm2npc <- grid::convertWidth(unit(1, "mm"), unitTo = "npc", valueOnly = T)
  if(size_unit=="mm"){
    # input 'size' and 'stroke' in 'mm', scaled to 'npc' by mm2npc
    offsets <- (nodes[["size"]]/2 + nodes[["stroke"]]/2 * 0.75) * mm2npc
  } else {
    # input 'size' in [0, 100], transformed to 'npc' by 0.01
    offsets <- (nodes[["size"]]/2 * 0.01) + ( nodes[["stroke"]]/2 * 0.75) * mm2npc
  }
  emode <- .get_emode(edges[["arrowType"]])
  edges$offset_start <- ifelse(emode %in% c(0,1), 0, offsets[edges[["vertex1"]]])
  edges$offset_end <- ifelse(emode %in% c(0,2), 0, offsets[edges[["vertex2"]]])
  
  return(edges)
  
}

#-------------------------------------------------------------------------------
.geom_adj_arrow_offsets <- function(edges){
  edges$offset_start <- edges[["offset_start"]] + edges[["arrow_offset"]]
  edges$offset_end <- edges[["offset_end"]] + edges[["arrow_offset"]]
  return(edges)
}

#-------------------------------------------------------------------------------
.geom_adj_arrow_size <- function(edges, size_unit){
  if(size_unit == "mm"){
    edges$arrow_size <- edges[["arrow_size"]] * ggplot2::.pt
  } else {
    edges$arrow_size <- edges[["arrow_size"]] * ggplot2::.pt * 0.01
  }
  return(edges)
}

#-------------------------------------------------------------------------------
# segmentsGrob
.get_edge_grobs <- function(edges, lineend = "butt", 
  linejoin = "mitre", size_unit = "npc"){
  
  if(.empty(edges)){
    return( zeroGrob() )
  }
  
  edges$colour <- scales::alpha(edges$colour, edges$alpha)
  
  edges <- .set_arrows(edges)
  
  arrows <- .get_arrows(edges, size_unit)
  
  grobs <- list()
  
  if (nrow(edges)>0) {
    gr <- grid::segmentsGrob(
      x0 = edges$x,
      y0 = edges$y,
      x1 = edges$xend,
      y1 = edges$yend,
      gp = ggplot2::gg_par(
        col = edges$colour,
        lwd = edges$linewidth, lty = edges$linetype,
        lineend = lineend, linejoin = linejoin
      )
    )
    gr$name <- grobName(gr, "edges")
    grobs[[length(grobs) + 1]] <- gr
  }
  
  if (!is.null(arrows$ends)) {
    gr <- grid::segmentsGrob(
      x0 = arrows$ends$exy$x,
      y0 = arrows$ends$exy$y,
      x1 = arrows$ends$exy$xend,
      y1 = arrows$ends$exy$yend,
      arrow = arrows$ends$arrow,
      gp = ggplot2::gg_par(
        col = arrows$ends$color,
        lwd = arrows$ends$linewidth, lty = "solid",
        lineend = lineend, linejoin = linejoin
      )
    )
    gr$name <- grobName(gr, "arrowends")
    grobs[[length(grobs) + 1]] <- gr
  }
  
  if (!is.null(arrows$starts)) {
    gr <- grid::segmentsGrob(
      x0 = arrows$starts$exy$x,
      y0 = arrows$starts$exy$y,
      x1 = arrows$starts$exy$xend,
      y1 = arrows$starts$exy$yend,
      arrow = arrows$starts$arrow,
      gp = ggplot2::gg_par(
        col = arrows$starts$color,
        lwd = arrows$starts$linewidth, lty = "solid",
        lineend = lineend, linejoin = linejoin
      )
    )
    gr$name <- grobName(gr, "arrowstarts")
    grobs[[length(grobs) + 1]] <- gr
  }
  
  do.call(grid::gList, grobs)
  
}

#-------------------------------------------------------------------------------
.empty <- function(df){
  is.null(df) || nrow(df) == 0 || ncol(df) == 0 || is_waiver(df)
}

################################################################################
### Adjust arrows
################################################################################
.set_arrows <- function(edges){
  edges <- .add_arrow_angle(edges)
  edges <- .adjust_arrow_size(edges)
  edges <- .adjust_arrow_position(edges)
  return(edges)
}
.add_arrow_angle <- function(edges){
  .a_start <- function(atype){
    a_angle <- rep(NA, length(atype))
    a_angle[atype %in% c(0, 1, -1)] <- 0
    a_angle[atype %in% c(2, 3, -4)] <- 30
    a_angle[atype %in% c(-2, -3, 4)] <- 90
    a_angle
  }
  .a_end <- function(atype){
    a_angle <- rep(NA, length(atype))
    a_angle[atype %in% c(0, 2, -2)] <- 0
    a_angle[atype %in% c(1, 3, 4)] <- 30
    a_angle[atype %in% c(-1, -3, -4)] <- 90
    a_angle
  }
  edges$arrowAngleStart <- .a_start(edges$arrowType)
  edges$arrowAngleEnd <- .a_end(edges$arrowType)
  return(edges)
}
.adjust_arrow_size <- function(edges){
  edges$arrowSize1 <- edges[["arrow_size"]]
  edges$arrowSize2 <- edges[["arrow_size"]]
  a_theta <- 60 #default angle to arrow sides
  a_theta <- a_theta / 180 * pi
  idx <- edges$arrowAngleStart==90
  if(any(idx, na.rm = TRUE)){
    l <- edges$arrowSize1[idx]/2
    b <- sqrt( (l^2 + l^2) - (2 * l^2) * cos(a_theta))
    edges$arrowSize1[idx] <- b + edges$linewidth[idx]/4
  }
  idx <- edges$arrowAngleEnd==90
  if(any(idx, na.rm = TRUE)){
    l <- edges$arrowSize2[idx]/2
    b <- sqrt( (l^2 + l^2) - (2 * l^2) * cos(a_theta))
    edges$arrowSize2[idx] <- b + edges$linewidth[idx]/4
  }
  return(edges)
}
.adjust_arrow_position <- function(edges){
  
  emode <- .get_emode(edges$arrowType)
  
  offset_start <- edges$offset_start * as.numeric(emode %in% c(2, 3))
  offset_end <- edges$offset_end * as.numeric(emode %in% c(1, 3))
  total_offset <- offset_start + offset_end
  
  dx <- edges$xend - edges$x
  dy <- edges$yend - edges$y
  L <- sqrt( dx^2 + dy^2 )
  L <- ifelse(L == 0, 1e-6, L)
  
  scale <- pmin(1, L / total_offset)
  offset_start <- offset_start * scale
  offset_end <- offset_end * scale
  
  edges$px <- dx/L
  edges$py <- dy/L
  edges$x <- edges$x + (edges$px * offset_start)
  edges$y <- edges$y + (edges$py * offset_start)
  edges$xend <- edges$xend - (edges$px * offset_end)
  edges$yend <- edges$yend - (edges$py * offset_end)
  
  return(edges)
}

################################################################################
### Construct arrows
################################################################################
.get_arrows <- function(edges, size_unit = "mm"){
  emode <- .get_emode(edges$arrowType)
  #--- add arrow starts
  idx <- emode==2 | emode==3
  if(any(idx)){
    starts <- .arrow_starts(edges[idx,], size_unit)
  } else {
    starts <- NULL
  }
  #--- add arrow ends
  idx <- emode==1 | emode==3
  if(any(idx)){
    ends <- .arrow_ends(edges[idx,], size_unit)
  } else {
    ends <- NULL
  }
  arrows <- list(ends = ends, starts = starts)
  return(arrows)
}
.arrow_starts <- function(edges, size_unit){
  exy <- edges[,c("x", "y", "xend", "yend")]
  exy$xend <- exy$x + (edges$px * 0.01)
  exy$yend <- exy$y + (edges$py * 0.01)
  arrow <- grid::arrow(angle = edges$arrowAngleStart,
    type = "open", ends = "first",
    length = grid::unit(edges$arrowSize1, size_unit))
  return(list(exy=exy, arrow = arrow, color = edges$colour, 
    linewidth = edges$linewidth))
}
.arrow_ends <- function(edges, size_unit){
  exy <- edges[,c("x", "y", "xend", "yend")]
  exy$x <- exy$xend - (edges$px * 0.01)
  exy$y <- exy$yend - (edges$py * 0.01)
  arrow <- grid::arrow(angle = edges$arrowAngleEnd,
    type = "open", ends = "last",
    length = grid::unit(edges$arrowSize2, size_unit))
  return(list(exy=exy, arrow = arrow, color = edges$colour, 
    linewidth = edges$linewidth))
}

#-------------------------------------------------------------------------------
#' Manipulate node features in a GraphSpace object
#'
#' Utilities for extracting and adding node-associated features stored in the
#' \code{fdata} container of a \code{GraphSpace} object.
#'
#' @param x A \code{GraphSpace} object.
#'
#' @param vars Character vector specifying feature names to extract.
#' If \code{NULL}, all features are returned.
#'
#' @param as_df Logical. If \code{TRUE}, returns a \code{data.frame}.
#' Otherwise returns the original backend representation.
#'
#' @param data A matrix-like or \code{data.frame} object containing node
#' features. Rows must correspond to node identifiers.
#'
#' @return
#' \itemize{
#'   \item \code{gs_fetch_features()} returns a matrix-like object or
#'   \code{data.frame} containing the selected node features.
#'
#'   \item \code{gs_add_features()} returns a modified
#'   \code{GraphSpace} object.
#' }
#'
#' @aliases gs_fetch_features
#' @aliases gs_add_features
#' @name gs_features-utils
NULL

#-------------------------------------------------------------------------------
#' @rdname gs_features-utils
#' @export
gs_fetch_features <- function(x, vars = NULL, as_df = FALSE) {
  
  if (!inherits(x, "GraphSpace")) {
    rlang::abort("'x' must be a GraphSpace object.")
  }
  if (!is.null(vars)) {
    .validate_gs_args("allCharacter", "vars", vars)
  }
  .validate_gs_args("singleLogical", "as_df", as_df)
  
  fdata <- gs_fdata(x)
  
  if (!inherits(fdata, "Matrix")) {
    return(NULL)
  }
  
  if (!is.null(vars)) {
    vars <- intersect(vars, colnames(fdata))
    if (length(vars) == 0) {
      return(NULL)
    }
    fdata <- fdata[, vars, drop = FALSE]
  }
  
  if (isTRUE(as_df)) {
    fdata <- as.data.frame(fdata, drop = FALSE)
  }
  
  return(fdata)
}

#' @rdname gs_features-utils
#' @importFrom utils head
#' @export
gs_add_features <- function(x, data) {
  
  if (!inherits(x, "GraphSpace")) {
    rlang::abort("'x' must be a GraphSpace object.")
  }
  
  if (length(dim(data)) != 2) {
    rlang::abort("'data' must be two-dimensional (e.g. matrix-like object).")
  }
  
  if (is.data.frame(data)) {
    data <- tryCatch(
      Matrix::Matrix(as.matrix(data)),
      error = function(e) {
        rlang::abort("'data' could not be coerced to a Matrix object.")
      }
    )
  } else if (!inherits(data, "Matrix")) {
    data <- tryCatch(
      Matrix::Matrix(data),
      error = function(e){
        rlang::abort("'data' could not be coerced to a Matrix object.")
      }
    )
  }
  
  if (is.null(rownames(data))) {
    rlang::abort("'data' must contain rownames matching node identifiers.")
  }
  
  if (is.null(colnames(data))) {
    rlang::abort("'data' must contain feature names as column names.")
  }
  
  node_ids <- gs_nodes(x)$name
  
  # auto-transpose if node IDs are more prevalent in columns than rows
  n_col_hits <- sum(colnames(data) %in% node_ids)
  n_row_hits <- sum(rownames(data) %in% node_ids)
  if (n_col_hits > n_row_hits) {
    rlang::inform(
      "Feature matrix transposed: more node IDs found in columns than rows."
      )
     data <- Matrix::t(data)
  }

  if (!any(rownames(data) %in% node_ids)) {
    rlang::abort("No GraphSpace node identifiers found in 'data'.")
  }

  if (anyDuplicated(rownames(data))) {
    rlang::abort("'data' contains duplicated identifiers.")
  }
  
  # subset and reorder to match node order
  data <- data[rownames(data) %in% node_ids, , drop = FALSE]
  data <- data[match(node_ids, rownames(data)), , drop = FALSE]
  
  # fill missing nodes with NA rows
  missing <- is.na(rownames(data))
  if (any(missing)) {
    rlang::warn(sprintf(
      "%d node(s) have no feature data and will be set to NA.", sum(missing)
    ))
  }
  rownames(data) <- node_ids
  
  # Load fdata slot
  x@fdata <- data
  x@pars$signal.layer <- TRUE
  
  return(x)
  
}

################################################################################
### Main constructor of GraphSpace-class objects
################################################################################
.buildGraphSpace <- function(g, layout = NULL, verbose = TRUE) {
    
    gg <- .validate_igraph(g, layout, verbose)
    nodes <- .get_nodes(gg)
    edges <- .get_edges(gg)
    
    if(verbose) message("Creating a 'GraphSpace' object...")
    instance_id <- .generate_gs_uuid()
    pars <- list(is.directed = is_directed(gg), 
        is.normalized = FALSE, image.layer = FALSE)
    gs <- new(Class = "GraphSpace", 
        nodes = nodes, 
        edges = edges, 
        graph = gg, 
        image = as.raster(matrix()), 
        pars = pars, 
        misc = list(g = g),
        uuid = instance_id
        )
    
    return(gs)
    
}

################################################################################
### Get nodes and edges in a df object
################################################################################
.get_nodes <- function(gg){
    lt <- vertex_attr(gg)
    nodes <- data.frame(row.names = seq_along(lt[[1]]))
    for(nm in names(lt)){
        nodes[[nm]] <- lt[[nm]]
    }
    vertex <- seq_len(igraph::vcount(gg))
    nodes <- cbind(vertex = vertex, nodes)
    rownames(nodes) <- nodes$name
    return(nodes)
}
.get_edges <- function(gg){
    if (igraph::is_directed(gg)) {
        edges <- .get_directed_edges(gg)
    } else {
        edges <- .get_undirected_edges(gg)
    }
    return(edges)
}

################################################################################
### Get edges in a df object
################################################################################
.get_undirected_edges <- function(g){
    if(ecount(g)>0){
        vertex <- igraph::V(g)$name
        edges <- igraph::as_edgelist(g, names = FALSE)
        rownames(edges) <- colnames(edges) <- NULL
        edges <- as.data.frame(edges)
        colnames(edges) <- c("vertex1", "vertex2")
        edges$name1 <- vertex[edges$vertex1]
        edges$name2 <- vertex[edges$vertex2]
        atts <- .get_eatt(g)
        if(!all(atts[,c(1,2)]==edges[,c(1,2)])){
            rlang::abort("unexpected indexing during edge attribute combination.")
        }
        edges <- cbind(edges, atts[,-c(1,2), drop = FALSE])
        edges <- edges[order(edges$vertex1,edges$vertex2), ]
        idx <- colnames(edges) %in% names(.get_empty_edgedf())
        edges <- edges[, c(which(idx), which(!idx))]
        rownames(edges) <- NULL
    } else {
        edges <- .get_empty_edgedf()
    }
    return(edges)
}
.get_eatt <- function(g){
    lt <- igraph::edge_attr(g)
    atts <- data.frame(row.names = seq_along(lt[[1]]))
    for(nm in names(lt)){
        atts[[nm]] <- lt[[nm]]
    }
    e <- igraph::as_edgelist(g, names = FALSE)
    colnames(e) <- c("vertex1", "vertex2")
    atts <- cbind(e, atts)
    return(atts)
}
.get_empty_edgedf <- function(){
    n <- numeric(); c <- character()
    edges <- data.frame(n, n, c, c, c, c, n, n, n)
    colnames(edges) <- c("vertex1","vertex2", "name1", "name2", 
        "edgeLineType", "edgeLineColor", "edgeLineWidth",
        "arrowType", "weight")
    return(edges)
}

################################################################################
### Get undirected edges in a df object
################################################################################
.get_directed_edges <- function(g) {
    if (ecount(g) > 0) {
        atts <- .extract_directed_att(g)
        vertex <- igraph::V(g)$name
        E(g)$emode <- 1
        E(g)$emode[igraph::which_mutual(g)] <- 3
        e <- emode <- .adjacency(g, attr = "emode")
        bl <- lower.tri(emode) & emode == 3
        emode[bl] <- 0
        edges <- arrayInd(seq_len(prod(dim(emode))), dim(emode), 
            useNames = TRUE)
        edges <- as.data.frame(edges)
        colnames(edges) <- c("vertex1", "vertex2")
        edges$emode <- as.numeric(emode)
        edges$name1 <- vertex[edges$vertex1]
        edges$name2 <- vertex[edges$vertex2]
        edges$e <- as.numeric(e > 0)
        eid <- e; eid[,] <- 0
        ut <- upper.tri(eid)
        eid[ut] <- seq_len(sum(ut))
        eid <- t(eid)
        eid[ut] <- seq_len(sum(ut))
        edges$eid <- as.numeric(eid)
        edges$ut <- as.numeric(upper.tri(e))
        edges$lt <- as.numeric(lower.tri(e))
        if (!all(atts[, c(1, 2)] == edges[, c(1, 2)])) {
            rlang::abort("unexpected indexing during edge attribute combination.")
        }
        edges <- cbind(edges, atts[, -c(1, 2), drop=FALSE])
        eid <- unique(edges$eid[edges$e > 0])
        edges <- edges[edges$eid %in% eid, ]
        edges <- edges[order(edges$eid), ]
        rownames(edges) <- NULL
        edges <- .set_arrowtype_dir(edges)
        idx <- colnames(edges) %in% names(.get_empty_edgedf())
        edges <- edges[, c(which(idx), which(!idx))]
        rownames(edges) <- NULL
    } else {
        edges <- .get_empty_edgedf()
    }
    return(edges)
}
.set_arrowtype_dir <- function(edges) {
    # Flip ut/lt from single-edge arrows; this
    # for collecting arrows from the same mtx side
    idx <- which(edges$emode == 1 & edges$lt == 1)
    if (length(idx) > 0) {
        for (i in idx) {
            ii <- which(edges$eid == edges$eid[i])
            edges[ii, c("ut", "lt")] <- edges[ii, c("lt", "ut")]
        }
    }
    # collect left-side arrows
    arrow1 <- edges[edges$lt == 1, "arrowType"]
    arrow1[is.na(arrow1)] <- 0
    # collect right-side arrows
    arrow2 <- edges[edges$ut == 1, "arrowType"]
    arrow2[is.na(arrow2)] <- 0
    # get single-edge assignments
    edges <- edges[edges$e == 1, ]
    eid <- sort(unique(edges$eid))
    edges <- edges[order(-edges$ut, edges$eid), ]
    edges <- edges[match(eid, edges$eid), ]
    # add arrows and remove intermediate columns
    edges <- .merge_arrowtypes_dir(edges, arrow1, arrow2)
    edges <- edges[, -which(colnames(edges) %in% 
            c("e", "eid", "ut", "lt","emode"))]
    return(edges)
}
.merge_arrowtypes_dir <- function(edges, arrow1, arrow2) {
    ##  0 = "---", 1 = "-->",  2 = "<--",  3 = "<->",  4 = "|->",
    ## -1 = "--|", -2 = "|--", -3 = "|-|", -4 = "<-|",
    atypes <- c(0, 1, 2, 3, 4, -1, -2, -3, -4)
    names(atypes) <- c("00","01","10","11","-11","0-1","-10","-1-1","1-1")
    arrowType <- paste0(format(arrow1, digits = 1, trim = TRUE),
        format(arrow2, digits = 1, trim = TRUE))
    edges$arrowType <- as.numeric(atypes[arrowType])
    return(edges)
}
.extract_directed_att <- function(g) {
    # e <- igraph::as_adjacency_matrix(g, sparse = FALSE)
    e <- .adjacency(g)
    atts <- arrayInd(seq_len(prod(dim(e))), dim(e), useNames = TRUE)
    atts <- as.data.frame(atts)
    colnames(atts) <- c("vertex1", "vertex2")
    atts$e <- as.numeric(e)
    # a_names <- names(.get.default.eatt())
    a_names <- igraph::edge_attr_names(g)
    ne <- e == 0
    for (at in a_names) {
        x <- .adjacency(g, attr = at)
        x[ne] <- NA
        if (is.data.frame(x)){
            atts[[at]] <- I(unlist(x, recursive=FALSE))
        } else {
            if (is.numeric(x)) {
                atts[[at]] <- as.numeric(x)
            } else if (is.character(x)) {
                atts[[at]] <- as.character(x)
            }
        }
    }
    rownames(atts) <- NULL
    atts <- atts[, c("vertex1", "vertex2", a_names)]
    return(atts)
}
# ..this is a fix for 'as_adjacency_matrix', when 'attr' is character
.adjacency <- function(g, attr = NULL) {
    if(is.null(attr)){
        exattr <- rep(1, ecount(g))
        x <- matrix(0, nrow = vcount(g), ncol = vcount(g))
    } else {
        exattr <- edge_attr(g, as.character(attr))
        x <- matrix(NA, nrow = vcount(g), ncol = vcount(g))
        if(is.list(exattr)) x <- as.data.frame(x)
    }
    e <- igraph::ends(g, seq_len(ecount(g)), names = FALSE)
    x[e] <- exattr
    if (!is_directed(g)) x[e[,c(2,1)]] <- exattr
    colnames(x) <- rownames(x) <- V(g)$name
    return(x)
}

################################################################################
### Other functions
################################################################################

.gs_edges <- function(gs){
    
    nodes <- gs@nodes
    edges <- gs@edges
    
    coord <- data.frame(
        x = nodes[edges$vertex1, "x"],
        y = nodes[edges$vertex1, "y"],
        xend = nodes[edges$vertex2, "x"],
        yend = nodes[edges$vertex2, "y"]
        )
    
    n_offsets <- nodes[["nodeSize"]]
    #emode <- .get_emode(edges[["arrowType"]])
    #coord$offset_start <- ifelse(emode %in% c(0,1), 0, n_offsets[edges[["vertex1"]]])
    #coord$offset_end <- ifelse(emode %in% c(0,2), 0, n_offsets[edges[["vertex2"]]])
    coord$offset_start <- n_offsets[edges[["vertex1"]]]
    coord$offset_end <- n_offsets[edges[["vertex2"]]]

    gs_id <- attr(edges, "gs_id")
    edges <- cbind(coord, edges)
    attr(edges, "gs_id") <- gs_id
    
    return(edges)
}

.get_emode <- function(arrow_type){
    emode <- abs(arrow_type)
    emode[emode>3] <- 3
    return(emode)
}

#-------------------------------------------------------------------------------
#' @title Create a GraphSpace object
#' 
#' @description \code{GraphSpace} is the main constructor for 
#' \linkS4class{GraphSpace} objects, designed to store graph data and 
#' metadata for optimized rendering in RGraphSpace.
#'
#' @param g A graph object inheriting from the \link[igraph]{igraph} class 
#' (such as \code{igraph} and \code{tbl_graph}) or a \code{data.frame} used 
#' to initialize a \code{GraphSpace} object. If a graph is provided, it 
#' should include vertex coordinates in \code{x} and \code{y} attributes, 
#' and vertex labels in the \code{name} attribute. If a \code{data.frame} 
#' is provided, it must contain at least \code{x} and \code{y} columns 
#' representing the node coordinates; additional columns will be treated as 
#' vertex attributes. For graphs requiring edge definitions, use the 
#' \code{igraph} initialization.
#' @param layout An optional numeric matrix with two columns for \code{x} and
#' \code{y} vertex coordinates. If provided, it overrides coordinates in \code{g}.
#' @param verbose A logical value. If \code{TRUE}, displays detailed messages.
#' @param ... Additional arguments passed to the \code{GraphSpace} constructor.
#' 
#' @return A \linkS4class{GraphSpace} class object.
#' 
#' @details
#' \code{GraphSpace} objects are designed to bridge the gap between network 
#' analysis (via \code{igraph}) and high-quality visualization (via \code{ggplot2}). 
#' The constructor ensures that all necessary aesthetics for 
#' \code{\link{geom_graphspace}} are pre-processed and validated.
#' 
#' \strong{Coordinate System and Normalization:}
#' By default, the constructor expects coordinates in the \code{x} and \code{y} 
#' vertex attributes, along with unique IDs in the \code{name} vertex 
#' attribute. If these are not provided, the constructor will generate 
#' sequential IDs and assign a layout using the 
#' \code{\link[igraph]{layout_nicely}} function. These coordinates define the 
#' relative positioning of nodes. For optimal rendering, it is recommended 
#' to pass the object through \code{\link{normalizeGraphSpace}} after 
#' construction. This converts vertex positions to Normalized Parent Coordinates 
#' (NPC), ensuring the graph remains centered and scaled relative to the 
#' plotting area.
#' 
#' \strong{Data Structure:}
#' The resulting object stores nodes and edges in separate internal slots, 
#' preserving metadata such as \code{nodeSize} and \code{edgeLineColor}. 
#' If an \code{igraph} object is provided without specific styling attributes, 
#' \code{GraphSpace} will assign the default values defined in the 
#' \code{\link{geom_graphspace}} aesthetics. Users can also specify custom 
#' variables in the input graph to be used as aesthetics within the 
#' \code{ggplot2} grammar.
#' 
#' \strong{Arrowhead Mapping:}
#' The \code{arrowType} attribute (see \emph{Arrowhead types} section) 
#' allows for a mapping between symbolic aliases (such as \code{"-->"}) 
#' and internal integer codes. This is useful for assigning interaction 
#' types in directed or undirected graphs (e.g., activation vs. inhibition).
#'  
#' @section Vertex attributes:
#' The following attributes in \code{g} are evaluated by the constructor:
#' 
#' \tabular{ll}{
#'   \code{nodeSize} \tab Numeric \code{[0, 100]}, representing % of the plotting space. \cr
#'   \code{nodeShape} \tab Integer code \code{[0-25]}; see \link[graphics]{points}. \cr
#'   \code{nodeColor} \tab A valid color name or hexadecimal code. \cr
#'   \code{nodeLineWidth} \tab Border thickness; see \link[grid]{gpar}. \cr
#'   \code{nodeLineColor} \tab A valid color name or hexadecimal code. \cr
#'   \code{nodeLabel} \tab Character string (\code{NA} will omit labels). \cr
#'   \code{nodeLabelSize} \tab Font size in \code{pts}; see \link[grid]{gpar}. \cr
#'   \code{nodeLabelColor} \tab A valid color name or hexadecimal code.
#' }
#' 
#' @section Edge attributes:
#' The following attributes in \code{g} are evaluated by the constructor:
#' 
#' \tabular{ll}{
#'   \code{edgeLineWidth} \tab Edge thickness; see \code{\link[grid]{gpar}}. \cr
#'   \code{edgeLineColor} \tab A valid color name or hexadecimal code. \cr
#'   \code{edgeLineType}  \tab Line style (e.g., "solid", "dashed"); see \code{\link[grid]{gpar}}. \cr
#'   \code{arrowType}     \tab Arrowhead style (see \emph{Arrowhead types} section).
#' }
#' 
#' @section Arrowhead types:
#' 
#' Arrowheads are controlled via the \code{arrowType} attribute using 
#' integer or character codes (see examples in the \emph{RGraphSpace} vignette).
#' 
#' In directed graphs, arrows follow the edge list orientation by default, 
#' representing forward directions (\emph{e.g.}, \code{A -> B}). 
#' While undirected graphs do not show arrows by default, specific styles 
#' can be manually assigned for detailed visualization, including forward, 
#' backward, or bidirectional arrowheads.
#' 
#' \subsection{Directed graphs (A -> B):}{
#' \tabular{lll}{
#'   \strong{Code} \tab \strong{Alias} \tab \strong{Description} \cr
#'   \code{0} \tab \code{"---"} \tab No arrow \cr
#'   \code{1} \tab \code{"-->"} \tab Forward arrow \cr
#'   \code{-1} \tab \code{"--|"} \tab Forward bar
#' }
#' }
#' 
#' \subsection{Undirected graphs (A -- B):}{
#' \tabular{lll}{
#'   \strong{Code} \tab \strong{Alias} \tab \strong{Description} \cr
#'   \code{0} \tab \code{"---"} \tab No arrow \cr
#'   \code{1} \tab \code{"-->"} \tab Forward arrow \cr
#'   \code{2} \tab \code{"<--"} \tab Backward arrow \cr
#'   \code{3} \tab \code{"<->"} \tab Bidirectional arrow \cr
#'   \code{4} \tab \code{"|->"} \tab Forward arrow / backward bar \cr
#'   \code{-1} \tab \code{"--|"} \tab Forward bar \cr
#'   \code{-2} \tab \code{"|--"} \tab Backward bar \cr
#'   \code{-3} \tab \code{"|-|"} \tab Bidirectional bar \cr
#'   \code{-4} \tab \code{"<-|"} \tab Backward arrow / forward bar
#' }
#' }
#' 
#' @author Sysbiolab.
#' 
#' @seealso \code{\link{geom_graphspace}}, \code{\link{plotGraphSpace}}
#' 
#' @examples
#' library(RGraphSpace)
#' library(igraph)
#' 
#' # Create a star graph
#' gtoy1 <- make_full_graph(15)
#' 
#' # Custom attributes
#' V(gtoy1)$nodeSize <- 5
#' E(gtoy1)$edgeLineColor <- "red"
#' E(gtoy1)$arrowType <- "-->"
#' 
#' # Create a GraphSpace
#' gs <- GraphSpace(gtoy1)
#' 
#' @importFrom igraph degree vcount ecount which_mutual any_loop any_multiple
#' @importFrom igraph as_edgelist as_adjacency_matrix is_simple 
#' @importFrom igraph graph_attr_names delete_graph_attr empty_graph
#' @importFrom igraph simplify V E 'V<-' 'E<-' is_directed vertex_attr
#' @importFrom igraph layout_nicely as_undirected delete_edge_attr
#' @importFrom igraph vertex_attr_names edge_attr edge_attr_names
#' @importFrom igraph ends delete_vertex_attr 'edge_attr<-' as.igraph
#' @importFrom tidygraph as.igraph
#' @importFrom scales rescale
#' @importFrom grDevices is.raster as.raster
#' @aliases GraphSpace
#' @rdname GraphSpace-methods
#' @export
setMethod("GraphSpace", signature(g = "ANY"),
  function(g, layout = NULL, verbose = TRUE, ...) {
    
    #--- validate argument values
    .validate_gs_args("singleLogical", "verbose", verbose)
    if(inherits(g, "layout_ggraph")){
      if (!inherits(attr(g, "graph"), "igraph")) {
        rlang::abort(
          message = c(
            "x" = "The 'layout_ggraph' object is missing a valid 'graph' attribute.",
            "i" = "RGraphSpace requires an 'igraph' object to be embedded in the layout.",
            "*" = "Ensure you are using a standard `ggraph::create_layout()` object."
          )
        )
      }
      layout <- tryCatch(
        as.matrix(g[, c("x", "y")]),
        error = function(e) NULL
      )
      g <- attr(g, "graph")
    }
    
    if (inherits(g, "igraph")) {
      class(g) <- "igraph"
    } else {
      rlang::abort(
        message = c(
          "x" = "Input 'g' must inherit from the 'igraph' class.",
          "i" = paste("Received an object of class:", paste(class(g), collapse = "/")),
          "*" = "Did you mean to use `tidygraph::as_tbl_graph()`?"
        )
      )
    }
    if(!is.null(layout)){
      .validate_gs_args("numeric_mtx", "layout", layout)
      if (ncol(layout) != 2) {
        rlang::abort(
          message = c(
            "x" = "The 'layout' matrix must have exactly two columns.",
            "i" = paste("Received a matrix with", ncol(layout), "columns."),
            "*" = "Ensure 'layout' represents (x, y) coordinates."
          )
        )
      } 
    }
    
    #--- validate igraph and build a gs object
    gs <- .buildGraphSpace(g, layout, verbose)
    
    return(gs)
  }
)

#' @rdname GraphSpace-methods
#' @export
setMethod("GraphSpace", signature(g = "data.frame"),
  function(g, verbose = TRUE, ...){
    
    if(!all(c("x", "y") %in% colnames(g))){
      ms_i <- c("i" = "GraphSpace requires x/y columns for 'data.frame' initialization.")
      rlang::abort(
        message = c("x" = "Input 'g' is missing 'x' and 'y' coordinates."),
        body = ms_i
      )
    }
    
    g <- .graphFromCoordinates(g)
    
    #--- build GraphSpace-class
    gs <- GraphSpace(g = g, verbose = verbose, ...)
    
    return(gs)
    
  }
)

#-------------------------------------------------------------------------------
.graphFromCoordinates <- function(coord){
  
  # Check attributes
  attr <- unique(colnames(coord))
  attr <- attr[!is.na(attr)]
  attr <- attr[!attr%in%c("x","y")]
  
  # Initialize a graph using 'coord' as vertices, with no edges
  g <- make_empty_graph(n = nrow(coord), directed = FALSE)
  if(!is.null(rownames(coord))){
    V(g)$name <- rownames(coord)
  }
  
  # Add coordinates
  V(g)$x <- coord$x
  V(g)$y <- coord$y
  
  if(length(attr)>0){
    for(name in attr){
      igraph::vertex_attr(g, name) <- coord[[name]]
    }
  }
  
  return(g)
  
}

#-------------------------------------------------------------------------------
#' @title Wrapper function to plot GraphSpace objects in ggplot2
#'
#' @description 
#' \code{plotGraphSpace()} is a High-level plotting interface that translates 
#' \code{igraph} and \code{GraphSpace} data objects into \code{ggplot2} layers.
#'
#' @param gs Either an \code{igraph} or \linkS4class{GraphSpace} class object.
#' If \code{gs} is an \code{igraph}, then it must include \code{x}, \code{y}, 
#' and \code{name}  vertex attributes (see \code{\link{GraphSpace}}).
#' @param theme Name of a custom RGraphSpace theme. These themes 
#' (from 'th0' to 'th3') consist of preconfigured ggplot settings, 
#' which can subsequently refine using \code{\link[ggplot2]{ggplot2}}.
#' @param xlab The title for the 'x' axis of a 2D-image space.
#' @param ylab The title for the 'y' axis of a 2D-image space.
#' @param font.size A single numeric value passed to ggplot themes.
#' @param bg.color A single color for background.
#' @param add.labels A logical value indicating whether to plot vertex labels.
#' @param node.labels A vector of vertex names to be highlighted in the graph
#' space. This argument overrides 'add.labels'.
#' @param label.size A size argument passed to \code{\link[ggplot2]{geom_text}}.
#' @param label.color A color passed to \code{\link[ggplot2]{geom_text}}.
#' @param add.image A logical value indicating whether to add a background 
#' image, when one is available (see \code{\link{GraphSpace}}).
#' @param raster A logical value indicating whether to rasterize the main plot.
#' See \code{\link[ggrastr]{rasterise}} for further specifications.
#' @param dpi Raster resolution, in dots per inch.
#' @param dev Device used in the \code{\link[ggrastr]{rasterise}} call.
#' @return A ggplot-class object.
#' @author Sysbiolab.
#' @seealso \code{\link{GraphSpace}}
#' @examples
#' library(RGraphSpace)
#' library(igraph)
#' 
#' # Load a demo igraph
#' data('gtoy1', package = 'RGraphSpace')
#' 
#' # Generate a ggplot for gtoy1
#' plotGraphSpace(gtoy1)
#' 
#' # Create a star graph
#' gtoy_star <- make_full_graph(15)
#' 
#' # Example of setting node and edge attributes
#' V(gtoy_star)$nodeSize <- 5
#' E(gtoy_star)$edgeLineColor <- "red"
#' E(gtoy_star)$arrowType <- "<->"
#' 
#' # Create a GraphSpace object
#' gs_star <- GraphSpace(gtoy_star)
#' 
#' # Normalize graph coordinates
#' gs_star <- normalizeGraphSpace(gs_star)
#' 
#' # Generate a ggplot for gs_star
#' plotGraphSpace(gs_star)
#' 
#' @import methods
#' @importFrom ggplot2 ggplot
#' @importFrom grDevices col2rgb
#' @importFrom ggrastr rasterize
#' @docType methods
#' @rdname plotGraphSpace-methods
#' @aliases plotGraphSpace
#' @export
setMethod("plotGraphSpace", "GraphSpace", 
    function(gs, theme = "th0", xlab = "Graph coordinates 1", 
        ylab = "Graph coordinates 2", font.size = 1,
        bg.color = "grey95", add.labels = FALSE,
        node.labels = NULL, label.size = 3, 
        label.color = "grey20", add.image = TRUE, 
        raster = FALSE, dpi = 300, dev = "cairo_png") {
        #--- validate the gs object and args
        .validate_gs_args("singleString", "xlab", xlab)
        .validate_gs_args("singleString", "ylab", ylab)
        .validate_gs_args("singleNumber", "font.size", font.size)
        .validate_gs_colors("singleColor", "bg.color", bg.color)
        .validate_gs_args("singleLogical", "add.labels", add.labels)
        .validate_gs_args("singleNumber", "label.size", label.size)
        .validate_gs_colors("singleColor", "label.color", label.color)
        .validate_gs_args("singleLogical", "add.image", add.image)
        .validate_gs_args("singleLogical", "raster", raster)
        .validate_gs_args("singleInteger", "dpi", dpi)
        .validate_gs_args("singleString", "dev", dev)
        if (!is.null(node.labels)) {
            .validate_gs_args("allCharacter", "node.labels", node.labels)
        }
        
        theme <- match.arg(theme, choices = c("th0", "th1", "th2", "th3"))
        
        #--- get gs slots
        nodes <- gs_nodes(gs)
        pars <- getGraphSpace(gs, "pars")
        
        #--- initialize a ggplot object
        ggp <- ggplot()
        
        #--- add image
        if(pars$image.layer){
            img <- getGraphSpace(gs, "image")
            if(add.image){
                ggp <- .add_image(ggp, img)
            } else {
                ggi <- .add_image(ggp, img)
                ggi <- ggi + theme_gspace_coords(theme = theme, 
                  is_norm = pars$is.normalized, xlab = xlab, ylab = ylab, 
                  txt_size = font.size, leg_size = font.size, 
                  bg_color = bg.color)
            }
        }

        #--- add graph
        ggp <- ggp + geom_graphspace(data = gs)
        
        #--- add node labels
        if (!is.null(node.labels)){
            ggp <- .add_labels1(ggp, nodes, node.labels, 
                label.size, label.color)
        } else if(add.labels){
            ggp <- .add_labels2(ggp, nodes)
        }
        
        #--- apply custom theme
        ggp <- ggp + theme_gspace_coords(theme = theme, 
          is_norm = pars$is.normalized, xlab = xlab, ylab = ylab, 
          txt_size = font.size, leg_size = font.size,
          bg_color = bg.color)
        
        if(raster){
          ggp <- ggrastr::rasterize(ggp, layers = "GraphSpace", 
              dpi = dpi, dev = dev)
        }
        
        if(pars$image.layer && !add.image){
            ggl <- list(image = ggi, graph = ggp)
            return(ggl)
        } else {
            return(ggp)
        }
        
    }
)

#-------------------------------------------------------------------------------
.plot_graph_wrapper <- function(gs, ...) {
  gs <- GraphSpace(gs, verbose = FALSE)
  gs <- normalizeGraphSpace(gs)
  plotGraphSpace(gs = gs, ...)
}

#' @param ... Additional arguments passed to the 
#' \code{\link{plotGraphSpace}} function.
#' @import methods
#' @docType methods
#' @rdname plotGraphSpace-methods
#' @aliases plotGraphSpace
#' @export
setMethod("plotGraphSpace", signature(gs = "igraph"), .plot_graph_wrapper)

#' @rdname plotGraphSpace-methods
#' @aliases plotGraphSpace
#' @export
setMethod("plotGraphSpace", signature(gs = "tbl_graph"), .plot_graph_wrapper)

#' @rdname plotGraphSpace-methods
#' @aliases plotGraphSpace
#' @export
setMethod("plotGraphSpace", signature(gs = "gs_graph"), .plot_graph_wrapper)

#-------------------------------------------------------------------------------
#' Plot GraphSpace objects
#' 
#' @param x A \linkS4class{GraphSpace} class object.
#' @param ... Additional arguments passed to the 
#' \code{\link{plotGraphSpace}} function.
#' @seealso \code{\link{plotGraphSpace}}
#' 
#' @importFrom graphics plot
#' @export
plot.GraphSpace <- function(x, ...) {
    plotGraphSpace(x, ...)
}

#-------------------------------------------------------------------------------
#' @title Accessors for fetching slots from a GraphSpace object
#'
#' @description \code{getGraphSpace} retrives information from
#' individual slots available in a GraphSpace object.
#'
#' @param gs A preprocessed \linkS4class{GraphSpace} class object
#' @param what A single character value specifying which slot to 
#' retrieve from a 'GraphSpace' object.
#' Options: "graph", "nodes", "edges", "pars", "misc", "image", and "fdata".
#' @return Content from slots in the \linkS4class{GraphSpace} object.
#' @examples
#' library(RGraphSpace)
#' library(igraph)
#' 
#' # Load a demo igraph
#' data('gtoy1', package = 'RGraphSpace')
#'
#' # Create a new GraphSpace object
#' gs <- GraphSpace(gtoy1)
#'
#' # Get the 'summary' slot in gs
#' getGraphSpace(gs, what = 'graph')
#'
#' @import methods
#' @docType methods
#' @rdname getGraphSpace-methods
#' @aliases getGraphSpace
#' @export
setMethod("getGraphSpace", "GraphSpace", function(gs, what = "graph") {
  .validate_gs_args("singleString", "what", what)
  opts <- c("graph", "nodes", "edges", "pars", "misc", "image", "fdata")
  if (!what %in% opts) {
    opts <- paste0(opts, collapse = ", ")
    stop("'what' must be one of:\n", opts, call. = FALSE)
  }
  if (what == "nodes") {
    obj <- gs@nodes
  } else if (what == "edges") {
    obj <- gs@edges
  } else if (what == "graph") {
    obj <- gs@graph
  } else if (what == "pars") {
    obj <- gs@pars
  } else if (what == "misc") {
    obj <- gs@misc
  } else if (what == "image") {
    obj <- gs@image
  } else if (what == "fdata") {
    obj <- gs@fdata
  } else {
    obj <- gs@graph
  }
  return(obj)
})

#-------------------------------------------------------------------------------
#' @title Accessors and attribute utilities for GraphSpace objects
#' 
#' @description Access and modify individual components of a
#' \linkS4class{GraphSpace} object. Selected \pkg{igraph} methods are
#' applied to the internal graph representation and propagated to
#' downstream node and edge components.
#' 
#' @param x A \linkS4class{GraphSpace} class object
#' @param name Name of the attribute.
#' @param value Replacement value for the selected slot or attribute.
#' @param ... Additional arguments passed to extraction methods. 
#' @details
#' For \code{gs_nodes()}, the optional \code{vars} argument specifies
#' node-associated features retrieved from the \code{fdata}
#' container. See also \code{\link{gs_fetch_features}}.
#' @return Updated \linkS4class{GraphSpace} object.
#' @seealso \code{\link[igraph]{vertex_attr}}, \code{\link[igraph]{edge_attr}}, 
#' \code{\link{gs_fetch_features}}
#' @examples
#' library(RGraphSpace)
#' library(igraph)
#' 
#' # Load a demo igraph
#' data('gtoy1', package = 'RGraphSpace')
#' 
#' # Create a new GraphSpace object
#' gs <- GraphSpace(gtoy1)
#' 
#' #--- Usage of GraphSpace attribute accessors:
#' 
#' # Vertex names
#' names(gs)
#' 
#' # Vertex attribute names
#' gs_names(gs)
#' 
#' # Get a data frame with nodes
#' gs_nodes(gs)
#' 
#' # Get a data frame with edges
#' gs_edges(gs)
#' 
#' # Get vertex count
#' gs_vcount(gs)
#' 
#' # Get edge count
#' gs_ecount(gs)
#' 
#' # Access all vertex attributes
#' gs_vertex_attr(gs)
#' 
#' # Access a specific vertex attribute
#' gs_vertex_attr(gs, "nodeLabel")
#' 
#' # Modify a single value within a vertex attribute
#' gs_vertex_attr(gs, "nodeSize")["n1"] <- 10
#' 
#' # Replace an entire vertex attribute
#' gs_vertex_attr(gs, "nodeSize") <- 10
#' 
#' # Alternative syntax using `$` accessor
#' gs_vertex_attr(gs)$nodeSize <- 10
#' 
#' # Access a specific edge attribute
#' gs_edge_attr(gs, "edgeLineColor")
#' 
#' # Replace an entire edge attribute
#' gs_edge_attr(gs, "edgeLineWidth") <- 1
#' 
#' # Alternative syntax using `$` for edge attributes
#' gs_edge_attr(gs)$edgeLineWidth <- 3
#' 
#' # Add an image and rescale graph coordinates to image space
#' # Images may be provided as a raster or numeric matrix
#' gs_image(gs) <- as_colorraster(volcano)
#' gs <- normalizeGraphSpace(gs)
#' 
#' @aliases names
#' @aliases gs_names
#' @aliases gs_nodes
#' @aliases gs_edges
#' @aliases gs_graph
#' @aliases gs_image
#' @aliases gs_image<-
#' @aliases gs_fdata
#' @aliases gs_fdata<-
#' @aliases gs_features
#' @aliases gs_nfeatures
#' @aliases gs_vcount
#' @aliases gs_ecount
#' @aliases gs_vertex_attr
#' @aliases gs_vertex_attr<-
#' @aliases gs_edge_attr
#' @aliases gs_edge_attr<-
#' @rdname GraphSpace-accessors
#' @export
setMethod("names", "GraphSpace", function(x) {
  x@nodes$name
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_names", "GraphSpace", function(x) {
  colnames(x@nodes)
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_nodes", "GraphSpace", function(x, ...) {
  
  args <- list(...)
  
  vars <- args$vars
  
  nodes <- x@nodes
  
  if (!is.null(vars) && .all_characterValues(vars)) {
    
    signal_df <- gs_fetch_features(x, vars = vars, as_df = TRUE)
    
    if (!is.null(signal_df)) {
      
      signal_vars <- setdiff(colnames(signal_df), colnames(nodes) )
      
      if (length(signal_vars) > 0) {
        signal_df <- signal_df[ rownames(nodes), signal_vars, drop = FALSE]
        nodes[, signal_vars] <- signal_df
      }
      
    }
    
  }
  
  attr(nodes, "gs_handler_type") <- "node"
  attr(nodes, "gs_id") <- x@uuid
  class(nodes) <- c("gs_nodes", class(nodes))
  return(nodes)
  
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_edges", "GraphSpace", function(x) {
  edges <- .gs_edges(x)
  attr(edges, "gs_id") <- x@uuid
  attr(edges, "gs_handler_type") <- "edge"
  class(edges) <- c("gs_edges", class(edges))
  return(edges)
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_image", "GraphSpace", function(x) {
  x <- x@image
  attr(x, "gs_handler_type") <- "image"
  class(x) <- c("gs_image", class(x))
  return(x)
})

#' @rdname GraphSpace-accessors
#' @export
setReplaceMethod("gs_image", "GraphSpace", function(x, value) {

  if(!is.raster(value)){
    .validate_gs_args("numeric_mtx", "value", value)
    rlang::inform(
      c("i" = "Rasterizing numeric matrix.",
        "*" = "Values outside [0,1] were rescaled before conversion.")
    )
    rng <- range(value, na.rm = TRUE)
    if (diff(rng) == 0) {
      if (rng[1] < 0 || rng[1] > 1) {
        value[] <- 0
      }
    } else if (rng[1] < 0 || rng[2] > 1) {
      value <- (value - rng[1]) / diff(rng)
    }
    value <- as.raster(value)
  }
  x@image <- value
  x@misc$image <- value
  x@pars$image.layer <- TRUE
  
  return(x)
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_graph", "GraphSpace", function(x) {
  g <- x@graph
  attr(g, "gs_handler_type") <- "graph"
  class(g) <- c("gs_graph", class(g))
  return(g)
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_fdata", "GraphSpace", function(x) {
  x@fdata
})

#' @rdname GraphSpace-accessors
#' @export
setReplaceMethod("gs_fdata", "GraphSpace", function(x, value) {
  
  x <- gs_add_features(x, value)
  
  validObject(x)
  
  return(x)
  
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_nfeatures", "GraphSpace", function(x) {
    ncol(x@fdata)
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_features", "GraphSpace", function(x) {
  colnames(x@fdata)
})


#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_vcount", "GraphSpace", function(x) {
  igraph::vcount(x@graph)
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_ecount", "GraphSpace", function(x) {
    igraph::ecount(x@graph)
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_vertex_attr", "GraphSpace", function(x, name, ...) {
    g <- x@graph
    if(missing(name)){
        att <- igraph::vertex_attr(graph = g, ...=...)
    } else {
        if(name %in% igraph::vertex_attr_names(g)){
            att <- igraph::vertex_attr(graph = g, name = name, ...=...)
            if(name!="name") names(att) <- V(g)$name
        } else {
            att <- NULL
        }
    }
    return(att)
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_vertex_attr<-", "GraphSpace", function(x, name, ..., value) {
    g <- x@graph
    if (missing(name) && is.list(value)){
        #workaround for analogy to igraph's "syntactic sugar" $<-
        len1 <- unlist(lapply(value, length))==1
        if(any(len1)){
            for(i in which(len1)){
                vl <- value[[i]]
                vl <- ifelse(.is_replicable(vl), vl, list(vl))
                value[[i]] <- rep(vl, igraph::vcount(g))
            }
        }
        igraph::vertex_attr(graph = g) <- value
    } else {
        if(length(value)==1){
            value <- ifelse(.is_replicable(value), value, list(value))
        }
        igraph::vertex_attr(graph = g, name = name, ...=...) <- value  
    }
    x <- .updateNodeSpace(x, g)
    return(x)
})
.is_replicable <- function(x) {
    tryCatch({
        rep(x, 2)
        TRUE
    }, error = function(e) FALSE)
}

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_edge_attr", "GraphSpace", function(x, name, ...) {
    g <- x@graph
    if(missing(name)){
      att <- igraph::edge_attr(graph = g, ...=...)
    } else {
      att <- igraph::edge_attr(graph = g, name = name, ...=...)
    }
    return(att)
})

#' @rdname GraphSpace-accessors
#' @export
setMethod("gs_edge_attr<-", "GraphSpace", function(x, name, ..., value) {
    g <- x@graph
    if (missing(name) && is.list(value)){
        #workaround for analogy to igraph's "syntactic sugar" $<-
        len1 <- unlist(lapply(value, length))==1
        if(any(len1)){
            for(i in which(len1)){
                vl <- value[[i]]
                vl <- ifelse(.is_replicable(vl), vl, list(vl))
                value[[i]] <- rep(vl, igraph::ecount(g))
            }
        }
        igraph::edge_attr(graph = g) <- value
    } else {
        if(length(value)==1){
            value <- ifelse(.is_replicable(value), value, list(value))
        }
        igraph::edge_attr(graph = g, name = name, ...=...) <- value  
    }
    x <- .updateEdgeSpace(x, g)
    return(x)
})
.updateEdgeSpace <- function(x, g){
  x@graph <- .validate_igraph(g)
  x@edges <- .get_edges(x@graph)
  return(x)
}
.updateNodeSpace <- function(x, g) {
    x@graph <- .validate_igraph(g)
    pars <- x@pars
    if(pars$is.normalized){
        if(pars$image.layer){
            x <- normalizeGraphSpace(x, 
                image = x@misc$image,
                mar = pars$mar,
                flip.x = pars$flip.x %||% FALSE, 
                flip.y = pars$flip.y %||% FALSE, 
                rotate.xy = pars$rotate.xy %||% FALSE, 
                flip.v = pars$flip.v %||% TRUE, 
                flip.h = pars$flip.h %||% FALSE,
                verbose = FALSE)
        } else {
            x <- normalizeGraphSpace(x, 
                mar = pars$mar,
                flip.x = pars$flip.x %||% FALSE, 
                flip.y = pars$flip.y %||% FALSE, 
                rotate.xy = pars$rotate.xy %||% FALSE,
                verbose = FALSE)
        }
    } else {
      x@nodes <- .get_nodes(x@graph)
    }
    return(x)
}

#' @rdname GraphSpace-accessors
#' @aliases $,GraphSpace-method
#' @export
setMethod("$", "GraphSpace", function(x, name) {
  
  nodes <- gs_nodes(x)
  
  if (!(name %in% names(nodes))) {
    rlang::warn(paste0("Column '", name, "' not found in nodes."))
    return(NULL)
  }
  
  nodes[[name]]
})

#' @rdname GraphSpace-accessors
#' @aliases $<-,GraphSpace-method [[<-,GraphSpace-method
#' @export
setReplaceMethod("$", "GraphSpace", function(x, name, value) {
  
  protected <- c("vertex", "name")
  
  if (name %in% protected) {
    rlang::abort(
      message = c(
        "x" = paste0("Column '", name, "' is read-only."),
        "i" = "Modifying vertex identifiers directly would break node-edge coherence.",
        "*" = "To change the graph structure, please modify the underlying igraph object."
      )
    )
  }
  
  gs_vertex_attr(x, name) <- value
  x
})

#' @rdname GraphSpace-accessors
#' @method as.igraph GraphSpace
#' @export
as.igraph.GraphSpace <- function(x, ...) {
  return(x@graph)
}

################################################################################
### Internal for GraphSpace objects
################################################################################
#' Internal methods for GraphSpace
#' 
#' @description 
#' Exported solely to enable RStudio auto-completion 
#' and should not be called directly by the user.
#' 
#' @param x,pattern Internal arguments.
#' @keywords internal
#' @name GraphSpace-internal
NULL

#' @rdname GraphSpace-internal
#' @importFrom utils .DollarNames
#' @method .DollarNames GraphSpace
#' @keywords internal
#' @export
.DollarNames.GraphSpace <- function(x, pattern = "") {
  nodes <- gs_nodes(x)
  grep(pattern, names(nodes), value = TRUE)
}

#-------------------------------------------------------------------------------
#' Using ggplot2 with GraphSpace objects
#' 
#' @description
#' 
#' GraphSpace objects can be used directly with \pkg{ggplot2}, allowing node 
#' attributes and high-dimensional feature data to be mapped through standard 
#' aesthetic mappings without manual data extraction.
#' This integration enables:
#' \itemize{
#'   \item Lazy evaluation of node attributes and feature data.
#'   \item Automatic synchronization of node metadata for standard ggplot2
#'   geoms such as \link[ggplot2]{geom_point}.
#'   \item Automatic propagation of node metadata required for edge clipping
#'   and arrow placement in GraphSpace-native geoms.
#' }
#' 
#' @param data A \link{GraphSpace} object.
#' @param mapping Set of aesthetic mappings created by \link[ggplot2]{aes}.
#' Passed to \link[ggplot2]{ggplot}.
#' @param ... Additional arguments passed to \link[ggplot2]{ggplot}.
#' 
#' @details
#' When a GraphSpace object is supplied to \code{ggplot()}, RGraphSpace 
#' extends the standard ggplot2 build process to automatically resolve 
#' GraphSpace variables and synchronize node metadata required for edge 
#' rendering.
#' 
#' When using \code{ggplot()}, neither \link{nodespace_handler} nor
#' \link{inject_nodespace} need to be called explicitly.
#' 
#' @return A \code{gspace_plot} object extending \link[ggplot2]{ggplot}.
#' 
#' @seealso
#' \link{GraphSpace}, \link{geom_nodespace}, \link{geom_edgespace},
#' \link{inject_nodespace}, \link{nodespace_handler}, \link{edgespace_handler}
#' 
#' @examples
#' library(RGraphSpace)
#' library(igraph)
#' library(ggplot2)
#' 
#' # Generate a toy star graph
#' gtoy1 <- make_star(15, mode = "out")
#' V(gtoy1)$my_node_var <- runif(vcount(gtoy1), 1, 20)
#' 
#' # Create a GraphSpace object
#' gs <- GraphSpace(gtoy1, layout = layout_in_circle(gtoy1))
#' 
#' \dontrun{
#'
#' # Example 1: Using RGraphSpace-native geoms
#' # Edge clipping metadata are injected automatically
#' ggplot(gs) +
#'   geom_edgespace(colour = "red") +
#'   geom_nodespace(aes(size = my_node_var), 
#'   fill = "steelblue", stroke = 2) +
#'   scale_size(range = c(2, 15))
#'
#' # Example 2: Mixing native and general geoms
#' # Note possible clipping mismatch when combining
#' # geom_edgespace() with generic ggplot2 node geoms.
#' # Since geom_point() does not expose the final rendered
#' # node radius to RGraphSpace, edge clipping is estimated
#' # from layer parameters and may not exactly match the
#' # displayed node geometry.
#' ggplot(gs) +
#'   geom_edgespace(colour = "red") +
#'   geom_point(aes(x, y, size = my_node_var), 
#'   fill = "steelblue", stroke = 2, shape = 21) +
#'   scale_size(range = c(2, 15))
#'
#' }
#' 
#' @name ggplot-GraphSpace
#' @rdname ggplot-GraphSpace
#' @importFrom ggplot2 ggplot
#' @method ggplot GraphSpace
#' @export
ggplot.GraphSpace <- function(data, mapping = NULL, ...) {

  p <- NextMethod()
  
  class(p) <- c("gspace_plot", class(p))
  
  p
  
}

#-------------------------------------------------------------------------------
#' @importFrom ggplot2 fortify
#' @export
fortify.GraphSpace <- function(model, data, ...) {
  res <- gs_nodes(model)
  attr(res, ".gs_graph") <- model
  class(res) <- c("gspace_data", class(res))
  return(res)
}

#-------------------------------------------------------------------------------
#' @importFrom ggplot2 ggplot_build
#' @method ggplot_build gspace_plot
#' @export
ggplot_build.gspace_plot <- function(plot, ...) {
  plot <- .inject_graphspace_handlers(plot)
  plot <- .inject_nodespace_internal(plot, verbose = FALSE) 
  NextMethod()
}

#-------------------------------------------------------------------------------
#' Dynamic Scale Injection for Edge Clipping
#'
#' @description
#' Utility function for **RGraphSpace** that enables edge layers to scan 
#' adjacent nodes and determine their dimensions. This information is used 
#' to compute arrow clipping offsets, preventing edge geometry from 
#' overlapping node symbols.
#' 
#' @param ... Additional parameters passed to other methods (currently ignored).
#'
#' @details
#' This function operates in two stages within the `ggplot2` workflow:
#' 
#' 1. **Capture:** It scans the `plot` layers for a \link{GeomNodeSpace} 
#' to extract both mapping variables (from `aes()`) and static parameters 
#' (specifically `size` and `stroke`).
#' 
#' 2. **Injection:** It locates \link{GeomEdgeSpace} layers and injects
#' scale rules, captured mappings, and fixed parameters into the geometry 
#' parameters.
#'    
#' This "lazy injection" calculates edge clipping based on the actual scales 
#' used by the nodes, even if scales are defined after the layers.
#' 
#' **Note:** `inject_nodespace()` must be called last in the `ggplot` chain to 
#' allow the function to correctly scan all previously added layers and scales.
#'
#' @return An object of class `inject_nodespace`, which interacts with the 
#' ggplot2 `+` operator.
#' 
#' @seealso
#' \link{geom_edgespace}, \link{geom_nodespace}
#' 
#' @examples
#' library(RGraphSpace)
#' library(igraph)
#' library(ggplot2)
#' 
#' # Generate a toy star graph
#' gtoy1 <- make_star(15, mode="out")
#'
#' # Set node and edge attributes
#' V(gtoy1)$my_node_var <- runif(vcount(gtoy1), 1, 20)
#' E(gtoy1)$my_edge_var <-  runif(ecount(gtoy1), 1, 20)
#'
#' # Create a GraphSpace object with a circular layout
#' gs <- GraphSpace(gtoy1, layout = layout_in_circle(gtoy1))
#'
#' \dontrun{
#' # Build the plot
#' # Note that inject_nodespace() is called at the end to
#' # synchronize node sizes with edge clipping.
#' ggplot() +
#'   geom_edgespace(aes(colour = my_edge_var), data = gs) +
#'   geom_nodespace(aes(size = my_node_var), data = gs) +
#'   scale_size(range = c(2, 15)) +
#'   inject_nodespace()
#' }
#' 
#' @importFrom ggplot2 ggplot_add
#' @rdname inject_nodespace
#' @export
inject_nodespace <- function(...) {
  structure(list(...), class = "inject_nodespace")
}

#' @method ggplot_add inject_nodespace
#' @export
ggplot_add.inject_nodespace <- function(object, plot, ...) {
  .inject_nodespace_internal(plot, verbose = TRUE)
}

#-------------------------------------------------------------------------------
.inject_graphspace_handlers <- function(plot){
  
  if(!inherits(plot@data, "gspace_data")){
    return(plot)
  }
  
  for(i in seq_along(plot@layers)){
    
    layer <- plot@layers[[i]]
    
    # Skip layers with explicit non-waiver data
    if(!inherits(layer$data, "waiver")){
      next
    }
    
    # For all other waiver-data layers (including general geoms),
    # inject a 'nodespace_handler' carrying the layer's own mapping
    if (inherits(layer$geom, "GeomEdgeSpace")) {
      layer$data <- edgespace_handler()
    } else {
      layer$data <- nodespace_handler(mapping = layer$mapping)
    }
    
    plot@layers[[i]] <- layer
    
  }
  
  plot
  
}

#-------------------------------------------------------------------------------
# Core injection logic - standalone internal function
.inject_nodespace_internal <- function(plot, verbose = TRUE) {
  
  ##--- ATTEMPT AUTO-DETECTION OF LAYERS
  
  p_layers <- list()
  has_nodes <- FALSE
  has_edges <- FALSE
  
  for (i in seq_along(plot@layers)){
    layer <- plot@layers[[i]]
    if (inherits(layer$geom, "GeomNodeSpace") && !has_nodes){
      has_nodes <- TRUE
      p_layers$nodes <- layer
      p_layers$n_idx <- i
    }
    if (inherits(layer$geom, "GeomEdgeSpace") && !has_edges){
      has_edges <- TRUE
      p_layers$edges <- layer
      p_layers$e_idx <- i
    }
  }
  
  if (!has_nodes || !has_edges) {
    return(plot)
  }
  
  ##--- CHECK COMPATIBILITY BETWEEN LAYERS
  
  node_data <- p_layers$nodes$data
  edge_data <- p_layers$edges$data
  if(.is_handler(node_data, "node") && .is_handler(edge_data, "edge")){
    node_data <- tryCatch(node_data(plot$data), error = function(e) NULL)
    edge_data <- tryCatch(edge_data(plot$data), error = function(e) NULL)
  } else if(identical(node_data, edge_data)){
    data <- .get_gs_graph(node_data)
    if(inherits(data, "GraphSpace")){
      node_data <- gs_nodes(data)
      edge_data <- gs_edges(data)
    }
  }
  
  if(!.is_data_compatible(node_data, edge_data)){
    rlang::warn(
      message = c(
        "!" = "`inject_nodespace()` failed to detect compatible edge/node layers.",
        "i" = "Both layers must originate from the same source graph.",
        "*" = "Check 'Interoperability' in the vignette for integration examples."
      )
    )
    return(plot)
  }
  
  ##--- EXTRACT NODE DATA, AES, AND PARAMS
  
  p_pars <- list(size_unit = "mm")
  has_aes <- FALSE
  
  # Extract aes
  if (inherits(p_layers$nodes$mapping, "uneval")) {
    m <- p_layers$nodes$mapping[names(p_layers$nodes$mapping) %in% c("size", "stroke")]
    if (length(m) > 0) {
      has_aes <- TRUE
      p_pars$size_aes <- m
    }
  }
  if(verbose && !has_aes){
    rlang::warn(
      message = c(
        "!" = "`inject_nodespace()` could not find a 'size' aesthetic.",
        "i" = "Fixed node 'size' scales with the viewport.",
        "*" = "Use aes(size = ...) for data-driven scaling."
      )
    )
  }
  # Extract params
  if(is.list(p_layers$nodes$aes_params)){
    p <- p_layers$nodes$aes_params[names(p_layers$nodes$aes_params) %in% c("size", "stroke")]
    if (length(p) > 0) p_pars$size_params <- p
  }
  if(is.list(p_layers$nodes$geom_params)){
    u <- p_layers$nodes$geom_params[[".size_unit"]]
    if (!is.null(u)) p_pars$size_unit <- u
  }
  
  ##--- MAP AES
  
  # Map 'size' and 'stroke' aesthetics if present
  if(!is.null(p_pars$size_aes)){
    node_data <- .geom_map_size_aes(node_data, size_aes = p_pars$size_aes)
  }
  
  ##--- PREPARE SCALES
  
  has_scales <- FALSE
  size_scales <- Filter(function(s) "size" %in% s$aesthetics, plot$scales$scales)
  if (length(size_scales) > 0){
    sz <- size_scales[[1]] 
    if (inherits(sz, "Scale") && !is.null(sz$palette)){
      has_scales <- TRUE
      p_pars$size_scales <- sz
    }
  }
  stroke_scales <- Filter(function(s) "stroke" %in% s$aesthetics, plot$scales$scales)
  if (length(stroke_scales) > 0){
    sz <- stroke_scales[[1]] 
    if (inherits(sz, "Scale") && !is.null(sz$palette)){
      has_scales <- TRUE
      p_pars$stroke_scales <- sz
    }
  }
  
  if(verbose && has_aes && !has_scales){
    rlang::warn(
      message = c(
        "!" = "`inject_nodespace()` could not find a 'size' scale.",
        "i" = "Clipping may be inaccurate without data-driven dimensions.",
        "*" = "Ensure a `scale_size()` is provided."
      )
    )
  }
  
  ##--- APPLY SCALES TO NODE DATA
  
  # adjust 'size' by user-defined scales if present
  if(!is.null(p_pars$size_scales$palette) && !is.null(node_data[["size"]])){
    size_range <- p_pars$size_scales$palette(c(0, 1))
    node_data$size <- .geom_node_to_offset(node_data[["size"]],
      range = size_range, size_unit = p_pars$size_unit)
  }
  if(!is.null(p_pars$stroke_scales$palette) && !is.null(node_data[["stroke"]])){
    stroke_range <- p_pars$stroke_scales$palette(c(0, 1))
    node_data$stroke <- .geom_node_to_offset(node_data[["stroke"]],
      range = stroke_range, size_unit = p_pars$size_unit)
  }
  # replace 'size' by user-defined params if present
  if (!is.null(p_pars$size_params$size)) {
    node_data$size <- p_pars$size_params$size
  }
  # replace 'stroke' by user-defined params if present
  if (!is.null(p_pars$size_params$stroke)) {
    node_data$stroke <- p_pars$size_params$stroke
  }
  
  # check if only one size params was included
  if( sum(c("size", "stroke") %in% colnames(node_data)) == 1){
    node_data <- .geom_map_size_params(node_data)
  }
  
  ##--- INJECT NODE METADATA INTO EDGE LAYERS
  required_att <- c("x", "y", "vertex", "size", "stroke")
  has_clipping_cols <- all(required_att %in% colnames(node_data))
  check_idx <- !is.null(p_layers$e_idx) && p_layers$e_idx <= length(plot@layers)
  has_unit <- !is.null(p_pars$size_unit)
  if (check_idx && (has_clipping_cols || has_unit)){
    if (has_clipping_cols){
      plot@layers[[p_layers$e_idx]]$geom_params$.nodes <- node_data 
    }
    if (has_unit){
      plot@layers[[p_layers$e_idx]]$geom_params$.size_unit <- p_pars$size_unit
    }
  } else {
    if(verbose){
      rlang::warn(
        message =  "`inject_nodespace()` found no clipping adjustments."
      )
    }
  }
  
  return(plot)
  
}

#-------------------------------------------------------------------------------
.get_gs_graph <- function(data) {
  if (inherits(data, "GraphSpace")) {
    return(data)
  }
  if (inherits(data, "gs_nodes")) {
    if(inherits(attr(data, ".gs_graph"), "GraphSpace")){
      data <- attr(data, ".gs_graph")
    }
    return(data)
  }
  if(inherits(data, c("igraph", "layout_ggraph"))){
    data <- GraphSpace(data, verbose = FALSE)
  }
  return(data)
}

#-------------------------------------------------------------------------------
.is_data_compatible <- function(node_data, edge_data){
  
  if(is.null(node_data) || is.null(edge_data)){
    return(FALSE)
  }
  
  # Check handlers
  if( !.was_handled(node_data, "node") || 
      !.was_handled(edge_data, "edge") ){
    return(FALSE)
  }
  
  # Quick check via UUID match
  n_id <- attr(node_data, "gs_id") 
  e_id <- attr(edge_data, "gs_id")
  if(.is_singleString(n_id) && .is_singleString(e_id)){
    if(n_id == e_id) return(TRUE)
  }
  
  # Compare data
  id_nodes <- node_data$vertex
  id_edges <- c(edge_data$vertex1, edge_data$vertex2)
  if (is.null(id_nodes) || is.null(id_edges)){
    return(FALSE)
  }
  id_nodes <- as.character(id_nodes)
  id_edges <- as.character(unique(id_edges))
  is_topo_ok <- all(id_edges %in% id_nodes) 
  if(is_topo_ok){
    rlang::message_cnd(
      message = c(
        "`inject_nodespace()`: layers synced via vertex IDs.",
        "i" = "Fallback triggered by UUID mismatch between edge and node data.",
        "*" = "Ensure graph objects remain identical across all plot layers."
      )
    )
  }
  
  return( is_topo_ok )
  
}

#-------------------------------------------------------------------------------
.is_handler <- function(f, type = "node") {
  if (!is.function(f)) return(FALSE)
  # Retrieve the original handler from inside ggplot2's data
  # wrapper; depends on ggplot2 storing it as 'data' in the
  # closure environment.
  inner_f <- tryCatch(environment(f)$data, error = function(e) NULL)
  if ( is.null(inner_f) ) return(FALSE)
  b1 <- identical(attr(inner_f, "gs_handler_type"), type)
  b2 <- inherits(inner_f, paste0(type, "space_handler"))
  return( b1 || b2 )
}

#-------------------------------------------------------------------------------
.was_handled <- function(dt, type = "node") {
  if (!is.data.frame(dt)) return(FALSE)
  return( identical(attr(dt, "gs_handler_type"), type) )
}

#-------------------------------------------------------------------------------
.geom_node_to_offset <- function(node_size, range = c(1, 6), size_unit = "npc"){
  # 1. Passed as a aesthetic, 'size' follows 'geom_point' behavior, in mm;
  # 2. Passed as a parameter, 'size' scales with the viewport (%) with
  # range expected in [0, 100], then later converted to 'npc'.
  if (is.null(node_size)) return(NULL)
  if (size_unit == "npc") {
    # Expect node_size already in [0,100] viewport percent
    node_size <- scales::squish(node_size, range = c(0, 100))
  } else {
    node_size <- scales::rescale(node_size, range = c(0, 1))
    node_size <- scales::rescale(sqrt(node_size), to = range)
  }
  return(node_size)
}

#-------------------------------------------------------------------------------
.geom_map_size_aes <- function(nodes, size_aes){
  
  required_clipping <- c("x", "y", "vertex", "size", "stroke")
  size_att <- c(size = "nodeSize", stroke = "nodeLineWidth")
  default_att <- .get_default_vatt()
  
  if (!is.null(size_aes) && !inherits(size_aes, "uneval")) {
    size_aes <- NULL
  }
  
  for(att in names(size_att)){
    if (.is_valid_aes(att, size_aes, nodes)) {
      nodes[[att]] <- tryCatch({
        rlang::eval_tidy(size_aes[[att]], nodes)
      }, error = function(e) {
        NULL
      })
    } else {
      if( !(att %in% colnames(nodes)) ){
        nodes[[att]] <- nodes[[size_att[att]]] %||% default_att[[size_att[att]]]
      }
    }
    if(is.null(nodes[[att]]) || !is.numeric(nodes[[att]])){
      nodes[[att]] <- default_att[[size_att[att]]]
    }
  }
  nodes <- nodes[, intersect(required_clipping, colnames(nodes)), drop = FALSE]
  
  return(nodes)
  
}
.is_valid_aes <- function(att, mapping, data_df) {
  if (is.null(mapping)) return(FALSE)
  expr <- mapping[[att]]
  if (is.null(expr)) return(FALSE)
  vars <- all.vars(expr)
  if (length(vars) == 0) return(TRUE)
  all(vars %in% names(data_df))
}

.geom_map_size_params <- function(nodes){
  
  required_clipping <- c("x", "y", "vertex", "size", "stroke")
  size_att <- c(size = "nodeSize", stroke = "nodeLineWidth")
  default_att <- .get_default_vatt()

  for(att in names(size_att)){
    if( !(att %in% colnames(nodes)) ){
      nodes[[att]] <- nodes[[size_att[att]]] %||% default_att[[size_att[att]]]
    }
    if(is.null(nodes[[att]]) || !is.numeric(nodes[[att]])){
      nodes[[att]] <- default_att[[size_att[att]]]
    }
  }
  nodes <- nodes[, intersect(required_clipping, colnames(nodes)), drop = FALSE]
  
  return(nodes)
  
}

## ----Load packages for quick start, eval=TRUE, message=FALSE------------------
#--- Load required packages
library("igraph")
library("ggplot2")
library("RGraphSpace")

## ----Toy igraph - 1, eval=TRUE, message=FALSE, results=FALSE------------------
# Make a 'toy' igraph with 5 nodes and 4 edges;
# ..either a directed or undirected graph
gtoy1 <- make_star(5, mode="out")

# Check whether the graph is directed or not
is_directed(gtoy1)
## [1] TRUE

# Check graph size
vcount(gtoy1)
## [1] 5
ecount(gtoy1)
## [1] 4

# Assign 'x' and 'y' coordinates to each vertex;
# ..this can be an arbitrary unit in (-Inf, +Inf)
V(gtoy1)$x <- c(0, 2, -2, -4, -8)
V(gtoy1)$y <- c(0, 0,  2, -4,  0)

# Assign a name to each vertex
V(gtoy1)$name <- paste0("n", 1:5)

## ----Toy igraph - 2, eval=TRUE, message=FALSE, out.width="100%"---------------
# Plot the 'gtoy1' using standard R graphics
plot(gtoy1)

## ----Toy igraph - 3, eval=TRUE, message=FALSE, out.width="80%"----------------
# Plot the 'gtoy1' using RGraphSpace
plotGraphSpace(gtoy1, add.labels = TRUE)

## ----Node attributes, eval=TRUE, message=FALSE--------------------------------
# Node size (numeric in [0, 100], as '%' of the plot space)
V(gtoy1)$nodeSize <- c(8, 5, 5, 10, 5)

# Node shape (integer code between 0 and 25; see 'help(points)')
V(gtoy1)$nodeShape <- c(21, 22, 23, 24, 25)

# Node color (Hexadecimal or color name)
V(gtoy1)$nodeColor <- c("red", "#00ad39", "grey80", "lightblue", "cyan")

# Node line width (as in 'lwd' standard graphics; see 'help(gpar)')
V(gtoy1)$nodeLineWidth <- 1

# Node line color (Hexadecimal or color name)
V(gtoy1)$nodeLineColor <- "grey20"

# Node labels ('NA' will omit the label)
V(gtoy1)$nodeLabel <- c("V1", "V2", "V3", "V4", "V5")

# Node label size (in pts)
V(gtoy1)$nodeLabelSize <- 8

# Node label color (Hexadecimal or color name)
V(gtoy1)$nodeLabelColor <- "black"

# Node transparency (in [0,1])
V(gtoy1)$nodeAlpha <- 1

## ----Edge attributes - 1, eval=TRUE, message=FALSE----------------------------
# Edge width (as in 'lwd' standard graphics; see 'help(gpar)')
E(gtoy1)$edgeLineWidth <- 0.8

# Edge color (Hexadecimal or color name)
E(gtoy1)$edgeLineColor <- c("red","green","blue","black")

# Edge type (as in 'lty' standard graphics; see 'help(gpar)')
E(gtoy1)$edgeLineType <- c("solid", "11", "dashed", "2124")

# Edge transparency (in [0,1])
E(gtoy1)$edgeAlpha <- 1

## ----Edge attributes - 2, eval=TRUE, message=FALSE----------------------------
# Arrowhead types in directed graphs
## Integer or character code:
## 0 = "---", 1 = "-->", -1 = "--|"
E(gtoy1)$arrowType <- 1

## ----Edge attributes - 3, eval=TRUE, message=FALSE----------------------------
# Arrowhead types in undirected graphs
## Integer or character code:
##  0 = "---"
##  1 = "-->",  2 = "<--",  3 = "<->",  4 = "|->",
## -1 = "--|", -2 = "|--", -3 = "|-|", -4 = "<-|", 
E(gtoy1)$arrowType <- 1
# Note: in undirected graphs, this attribute overrides 
# the edge's orientation in the edge list

## ----A shortcut for RGraphSpace, eval=TRUE, message=FALSE, out.width="80%"----
# Plot the updated 'gtoy1' using RGraphSpace
plotGraphSpace(gtoy1, add.labels = TRUE)

## ----Using geoms, eval=TRUE, message=FALSE, out.width="80%"-------------------
data('gtoy1', package = 'RGraphSpace')
gs <- GraphSpace(gtoy1)
gs <- normalizeGraphSpace(gs)
ggplot(gs) + 
  geom_edgespace() + 
  geom_nodespace() + 
  geom_text(aes(x = x, y = y, label = nodeLabel), size = 2) + 
  theme_gspace_coords(is_norm = TRUE)

## ----label='Session information', eval=TRUE, echo=FALSE-----------------------
sessionInfo()

################################################################################
### Plotting adjusts for GraphSpace-class methods
################################################################################

#-------------------------------------------------------------------------------
.add_labels1 <- function(ggp, nodes, node.labels, label.size, label.color){
  node.labels <- node.labels[!duplicated(node.labels)]
  if (is.null(names(node.labels))) names(node.labels) <- node.labels
  names(node.labels) <- ifelse(is.na(names(node.labels)), 
    node.labels, names(node.labels))
  names(node.labels) <- ifelse(names(node.labels) == "",
    node.labels, names(node.labels))
  idx_df <- .label_idx(node.labels, nodes)
  if(any(is.na(idx_df$idx))){
    stop("All 'node.labels' should be annotated in the 'GraphSpace' object.",
      call. = FALSE)
  }
  nodes_ft <- nodes[idx_df$idx, , drop = FALSE]
  nodes_ft$ID <- idx_df$label
  x <- y <- ID <- NULL
  ggp <- ggp + ggplot2::geom_text(
    mapping = aes(x = x, y = y, label = ID), 
    data = nodes_ft, fontface = "bold", size.unit = "mm",
    size = label.size, colour = label.color, 
    vjust="center", hjust="center")
  return(ggp)
}
.label_idx <- function(node.labels, nodes){
  idx_df <- data.frame(name=node.labels, label=names(node.labels))
  idx_df$name1 <- match(node.labels, nodes$name)
  idx_df$name2 <- match(names(node.labels), nodes$name)
  idx_df$labl1  <- match(node.labels, nodes$nodeLabel)
  idx_df$labl2  <- match(names(node.labels), nodes$nodeLabel)
  idx_df$idx <- sapply(seq_len(nrow(idx_df)), function(i){
    idx <- idx_df[i,seq.int(3,6)]
    idx[!is.na(idx)][1]
  })
  idx_df <- idx_df[,c(1,2,7)]
  return(idx_df)
}

#-------------------------------------------------------------------------------
.add_labels2 <- function(ggp, nodes){
  vatt <- .get_default_vatt()
  nodes_ft <- nodes[!is.na(nodes$nodeLabel), ]
  x <- y <- nodeLabel <- NULL
  ggp <- ggp + geom_text(
    mapping = aes(x = x, y = y, label = nodeLabel), 
    data = nodes_ft, fontface = "bold",
    size = nodes_ft$nodeLabelSize %||% vatt$nodeLabelSize, 
    colour = nodes_ft$nodeLabelColor %||% vatt$nodeLabelColor, 
    size.unit = "pt", vjust="center", hjust="center")
  return(ggp)
}

#-------------------------------------------------------------------------------
.add_image <- function(ggp, image){
  ggp <- ggp + annotation_raster(image, xmin = 0, xmax = 1, ymin = 0, ymax = 1)
  return(ggp)
}

#-------------------------------------------------------------------------------
#' RGraphSpace ggplot2 themes
#'
#' A set of \pkg{ggplot2} themes used by \pkg{RGraphSpace} plots.
#' 
#' @param txt_size Numeric value to scale plot- and axis-related text elements.
#' @param leg_size Numeric value to scale legend-related elements.
#' @param bg_colour A color name or hex code specifying the panel background.
#' @param discrete_fill Logical; if TRUE, treats the fill legend as discrete 
#' to adjust key size.
#' @param discrete_colour Logical; if TRUE, treats the colour legend as discrete 
#' to adjust key size.
#' @param ... Additional arguments passed to \code{theme_gspace_th*} and
#' \link[ggplot2]{ggtheme}.
#' @details
#' \code{theme_gspace_th0()} is a minimal wrapper around 
#' \link[ggplot2]{theme_gray} that simplifies axis and legend scaling.
#' The \code{txt_size} and \code{leg_size} arguments aggregate related 
#' \link[ggplot2]{theme} parameters for quick thematic overrides.
#' 
#' \code{theme_gspace_th1()} builds on \code{theme_gspace_th0()} and 
#' modifies grid lines, axis appearance, and panel borders.
#'
#' \code{theme_gspace_th2()} is similar to \code{theme_gspace_th1()} with
#' simplified grid elements and a customizable panel background.
#'
#' \code{theme_gspace_th3()} is similar to \code{theme_gspace_th2()} but with
#' slightly adjusted margins, tick appearance, and legend formatting.
#'
#' The \code{theme_gspace_coords()} is a helper function that also adds axes
#' scales for normalized coordinates. It configures axis breaks, limits, 
#' and expansion for graph layouts. Plot coordinates are ideally normalized 
#' to the interval \code{[0, 1]}.
#' 
#' @return
#' \code{theme_gspace_th*()} return a \code{ggplot2} theme object.
#' 
#' \code{theme_gspace_coords()} returns a list containing scale and theme
#' components that can be added to a \pkg{ggplot2} plot.
#'
#' @seealso \link[ggplot2]{ggtheme}, \link[ggplot2]{theme}
#' 
#' @examples
#' library(RGraphSpace)
#' library(ggplot2)
#' 
#' ggplot(mtcars, aes(wt, mpg)) +
#'   geom_point() +
#'   theme_gspace_th0()
#'   
#' ggplot(mtcars, 
#'   aes(scales::rescale(wt), 
#'       scales::rescale(mpg))) +
#'   geom_point() +
#'   theme_gspace_coords("th2", is_norm = TRUE)
#'   
#' @importFrom ggplot2 "%+replace%" 
#' @name theme_gspace
#' @export
theme_gspace_th0 <- function(txt_size = 1, leg_size = 1, 
  bg_colour = "grey95", discrete_fill = FALSE,
  discrete_colour = FALSE, ...) {
  
  et0 <- ggplot2::element_text(size = 14 * txt_size)
  et1 <- ggplot2::element_text(size = 12 * txt_size)
  et2 <- ggplot2::element_text(size = 11 * txt_size)
  et3 <- ggplot2::element_text(size = 9 * leg_size)
  et4 <- ggplot2::element_text(size = 8 * leg_size)
  et1s <- ggplot2::element_text(size = 12 * txt_size, hjust = 0)
  
  l_th <- list()
  l_th[[1]] <- ggplot2::theme_gray() %+replace%
    ggplot2::theme(
      plot.title = et0, plot.subtitle = et1s,
      axis.title = et1, axis.text = et2,
      legend.title = et3, legend.text = et4,
      legend.key.spacing = unit(leg_size, "mm"),
      legend.spacing = unit(leg_size, "mm"),
      legend.key.size = unit(4 * leg_size, "mm"),
      legend.margin = ggplot2::margin(leg_size, leg_size, 
        leg_size, leg_size),
      legend.background = element_blank(),
      legend.box.background = element_blank(),
      panel.background = element_rect(fill = bg_colour, colour = NA),
      aspect.ratio = 1, complete = TRUE) +
    do.call(ggplot2::theme, .clean_args(...))
  
  l_args <- .guide_args(leg_size, discrete_fill, discrete_colour)
  
  if (length(l_args) > 0) {
    l_th[[2]] <- do.call(guides, l_args)
  }
  
  return(l_th)
  
}

#-------------------------------------------------------------------------------
#' @rdname theme_gspace
#' @export
theme_gspace_th1 <- function(txt_size = 1, leg_size = 1, 
  bg_colour = "grey95", discrete_fill = FALSE,
  discrete_colour = FALSE, ...) {
  
  et0 <- ggplot2::element_text(size = 14 * txt_size)
  et1 <- ggplot2::element_text(size = 12 * txt_size)
  et2 <- ggplot2::element_text(size = 11 * txt_size)
  et3 <- ggplot2::element_text(size = 9 * leg_size)
  et4 <- ggplot2::element_text(size = 8 * leg_size)
  et1s <- ggplot2::element_text(size = 12 * txt_size, hjust = 0)
  
  l_th <- list()
  l_th[[1]] <- ggplot2::theme_gray() %+replace%
    ggplot2::theme(
      plot.title = et0, plot.subtitle = et1s,
      axis.title = et1, axis.text = et2,
      legend.title = et3, legend.text = et4,
      legend.key.spacing = unit(leg_size, "mm"),
      legend.spacing = unit(leg_size, "mm"),
      legend.key.size = unit(4 * leg_size, "mm"),
      legend.margin = ggplot2::margin(leg_size, leg_size, 
        leg_size, leg_size),
      legend.background = element_blank(),
      legend.box.background = element_blank(),
      panel.background = element_rect(fill = bg_colour, colour = NA),
      panel.grid.minor = element_line(linewidth = 0.7),
      panel.grid.major = element_line(linewidth = 0.7),
      panel.border = element_rect(linewidth = 1.2),
      axis.ticks = element_line(linewidth = 0.7),
      axis.line = element_blank(),
      plot.margin = ggplot2::margin(1, 1, 1, 1), 
      aspect.ratio = 1, complete = TRUE) +
    do.call(ggplot2::theme, .clean_args(...))
  
  l_args <- .guide_args(leg_size, discrete_fill, discrete_colour)
  
  if (length(l_args) > 0) {
    l_th[[2]] <- do.call(guides, l_args)
  }
  
  return(l_th)
  
}

#-------------------------------------------------------------------------------
#' @rdname theme_gspace
#' @export
theme_gspace_th2 <- function(txt_size = 1, leg_size = 1, 
  bg_colour = "grey95", discrete_fill = FALSE, 
  discrete_colour = FALSE, ...) {
  
  et0 <- ggplot2::element_text(size = 14 * txt_size)
  et1 <- ggplot2::element_text(size = 12 * txt_size)
  et2 <- ggplot2::element_text(size = 11 * txt_size)
  et3 <- ggplot2::element_text(size = 9 * leg_size)
  et4 <- ggplot2::element_text(size = 8 * leg_size)
  et1s <- ggplot2::element_text(size = 12 * txt_size, hjust = 0)
  
  l_th <- list()
  l_th[[1]] <- ggplot2::theme_gray() %+replace%
    ggplot2::theme(
      plot.title = et0, plot.subtitle = et1s,
      axis.title = et1, axis.text = et2,
      legend.title = et3, legend.text = et4,
      legend.key.spacing = unit(leg_size, "mm"),
      legend.spacing = unit(leg_size, "mm"),
      legend.key.size = unit(4 * leg_size, "mm"),
      legend.margin = ggplot2::margin(leg_size, leg_size, 
        leg_size, leg_size),
      legend.background = element_blank(),
      legend.box.background = element_blank(),
      panel.background = element_rect(fill = bg_colour, colour = NA),
      panel.grid.minor = element_blank(),
      panel.grid.major = element_line(linewidth = 0.75),
      axis.ticks = element_line(linewidth = 0.7),
      axis.line = element_blank(), panel.border = element_blank(),
      plot.margin = ggplot2::margin(5, 10, 0, 10), 
      aspect.ratio = 1, complete = TRUE) +
    do.call(ggplot2::theme, .clean_args(...))
  
  l_args <- .guide_args(leg_size, discrete_fill, discrete_colour)
  
  if (length(l_args) > 0) {
    l_th[[2]] <- do.call(guides, l_args)
  }
  
  return(l_th)
  
}

#-------------------------------------------------------------------------------
#' @rdname theme_gspace
#' @export
theme_gspace_th3 <- function(txt_size = 1, leg_size = 1, 
  bg_colour = "grey95", discrete_fill = FALSE, 
  discrete_colour = FALSE, ...) {
  
  et0 <- ggplot2::element_text(size = 14 * txt_size)
  et1 <- ggplot2::element_text(size = 12 * txt_size)
  et2 <- ggplot2::element_text(size = 11 * txt_size)
  et3 <- ggplot2::element_text(size = 9 * leg_size)
  et4 <- ggplot2::element_text(size = 8 * leg_size)
  et1s <- ggplot2::element_text(size = 12 * txt_size, hjust = 0)
  
  l_th <- list()
  l_th[[1]] <- ggplot2::theme_gray() %+replace%
    ggplot2::theme(
      plot.title = et0, plot.subtitle = et1s,
      axis.title = et1, axis.text = et2,
      legend.title = element_text(size = 9 * leg_size, vjust = 1), 
      legend.text = element_text(size = 8 * leg_size, vjust = 1),
      legend.key.spacing = unit(leg_size, "mm"), 
      legend.spacing = unit(leg_size, "mm"),
      legend.key.size = unit(4 * leg_size, "mm"),
      legend.margin = ggplot2::margin(leg_size, leg_size, 
        leg_size, leg_size),
      legend.background = element_blank(),
      legend.box.background = element_blank(),
      legend.box.margin = ggplot2::margin(0, 0, 0, 0), 
      legend.position = "bottom",
      panel.background = element_rect(fill = bg_colour, colour = NA),
      panel.grid.minor = element_blank(),
      panel.grid.major = element_line(linewidth = 0.75),
      axis.ticks = element_line(linewidth = 0.5),
      axis.line = element_blank(), panel.border = element_blank(),
      plot.margin = ggplot2::margin(5, 5, 5, 5),
      aspect.ratio = 1, complete = TRUE) +
    do.call(ggplot2::theme, .clean_args(...))
  
  l_args <- .guide_args(leg_size, discrete_fill, discrete_colour)
  
  if (length(l_args) > 0) {
    l_th[[2]] <- do.call(guides, l_args)
  }
  
  return(l_th)
  
}

#' @param theme Character string specifying the GraphSpace theme variant. 
#' Options: `th0`, `th1`, `th2`, and `th3`.
#' @param is_norm Logical; if TRUE, assumes plot coordinates are 
#' already normalized in \code{[0, 1]}.
#' @param xlab The title for the 'x' axis.
#' @param ylab The title for the 'y' axis.
#' @param expand A range expansion factor applied to both the lower and 
#' upper limits of the 'x' and 'y' scales.
#' @rdname theme_gspace
#' @export
theme_gspace_coords <- function(theme = "th0", is_norm = FALSE,
  xlab = "Graph coordinates 1", ylab = "Graph coordinates 2", 
  expand = NULL, ...) {
  
  theme <- match.arg(theme, choices = c("th0", "th1", "th2", "th3"))
  bks <- .set_theme_bks(theme)
  bks$expand <- expand %||% bks$expand
  th <- list()
  if(is_norm){
    th[[length(th)+1]] <- scale_x_continuous(breaks = bks$axis.ticks,
      labels = format(bks$axis.ticks), position = bks$x.position,
      limits = bks$xylim, expand = ggplot2::expansion(mult = bks$expand))
    th[[length(th)+1]] <- scale_y_continuous(breaks = bks$axis.ticks,
      labels = format(bks$axis.ticks), limits = bks$xylim,
      expand = ggplot2::expansion(mult = bks$expand))
  } else {
    th[[length(th)+1]] <- scale_x_continuous(position = bks$x.position)
  }
  ftheme <- .get_gspace_theme(theme)
  th[[length(th)+1]] <- ftheme(...)
  th[[length(th)+1]] <- labs(x = xlab, y = ylab)
  attr(th, "gspace_pars") <- bks
  return(th)
}
.get_gspace_theme <- function(theme) {
  if (theme == "th3") {
    ftheme <- theme_gspace_th3
  } else if (theme == "th2") {
    ftheme <- theme_gspace_th2
  } else if (theme == "th1") {
    ftheme <- theme_gspace_th1 
  } else {
    ftheme <- theme_gspace_th0
  }
  return(ftheme)
}
.set_theme_bks <- function(theme){
  bks <- list()
  if (theme %in% c("th3")) {
    bks$axis.ticks <- c(0.25, 0.5, 0.75)
    bks$xylim <- c(0, 1)
    bks$expand <- 0.01
    bks$x.position <- "top"
    bks$justify <- "centre"
    bks$leg.position <- "bottom"
  } else if (theme %in% c("th2")) {
    bks$axis.ticks <- seq(0.1, 0.9, 0.2)
    bks$xylim <- c(0, 1)
    bks$expand <- 0.02
    bks$x.position <- "bottom"
    bks$justify <- "right"
    bks$leg.position <- "right"
  } else {
    bks$axis.ticks <- seq(0, 1, 0.2)
    bks$xylim <- c(0, 1)
    bks$expand <- 0.05
    bks$x.position <- "bottom"
    bks$justify <- "right"
    bks$leg.position <- "right"
  }
  return(bks)
}

#-------------------------------------------------------------------------------
#' @details
#' \code{theme_gspace_legend()} is helper function that adjusts legend text,
#' title, and key sizes by a single scaling factor.
#' 
#' @return
#' \code{theme_gspace_legend()} returns a list of theme and guide components.
#' 
#' @examples
#' # Reduce legend element sizes
#' ggplot(mtcars, aes(wt, mpg, fill = factor(cyl))) + 
#'   geom_point(shape = 21) + 
#'   theme_gspace_legend(0.8)
#'
#' @importFrom ggplot2 theme element_text unit guides guide_legend
#' @rdname theme_gspace
#' @export
theme_gspace_legend <- function(leg_size = 1, 
  discrete_fill = FALSE, discrete_colour = FALSE, ...) {
  
  et1 <- ggplot2::element_text(size = 9 * leg_size)
  et2 <- ggplot2::element_text(size = 8 * leg_size)
  
  l_th <- list()
  l_th[[1]] <- ggplot2::theme(
    legend.title = et1, legend.text = et2,
    legend.key.spacing = unit(2*leg_size, "mm"),
    legend.spacing = unit(leg_size, "mm"),
    legend.key.size = unit(4 * leg_size, "mm"),
    legend.margin = ggplot2::margin(2*leg_size, 2*leg_size, 
      2*leg_size, 2*leg_size), 
    legend.background = element_blank(),
    legend.box.background = element_blank(),
    legend.box.margin = ggplot2::margin(0, 0, 0, 0)) + 
    ggplot2::theme(...)
  
  l_args <- .guide_args(leg_size, discrete_fill, discrete_colour)
  
  if (length(l_args) > 0) {
    l_th[[length(l_th)+1]] <- do.call(guides, l_args)
  }
  
  return(l_th)
  
}

#-------------------------------------------------------------------------------
.guide_args <- function(leg_size = 1, discrete_fill = FALSE, 
  discrete_colour = FALSE) {
  l_args <- list()
  l_args$shape <- guide_legend(override.aes = list(size = 3 * leg_size))
  l_args$linetype <- guide_legend(
    theme = theme(legend.key.width = unit(10 * leg_size, "mm")),
    override.aes = list(size = 3 * leg_size))
  if (discrete_fill) {
    l_args$fill <- guide_legend(override.aes = list(size = 4 * leg_size))
  }
  if (discrete_colour) {
    l_args$colour <- guide_legend(override.aes = list(size = 4 * leg_size))
  }
  l_args
}


.clean_args <- function(...) {
  args <- list(...)
  args[c("key_fill", "key_colour", "bg_color")] <- NULL
  args
}

#-------------------------------------------------------------------------------
#' Convert objects to GraphSpace
#' 
#' S3 generic function for coercing objects into a \code{GraphSpace} object.
#' 
#' @details
#' Unified entry point for converting graph, spatial, and high-dimensional 
#' data into a \code{GraphSpace} object.
#' 
#' Graph objects are imported either through native methods or via
#' \link[tidygraph]{as_tbl_graph} when available.
#' 
#' @param x An object to be converted.
#' @param ... Additional arguments passed to methods.
#'
#' @return
#' A \code{GraphSpace} object.
#' 
#' @seealso
#' \code{\linkS4class{GraphSpace}}
#' 
#' @export
as.GraphSpace <- function(x, ...) {
  UseMethod("as.GraphSpace")
}

#-------------------------------------------------------------------------------
#' @rdname as.GraphSpace
#' @importFrom tidygraph as_tbl_graph
#' @export
as.GraphSpace.default <- function(x, ...) {
  
  y <- tryCatch( as_tbl_graph(x),
    error = function(e) NULL
  )
  
  if (!is.null(y)) {
    return(GraphSpace(y, ...))
  }
  
  rlang::abort(paste0(
    "No 'as.GraphSpace' method available for class ",
    paste(class(x), collapse = ", "),
    ", and conversion via 'tidygraph::as_tbl_graph()' failed."
  ))
  
}

#-------------------------------------------------------------------------------
#' @rdname as.GraphSpace
#' @export
as.GraphSpace.igraph <- function(x, ...) {
  GraphSpace(x, ...)
}

#-------------------------------------------------------------------------------
#' @rdname as.GraphSpace
#' @export
as.GraphSpace.tbl_graph <- function(x, ...) {
  GraphSpace(x, ...)
}

#-------------------------------------------------------------------------------
#' @rdname as.GraphSpace
#' @export
as.GraphSpace.data.frame <- function(x, ...) {
  GraphSpace(x, ...)
}

#-------------------------------------------------------------------------------
#' @rdname as.GraphSpace
#'
#' @param space Character specifying the coordinate space used for node
#' geometry. Either \code{"embedding"} or \code{"spatial"}. See details.
#' @param layer Name of the layer from which node data should be retrieved
#' (see \code{\link[SeuratObject]{LayerData}}).
#' @param ... Additional arguments passed to methods associated with the
#' selected \code{space}.
#' 
#' @details
#' For \strong{Seurat} objects, coordinate extraction depends on the 
#' selected \code{space}:
#' \itemize{
#'   \item \code{space = "embedding"} uses the first two dimensions returned by
#'   \code{\link[SeuratObject]{Embeddings}}.
#'   
#'   \item \code{space = "spatial"} uses tissue coordinates returned by
#'   \code{\link[SeuratObject]{GetTissueCoordinates}}.
#' }
#' 
#' Assay data are stored in the \code{data} slot of the resulting
#' \code{GraphSpace} object. Node metadata from \code{x@meta.data} are
#' appended to the node table.
#' 
#' @export
as.GraphSpace.Seurat <- function(x, 
  space = c("embedding","spatial"), 
  layer = NULL, ...) {
  
  space <- match.arg(space)
  
  if (!requireNamespace("SeuratObject", quietly = TRUE)) {
    rlang::abort("Package 'SeuratObject' is required for Seurat conversion.")
  }
  
  if (!inherits(x, "Seurat")){
    rlang::abort("'x' must be a Seurat object.")
  }
  
  # Get node coordinates
  if (space == "embedding"){
    
    coords <- SeuratObject::Embeddings(x, ...)
    
    if (is.null(coords) || length(dim(coords)) != 2L ||
        nrow(coords) == 0L || ncol(coords) < 2L) {
      rlang::abort(
        "Reduction must contain at least two dimensions."
      )
    }
    coords <- coords[, seq_len(2), drop = FALSE]
    colnames(coords) <- c("x", "y")
    coords <- as.data.frame(coords)
    
  } else {
    
    if (length(SeuratObject::Images(x)) == 0){
      rlang::abort("No spatial coordinates found.")
    }
    
    coords <- SeuratObject::GetTissueCoordinates(object = x, ...)
    
    if (is.null(coords) || (!is.matrix(coords) && !is.data.frame(coords)) ) {
      rlang::abort(
        "Spatial coordinates must be returned as a matrix or data frame."
      )
    }
    coords <- as.data.frame(coords)
    
    # Remove unnamed columns occasionally returned by some methods
    coords[, !nzchar(names(coords))] <- NULL
    
    if (nrow(coords) == 0L || ncol(coords) < 2L){
      rlang::abort(
        "No valid spatial coordinates found."
      )
    }
    
    if ( !all(c("x", "y") %in% colnames(coords))){
      rlang::abort(
        "Spatial coordinates must contain 'x' and 'y' columns."
      )
    }
    
  }
  
  if (is.null(rownames(coords))) {
    if (nrow(coords) != ncol(x)) {
      rlang::abort("Unable to infer node identifiers from coordinate space.")
    }
    rownames(coords) <- colnames(x)
  }
  
  # Add metadata
  metadata <- x[[]]
  if(inherits(metadata, "data.frame") && ncol(metadata) > 0){
    cids <- setdiff(colnames(metadata), colnames(coords))
    if (length(cids) > 0){
      metadata <- metadata[rownames(coords), cids, drop = FALSE]
      coords <- cbind(coords, metadata)
    }
  }
  
  # Create GraphSpace
  coords$nodeSize <- 1
  gs <- GraphSpace(coords)
  
  # Add fdata
  fdata <- SeuratObject::LayerData(x, layer = layer)
  fdata <- Matrix::t(fdata)
  gs_fdata(gs) <- fdata
  
  return(gs)
  
}

#' @importFrom methods setOldClass setClass
#' @importFrom grDevices as.raster
#' @importFrom igraph empty_graph
#' @importFrom tidygraph tbl_graph
setOldClass("raster")
setOldClass("igraph")
setOldClass("tbl_graph")
setOldClass("gs_graph")

#-------------------------------------------------------------------------------
#' @title GraphSpace: An S4 class for igraph objects
#'
#' @slot nodes A data frame containing node coordinates, attributes, and metadata.
#' @slot edges  A data frame containing edge relationships and attributes.
#' @slot graph An \code{\link[igraph]{igraph}} object representing the graph 
#' structure.
#' @slot image A \code{raster} object (see \code{\link[grDevices]{as.raster}}) used 
#' as background image.
#' @slot fdata A \code{\link[Matrix]{Matrix}} object storing high-dimensional 
#' feature data associated with graph nodes.
#' @slot pars A list with parameters.
#' @slot misc A list with intermediate objects for downstream methods.
#' @slot uuid A Universally Unique Identifier (UUID) for the object instance.
#' 
#' @method plotGraphSpace \link{plotGraphSpace}
#' @method getGraphSpace \link{getGraphSpace}
#' @aliases GraphSpace-class
#' @return An S4 class object.
#' @section Constructor:
#' see \code{\link{GraphSpace}} constructor.
#' @import igraph
#' @exportClass GraphSpace
#'
## Class GraphSpace
setClass("GraphSpace",
  slot = c(
    nodes = "data.frame",
    edges = "data.frame",
    graph = "igraph",
    image = "raster",
    fdata = "Matrix",
    pars = "list",
    misc = "list",
    uuid = "character"
  ),
  prototype = list(
    nodes = data.frame(),
    edges = data.frame(),
    graph = igraph::empty_graph(),
    image = as.raster(matrix()),
    fdata = Matrix::Matrix(nrow = 0, ncol = 0),
    pars = list(),
    misc = list(),
    uuid = character()
  )
)

setValidity("GraphSpace", function(object) {
  
  errors <- character()
  
  if (!is.data.frame(object@nodes)) {
    errors <- c(errors, "'@nodes' slot must be a data.frame.")
  }
  
  if (nrow(object@nodes) > 0 && is.null(rownames(object@nodes))) {
    errors <- c(errors, "'@nodes' slot must have row names.")
  }
  
  if (!is.data.frame(object@edges)) {
    errors <- c(errors, "'@edges' slot must be a data.frame.")
  }
  
  if (!is(object@fdata, "Matrix")) {
    errors <- c(errors, "'@fdata' slot  must be a Matrix object.")
  }
  
  if (nrow(object@fdata) > 0 && is.null(rownames(object@fdata))) {
    errors <- c(errors, "'@fdata' slot must have row names.")
  }
  
  if (ncol(object@fdata) > 0 && is.null(colnames(object@fdata))) {
    errors <- c(errors, "'@fdata' slot must have column names.")
  }
  
  # fdata <-> nodes consistency
  if (nrow(object@nodes) > 0 && nrow(object@fdata) > 0) {
    if (!identical(rownames(object@nodes), rownames(object@fdata))) {
      errors <- c(errors, "Row names in '@fdata' slot must match row names in '@nodes' slot.")
    }
  }
  
  # graph <-> nodes consistency
  if (!inherits(object@graph, "igraph")) {
    errors <- c(errors, "'@graph' slot must be an igraph object.")
  } else {
    if(igraph::vcount(object@graph) > 0){
      g_vertex_names <- igraph::V(object@graph)$name
      if (is.null(g_vertex_names)) {
        errors <- c(errors, "'@graph' slot must have a 'name' attribute.")
      } else if (nrow(object@nodes) > 0) {
        if (!setequal(rownames(object@nodes), g_vertex_names)) {
          errors <- c(errors, 
            "Vertex names in '@graph' slot must match row names in '@nodes' slot.")
        }
      }
    }
  }
  
  if (length(errors) == 0) TRUE else errors
  
})

#-------------------------------------------------------------------------------
#' Generate a unique identifier for GraphSpace objects
#' 
#' This helper function creates a unique ID without relying on the R 
#' Random Number Generator (RNG), making it immune to `set.seed()`.
#' It combines the Process ID (PID), high-precision system time, and 
#' a system-level temporary identifier to ensure uniqueness across 
#' parallel processes and rapid sequential calls.
#' 
#' @return A character string containing a unique alphanumeric ID.
#' @keywords internal
.generate_gs_uuid <- function() {
  # Capture current PID (uniqueness across different R sessions)
  pid <- Sys.getpid()
  
  # Capture high-precision time
  # We use %OS6 for microsecond precision
  time_stmp <- format(Sys.time(), "%d%H%M%OS6")
  
  # Capture a system-level unique string
  # tempfile() calls the OS to generate a unique name, bypassing R's RNG
  sys_id <- basename(tempfile(pattern = ""))
  
  # Combine and sanitize
  raw_id <- paste0("gs", pid, time_stmp, sys_id)
  uuid <- gsub("[^a-zA-Z0-9]", "", raw_id)
  
  return(uuid)
  
}

#-------------------------------------------------------------------------------
# show summary information on screen
setMethod("show", "GraphSpace", 
  function(object) {
    message("A GraphSpace-class object for:")
    summary(object@graph)
    nfeat <- ncol(object@fdata)
    if (nfeat > 0) {
      feat <- .gs_preview(colnames(object@fdata))
      cat(
        "+ features: ", nfeat, " (",
        paste(feat, collapse = ", "),
        ")\n",
        sep = ""
      )
    }
    invisible(object)
    
  }
)

.gs_preview <- function(x, n = 4) {
  if (length(x) == 0) return("<empty>")
  out <- head(x, n)
  if (length(x) > n) out <- c(out, "...")
  paste(out, collapse = ", ")
}

#-------------------------------------------------------------------------------
#' Normalize node coordinates to graph and image spaces
#' 
#' Accessory functions to normalize node coordinates in GraphSpace, 
#' either by centering them within the graph boundaries 
#' or by mapping them to pixel coordinates of a background image.
#' 
#' @param gs A \code{GraphSpace} object to be normalized.
#' @param mar A single numeric value in \code{[0, 0.5]} controlling the size of
#' the outer margins around the graph. Without an image, \code{mar} specifies
#' symmetric margins as a fraction of the graph space. With an image,
#' \code{mar} is interpreted as a fraction of the available image margins
#' surrounding the graph.
#' @param use_image Logical; if an image is available, whether to use it as 
#' a background reference map. When enabled, \code{x} and \code{y} graph 
#' coordinates are interpreted as pixel coordinates in the image matrix. 
#' Images can be inspected and assigned with \code{\link{gs_image}}.
#' @param flip.y Logical; whether to flip the node coordinates along the y-axis. 
#' Useful for aligning nodes with image backgrounds, which often use an 
#' inverted coordinate system.
#' @param flip.x Logical; whether to flip the node coordinates along the x-axis.
#' @param rotate.xy Logical; whether to rotate x-y coordinates.
#' @param flip.v Logical; whether to vertically flip the background image  
#' matrix (top-to-bottom) to align with the graph coordinate system.
#' @param flip.h Logical; whether to horizontally flip the background image  
#' matrix (left-to-right) to align with the graph coordinate system.
#' @param crop.coord An optional numeric vector of length four specifying a  
#' cropping region (xmin, xmax, ymin, ymax), with values in normalized 
#' coordinates \code{[0,1]}.
#' @param verbose A single logical value specifying to display detailed 
#' messages (when \code{verbose=TRUE}) or not (when \code{verbose=FALSE}).
#' @param image Deprecated from RGraphSpace 1.3.0; use \link{gs_image} instead.
#' 
#' @details 
#' These functions provide different strategies for coordinate transformation:
#' \itemize{
#'   \item \bold{normalizeGraphSpace}: Re-scales node coordinates to a 
#'   \code{[0, 1]} unit square based on the graph's bounding box (when 
#'   \code{use_image = FALSE}) or maps them to pixel coordinates (when 
#'   \code{use_image = TRUE} and an image is provided; see \link{gs_image}). 
#'   It handles image-to-graph alignment via \code{flip.\*} and 
#'   \code{rotate.\*} arguments, used to adjust the graph origin with the 
#'   image matrix layout. Users should be aware of the potential discrepancy 
#'   between image matrix orientation (top-down) and graph coordinates 
#'   (bottom-up). The function attempts to automatically adjust the y-axis to 
#'   align the graph's bottom-up coordinates with the image's top-down layout, 
#'   but further manual adjustments might be required. 
#'   \item \bold{cropGraphSpace}: Subsets the normalized graph space into a 
#'   specific region defined by \code{crop.coord}. 
#'   It recalculates node positions and background image boundaries to maintain 
#'   spatial consistency after cropping. This function requires a previously 
#'   normalized \code{GraphSpace} object.
#' }
#' 
#' @return A \code{GraphSpace} object with updated \code{nodes} 
#' and \code{image} slots.
#' 
#' @note This is an accessory function typically called during 
#' the preprocessing of \code{GraphSpace} objects before rendering.
#' 
#' @seealso \code{\link{gs_image}}
#' 
#' @examples
#' library(RGraphSpace)
#' library(igraph)
#' 
#' # Create a star graph
#' gtoy1 <- make_full_graph(30)
#' 
#' # Create a GraphSpace
#' gs <- GraphSpace(gtoy1)
#' 
#' gs <- normalizeGraphSpace(gs)
#' 
#' gs_crop <- cropGraphSpace(gs, 
#'          crop.coord = c(0, 0.75, 0, 0.75))
#' 
#' @aliases normalizeGraphSpace
#' @rdname normalizeGraphSpace-methods
#' @export
setMethod("normalizeGraphSpace", "GraphSpace",
  function(gs, mar = 0.1, use_image = FALSE, flip.x = FALSE, flip.y = FALSE, 
    rotate.xy = FALSE, flip.v = FALSE, flip.h = FALSE, verbose = TRUE, 
    image = deprecated()){
    
    if (lifecycle::is_present(image)) {
      deprecate_soft("1.3.0", "normalizeGraphSpace(image)", I("'gs_image(gs)<-image'"))
      gs_image(gs) <- image
      use_image <- TRUE
    }
    
    .validate_gs_args("singleLogical", "use_image", use_image)
    .validate_gs_args("singleLogical", "flip.v", flip.v)
    .validate_gs_args("singleLogical", "flip.h", flip.h)
    .validate_gs_args("singleLogical", "flip.x", flip.x)
    .validate_gs_args("singleLogical", "flip.y", flip.y)
    .validate_gs_args("singleLogical", "rotate.xy", rotate.xy)
    .validate_gs_args("singleLogical", "verbose", verbose)
    .validate_gs_args("singleNumber", "mar", mar)
    
    if (mar < 0 || mar > 0.5) {
      rlang::warn("'mar' should be in [0, 0.5]")
      mar <- max(0, min(mar, 0.5))
    }
    if(use_image && !gs@pars$image.layer){
      rlang::warn(
        message = c(
          "!" = "'use_image = TRUE' requested, but no image is available.",
          "i" = "Proceeding without the image layer.",
          "*" = "Use `gs_image()` to inspect the image slot.",
          "*" = "Use `gs_image()<-` to add an image."
        )
      )
      use_image <- FALSE
    }
    
    if(gs_vcount(gs)>0){
      if(use_image){
        gs <- .normalizeGraphSpace.image(gs, mar, flip.x, flip.y, 
          rotate.xy, flip.v, flip.h, verbose)
      } else {
        gs <- .normalizeGraphSpace.graph(gs, mar, flip.x, flip.y, 
          rotate.xy, verbose)
      }
    }
    
    return(gs)
    
  })

.normalizeGraphSpace.graph <- function(gs, mar, flip.x, flip.y, 
  rotate.xy, verbose){
  
  nodes <- .get_nodes(gs@graph)
  nodes <- .setCoordToGraph(nodes, flip.x, flip.y, rotate.xy, verbose)
  if(verbose) message("Normalizing node coordinates to graph space...")
  gs@nodes <- .fit_graph_space(nodes, mar)
  gs@pars$is.normalized <- TRUE
  gs@pars$flip.x <- flip.x
  gs@pars$flip.y <- flip.y
  gs@pars$rotate.xy <- rotate.xy
  gs@pars$mar <- mar
  
  return(gs)
  
}

.normalizeGraphSpace.image <- function(gs, mar, flip.x, flip.y, 
  rotate.xy, flip.v, flip.h, verbose){
  
  nodes <- .get_nodes(gs@graph)
  image <- gs@misc$image
  
  if(verbose) message("Normalizing node coordinates to image space...")
  
  if(!is.raster(image)) image <- as.raster(image)
  
  if(flip.v){
    if(verbose) message("Flipping image top-to-bottom...")  
    image <- image[rev(seq_len(nrow(image))), , drop = FALSE]
  } 
  
  if(flip.h){
    if(verbose) message("Flipping image left-to-right...")  
    image <- image[, rev(seq_len(ncol(image))), drop = FALSE]
  } 
  
  nodes <- .setCoordToImage(nodes, image, flip.x, flip.y, rotate.xy, verbose)
  
  l_temp <- .fitImageNodes(nodes, image, mar)
  
  gs@image <- l_temp$image
  gs@nodes <- l_temp$nodes
  gs@pars$is.normalized <- TRUE
  gs@pars$flip.v <- flip.v
  gs@pars$flip.x <- flip.x
  gs@pars$flip.y <- flip.y
  gs@pars$rotate.xy <- rotate.xy
  gs@pars$mar <- mar
  
  return(gs)
}

#-------------------------------------------------------------------------------
#' @aliases cropGraphSpace
#' @rdname normalizeGraphSpace-methods
#' @export
setMethod("cropGraphSpace", "GraphSpace", 
  function(gs, crop.coord = c(0, 1, 0, 1), verbose = TRUE){
    .validate_gs_args("numeric_vec", "crop.coord", crop.coord)
    if(length(crop.coord)!=4){
      rlang::abort("'crop.coord' should be a numeric vector of length = 4.")
    }
    if(any(crop.coord < 0) || any(crop.coord > 1)){
      rlang::abort("'crop.coord' should be in [0,1].")
    }
    if(!gs@pars$is.normalized){
      rlang::abort(
        message = c(
          "The 'GraphSpace' object must be normalized before cropping.",
          "i" = "Please run 'normalizeGraphSpace(gs)' first."
        )
      )
    }
    
    gs <- .crop_gspace(gs, crop.coord)
    
    return(gs)
    
  })

################################################################################
### Graph adjusts
################################################################################
.setCoordToGraph <- function(nodes, flip.x = FALSE, flip.y = FALSE, 
  rotate.xy= FALSE, verbose = TRUE){
  
  # Rotated coordinates
  coord <- nodes[,c("x","y")]
  if(rotate.xy){
    if(verbose) message("Rotating xy-coordinates...")
    coord$x2 <- coord$y
    coord$y2 <- coord$x
  } else {
    coord$x2 <- coord$x
    coord$y2 <- coord$y
  }
  
  # Flip y-coordinates
  if(flip.y){
    if(verbose) message("Flipping y-coordinates...")
    y <- coord$y2
    coord$y2 <- -(y - max(y)) - max(y) + 1
  }
  
  # Flip x-coordinates
  if(flip.x){
    if(verbose) message("Flipping x-coordinates...")
    x <- coord$x2
    coord$x2 <- -(x - max(x)) - max(x) + 1
  }
  # Update coordinates
  nodes$x <- coord$x2
  nodes$y <- coord$y2
  
  return(nodes)
}

################################################################################
### Adjust node coordinates
################################################################################
# Fit graph in a [0, 1] space with focus on adjusting margins
.fit_graph_space <- function(nodes, mar = 0.1){
  
  mar <- max(0, min(mar, 0.49))
  
  nds <- nodes
  
  if(nrow(nds)>0){
    
    nds$x <- nds$x - mean(range(nds$x))
    nds$y <- nds$y - mean(range(nds$y))
    
    from <- range(c(nds$x, nds$y))
    to <- c(mar, 1 - mar)
    
    nds$x <- scales::rescale(nds$x, from = from, to=to)
    nds$y <- scales::rescale(nds$y, from = from, to=to)
    
  }
  
  return(nds)
}

################################################################################
### Graph-to-image adjusts
################################################################################
.setCoordToImage <- function(nodes, image, 
  flip.x = FALSE, flip.y = FALSE, rotate.xy = FALSE, 
  verbose = TRUE){
  
  # Rotated coordinates
  coord <- nodes[,c("x","y")]
  if(rotate.xy){
    if(verbose) message("Rotating xy-coordinates...")
    coord$x2 <- coord$y
    coord$y2 <- coord$x
  } else {
    coord$x2 <- coord$x
    coord$y2 <- coord$y
  }
  
  if(flip.y){
    if(verbose) message("Flipping y-coordinates...")
  } else {
    # Flip 'y' by default if an image is provided
    y <- coord$y2
    y <- -(y - max(y)) + nrow(image) - max(y) + 1
    coord$y2 <- y
  }
  
  # Flip x-coordinates over image axis
  if(flip.x){
    if(verbose) message("Flipping x-coordinates...")
    x <- coord$x2
    x <- -(x - max(x)) + ncol(image) - max(x) + 1
    coord$x2 <- x
  }
  
  # Update coordinates
  nodes$x <- coord$x2
  nodes$y <- coord$y2
  
  .check_final_coords(nodes, image)
  
  return(nodes)
  
}

#-------------------------------------------------------------------------------
.check_final_coords <- function(nodes, image){
  
  d <- dim(image)
  xr <- range(nodes$x, na.rm = TRUE)
  yr <- range(nodes$y, na.rm = TRUE)
  
  xr_int <- c(floor(xr[1]), ceiling(xr[2]))
  yr_int <- c(floor(yr[1]), ceiling(yr[2]))
  
  out_x <- (xr_int[1] < 1) || (xr_int[2] > d[2])
  out_y <- (yr_int[1] < 1) || (yr_int[2] > d[1])
  
  if( out_x || out_y ){
    msg <- "Graph coordinates outside the image boundaries."
    
    ms_i <- c("i" = "Note: node 'x' or 'y' coordinates are treated as indices of the image matrix.")
    ms_x <- c("x" = sprintf(
      "Observed ranges: x[%s, %s], y[%s, %s]. Image dimensions: %s x %s (yx).",
      xr_int[1], xr_int[2], yr_int[1], yr_int[2], d[1], d[2]
    ))
    ms_a <- c("*" = "Try adjusting 'flip' and 'rotate' options in `normalizeGraphSpace()`.")
    ms_f <- "See `vignette('RGraphSpace')` for more information on coordinate normalization."
    
    rlang::abort(message = msg, 
      body = c(ms_i, ms_x, ms_a), footer = ms_f, 
      call = rlang::caller_env())
  }
  
  invisible(TRUE)
  
}

################################################################################
### Adjust image to node coordinates
################################################################################

#-------------------------------------------------------------------------------
.fitImageNodes <- function(nodes, image, mar = 0.1){
  
  l_temp <- .fit_image_nodes(nodes, image, mar)
  l_temp <- .adjust_aspect_ratio(l_temp)
  l_temp <- .normalize_image_nodes(l_temp)
  
  return(l_temp)
}

#-------------------------------------------------------------------------------
# Fit image to nodes with focus on adjusting graph margins
.fit_image_nodes <- function(nodes, image, mar = 0.1) {
  
  nds <- nodes
  img <- image
  d <- dim(img)
  mar <- max(0, min(mar, 0.49))
  
  # bounding box around nodes
  xl_nds <- range(nds$x)
  yl_nds <- range(nds$y)
  center_x <- mean(xl_nds)
  center_y <- mean(yl_nds)
  
  # target dimension centered on nodes;
  # side_length is calculated so that max_d is 
  # exactly (1 - 2*mar) of the total width
  max_d <- max(diff(xl_nds), diff(yl_nds))
  side_length <- max_d / (1 - 2 * mar)
  half_side <- side_length / 2
  
  # initial crop coordinates
  x_start <- center_x - half_side
  x_end   <- x_start + side_length
  y_start <- center_y - half_side
  y_end   <- y_start + side_length
  
  # shift crop coordinates to the image boundaries
  if (x_start < 1) { 
    shift <- 1 - x_start
    x_start <- 1
    x_end <- min(d[2], x_end + shift)
  }
  if (x_end > d[2]) { 
    shift <- x_end - d[2]
    x_end <- d[2]
    x_start <- max(1, x_start - shift)
  }
  if (y_start < 1) { 
    shift <- 1 - y_start
    y_start <- 1
    y_end <- min(d[1], y_end + shift)
  }
  if (y_end > d[1]) { 
    shift <- y_end - d[1]
    y_end <- d[1]
    y_start <- max(1, y_start - shift)
  }
  
  # force the limits to include the node bounding box
  x_start <- max(1, min(x_start, xl_nds[1]))
  x_end   <- min(d[2], max(x_end, xl_nds[2]))
  y_start <- max(1, min(y_start, yl_nds[1]))
  y_end   <- min(d[1], max(y_end, yl_nds[2]))
  
  # convert to indices
  x_s_idx <- floor(x_start)
  x_e_idx   <- ceiling(x_end)
  y_s_idx <- floor(y_start)
  y_e_idx   <- ceiling(y_end)
  
  # final validity check
  x_s_idx <- max(1L, x_s_idx)
  x_e_idx   <- min(d[2], x_e_idx)
  y_s_idx <- max(1L, y_s_idx)
  y_e_idx   <- min(d[1], y_e_idx)
  
  # execute crop on flipped image
  img_res <- img[seq.int(d[1], 1), ]
  img_res <- img_res[seq.int(y_s_idx, y_e_idx), seq.int(x_s_idx, x_e_idx)]
  img_res <- img_res[seq.int(nrow(img_res), 1), ]
  
  # update node coordinates
  nds$x <- nds$x - x_s_idx + 1
  nds$y <- nds$y - y_s_idx + 1
  
  # calculate final side_length
  x_side_length <- x_end - x_start
  y_side_length <- y_end - y_start
  x_side_length <- x_side_length + (x_start - x_s_idx)
  y_side_length <- y_side_length + (y_start - y_s_idx)
  side_length <- max(x_side_length, y_side_length)
  
  return(list(nodes = nds, image = img_res, 
    side_length = side_length))
}

#-------------------------------------------------------------------------------
.adjust_aspect_ratio <- function(l_temp){
  d <- dim(l_temp$image)
  if(d[1] > d[2]){
    n <- ceiling( (d[1] - d[2]) )/2
    img_d <- matrix(NA, nrow = d[1], ncol = d[1])
    img_d[ , seq(n + 1, n + d[2])] <- as.matrix(l_temp$image)
    l_temp$nodes$x <- l_temp$nodes$x + n
    l_temp$image  <- as.raster(img_d)
  } else if(d[1] < d[2]){
    n <- ceiling( (d[2] - d[1])/2 )
    img_d <- matrix(NA, nrow = d[2], ncol = d[2])
    img_d[seq(n + 1, n + d[1]), ] <- as.matrix(l_temp$image)
    l_temp$nodes$y <- l_temp$nodes$y + n
    l_temp$image  <- as.raster(img_d)
  }
  return(l_temp)
}

#-------------------------------------------------------------------------------
.normalize_image_nodes <- function(l_temp){
  d <- dim(l_temp$image)
  l_temp$nodes$x <- .rescale_direct(l_temp$nodes$x, d[2], 0.5 / d[2])
  l_temp$nodes$y <- .rescale_direct(l_temp$nodes$y, d[1], 0.5 / d[1])
  return(l_temp)
}
.rescale_direct <- function(x, n, half_pixel) {
  ((x - 1) / (n - 1)) * (1 - 2 * half_pixel) + half_pixel
}

################################################################################
### Crop graph and image
################################################################################

#-------------------------------------------------------------------------------
.crop_gspace <- function(gs, crop.coord) {
  if (gs@pars$image.layer) {
    gs <- .crop_gspace_image(gs, crop.coord)
  } else {
    gs <- .crop_gspace_graph(gs, crop.coord)
  }
  return(gs)
}

#-------------------------------------------------------------------------------
.crop_gspace_graph <- function(gs, crop.coord) {
  
  xmin <- crop.coord[1]; xmax <- crop.coord[2]
  ymin <- crop.coord[3]; ymax <- crop.coord[4]
  
  # Filter nodes within the crop window
  nodes <- gs@nodes
  cx <- nodes$x >= xmin & nodes$x <= xmax
  cy <- nodes$y >= ymin & nodes$y <= ymax
  nodes <- nodes[which(cx & cy), ]
  
  # Rescale node coordinates relative to the crop window
  # Maps [xmin, xmax] -> [0, 1] and [ymin, ymax] -> [0, 1]
  nodes$x <- (nodes$x - xmin) / (xmax - xmin)
  nodes$y <- (nodes$y - ymin) / (ymax - ymin)
  
  gs <- .crop_update_graph(gs, nodes)
  
  return(gs)
  
}

#-------------------------------------------------------------------------------
.crop_gspace_image <- function(gs, crop.coord) {
  
  xmin <- crop.coord[1]; xmax <- crop.coord[2]
  ymin <- crop.coord[3]; ymax <- crop.coord[4]
  
  # Filter nodes within the crop window
  nodes <- gs@nodes
  
  # Compute image crop indices
  nrow_mat <- nrow(gs@image)
  ncol_mat <- ncol(gs@image)
  col_s <- max(1L, ceiling(xmin * ncol_mat))
  col_e <- min(ncol_mat, floor(xmax * ncol_mat))
  row_s <- max(1L, ceiling((1 - ymax) * nrow_mat))
  row_e <- min(nrow_mat, floor((1 - ymin) * nrow_mat))
  d <- c(row_e - row_s + 1L, col_e - col_s + 1L)
  
  # Reverse pixel-center encoding to recover 1-based pixel indices
  # x: shift to crop-window origin (left edge)
  nodes$x <- .rescale_direct_inv(nodes$x, ncol_mat, 0.5 / ncol_mat) - (col_s - 1)
  # y: shift to crop-window origin (top edge, image convention)
  nodes$y <- .rescale_direct_inv(nodes$y, nrow_mat, 0.5 / nrow_mat) - (nrow_mat - row_e)
  
  # Re-encode to pixel centers in the cropped image
  nodes$x <- .rescale_direct(nodes$x, d[2], 0.5 / d[2])
  nodes$y <- .rescale_direct(nodes$y, d[1], 0.5 / d[1])
  
  # Crop nodes
  cx <- nodes$x >= 0 & nodes$x <= 1
  cy <- nodes$y >= 0 & nodes$y <= 1
  nodes <- nodes[which(cx & cy), ]
  
  # Crop image
  gs@image <- gs@image[row_s:row_e, col_s:col_e, drop = FALSE]
  
  gs <- .crop_update_graph(gs, nodes)
  
  return(gs)
  
}

#-------------------------------------------------------------------------------
.rescale_direct_inv <- function(x, n, half_pixel) {
  ((x - half_pixel) / (1 - 2 * half_pixel)) * (n - 1) + 1
}

#-------------------------------------------------------------------------------
.crop_update_graph <- function(gs, nodes) {
  
  # Remove edges whose endpoints are no longer in the node set
  idx <- (gs@edges$name1 %in% nodes$name) &
    (gs@edges$name2 %in% nodes$name)
  gs@edges <- gs@edges[idx, ]
  
  # Update graph vertices
  idx <- V(gs@graph)$name %in% rownames(nodes)
  gs@graph <- igraph::delete_vertices(gs@graph, which(!idx))
  idx <- match(rownames(nodes), V(gs@graph)$name)
  V(gs@graph)$x[idx] <- nodes$x
  V(gs@graph)$y[idx] <- nodes$y
  gs@nodes <- nodes
  
  return(gs)
  
}

setGeneric("GraphSpace", function(g, ...) standardGeneric("GraphSpace"))

setGeneric("plotGraphSpace", function(gs, ...)
  standardGeneric("plotGraphSpace"),
  package = "RGraphSpace"
)

setGeneric("getGraphSpace", function(gs, what = "summary") 
  standardGeneric("getGraphSpace"), 
  package = "RGraphSpace"
)

setGeneric("normalizeGraphSpace", function(gs, ...) 
  standardGeneric("normalizeGraphSpace"),
  package = "RGraphSpace"
)

setGeneric("cropGraphSpace", function(gs, ...) 
  standardGeneric("cropGraphSpace"),
  package = "RGraphSpace"
)

setGeneric("gs_graph", function(x) 
  standardGeneric("gs_graph"),
  package = "RGraphSpace"
)

setGeneric("gs_names", function(x) 
  standardGeneric("gs_names"),
  package = "RGraphSpace"
)

setGeneric("gs_nodes", function(x, ...) 
  standardGeneric("gs_nodes"),
  package = "RGraphSpace"
)

setGeneric("gs_edges", function(x) 
  standardGeneric("gs_edges"),
  package = "RGraphSpace"
)

setGeneric("gs_image", function(x) 
  standardGeneric("gs_image"),
  package = "RGraphSpace"
)

setGeneric("gs_image<-", function(x, value) 
  standardGeneric("gs_image<-"),
  package = "RGraphSpace"
)

setGeneric("gs_fdata", function(x) 
  standardGeneric("gs_fdata"),
  package = "RGraphSpace"
)

setGeneric("gs_fdata<-", function(x, value)
  standardGeneric("gs_fdata<-"),
  package = "RGraphSpace"
)

setGeneric("gs_nfeatures", function(x) 
  standardGeneric("gs_nfeatures"),
  package = "RGraphSpace"
)

setGeneric("gs_features", function(x) 
  standardGeneric("gs_features"),
  package = "RGraphSpace"
)

setGeneric("gs_vcount", function(x) 
  standardGeneric("gs_vcount"),
  package = "RGraphSpace"
)

setGeneric("gs_ecount", function(x) 
  standardGeneric("gs_ecount"),
  package = "RGraphSpace"
)

setGeneric("gs_vertex_attr", function(x, ...) 
  standardGeneric("gs_vertex_attr"),
  package = "RGraphSpace"
)

setGeneric("gs_vertex_attr<-", function(x, ..., value) 
  standardGeneric("gs_vertex_attr<-"),
  package = "RGraphSpace"
)

setGeneric("gs_edge_attr", function(x, ...) 
  standardGeneric("gs_edge_attr"),
  package = "RGraphSpace"
)

setGeneric("gs_edge_attr<-", function(x, ..., value) 
  standardGeneric("gs_edge_attr<-"),
  package = "RGraphSpace"
)

#-------------------------------------------------------------------------------
#' @title Draw edge elements in a 2D graph layout
#' 
#' @description
#' 
#' Constructor for \link{GeomEdgeSpace} ggproto objects.
#' 
#' A wrapper around \link[ggplot2]{geom_segment} that enables direct use of
#' edge attributes stored in \link{GraphSpace} objects as aesthetics.
#'
#' This \code{geom} is designed to create edge-level aesthetics such as
#' \code{colour} and \code{linewidth}, or any custom aesthetics defined 
#' in \link{GeomEdgeSpace}.
#'
#' @param mapping Set of aesthetic mappings created by [ggplot2::aes()].
#' These mappings override global aesthetics and are not inherited 
#' from the top-level plot.
#'
#' @param data The data to be displayed in this layer. It can be a 
#' \link{GraphSpace} object, an \link[igraph]{igraph} object, or the 
#' \code{edgespace_handler()} closure. When \code{NULL} (default),
#' a handler is created internally.
#'
#' @param stat The statistical transformation to use on the data.
#' Defaults to \code{identity}.
#'
#' @param position Position adjustment, either as a string or
#' the result of a call to a position adjustment function.
#'
#' @param ... Additional parameters passed to the underlying
#' drawing function in \link{GeomEdgeSpace}.
#'
#' @param na.rm Logical. Should missing values be removed?
#' Defaults to \code{FALSE}.
#' 
#' @param show.legend Logical or a named logical vector indicating
#' whether this layer should be included in legends.
#'
#' @param inherit.aes Logical. If \code{FALSE} (default), the layer will use 
#' aesthetics defined in \code{mapping}.
#' 
#' @param arrow_size Numeric scaling factor controlling arrowhead 
#' geometry (see 'drawing' section).
#' 
#' @param arrow_offset Numeric value controlling the base offset of arrows  
#' at edge endpoints (see 'drawing' section).
#'
#' @param lineend Line end style (round, butt, square). Supplied for 
#' compatibility with \link[ggplot2]{geom_segment}.
#' 
#' @param linejoin Line join style (round, mitre, bevel). Supplied for 
#' compatibility with \link[ggplot2]{geom_segment}.
#' 
#' @param raster Logical. Should node glyphs be rasterized? 
#' Rasterization support is based on \code{\link[ggrastr]{rasterise}}.
#' 
#' @param dpi Numeric. Rasterization resolution.
#' 
#' @param dev Character. Rasterization backend. One of `"cairo"`,
#' `"ragg"`, `"ragg_png"`, or `"cairo_png"`.
#' 
#' @param scale Numeric. Rasterization scaling factor
#' (see \code{\link[ggrastr]{rasterise}}).
#' 
#' @return A ggplot2 layer that renders edge segments defined by
#' \link{GeomEdgeSpace}.
#' 
#' @section Aesthetics:
#' 
#' \code{geom_edgespace()} understands \link[ggplot2]{geom_segment} aesthetics.
#' 
#' If these aesthetics are not explicitly provided in \code{aes()}, they 
#' are automatically retrieved from the \link{GraphSpace} object.
#'
#' \tabular{ll}{
#'   \strong{\code{x}, \code{y}, \code{xend}, \code{yend}} \tab Required (automatically supplied).\cr
#'   \code{colour} \tab Edge colour (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
#'   \code{alpha} \tab Transparency (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
#'   \code{linetype} \tab Edge line type (see \link[ggplot2]{aes_linetype_size_shape}).\cr
#'   \code{linewidth} \tab Edge line width (see \link[ggplot2]{aes_linetype_size_shape}).
#' }
#' 
#' Required aesthetics (\code{x}, \code{y}, \code{xend}, \code{yend}, ...)  
#' are supplied from the \link{GraphSpace} object and do not need to be 
#' manually mapped.
#' 
#' Additional parameters can be passed to control fixed values for the layer.
#' For example: `colour = "grey"`, `linetype = 2`, `linewidth = 1`.
#' 
#' Arrows can be further adjusted by \code{arrow_size} and \code{arrow_offset} 
#' arguments (see *details*).
#' 
#' @section Integration with ggraph:
#' 
#' \code{geom_edgespace} is compatible with the \code{ggraph} methods.
#' When used within a \code{ggraph()} call, the default \code{edgespace_handler()} 
#' automatically:
#' \itemize{
#'   \item Identifies the current \code{layout_ggraph}.
#'   \item Extracts the \code{x} and \code{y} coordinates calculated by \code{ggraph}.
#'   \item Reconstructs a temporary \code{GraphSpace} object to inject spatial 
#'   metadata and user-chosen \code{ggraph} layout.
#' }
#' 
#' @details
#' 
#' **arrow_size** is a numeric scaling factor controlling arrowhead geometry. 
#' The value is interpreted in the same numeric space as line width (`lwd`), 
#' ensuring consistent scaling between edge strokes and arrowheads.
#' 
#' **arrow_offset** is an additive term that offsets arrow endpoints 
#' uniformly in graph space and is bounded by the edge length, in NPC units.
#' 
#' Arrowhead types are specified in the \link{GraphSpace} constructor.
#' 
#' @seealso
#' \link{GraphSpace}, \link{geom_nodespace}, \link{geom_graphspace}, 
#' \link[ggplot2]{geom_segment}
#'
#' @examples
#' library(RGraphSpace)
#' library(igraph)
#' library(ggplot2)
#' 
#' # Load a demo igraph
#' data('gtoy1', package = 'RGraphSpace')
#' 
#' # Create a GraphSpace object
#' gs <- GraphSpace(gtoy1)
#' 
#' \dontrun{
#' 
#' ggplot() +
#'   geom_edgespace(data = gs) +
#'   geom_nodespace(data = gs) +
#'   theme(aspect.ratio = 1)
#' 
#' }
#' 
#' @export
geom_edgespace <- function(mapping = NULL, data = NULL,
  stat = StatEdgeSpace, position = "identity", ..., 
  na.rm = FALSE, show.legend = NA, inherit.aes = FALSE,
  arrow_size = 1, arrow_offset = 0.01,
  lineend = "butt", linejoin = "mitre",
  raster = FALSE, dpi = NULL, dev = "cairo", scale = 1) {
  
  # Check custom params
  .validate_gs_args("singleLogical", "na.rm", na.rm)
  .validate_gs_args("singleNumber", "arrow_size", arrow_size)
  .validate_gs_args("singleNumber", "arrow_offset", arrow_offset)
  .validate_gs_args("singleString", "lineend", lineend)
  .validate_gs_args("singleString", "linejoin", linejoin)
  
  if (is.null(data)){
    data <- edgespace_handler()
  } else if (!inherits(data, "edgespace_handler")){
    if (is.function(data)){
      rlang::abort(
        message = c(
          "x" = "Invalid handler function provided to `data`.",
          "*" = "Use `edgespace_handler()` to create a compatible handler."
        )
      )
    }
    data <- edgespace_handler()(data)
  }
  
  mapping <- .mapping_edgespace(mapping)
  
  params <- list2(
    na.rm = na.rm, 
    arrow_size = arrow_size,
    arrow_offset = arrow_offset,
    lineend = lineend,
    linejoin = linejoin,
    .size_unit = "mm",
    .nodes = NULL, 
    raster = raster, 
    dpi = dpi, 
    dev = dev, 
    scale = scale,
    ...)
  
  ggplot2::layer(
    geom = GeomEdgeSpace,
    stat = stat,
    mapping = mapping,
    data = data,
    position = position,
    show.legend = show.legend,
    inherit.aes = inherit.aes,
    params = params
  )
  
}

#-------------------------------------------------------------------------------
#' Attribute Processing for GeomEdgeSpace
#'
#' Manage visual attribute precedence (color, size, shape) for `GeomEdgeSpace` 
#' objects.
#'
#' @section Attribute Priority:
#' 1. Explicit `aes()` mappings.
#' 2. Fixed `geom_edgespace()` arguments.
#' 3. Original graph attributes (via `optional_aes`).
#' 
#' During the `setup_data` stage, the Stat invokes internal functions 
#' to resolve value priority:
#' \enumerate{
#'   \item **Explicit Mapping**: Values defined by the user inside `aes()`.
#'   \item **Fixed Parameters**: Constant values passed as arguments in the `geom_edgespace()` call.
#'   \item **Graph Attributes**: Original attributes stored within the GraphSpace 
#'   object, retrieved from the data columns.
#' }
#'
#' @format A \code{ggproto} object.
#' @seealso \code{\link{geom_edgespace}}
#' @export
StatEdgeSpace <- ggproto(
  "StatEdgeSpace", ggplot2::Stat,
  optional_aes = c("edgeLineColor", "edgeLineWidth", "edgeLineType", "edgeAlpha"),
  setup_data = function(data, params) {
    data <- .params_edgespace(params, data)
    return(data)
  },
  compute_panel = function(data, scales){
    return(data)
  }
)

#-------------------------------------------------------------------------------
#' @rdname geom_edgespace
#' @export
edgespace_handler <- function() {
  
  fn <- function(data) {
    
    if (is_waiver(data)) return(NULL)
    
    if ( inherits(data, c("igraph", "layout_ggraph")) ) {
      data <- gs_edges(GraphSpace(data, verbose = FALSE))
    } else if (inherits(data, "GraphSpace")){
      data <- gs_edges(data)
    } else if (inherits(data, "gs_nodes")){
      if(inherits(attr(data, ".gs_graph"), "GraphSpace")){
        data <- gs_edges(attr(data, ".gs_graph"))
      } else {
        rlang::warn(
          message = c(
            "x" = "`edgespace_handler()` found no edges in the input data.",
            "i" = "Input must be a 'GraphSpace', 'gs_edges', 'igraph', 'tidygraph', or 'ggraph' layout."
          )
        )
        data <- NULL
      }
    } else if (!inherits(data, "gs_edges")){
      rlang::abort(
        message = c(
          "x" = "`edgespace_handler()` received an unsupported object type.",
          "i" = "Input must be a 'GraphSpace', 'gs_edges', 'igraph', 'tidygraph', or 'ggraph' layout."
        )
      )
    }
    
    return(data)
    
  }
  
  attr(fn, "gs_handler_type") <- "edge"
  
  class(fn) <- c("edgespace_handler", class(fn))
  
  return(fn)
  
}

#-------------------------------------------------------------------------------
.mapping_edgespace <- function(mapping) {
  
  x <- y <- xend <- yend <- arrowType <- NULL
  
  offset_start <- offset_end <- vertex1 <- vertex2 <- NULL
  
  edgeLineColor <- edgeLineWidth <- edgeLineType <- edgeAlpha <- NULL
  
  default_mapping <- ggplot2::aes(
    x = x, y = y, xend = xend, yend = yend,
    offset_start = offset_start,
    offset_end = offset_end,
    arrowType = arrowType, 
    vertex1 = vertex1, vertex2 = vertex2
  )
  
  optional_mapping <- ggplot2::aes(
    edgeLineColor = edgeLineColor, 
    edgeLineWidth = edgeLineWidth,
    edgeLineType = edgeLineType,
    edgeAlpha = edgeAlpha)
  
  if (is.null(mapping)) {
    mapping <- utils::modifyList(
      default_mapping, optional_mapping)
  } else {
    mapping <- utils::modifyList(utils::modifyList(
      default_mapping, optional_mapping), mapping)
  }
  return(mapping)
}

#-------------------------------------------------------------------------------
.params_edgespace <- function(params, edges, mapping){
  
  # Note: 'mapping' is read from colnames(edges) because, at this
  # stage, ggplot2 has already evaluated aes() and pruned unmapped
  # columns. Any standard aesthetic name present here (e.g. "colour") 
  # was explicitly mapped by the user.
  if(missing(mapping)){
    mapping <- colnames(edges)
  } else {
    mapping <- names(mapping)
  }
  
  if(is.null(params[["colour"]]) &&  !"colour" %in% mapping){
    if("edgeLineColor" %in% names(edges) ){
      edges[["colour"]] <- edges[["edgeLineColor"]]
    }
  }
  
  if(is.null(params[["linewidth"]]) && !"linewidth" %in% mapping ){
    if("edgeLineWidth" %in% names(edges) ){
      edges[["linewidth"]] <- edges[["edgeLineWidth"]]
    }
  }
  
  if(is.null(params[["linetype"]]) && !"linetype" %in% mapping ){
    if("edgeLineType" %in% names(edges) ){
      edges[["linetype"]] <- edges[["edgeLineType"]]
    }
  }
  
  if(is.null(params[["alpha"]]) && !"alpha" %in% mapping ){
    if("edgeAlpha" %in% names(edges) ){
      edges[["alpha"]] <- edges[["edgeAlpha"]]
    }
  }
  
  return(edges)
  
}

#-------------------------------------------------------------------------------
#' @title GeomEdgeSpace: a ggplot2 prototype for GraphSpace-class methods
#'
#' @description
#' 
#' \code{GeomEdgeSpace} is the underlying \link[ggplot2]{ggproto} object 
#' used by \link{geom_edgespace} to draw edge elements in a graph layout. 
#'
#' This geom is designed for network diagrams, where graph attributes 
#' are often already in their final form (e.g., hex colors).
#' 
#' @section Aesthetics:
#'
#' \code{GeomEdgeSpace} understands ggplot2's conventions for segment-like geoms.
#' 
#' @seealso
#' \link{geom_edgespace}, \link[ggplot2]{geom_segment}
#'
#' @importFrom ggplot2 draw_key_path
#' @export
GeomEdgeSpace <- ggproto(
  
  "GeomEdgeSpace", ggplot2::GeomSegment,
  
  required_aes = c("x", "y", "xend", "yend", "arrowType",
    "vertex1", "vertex2", "offset_start", "offset_end"),
  
  non_missing_aes = c("linewidth", "linetype", "colour"),
  
  default_aes = ggplot2::aes(
    linewidth = 0.5,
    linetype = "solid",
    colour = "grey80",
    alpha = NA
  ),
  
  draw_panel = function(self, data, panel_params, coord,   
    arrow_size = 1, arrow_offset = 0.01, lineend = "butt", 
    linejoin = "mitre", na.rm = FALSE, raster = FALSE, 
    dpi = NULL, dev = "cairo", scale = 1, .size_unit = "mm", 
    .nodes = NULL) {
    
    required_att <- c("x", "y", "vertex", "size", "stroke")
    if(!is.null(.nodes) && all(required_att %in% colnames(.nodes))){
      data <- .geom_remap_edge_offsets(data, .nodes, size_unit = .size_unit)
    } else {
      data <- .geom_adj_edge_offsets(data, size_unit = .size_unit)
    }
    
    data$arrow_size <- (arrow_size %||% 1)
    data$arrow_offset <- arrow_offset %||% 0
    
    coords <- coord$transform(data, panel_params)
    
    coords <- .geom_adj_arrow_offsets(coords)
    
    coords <- .geom_adj_arrow_size(coords, size_unit = .size_unit)
    
    # Create edge grobs
    grobs <- .get_edge_grobs(coords, lineend = lineend, 
      linejoin = linejoin, size_unit = .size_unit)
    
    edge_grob <- grid::gTree(children = grobs,
      name = grid::grobName(prefix = "geom_edgespace")
    )
    
    if (raster) {
      edge_grob <- .as_rasteriser(edge_grob, dpi = dpi, 
        dev = dev, scale = scale)
    }
    
    edge_grob
    
  },
  draw_key = draw_key_path
)

################################################################################
### segmentsGrob
################################################################################
.get_edge_grobs <- function(edges, lineend = "butt", 
  linejoin = "mitre", size_unit = "npc"){
  
  if(.empty(edges)){
    return( zeroGrob() )
  }
  
  edges$colour <- scales::alpha(edges$colour, edges$alpha)
  
  edges <- .set_arrows(edges, size_unit)
  
  arrows <- .get_arrows(edges, size_unit)
  
  grobs <- list()
  
  if (nrow(edges)>0) {
    gr <- grid::segmentsGrob(
      x0 = edges$x,
      y0 = edges$y,
      x1 = edges$xend,
      y1 = edges$yend,
      gp = ggplot2::gg_par(
        col = edges$colour,
        lwd = edges$linewidth, lty = edges$linetype,
        lineend = lineend, linejoin = linejoin
      )
    )
    gr$name <- grobName(gr, "edges")
    grobs[[length(grobs) + 1]] <- gr
  }
  
  if (!is.null(arrows$ends)) {
    gr <- grid::segmentsGrob(
      x0 = arrows$ends$exy$x,
      y0 = arrows$ends$exy$y,
      x1 = arrows$ends$exy$xend,
      y1 = arrows$ends$exy$yend,
      arrow = arrows$ends$arrow,
      gp = ggplot2::gg_par(
        col = arrows$ends$color,
        lwd = arrows$ends$linewidth, lty = "solid",
        lineend = lineend, linejoin = linejoin
      )
    )
    gr$name <- grobName(gr, "arrowends")
    grobs[[length(grobs) + 1]] <- gr
  }
  
  if (!is.null(arrows$starts)) {
    gr <- grid::segmentsGrob(
      x0 = arrows$starts$exy$x,
      y0 = arrows$starts$exy$y,
      x1 = arrows$starts$exy$xend,
      y1 = arrows$starts$exy$yend,
      arrow = arrows$starts$arrow,
      gp = ggplot2::gg_par(
        col = arrows$starts$color,
        lwd = arrows$starts$linewidth, lty = "solid",
        lineend = lineend, linejoin = linejoin
      )
    )
    gr$name <- grobName(gr, "arrowstarts")
    grobs[[length(grobs) + 1]] <- gr
  }
  
  do.call(grid::gList, grobs)
  
}

#-------------------------------------------------------------------------------
.empty <- function(df){
  is.null(df) || nrow(df) == 0 || ncol(df) == 0 || is_waiver(df)
}

################################################################################
### Edge Offsets
################################################################################
# Here, the final node sizes computed by ggplot2 are available.
# Clipping offsets are adjusted using the effective node radius,
# stroke, and linewidth;
# The effective node radius is: (size / 2) + (stroke / 2); 'size' 
# represents the node diameter in points (mm * .pt) and is converted
# to 'npc' for alignment with grid coordinates;
# For stroke, see .stroke_offset_estimate();
# For linewidth, see .lwd_offset_estimate().
.geom_remap_edge_offsets <- function(edges, nodes, size_unit){
  
  # size-to-npc conversion factor (1 mm expressed in npc units)
  sz2npc <- grid::convertWidth(unit(1, "mm"), unitTo = "npc", valueOnly = T)
  
  if(size_unit=="mm"){
    # ggplot2 node 'size' in 'mm', scaled to 'npc'
    n_offsets <- nodes[["size"]]/2 * sz2npc
  } else {
    # gspace node 'size' in [0, 100], transformed to 'npc'
    n_offsets <- nodes[["size"]]/2 * .gs_nsz_to_npc()
  }
  # 'stroke' and 'linewidth' in 'mm', scaled to 'npc'
  n_offsets <- n_offsets + (nodes[["stroke"]] * .stroke_offset_estimate(sz2npc))
  e_offsets <-  edges[["linewidth"]] * .lwd_offset_estimate(sz2npc)
  
  edges$offset_start <- n_offsets[edges[["vertex1"]]] + e_offsets
  edges$offset_end <- n_offsets[edges[["vertex2"]]] + e_offsets
  
  return(edges)
  
}

#-------------------------------------------------------------------------------
# Here, the final node sizes computed by ggplot2 are not available;
# Pre-computed clipping offsets are therefore adjusted using 
# 'size_unit', together with linewidth and a default stroke estimate
.geom_adj_edge_offsets <- function(edges, size_unit){
  
  # size-to-npc conversion factor (1 mm expressed in npc units)
  sz2npc <- grid::convertWidth(unit(1, "mm"), unitTo = "npc", valueOnly = T)
  
  if(size_unit=="mm"){
    # ggplot2 node 'size' in 'mm', scaled to 'npc'
    n_offsets <- sz2npc
  } else {
    # gspace node 'size' in [0, 100], transformed to 'npc'
    n_offsets <- .gs_nsz_to_npc()
  }
  stroke_offset <- .stroke_offset_estimate(sz2npc)
  lwd_offset <- edges[["linewidth"]] * .lwd_offset_estimate(sz2npc)
  
  edges$offset_start <- (edges[["offset_start"]]/2 * n_offsets) + 
    lwd_offset + stroke_offset
  edges$offset_end <- (edges[["offset_end"]]/2 * n_offsets) + 
    lwd_offset + stroke_offset

  return(edges)
  
}

#-------------------------------------------------------------------------------
# Estimate the stroke thickness added to the node radius.
# 1. 'stroke' is approximately 0.75 mm per unit (see 'aes_linetype_size_shape') 
#     and is pre-processed by gg_par() as (stroke * .stroke / 2).
# 2. Because the border is centered on the node boundary, only half of the
#    stroke extends outward. This correction estimates the effective increase
#    in node radius attributable to the stroke.
# sz2npc: size-to-npc conversion factor used to calculate node size
.stroke_offset_estimate <- function(sz2npc){
  0.5 * 0.75 * sz2npc
}
# Full linewidth contribution to the clipping offset
# Unlike 'stroke', no half-width correction is required.
.lwd_offset_estimate <- function(sz2npc){
  0.75 * sz2npc
}

#-------------------------------------------------------------------------------
.geom_adj_arrow_offsets <- function(edges){
  edges$offset_start <- edges[["offset_start"]] + edges[["arrow_offset"]]
  edges$offset_end <- edges[["offset_end"]] + edges[["arrow_offset"]]
  return(edges)
}

#-------------------------------------------------------------------------------
.geom_adj_arrow_size <- function(edges, size_unit){
  if(size_unit == "mm"){
    # ggplot2 'size' in 'mm', scaled to 'npc'
    edges$arrow_size <- edges[["arrow_size"]] * ggplot2::.pt
  } else {
    # gspace 'size' in [0, 100], transformed to 'npc'
    edges$arrow_size <- edges[["arrow_size"]] * ggplot2::.pt * 
      .gs_nsz_to_npc()
  }
  return(edges)
}

################################################################################
### Adjust arrows
################################################################################
.set_arrows <- function(edges, size_unit){
  edges <- .add_arrow_angle(edges)
  edges <- .adjust_arrow_size(edges, size_unit)
  edges <- .adjust_arrow_position(edges)
  return(edges)
}
.add_arrow_angle <- function(edges){
  .a_start <- function(atype){
    a_angle <- rep(NA, length(atype))
    a_angle[atype %in% c(0, 1, -1)] <- 0
    a_angle[atype %in% c(2, 3, -4)] <- 30
    a_angle[atype %in% c(-2, -3, 4)] <- 90
    a_angle
  }
  .a_end <- function(atype){
    a_angle <- rep(NA, length(atype))
    a_angle[atype %in% c(0, 2, -2)] <- 0
    a_angle[atype %in% c(1, 3, 4)] <- 30
    a_angle[atype %in% c(-1, -3, -4)] <- 90
    a_angle
  }
  edges$arrowAngleStart <- .a_start(edges$arrowType)
  edges$arrowAngleEnd <- .a_end(edges$arrowType)
  return(edges)
}
.adjust_arrow_size <- function(edges, size_unit){
  edges$arrowSize1 <- edges[["arrow_size"]]
  edges$arrowSize2 <- edges[["arrow_size"]]
  if(size_unit=="npc"){
    # gspace 'size' in [0, 100], transformed to 'npc'
    lw <- edges$linewidth * .gs_nsz_to_npc()
  } else {
    lw <- edges$linewidth
  }
  a_theta <- 60 # default arrowhead opening angle;
  # grid::arrow() expects the half-angle
  a_theta <- a_theta / 180 * pi
  idx <- edges$arrowAngleStart==90
  if(any(idx, na.rm = TRUE)){
    l <- edges$arrowSize1[idx]/2
    b <- sqrt( (l^2 + l^2) - (2 * l^2) * cos(a_theta))
    edges$arrowSize1[idx] <- b + lw[idx]/4
  }
  idx <- edges$arrowAngleEnd==90
  if(any(idx, na.rm = TRUE)){
    l <- edges$arrowSize2[idx]/2
    b <- sqrt( (l^2 + l^2) - (2 * l^2) * cos(a_theta))
    edges$arrowSize2[idx] <- b + lw[idx]/4
  }
  return(edges)
}
.adjust_arrow_position <- function(edges){
  
  emode <- .get_emode(edges$arrowType)
  
  offset_start <- edges$offset_start * as.numeric(emode %in% c(2, 3))
  offset_end <- edges$offset_end * as.numeric(emode %in% c(1, 3))
  total_offset <- (offset_start + offset_end)
  
  dx <- edges$xend - edges$x
  dy <- edges$yend - edges$y
  L <- sqrt( dx^2 + dy^2 )
  L <- ifelse(L == 0, 1e-6, L)
  
  edge_body_len <- pmax(0.02, L * 0.2)
  available_space <- L - edge_body_len
  
  excess <- pmax(0, total_offset - available_space)
  
  adj_start <- offset_start - (excess / 2)
  adj_end <- offset_end - (excess / 2)
  
  final_start <- pmax(0, adj_start + pmin(0, adj_end))
  final_end <- pmax(0, adj_end + pmin(0, adj_start))
  
  idx <- which(excess > 0 & offset_start > 0)
  offset_start[idx] <- final_start[idx]
  
  idx <- which(excess > 0 & offset_end > 0)
  offset_end[idx] <- final_end[idx]
  
  edges$px <- dx/L
  edges$py <- dy/L
  edges$x <- edges$x + (edges$px * offset_start)
  edges$y <- edges$y + (edges$py * offset_start)
  edges$xend <- edges$xend - (edges$px * offset_end)
  edges$yend <- edges$yend - (edges$py * offset_end)
  
  return(edges)
}

################################################################################
### Arrow constructor
################################################################################
.get_arrows <- function(edges, size_unit = "mm"){
  emode <- .get_emode(edges$arrowType)
  #--- add arrow starts
  idx <- emode==2 | emode==3
  if(any(idx)){
    starts <- .arrow_starts(edges[idx,], size_unit)
  } else {
    starts <- NULL
  }
  #--- add arrow ends
  idx <- emode==1 | emode==3
  if(any(idx)){
    ends <- .arrow_ends(edges[idx,], size_unit)
  } else {
    ends <- NULL
  }
  arrows <- list(ends = ends, starts = starts)
  return(arrows)
}
.arrow_starts <- function(edges, size_unit){
  exy <- edges[,c("x", "y", "xend", "yend")]
  # a tiny segment (0.01 npc) is used to anchor the arrowhead
  exy$xend <- exy$x + (edges$px * 0.01)
  exy$yend <- exy$y + (edges$py * 0.01)
  arrow <- grid::arrow(angle = edges$arrowAngleStart,
    type = "open", ends = "first",
    length = grid::unit(edges$arrowSize1, size_unit))
  return(list(exy=exy, arrow = arrow, color = edges$colour, 
    linewidth = edges$linewidth))
}
.arrow_ends <- function(edges, size_unit){
  exy <- edges[,c("x", "y", "xend", "yend")]
  # a tiny segment (0.01 npc) is used to anchor the arrowhead
  exy$x <- exy$xend - (edges$px * 0.01)
  exy$y <- exy$yend - (edges$py * 0.01)
  arrow <- grid::arrow(angle = edges$arrowAngleEnd,
    type = "open", ends = "last",
    length = grid::unit(edges$arrowSize2, size_unit))
  return(list(exy=exy, arrow = arrow, color = edges$colour, 
    linewidth = edges$linewidth))
}

---
title: "Getting started with RGraphSpace"
author: "Sysbiolab - Bioinformatics and Systems Biology Laboratory"
date: "`r Sys.Date()`"
bibliography: bibliography.bib
output: 
  html_document:
    theme: cerulean
    self_contained: yes
    toc: true
    toc_float: true
    toc_depth: 2
    css: custom.css
vignette: >
  %\VignetteIndexEntry{"RGraphSpace: igraph to ggplot2 graphics"}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE, purl=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  collapse = TRUE,
  comment = "#>",
  fig.align = "center",
  fig.width = 7,
  fig.height = 5
  )
```

<br/>
**Package**: RGraphSpace `r packageVersion('RGraphSpace')`

# Overview

*RGraphSpace* is an R package that generates *ggplot2* graphics for *igraph* objects [@Nepusz2006], scaling nodes and edges to a unit space. The package implements new *ggplot2* geometric prototypes [@Wickham2016], optimized for representing large networks. This enables extensive customization of aesthetics and visual style, including colors, shapes, and line types. For extended documentation and use cases, see the [online tutorials](https://sysbiolab.github.io/RGraphSpace/).

# Quick start

This section will create a toy `igraph` object to demonstrate the basic *RGraphSpace* workflow. The input data is configured manually to ensure that users can easily view the basic graph attributes. We will use the igraph's `make_star()` function to create a simple star-like graph and then the `V()` and `E()` functions to set attributes for vertices and edges, respectively. *RGraphSpace* will require that all vertices have `x`, `y`, and `name` attributes.

```{r Load packages for quick start, eval=TRUE, message=FALSE}
#--- Load required packages
library("igraph")
library("ggplot2")
library("RGraphSpace")
```

```{r Toy igraph - 1, eval=TRUE, message=FALSE, results=FALSE}
# Make a 'toy' igraph with 5 nodes and 4 edges;
# ..either a directed or undirected graph
gtoy1 <- make_star(5, mode="out")

# Check whether the graph is directed or not
is_directed(gtoy1)
## [1] TRUE

# Check graph size
vcount(gtoy1)
## [1] 5
ecount(gtoy1)
## [1] 4

# Assign 'x' and 'y' coordinates to each vertex;
# ..this can be an arbitrary unit in (-Inf, +Inf)
V(gtoy1)$x <- c(0, 2, -2, -4, -8)
V(gtoy1)$y <- c(0, 0,  2, -4,  0)

# Assign a name to each vertex
V(gtoy1)$name <- paste0("n", 1:5)
```

```{r Toy igraph - 2, eval=TRUE, message=FALSE, out.width="100%"}
# Plot the 'gtoy1' using standard R graphics
plot(gtoy1)
```

```{r Toy igraph - 3, eval=TRUE, message=FALSE, out.width="80%"}
# Plot the 'gtoy1' using RGraphSpace
plotGraphSpace(gtoy1, add.labels = TRUE)
```

# *RGraphSpace* attributes

Next, we list all vertex and edge attributes that can be passed to *RGraphSpace* methods.

## Vertex attributes

```{r Node attributes, eval=TRUE, message=FALSE}
# Node size (numeric in [0, 100], as '%' of the plot space)
V(gtoy1)$nodeSize <- c(8, 5, 5, 10, 5)

# Node shape (integer code between 0 and 25; see 'help(points)')
V(gtoy1)$nodeShape <- c(21, 22, 23, 24, 25)

# Node color (Hexadecimal or color name)
V(gtoy1)$nodeColor <- c("red", "#00ad39", "grey80", "lightblue", "cyan")

# Node line width (as in 'lwd' standard graphics; see 'help(gpar)')
V(gtoy1)$nodeLineWidth <- 1

# Node line color (Hexadecimal or color name)
V(gtoy1)$nodeLineColor <- "grey20"

# Node labels ('NA' will omit labels)
V(gtoy1)$nodeLabel <- c("V1", "V2", "V3", "V4", NA)

# Node label size (in pts)
V(gtoy1)$nodeLabelSize <- 8

# Node label color (Hexadecimal or color name)
V(gtoy1)$nodeLabelColor <- "black"

# Node transparency (in [0,1])
V(gtoy1)$nodeAlpha <- 1
```

## Edge attributes

Given a list of edges, *RGraphSpace* represents only one edge for each pair of connected vertices. If there are multiple edges connecting the same node pair, it will display the attributes of the first occurrence in the data.

```{r Edge attributes - 1, eval=TRUE, message=FALSE}
# Edge width (as in 'lwd' standard graphics; see 'help(gpar)')
E(gtoy1)$edgeLineWidth <- 0.8

# Edge color (Hexadecimal or color name)
E(gtoy1)$edgeLineColor <- c("red","green","blue","black")

# Edge type (as in 'lty' standard graphics; see 'help(gpar)')
E(gtoy1)$edgeLineType <- c("solid", "11", "dashed", "2124")

# Edge transparency (in [0,1])
E(gtoy1)$edgeAlpha <- 1
```

## Arrowhead attributes

**Arrowhead in directed graphs**: By default, an arrow will be drawn for each edge according to its left-to-right orientation in the edge list (*e.g.* `A -> B`). If there are mutual connections, the package will recode the mutual edges to represent a bidirectional flow.

```{r Edge attributes - 2, eval=TRUE, message=FALSE}
# Arrowhead types in directed graphs
## Integer or character code:
## 0 = "---", 1 = "-->", -1 = "--|"
E(gtoy1)$arrowType <- 1
``` 

**Arrowhead in undirected graphs**: By default, no arrow will be drawn for undirected graphs. However, arrowheads may be assigned according to the coding below.

```{r Edge attributes - 3, eval=TRUE, message=FALSE}
# Arrowhead types in undirected graphs
## Integer or character code:
##  0 = "---"
##  1 = "-->",  2 = "<--",  3 = "<->",  4 = "|->",
## -1 = "--|", -2 = "|--", -3 = "|-|", -4 = "<-|", 
E(gtoy1)$arrowType <- 1
# Note: in undirected graphs, this attribute overrides 
# the edge's orientation in the edge list
```

... and plot the updated `igraph` object. The most straightforward call simply passes an `igraph` directly to the wrapper `plotGraphSpace()` function.
 
```{r A shortcut for RGraphSpace, eval=TRUE, message=FALSE, out.width="80%"}
# Plot the updated 'gtoy1' using RGraphSpace
plotGraphSpace(gtoy1, add.labels = TRUE)
```

```{r, eval=FALSE, message=FALSE, echo=FALSE, include=FALSE, purl=FALSE}
# p <- plotGraphSpace(gtoy1, add.labels = TRUE)
# ggsave(filename = "./vignettes/cards/building_gspace.png", height=4, width=4,
#   units="in", device="png", dpi=200, plot = p)
```


# Passing graphs to *geoms*

Alternatively, the `igraph` object can be transformed into a `GraphSpace` container and passed to the underlying `geoms`:

```{r Using geoms, eval=TRUE, message=FALSE, out.width="80%"}
gs <- GraphSpace(gtoy1)
gs <- normalizeGraphSpace(gs)
ggplot(gs) + 
  geom_edgespace() +
  geom_nodespace() +
  theme_gspace_coords(is_norm = TRUE)
```

# Online tutorials

For detailed integration with the *ggplot2* ecosystem, see the online documentation available at:

* https://sysbiolab.github.io/RGraphSpace/

# Other examples

Vignettes illustrating how *RGraphSpace* can be used in combination with *PathwaySpace* to project network signals.

* https://sysbiolab.github.io/PathwaySpace/

# Citation

If you use *RGraphSpace*, please cite:

* Sysbiolab Team. "RGraphSpace: A lightweight interface between igraph and ggplot2 graphics." R package, 2023. Doi: 10.32614/CRAN.package.RGraphSpace

# Session information
```{r label='Session information', eval=TRUE, echo=FALSE}
sessionInfo()
```


# References

---
title: "RGraphSpace: A lightweight interface between 'igraph' and 'ggplot2' graphics"
author: "Sysbiolab - Bioinformatics and Systems Biology Laboratory"
date: "`r Sys.Date()`"
bibliography: bibliography.bib
abstract: "An interface to integrate 'igraph' and 'ggplot2' graphics within a normalized coordinate system. 'RGraphSpace' implements geometric objects based on 'ggplot2' prototypes, optimized for the representation of large networks. The package aims to facilitate side-by-side visualization of multiple graphs spatially aligned with reference maps and images."
output: 
  html_document:
    theme: cerulean
    self_contained: yes
    toc: true
    toc_float: true
    toc_depth: 2
    css: custom.css
vignette: >
  %\VignetteIndexEntry{"RGraphSpace: igraph to ggplot2 graphics"}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE, purl=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  collapse = TRUE,
  comment = "#>",
  fig.align = "center",
  fig.width = 7,
  fig.height = 5
  )
```

<br/>
**Package**: RGraphSpace `r packageVersion('RGraphSpace')`

# Highlights
* Native *ggplot2* interface for *igraph* objects
* Optimized *geoms* for large-scale network visualization
* Dual-anchor normalization for precise spatial alignment
* Aligns networks with background spatial references

# Overview

*RGraphSpace* is an R package that generates *ggplot2* graphics for *igraph* objects [@Nepusz2006], scaling nodes and edges to a unit space. The package implements new *ggplot2* geometric prototypes [@Wickham2016], optimized for representing large networks. This enables extensive customization of aesthetics and visual style, including colors, shapes, and line types. Three specialized `geoms` translate graph data into geometric layers (see [*using geoms*](#using-geoms)). These `geoms` use a dual-anchor normalization approach to align layers, which is critical for analysis where network elements must be accurately referenced to a spatial map. Section [*mapping graphs to images*](#mapping-graphs-to-images) illustrates how this alignment achieves pixel-level precision. In what follows, this tutorial demonstrates how to use *RGraphSpace* for side-by-side visualization of multiple graphs.

# Quick start

This section will create a toy *igraph* object to demonstrate the *RGraphSpace* workflow. The graph layout is configured manually to ensure that users can easily view all the relevant arguments needed to prepare the input data. We will use the igraph's `make_star()` function to create a simple star-like graph and then the `V()` and `E()` functions to set attributes for vertices and edges, respectively. The *RGraphSpace* package will require that all vertices have `x`, `y`, and `name` attributes.

```{r Load packages for quick start, eval=TRUE, message=FALSE}
#--- Load required packages
library("igraph")
library("ggplot2")
library("RGraphSpace")
```

```{r Toy igraph - 1, eval=TRUE, message=FALSE, results=FALSE}
# Make a 'toy' igraph with 5 nodes and 4 edges;
# ..either a directed or undirected graph
gtoy1 <- make_star(5, mode="out")

# Check whether the graph is directed or not
is_directed(gtoy1)
## [1] TRUE

# Check graph size
vcount(gtoy1)
## [1] 5
ecount(gtoy1)
## [1] 4

# Assign 'x' and 'y' coordinates to each vertex;
# ..this can be an arbitrary unit in (-Inf, +Inf)
V(gtoy1)$x <- c(0, 2, -2, -4, -8)
V(gtoy1)$y <- c(0, 0,  2, -4,  0)

# Assign a name to each vertex
V(gtoy1)$name <- paste0("n", 1:5)
```

```{r Toy igraph - 2, eval=TRUE, message=FALSE, out.width="100%"}
# Plot the 'gtoy1' using standard R graphics
plot(gtoy1)
```

```{r Toy igraph - 3, eval=TRUE, message=FALSE, out.width="80%"}
# Plot the 'gtoy1' using RGraphSpace
plotGraphSpace(gtoy1, add.labels = TRUE)
```

# *RGraphSpace* attributes

Next, we will demonstrate all vertex and edge attributes that can be passed to *RGraphSpace* methods.

## Vertex attributes

```{r Node attributes, eval=TRUE, message=FALSE}
# Node size (numeric in [0, 100], as '%' of the plot space)
V(gtoy1)$nodeSize <- c(8, 5, 5, 10, 5)

# Node shape (integer code between 0 and 25; see 'help(points)')
V(gtoy1)$nodeShape <- c(21, 22, 23, 24, 25)

# Node color (Hexadecimal or color name)
V(gtoy1)$nodeColor <- c("red", "#00ad39", "grey80", "lightblue", "cyan")

# Node line width (as in 'lwd' standard graphics; see 'help(gpar)')
V(gtoy1)$nodeLineWidth <- 1

# Node line color (Hexadecimal or color name)
V(gtoy1)$nodeLineColor <- "grey20"

# Node labels ('NA' will omit labels)
V(gtoy1)$nodeLabel <- c("V1", "V2", "V3", "V4", NA)

# Node label size (in pts)
V(gtoy1)$nodeLabelSize <- 8

# Node label color (Hexadecimal or color name)
V(gtoy1)$nodeLabelColor <- "black"

# Node transparency (in [0,1])
V(gtoy1)$nodeAlpha <- 1
```

## Edge attributes

Given a list of edges, *RGraphSpace* represents only one edge for each pair of connected vertices. If there are multiple edges connecting the same vertex pairs, it will display the line attributes of the first edge in the list.

```{r Edge attributes - 1, eval=TRUE, message=FALSE}
# Edge width (as in 'lwd' standard graphics; see 'help(gpar)')
E(gtoy1)$edgeLineWidth <- 0.8

# Edge color (Hexadecimal or color name)
E(gtoy1)$edgeLineColor <- c("red","green","blue","black")

# Edge type (as in 'lty' standard graphics; see 'help(gpar)')
E(gtoy1)$edgeLineType <- c("solid", "11", "dashed", "2124")

# Edge transparency (in [0,1])
E(gtoy1)$edgeAlpha <- 1
```

## Arrowhead attributes

**Arrowhead in directed graphs**: By default, an arrow will be drawn for each edge according to its left-to-right orientation in the edge list (*e.g.* `A -> B`).

```{r Edge attributes - 2, eval=TRUE, message=FALSE}
# Arrowhead types in directed graphs (integer code or character)
## 0 = "---", 1 = "-->", -1 = "--|"
E(gtoy1)$arrowType <- 1
``` 

**Arrowhead in undirected graphs**: By default, no arrow will be drawn in undirected graphs.

```{r Edge attributes - 3, eval=TRUE, message=FALSE}
# Arrowhead types in undirected graphs (integer or character code)
##  0 = "---"
##  1 = "-->",  2 = "<--",  3 = "<->",  4 = "|->",
## -1 = "--|", -2 = "|--", -3 = "|-|", -4 = "<-|", 
E(gtoy1)$arrowType <- 1
# Note: in undirected graphs, this attribute overrides the 
# edge's orientation in the edge list
```

... and plot the updated *igraph* object with *RGraphSpace*:
 
```{r A shortcut for RGraphSpace, eval=TRUE, message=FALSE, out.width="80%"}
# Plot the updated 'gtoy1' using RGraphSpace
plotGraphSpace(gtoy1, add.labels = TRUE)
```


# Using *ggplot2* geoms {#using-geoms}

## Visual integration and aesthetics mapping

This section illustrates how *RGraphSpace* integrates with the *ggplot2* using `geoms` building blocks. Graph attributes stored within the `GraphSpace` object can be handled in two ways:

* **Identity mapping:** Graph attributes are interpreted as "identity values" (such as `nodeColor`, `nodeSize`, or `nodeShape`) and are displayed exactly as they are, without further scaling or mapping.

* **Dynamic aesthetic mapping:** Graph attributes are mapped to  *aesthetics* (such as `colour`, `size`, and `shape`) and rendered through standard *ggplot2* scales, which automatically generate synchronized legends.

### The *GraphSpace* geoms

To facilitate this integration, *RGraphSpace* implements three specialized `geoms` designed to handle graph data types within a *ggplot2* workflow:

1. **`geom_graphspace()`**: A high-level convenience layer that processes both nodes and edges in a single call. 
2. **`geom_nodespace()`**: Dedicated to rendering nodes. Inherits `GeomPoint` aesthetic mappings, modified to inform the edge layer on node states. It can be used with the `inject_nodespace()` function to adjust edge offsets when the `size` aesthetic is applied to nodes.
3. **`geom_edgespace()`**: Handles the relational data between nodes. Inherits `GeomSegment` aesthetic mappings; unlike standard segments, it is "node-aware" and dynamically calibrates start and end points to connected nodes. 

In the following example, we create a small modular graph containing variables of different types in order to demonstrate these *geoms*.

```{r Load a toy graph, eval=TRUE, message=FALSE, out.width="80%"}
# Make a toy modular graph
library("igraph")
gtoy3 <- sample_islands(
  islands.n = 3,       # number of modules
  islands.size = 30,   # nodes per module
  islands.pin = 0.25,  # probability of edges within modules
  n.inter = 2)         # edges between modules

# Assign module membership to nodes
V(gtoy3)$module <- rep(1:3, each = 30)

# Assign colors to nodes
V(gtoy3)$nodeColor <- rainbow(3)[V(gtoy3)$module]

# Assign a categorical variable to nodes
V(gtoy3)$node_group <- c("A", "B", "C")[V(gtoy3)$module]

# Assign numeric variables to nodes and edges
V(gtoy3)$node_var <- runif(vcount(gtoy3))
E(gtoy3)$edge_var <- runif(ecount(gtoy3))

# Create a GraphSpace from the toy igraph
gs <- GraphSpace(gtoy3)
```

## Plotting identity values

In this example, `nodeColor` already contains the final colour values stored in the `GraphSpace` object. The colours will be displayed as-is by the `geom_graphspace()` function. This approach is particularly useful when nodes have been pre-processed with specific color schemes and you want the visual output without further mapping.

```{r, eval=TRUE, message=FALSE, include = FALSE}
edge_var <- node_var <- nodeColor <- node_group <- NULL
```

```{r Plot identity values, eval=TRUE, message=FALSE, out.width="70%"}
ggplot() + 
  geom_graphspace(colour = "grey", data = gs) +
  theme(aspect.ratio = 1)
```

The trade-off on this approach is that, on one hand, all attributes reflect the original data directly, but no legend is accessible. This is because identity scales bypass the scaling and guide-building process of **ggplot2**. If a legend is required to explain the meaning of these colors, the attribute should be mapped as a variable (e.g., `aes(fill = attribute)`) using standard discrete or continuous scales.

## Mapping categorical variables

In this example, the node categorical variable `node_group` is mapped to the `fill` aesthetic.

```{r Map aesthetics to categorical variables, eval=TRUE, message=FALSE, out.width="70%"}
ggplot() + 
  geom_graphspace(aes(fill = node_group), 
    colour = "grey", data = gs) +
  scale_fill_viridis_d(option = "viridis") +
  theme_gspace_coords()
```

## Mapping numeric variables

In this example, node and edge numeric variables are mapped to `fill` and `colour` aesthetics, respectively.

```{r Map aesthetics to numeric variables, eval=TRUE, message=FALSE, out.width="70%"}
# Map aesthetics to numeric variables
ggplot() + 
  geom_edgespace(aes(colour = edge_var), data = gs) +
  geom_nodespace(aes(fill = node_var), 
    colour = "grey", data = gs) +
  scale_colour_continuous(palette = c("cyan","blue")) +
  scale_fill_continuous(palette = c("white","purple")) +
  theme_gspace_coords()
```

## Using separate colour scales

When multiple geoms use the same aesthetic (for example `colour`) but require mapping to different variables with independent scales, the *ggnewscale* package can be used to introduce a new scale [@Campitelli2025].

```{r Map aesthetics to separate colour scales, eval=FALSE, message=FALSE, out.width="70%"}
if (!require("ggnewscale", quietly = TRUE)) {
  install.packages("ggnewscale")
}
library("ggnewscale")
ggplot() + 
  geom_edgespace(aes(colour = edge_var), data = gs) +
  scale_colour_continuous(palette = c("cyan","blue")) +
  ggnewscale::new_scale_colour() +
  geom_nodespace(aes(colour = node_var), 
    data = gs, stroke = 2, fill = NA) +
  scale_colour_continuous(palette = c("white","purple")) +
  theme_gspace_coords()
```

```{r toy_newscale.png, eval=FALSE, message=FALSE, echo=FALSE, include=FALSE, purl=FALSE}
# gg <- ggplot() + 
#   geom_edgespace(aes(colour = edge_var), data = gs) +
#   scale_colour_continuous(palette = c("cyan","blue")) +
#   ggnewscale::new_scale_colour() +
#   geom_nodespace(aes(colour = node_var), 
#     data = gs, stroke = 2, fill = NA) +
#   scale_colour_continuous(palette = c("white","purple")) +
#   theme_gspace_coords()
# ggsave(filename = "./figs/toy_newscale.png", height=NA, width=NA,
#   units="in", device="png", dpi=200, plot=gg)
```

```{r toy_newscale, echo=FALSE, out.width = '70%', purl=FALSE}
knitr::include_graphics("figs/toy_newscale.png")
```


# Mapping graphs to images

Images can be used as spatial references for graphs. When a raster image is provided, pixel coordinates define where nodes are positioned, supporting the construction of graphs from image features.

As an example, topographic features extracted from the `volcano` matrix are mapped to graph nodes and visualized over a raster image.

```{r Mapping images to graph space, eval=TRUE, message=FALSE, out.width="80%"}
# Extract pixel coordinates for a specific intensity quantile.
coords <- which(volcano == quantile(volcano, 0.85), arr.ind = TRUE)

# Mark target pixels with '0'; it will appear as black in the background. 
# This creates a visual anchor to verify the alignment precision.
volcano2 <- volcano
volcano2[coords] <- 0

# Create an igraph object from the pixel coordinates; 
# note that at this stage, 'y' represents matrix row indices.
gtoy2 <- igraph::make_empty_graph(n = nrow(coords))
igraph::V(gtoy2)$y <- coords[,1]
igraph::V(gtoy2)$x <- coords[,2]

# Highlight the bottom-row vertex (max 'y' index) to demonstrate alignment; 
# since matrix indexing is top-down, this accounts for the default flip 
# between matrix and plot coordinate systems.
igraph::V(gtoy2)$nodeColor <- NA
bottom_row <- which.max(igraph::V(gtoy2)$y)
igraph::V(gtoy2)$nodeColor[bottom_row] <- adjustcolor("red", 0.4)

# Initialize a GraphSpace object
gs <- GraphSpace(gtoy2)

# Map graph coordinates to the image space; by default,
# 'y' row indices will be flipped (see comments below).
gs <- normalizeGraphSpace(gs, image = as_colorraster(volcano2) )

# Render the graph with the raster as background
plotGraphSpace(gs, add.image = TRUE)
```

**Note on image alignment**: Proper spatial alignment between nodes and the background image requires consistent coordinate conventions. Spatial misalignment may occur if the input image and node coordinates differ in axis orientation (e.g., top-left versus bottom-left origins). To accommodate these differences, `normalizeGraphSpace()` provides orientation controls through the `rotate.xy`, `flip.x`, and `flip.y` arguments. If the nodes appear misaligned with the input image, try combinations of these parameters to correct the alignment. Alternatively, `flip.v` and `flip.h` arguments can be used to apply flipping directly to the background image.


# Interoperability with other packages

## Geospatial data

The following example demonstrates interoperability between *RGraphSpace* and *sf*, a well-established infrastructure package for spatial data analysis [@Pebesma2023]. We will use a spatial network of cities to show how *RGraphSpace* geoms can be plugged into *sf* workflows.

```{r Maps with sf, eval=FALSE, message=FALSE, out.width="80%"}
if(!require("sf", quietly = TRUE)){
  install.packages("sf")
}
if(!require("rnaturalearth", quietly = TRUE)){
  install.packages("rnaturalearth")
}
if(!require("maps", quietly = TRUE)){
  install.packages("maps")
}
if(!require("geometry", quietly = TRUE)){
  install.packages("geometry")
}
library("RGraphSpace")
library("igraph")
library("sf")
library("maps")
library("geometry")
library("rnaturalearth")

# Load and project map
map_sf <- ne_countries(country = "Brazil", returnclass = "sf")
map_proj <- st_transform(map_sf)

# Filter major cities by regional capitals
data(world.cities, package = "maps")
r_capitals <- c(
  "Aracaju", "Belem", "Belo Horizonte", "Boa Vista", "Brasilia", 
  "Campo Grande", "Cuiaba", "Curitiba", "Florianopolis", "Fortaleza", 
  "Goiania", "Joao Pessoa", "Macapa", "Maceio", "Manaus", "Natal", 
  "Palmas", "Porto Alegre", "Porto Velho", "Recife", "Rio Branco", 
  "Rio de Janeiro", "Salvador", "Sao Luis", "Sao Paulo", "Teresina", 
  "Vitoria"
)
cities <- subset(world.cities, country.etc == "Brazil" & 
    name %in% r_capitals & pop > 1000000)

# Create Delaunay triangulation edges
# Note: the edges hold no particular meaning beyond
# demonstrating integration between coordinate systems
tri <- delaunayn(cities[,c("lat","long")])
edges <- unique(rbind(tri[,c(1,2)], tri[,c(2,3)], tri[,c(1,3)] ))

# Build igraph with coordinates
gtoy1 <- igraph::graph_from_edgelist(edges, directed = FALSE)
igraph::V(gtoy1)$x <- cities$long
igraph::V(gtoy1)$y <- cities$lat
igraph::V(gtoy1)$Cities <- cities$name
igraph::V(gtoy1)$`Population (M)` <- cities$pop/1000000
igraph::E(gtoy1)$arrowType <- 3

# Make a GraphSpace
gs1 <- GraphSpace(gtoy1)

# Plot
ggplot() + 
  geom_sf(data = map_proj, fill = "grey95", color = "grey60") +
  geom_edgespace(color = "grey40", arrow_size = 0.5, 
    arrow_offset = 0.01, data = gs1) + 
  geom_nodespace(aes(fill = Cities, size = `Population (M)`), 
    data = gs1) +
  scale_size(range = c(3,9)) +
  scale_fill_discrete() + 
  inject_nodespace() +
  theme_gspace_legend(key_fill = TRUE)
```

```{r toy_sf.png, eval=FALSE, message=FALSE, echo=FALSE, include=FALSE, purl=FALSE}
# gg <- ggplot() + 
#   geom_sf(data = map_proj, fill = "grey95", color = "grey60") +
#   geom_edgespace(color = "grey40", arrow_size = 0.5, 
#     arrow_offset = 0.01, data = gs1) + 
#   geom_nodespace(aes(fill = Cities, size = `Population (M)`), 
#     data = gs1) +
#   scale_size(range = c(3,9)) +
#   scale_fill_discrete() + 
#   inject_nodespace() +
#   theme_gspace_legend(key_fill = TRUE)
# ggsave(filename = "./figs/toy_sf.png", height=3.5, width=5,
#   units="in", device="png", dpi=200, plot=gg)
```

```{r toymap, echo=FALSE, out.width = '75%', purl=FALSE}
knitr::include_graphics("figs/toy_sf.png")
```

## Interactive visualization

The following example demonstrates interoperability between *RGraphSpace* and *RedeR*, an R/Bioconductor package for interactive network visualization and manipulation.

```{r A shortcut for RedeR, eval=FALSE, message=FALSE}
# Load RedeR, a graph package for interactive visualization
## Note: this example requires Bioc >= 3.19
if(!require("BiocManager", quietly = TRUE)){
  install.packages("BiocManager")
  #BiocManager::install(version = "3.19")
}
if(!require("RedeR", quietly = TRUE)){
  BiocManager::install("RedeR")
}

# Launch the RedeR application
library("RedeR")
startRedeR()
resetRedeR()
data(gtoy1, package = "RGraphSpace")

# Send 'gtoy1' to the RedeR interface
addGraphToRedeR(gtoy1, unit="npc")
relaxRedeR()

# Fetch 'gtoy1' with a fresh layout
gtoy2 <- getGraphFromRedeR(unit="npc")

# Check the round trip...
plotGraphSpace(gtoy2, add.labels = TRUE)

## Note that for the round trip, shapes and line types are
## partially compatible between ggplot2 and RedeR.

# ...alternatively, just update the graph layout
gtoy2 <- updateLayoutFromRedeR(g=gtoy1)

# ...check the updated layout
plotGraphSpace(gtoy2, add.labels = TRUE)
```

```{r toygraph, echo=FALSE, out.width = '95%', purl=FALSE}
knitr::include_graphics("figs/toy_reder.png")
```


## Other examples

The following vignettes illustrate how *RGraphSpace* can be used in combination with *PathwaySpace* to project network signals on landscape images.

[Projection of network signals](https://sysbiolab.github.io/PathwaySpace/){target="_blank" rel="noopener"}

# Citation

If you use *RGraphSpace*, please cite:

* Sysbiolab Team. "RGraphSpace: A lightweight interface between igraph and ggplot2 graphics." R package, 2023. Doi: 10.32614/CRAN.package.RGraphSpace

* Castro MA, Wang X, Fletcher MN, Meyer KB, Markowetz F (2012). "RedeR: R/Bioconductor package for representing modular structures, nested networks and multiple levels of hierarchical associations." *Genome Biology*, 13(4), R29. Doi: 10.1186/gb-2012-13-4-r29


# Session information
```{r label='Session information', eval=TRUE, echo=FALSE}
sessionInfo()
```


# References

---
title: "Getting started with RGraphSpace"
author: "Sysbiolab Team"
date: "`r Sys.Date()`"
bibliography: bibliography.bib
output: 
  html_document:
    theme: cerulean
    self_contained: yes
    toc: true
    toc_float: true
    toc_depth: 2
    css: custom.css
vignette: >
  %\VignetteIndexEntry{"RGraphSpace: igraph to ggplot2 graphics"}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE, purl=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  collapse = TRUE,
  comment = "#>",
  fig.align = "center",
  fig.width = 7,
  fig.height = 5
  )
```

<br/>
**Package**: RGraphSpace `r packageVersion('RGraphSpace')`

# Overview

*RGraphSpace* is an R package that generates *ggplot2* graphics for *igraph* objects [@Nepusz2006], scaling nodes and edges to a unit space. The package implements new *ggplot2* geometric prototypes [@Wickham2016], optimized for representing large networks. This enables extensive customization of aesthetics and visual style, including colors, shapes, and line types. For extended documentation and use cases, see the [online tutorials](https://sysbiolab.github.io/RGraphSpace/).

# Quick start

This section will create a toy `igraph` object to demonstrate the basic *RGraphSpace* workflow. The input data is configured manually to ensure that users can easily view the basic graph attributes. We will use the igraph's `make_star()` function to create a simple star-like graph and then the `V()` and `E()` functions to set attributes for vertices and edges, respectively. *RGraphSpace* will require that all vertices have `x`, `y`, and `name` attributes.

```{r Load packages for quick start, eval=TRUE, message=FALSE}
#--- Load required packages
library("igraph")
library("ggplot2")
library("RGraphSpace")
```

```{r Toy igraph - 1, eval=TRUE, message=FALSE, results=FALSE}
# Make a 'toy' igraph with 5 nodes and 4 edges;
# ..either a directed or undirected graph
gtoy1 <- make_star(5, mode="out")

# Check whether the graph is directed or not
is_directed(gtoy1)
## [1] TRUE

# Check graph size
vcount(gtoy1)
## [1] 5
ecount(gtoy1)
## [1] 4

# Assign 'x' and 'y' coordinates to each vertex;
# ..this can be an arbitrary unit in (-Inf, +Inf)
V(gtoy1)$x <- c(0, 2, -2, -4, -8)
V(gtoy1)$y <- c(0, 0,  2, -4,  0)

# Assign a name to each vertex
V(gtoy1)$name <- paste0("n", 1:5)
```

```{r Toy igraph - 2, eval=TRUE, message=FALSE, out.width="100%"}
# Plot the 'gtoy1' using standard R graphics
plot(gtoy1)
```

```{r Toy igraph - 3, eval=TRUE, message=FALSE, out.width="80%"}
# Plot the 'gtoy1' using RGraphSpace
plotGraphSpace(gtoy1, add.labels = TRUE)
```

# *RGraphSpace* attributes

Next, we list all vertex and edge attributes that can be passed to *RGraphSpace* methods.

## Vertex attributes

```{r Node attributes, eval=TRUE, message=FALSE}
# Node size (numeric in [0, 100], as '%' of the plot space)
V(gtoy1)$nodeSize <- c(8, 5, 5, 10, 5)

# Node shape (integer code between 0 and 25; see 'help(points)')
V(gtoy1)$nodeShape <- c(21, 22, 23, 24, 25)

# Node color (Hexadecimal or color name)
V(gtoy1)$nodeColor <- c("red", "#00ad39", "grey80", "lightblue", "cyan")

# Node line width (as in 'lwd' standard graphics; see 'help(gpar)')
V(gtoy1)$nodeLineWidth <- 1

# Node line color (Hexadecimal or color name)
V(gtoy1)$nodeLineColor <- "grey20"

# Node labels ('NA' will omit the label)
V(gtoy1)$nodeLabel <- c("V1", "V2", "V3", "V4", "V5")

# Node label size (in pts)
V(gtoy1)$nodeLabelSize <- 8

# Node label color (Hexadecimal or color name)
V(gtoy1)$nodeLabelColor <- "black"

# Node transparency (in [0,1])
V(gtoy1)$nodeAlpha <- 1
```

## Edge attributes

Given a list of edges, *RGraphSpace* represents only one edge for each pair of connected vertices. If there are multiple edges connecting the same node pair, it will display the attributes of the first occurrence in the data.

```{r Edge attributes - 1, eval=TRUE, message=FALSE}
# Edge width (as in 'lwd' standard graphics; see 'help(gpar)')
E(gtoy1)$edgeLineWidth <- 0.8

# Edge color (Hexadecimal or color name)
E(gtoy1)$edgeLineColor <- c("red","green","blue","black")

# Edge type (as in 'lty' standard graphics; see 'help(gpar)')
E(gtoy1)$edgeLineType <- c("solid", "11", "dashed", "2124")

# Edge transparency (in [0,1])
E(gtoy1)$edgeAlpha <- 1
```

## Arrowhead attributes

**Arrowhead in directed graphs**: By default, an arrow will be drawn for each edge according to its left-to-right orientation in the edge list (*e.g.* `A -> B`). If there are mutual connections, the package will recode the mutual edges to represent a bidirectional flow.

```{r Edge attributes - 2, eval=TRUE, message=FALSE}
# Arrowhead types in directed graphs
## Integer or character code:
## 0 = "---", 1 = "-->", -1 = "--|"
E(gtoy1)$arrowType <- 1
``` 

**Arrowhead in undirected graphs**: By default, no arrow will be drawn for undirected graphs. However, arrowheads may be assigned according to the coding below.

```{r Edge attributes - 3, eval=TRUE, message=FALSE}
# Arrowhead types in undirected graphs
## Integer or character code:
##  0 = "---"
##  1 = "-->",  2 = "<--",  3 = "<->",  4 = "|->",
## -1 = "--|", -2 = "|--", -3 = "|-|", -4 = "<-|", 
E(gtoy1)$arrowType <- 1
# Note: in undirected graphs, this attribute overrides 
# the edge's orientation in the edge list
```

... and plot the updated `igraph` object. The most straightforward call simply passes an `igraph` directly to the wrapper `plotGraphSpace()` function.
 
```{r A shortcut for RGraphSpace, eval=TRUE, message=FALSE, out.width="80%"}
# Plot the updated 'gtoy1' using RGraphSpace
plotGraphSpace(gtoy1, add.labels = TRUE)
```

```{r, eval=FALSE, message=FALSE, echo=FALSE, include=FALSE, purl=FALSE}
# p <- plotGraphSpace(gtoy1, add.labels = TRUE)
# ggsave(filename = "./vignettes/cards/building_gspace.png", height=4, width=4,
#   units="in", device="png", dpi=200, plot = p)
```


# Passing graphs to *geoms*

Alternatively, the `igraph` object can be transformed into a `GraphSpace` container and passed to the underlying `geoms`:

```{r Using geoms, eval=TRUE, message=FALSE, out.width="80%"}
data('gtoy1', package = 'RGraphSpace')
gs <- GraphSpace(gtoy1)
gs <- normalizeGraphSpace(gs)
ggplot(gs) + 
  geom_edgespace() + 
  geom_nodespace() + 
  geom_text(aes(x = x, y = y, label = nodeLabel), size = 2) + 
  theme_gspace_coords(is_norm = TRUE)
```

# Online tutorials

For detailed integration with the *ggplot2* ecosystem, see the online documentation available at:

* https://sysbiolab.github.io/RGraphSpace/

# Other examples

Vignettes illustrating how *RGraphSpace* can be used in combination with *PathwaySpace* to project network signals.

* https://sysbiolab.github.io/PathwaySpace/

# Citation

If you use *RGraphSpace*, please cite:

* Sysbiolab Team. "RGraphSpace: A lightweight interface between igraph and ggplot2 graphics." R package, 2023. Doi: 10.32614/CRAN.package.RGraphSpace

# Session information
```{r label='Session information', eval=TRUE, echo=FALSE}
sessionInfo()
```


# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-misc.R
\docType{package}
\name{RGraphSpace-package}
\alias{RGraphSpace}
\alias{RGraphSpace-package}
\title{RGraphSpace: A Lightweight Interface Between 'igraph' and 'ggplot2' Graphics}
\description{
An interface to integrate 'igraph' and 'ggplot2' graphics within a normalized coordinate system. 'RGraphSpace' implements geometric objects based on 'ggplot2' prototypes, optimized for the representation of large networks. The package provides three specialized 'geoms' to translate graph data into geometric layers, supporting customization of aesthetics and visual styles. These 'geoms' use a dual-anchor normalization approach to align layers, required for analyses where network elements must be referenced to a spatial map. 'RGraphSpace' aims to facilitate side-by-side visualization of multiple graphs spatially aligned with reference maps and images.
}
\section{Index}{

\tabular{ll}{
\link{GraphSpace}:
\tab Constructor of GraphSpace-class objects.\cr
\link{plotGraphSpace}:
\tab Wrapper function to plot GraphSpace objects in ggplot2.
}
Further information is available in the vignettes by typing
\code{vignette('RGraphSpace')}. Documented topics are also available in
HTML by typing \code{help.start()} and selecting the RGraphSpace package
from the menu.
}

\references{
Sysbiolab Team (2026). \emph{RGraphSpace: A lightweight interface between
igraph and ggplot2 graphics}. R package version 1.2.0 (Doi:
10.32614/CRAN.package.RGraphSpace),
\url{https://CRAN.R-project.org/package=RGraphSpace}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/sysbiolab/RGraphSpace}
  \item Report bugs at \url{https://github.com/sysbiolab/RGraphSpace/issues}
}

}
\author{
\strong{Maintainer}: Mauro Castro \email{mauro.a.castro@gmail.com} (\href{https://orcid.org/0000-0003-4942-8131}{ORCID})

Authors:
\itemize{
  \item Sysbiolab Team
}

Other contributors:
\itemize{
  \item Flávio Kessler [contributor]
  \item Jonathan Back [contributor]
  \item Lana Querne [contributor]
  \item Victor Apolonio [contributor]
  \item Vinicius Chagas [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-misc.R
\docType{package}
\name{RGraphSpace-package}
\alias{RGraphSpace}
\alias{RGraphSpace-package}
\title{RGraphSpace: A Lightweight Interface Between 'igraph' and 'ggplot2' Graphics}
\description{
An interface to integrate 'igraph' and 'ggplot2' graphics within a normalized coordinate system. 'RGraphSpace' extends 'ggplot2' with graph-aware geometries optimized for large networks. The 'GraphSpace' class integrates directly with 'ggplot2' through specialized 'geoms' and lazy resolution of node attributes, supporting customization of aesthetics and visual styles. These 'geoms' use a dual-anchor normalization approach to align layers, particularly useful for analyses in which network elements must be spatially aligned with reference maps and images.
}
\section{Index}{

\tabular{ll}{
\link{GraphSpace}:
\tab Constructor of GraphSpace-class objects.\cr
\link{plotGraphSpace}:
\tab Wrapper function to plot GraphSpace objects in ggplot2.
}
Further information is available in the vignettes by typing
\code{vignette('RGraphSpace')}. Documented topics are also available in
HTML by typing \code{help.start()} and selecting the RGraphSpace package
from the menu.
}

\references{
Sysbiolab Team (2026). \emph{RGraphSpace: A lightweight interface between
'igraph' and 'ggplot2' graphics}. R package version 1.4.0 (Doi:
10.32614/CRAN.package.RGraphSpace),
\url{https://CRAN.R-project.org/package=RGraphSpace}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/sysbiolab/RGraphSpace}
  \item \url{https://sysbiolab.github.io/RGraphSpace/}
  \item Report bugs at \url{https://github.com/sysbiolab/RGraphSpace/issues}
}

}
\author{
\strong{Maintainer}: Mauro Castro \email{mauro.a.castro@gmail.com} (\href{https://orcid.org/0000-0003-4942-8131}{ORCID})

Authors:
\itemize{
  \item Sysbiolab Team
}

Other contributors:
\itemize{
  \item Flávio Kessler [contributor]
  \item Jonathan Back [contributor]
  \item Lana Querne [contributor]
  \item Victor Apolonio [contributor]
  \item Vinicius Chagas [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-nodespace.R
\docType{data}
\name{GeomNodeSpace}
\alias{GeomNodeSpace}
\title{GeomNodeSpace: a ggplot2 prototype for GraphSpace-class methods}
\format{
An object of class \code{GeomNodeSpace} (inherits from \code{GeomPoint}, \code{Geom}, \code{ggproto}, \code{gg}) of length 6.
}
\usage{
GeomNodeSpace
}
\description{
\code{GeomNodeSpace} is the underlying \link[ggplot2]{ggproto} object
used by \link{geom_nodespace} to draw node elements in a graph layout.

This geom is designed for network diagrams, where graph attributes
are often already in their final form (e.g., hex colors).
}
\section{Aesthetics}{


\code{GeomNodeSpace} understands ggplot2's conventions for point-like geoms.
}

\seealso{
\link{geom_nodespace}, \link[ggplot2]{geom_point}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-nodespace.R
\name{GeomNodeSpace}
\alias{GeomNodeSpace}
\title{GeomNodeSpace: a ggplot2 prototype for GraphSpace-class methods}
\usage{
GeomNodeSpace
}
\description{
\code{GeomNodeSpace} is the underlying \link[ggplot2]{ggproto} object
used by \link{geom_nodespace} to draw node elements in a graph layout.

This geom is designed for network diagrams, where graph attributes
are often already in their final form (e.g., hex colors).
}
\section{Aesthetics}{


\code{GeomNodeSpace} understands ggplot2's conventions for point-like geoms.
}

\seealso{
\link{geom_nodespace}, \link[ggplot2]{geom_point}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-methods.R
\name{plot.GraphSpace}
\alias{plot.GraphSpace}
\title{Plot GraphSpace objects}
\usage{
\method{plot}{GraphSpace}(x, ...)
}
\arguments{
\item{x}{A \linkS4class{GraphSpace} class object.}

\item{...}{Additional arguments passed to the
\code{\link{plotGraphSpace}} function.}
}
\description{
Plot GraphSpace objects
}
\seealso{
\code{\link{plotGraphSpace}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-edgespace.R
\name{geom_edgespace}
\alias{geom_edgespace}
\alias{edgespace_handler}
\title{Draw edge elements in a 2D graph layout}
\usage{
geom_edgespace(
  mapping = NULL,
  data = NULL,
  stat = StatEdgeSpace,
  position = "identity",
  ...,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = FALSE,
  arrow_size = 1,
  arrow_offset = 0.01,
  lineend = "butt",
  linejoin = "mitre",
  raster = FALSE,
  dpi = NULL,
  dev = "cairo",
  scale = 1
)

edgespace_handler()
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[ggplot2:aes]{ggplot2::aes()}}.
These mappings override global aesthetics and are not inherited
from the top-level plot.}

\item{data}{The data to be displayed in this layer. It can be a
\link{GraphSpace} object, an \link[igraph]{igraph} object, or the
\code{edgespace_handler()} closure. When \code{NULL} (default),
a handler is created internally.}

\item{stat}{The statistical transformation to use on the data.
Defaults to \code{identity}.}

\item{position}{Position adjustment, either as a string or
the result of a call to a position adjustment function.}

\item{...}{Additional parameters passed to the underlying
drawing function in \link{GeomEdgeSpace}.}

\item{na.rm}{Logical. Should missing values be removed?
Defaults to \code{FALSE}.}

\item{show.legend}{Logical or a named logical vector indicating
whether this layer should be included in legends.}

\item{inherit.aes}{Logical. If \code{FALSE} (default), the layer will use
aesthetics defined in \code{mapping}.}

\item{arrow_size}{Numeric scaling factor controlling arrowhead
geometry (see 'drawing' section).}

\item{arrow_offset}{Numeric value controlling the base offset of arrows
at edge endpoints (see 'drawing' section).}

\item{lineend}{Line end style (round, butt, square). Supplied for
compatibility with \link[ggplot2]{geom_segment}.}

\item{linejoin}{Line join style (round, mitre, bevel). Supplied for
compatibility with \link[ggplot2]{geom_segment}.}

\item{raster}{Logical. Should node glyphs be rasterized?
Rasterization support is based on \code{\link[ggrastr]{rasterise}}.}

\item{dpi}{Numeric. Rasterization resolution.}

\item{dev}{Character. Rasterization backend. One of \code{"cairo"},
\code{"ragg"}, \code{"ragg_png"}, or \code{"cairo_png"}.}

\item{scale}{Numeric. Rasterization scaling factor
(see \code{\link[ggrastr]{rasterise}}).}
}
\value{
A ggplot2 layer that renders edge segments defined by
\link{GeomEdgeSpace}.
}
\description{
Constructor for \link{GeomEdgeSpace} ggproto objects.

A wrapper around \link[ggplot2]{geom_segment} that enables direct use of
edge attributes stored in \link{GraphSpace} objects as aesthetics.

This \code{geom} is designed to create edge-level aesthetics such as
\code{colour} and \code{linewidth}, or any custom aesthetics defined
in \link{GeomEdgeSpace}.
}
\details{
\strong{arrow_size} is a numeric scaling factor controlling arrowhead geometry.
The value is interpreted in the same numeric space as line width (\code{lwd}),
ensuring consistent scaling between edge strokes and arrowheads.

\strong{arrow_offset} is an additive term that offsets arrow endpoints
uniformly in graph space and is bounded by the edge length, in NPC units.

Arrowhead types are specified in the \link{GraphSpace} constructor.
}
\section{Aesthetics}{


\code{geom_edgespace()} understands \link[ggplot2]{geom_segment} aesthetics.

If these aesthetics are not explicitly provided in \code{aes()}, they
are automatically retrieved from the \link{GraphSpace} object.

\tabular{ll}{
\strong{\code{x}, \code{y}, \code{xend}, \code{yend}} \tab Required (automatically supplied).\cr
\code{colour} \tab Edge colour (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
\code{alpha} \tab Transparency (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
\code{linetype} \tab Edge line type (see \link[ggplot2]{aes_linetype_size_shape}).\cr
\code{linewidth} \tab Edge line width (see \link[ggplot2]{aes_linetype_size_shape}).
}

Required aesthetics (\code{x}, \code{y}, \code{xend}, \code{yend}, ...)
are supplied from the \link{GraphSpace} object and do not need to be
manually mapped.

Additional parameters can be passed to control fixed values for the layer.
For example: \code{colour = "grey"}, \code{linetype = 2}, \code{linewidth = 1}.

Arrows can be further adjusted by \code{arrow_size} and \code{arrow_offset}
arguments (see \emph{details}).
}

\section{Integration with ggraph}{


\code{geom_edgespace} is compatible with the \code{ggraph} methods.
When used within a \code{ggraph()} call, the default \code{edgespace_handler()}
automatically:
\itemize{
\item Identifies the current \code{layout_ggraph}.
\item Extracts the \code{x} and \code{y} coordinates calculated by \code{ggraph}.
\item Reconstructs a temporary \code{GraphSpace} object to inject spatial
metadata and user-chosen \code{ggraph} layout.
}
}

\examples{
library(RGraphSpace)
library(igraph)
library(ggplot2)

# Load a demo igraph
data('gtoy1', package = 'RGraphSpace')

# Create a GraphSpace object
gs <- GraphSpace(gtoy1)

\dontrun{

ggplot() +
  geom_edgespace(data = gs) +
  geom_nodespace(data = gs) +
  theme(aspect.ratio = 1)

}

}
\seealso{
\link{GraphSpace}, \link{geom_nodespace}, \link{geom_graphspace},
\link[ggplot2]{geom_segment}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-edgespace.R
\name{geom_edgespace}
\alias{geom_edgespace}
\title{Draw edge elements in a 2D graph layout}
\usage{
geom_edgespace(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = FALSE,
  size_aes = NULL,
  range = c(1, 6),
  arrow_size = 1,
  arrow_offset = 0.01,
  lineend = "butt",
  linejoin = "mitre"
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[ggplot2:aes]{ggplot2::aes()}}.
These mappings override global aesthetics and are not inherited
from the top-level plot.}

\item{data}{A \link{GraphSpace} object.}

\item{stat}{The statistical transformation to use on the data.
Defaults to \code{identity}.}

\item{position}{Position adjustment, either as a string or
the result of a call to a position adjustment function.}

\item{...}{Additional parameters passed to the underlying
drawing function in \link{GeomEdgeSpace}.}

\item{na.rm}{Logical. Should missing values be removed?
Defaults to \code{FALSE}.}

\item{show.legend}{Logical or a named logical vector indicating
whether this layer should be included in legends.}

\item{inherit.aes}{Logical. If \code{FALSE} (default), the layer will use
aesthetics defined in \code{mapping}.}

\item{size_aes}{Optional node size mapping, created with \code{\link[ggplot2:aes]{ggplot2::aes()}}.
Can be used to synchronize \link{geom_nodespace} and \link{geom_edgespace}
to ensure directed edges and arrows are offset from node boundaries.
This prevents overlaps when \link{geom_nodespace} includes a dynamic
\code{size} or \code{stroke} mappings. For automated scale detection, use
\link{inject_nodespace} instead.}

\item{range}{Optional numeric vector of length 2 (default \code{c(1, 6)})
specifying the min/max node diameters in mm. Used with \code{size_aes} and
should match the \code{range} in \code{\link[ggplot2:scale_size]{ggplot2::scale_size_continuous()}} for
accurate edge clipping. For automated scale detection, use
\link{inject_nodespace} instead.}

\item{arrow_size}{Numeric scaling factor controlling arrowhead
geometry (see 'drawing' section).}

\item{arrow_offset}{Numeric value controlling the base offset of arrows
at edge endpoints (see 'drawing' section).}

\item{lineend}{Line end style (round, butt, square). Supplied for
compatibility with \link[ggplot2]{geom_segment}.}

\item{linejoin}{Line join style (round, mitre, bevel). Supplied for
compatibility with \link[ggplot2]{geom_segment}.}
}
\value{
A ggplot2 layer that renders edge segments defined by
\link{GeomEdgeSpace}.
}
\description{
Constructor for \link{GeomEdgeSpace} ggproto objects.

A wrapper around \link[ggplot2]{geom_segment} that enables direct use of
edge attributes stored in \link{GraphSpace} objects as aesthetics.

This \code{geom} is designed to create edge-level aesthetics such as
\code{colour} and \code{linewidth}, or any custom aesthetics defined
in \link{GeomEdgeSpace}.
}
\details{
\strong{arrow_size} is a numeric scaling factor controlling arrowhead geometry.
The value is interpreted in the same numeric space as line width (\code{lwd}),
ensuring consistent scaling between edge strokes and arrowheads.

\strong{arrow_offset} is an additive term that offsets arrow endpoints
uniformly in graph space and is bounded by the edge length, in NPC units.

Arrowhead types are specified in the \link{GraphSpace} constructor.
}
\section{Aesthetics}{


\code{geom_edgespace()} understands \link[ggplot2]{geom_segment} aesthetics.

If these aesthetics are not explicitly provided in \code{aes()}, they
are automatically retrieved from the \link{GraphSpace} object.

\tabular{ll}{
\strong{\code{x}, \code{y}, \code{xend}, \code{yend}} \tab Required (automatically supplied).\cr
\code{colour} \tab Node border colour (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
\code{alpha} \tab Transparency (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
\code{linetype} \tab Edge line type (see \link[ggplot2]{aes_linetype_size_shape}).\cr
\code{linewidth} \tab Edge line width (see \link[ggplot2]{aes_linetype_size_shape}).
}

Required aesthetics (\code{x}, \code{y}, \code{xend}, \code{yend}, ...)
are supplied from the \link{GraphSpace} object and do not need to be
manually mapped.

Additional parameters can be passed to control fixed values for the layer.
For example: \code{colour = "grey"}, \code{linetype = 2}, \code{linewidth = 1}.

Arrows can be further adjusted by \code{arrow_size} and \code{arrow_offset}
arguments (see \emph{details}).
}

\examples{

# Load a demo igraph
data('gtoy1', package = 'RGraphSpace')

# Create a GraphSpace object
gs <- GraphSpace(gtoy1)

\dontrun{

ggplot() +
  geom_edgespace(data = gs) +
  geom_nodespace(data = gs) +
  theme(aspect.ratio = 1)

}

}
\seealso{
\link{GraphSpace}, \link{geom_nodespace}, \link{geom_graphspace},
\link[ggplot2]{geom_segment}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-methods.R
\docType{methods}
\name{getGraphSpace,GraphSpace-method}
\alias{getGraphSpace,GraphSpace-method}
\alias{getGraphSpace}
\title{Accessors for fetching slots from a GraphSpace object}
\usage{
\S4method{getGraphSpace}{GraphSpace}(gs, what = "graph")
}
\arguments{
\item{gs}{A preprocessed \linkS4class{GraphSpace} class object}

\item{what}{A single character value specifying which slot to
retrieve from a 'GraphSpace' object.
Options: "graph", "nodes", "edges", "pars", "misc", "image", and "fdata".}
}
\value{
Content from slots in the \linkS4class{GraphSpace} object.
}
\description{
\code{getGraphSpace} retrives information from
individual slots available in a GraphSpace object.
}
\examples{
library(RGraphSpace)
library(igraph)

# Load a demo igraph
data('gtoy1', package = 'RGraphSpace')

# Create a new GraphSpace object
gs <- GraphSpace(gtoy1)

# Get the 'summary' slot in gs
getGraphSpace(gs, what = 'graph')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-methods.R
\docType{methods}
\name{getGraphSpace,GraphSpace-method}
\alias{getGraphSpace,GraphSpace-method}
\alias{getGraphSpace}
\title{Accessors for fetching slots from a GraphSpace object}
\usage{
\S4method{getGraphSpace}{GraphSpace}(gs, what = "graph")
}
\arguments{
\item{gs}{A preprocessed \linkS4class{GraphSpace} class object}

\item{what}{A single character value specifying which information should
be retrieved from the slots.
Options: 'graph','gxy','gxyz','pars','misc','status','summits',
'summit_mask', and 'summit_contour'.}
}
\value{
Content from slots in the \linkS4class{GraphSpace} object.
}
\description{
\code{getGraphSpace} retrives information from
individual slots available in a GraphSpace object.
}
\examples{
# Load a demo igraph
data('gtoy1', package = 'RGraphSpace')

# Create a new GraphSpace object
gs <- GraphSpace(gtoy1)

# Get the 'summary' slot in gs
getGraphSpace(gs, what = 'graph')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-methods.R
\name{GraphSpace,ANY-method}
\alias{GraphSpace,ANY-method}
\alias{GraphSpace}
\alias{GraphSpace,data.frame-method}
\title{Create a GraphSpace object}
\usage{
\S4method{GraphSpace}{ANY}(g, layout = NULL, verbose = TRUE, ...)

\S4method{GraphSpace}{data.frame}(g, verbose = TRUE, ...)
}
\arguments{
\item{g}{A graph object inheriting from the \link[igraph]{igraph} class
(such as \code{igraph} and \code{tbl_graph}) or a \code{data.frame} used
to initialize a \code{GraphSpace} object. If a graph is provided, it
should include vertex coordinates in \code{x} and \code{y} attributes,
and vertex labels in the \code{name} attribute. If a \code{data.frame}
is provided, it must contain at least \code{x} and \code{y} columns
representing the node coordinates; additional columns will be treated as
vertex attributes. For graphs requiring edge definitions, use the
\code{igraph} initialization.}

\item{layout}{An optional numeric matrix with two columns for \code{x} and
\code{y} vertex coordinates. If provided, it overrides coordinates in \code{g}.}

\item{verbose}{A logical value. If \code{TRUE}, displays detailed messages.}

\item{...}{Additional arguments passed to the \code{GraphSpace} constructor.}
}
\value{
A \linkS4class{GraphSpace} class object.
}
\description{
\code{GraphSpace} is the main constructor for
\linkS4class{GraphSpace} objects, designed to store graph data and
metadata for optimized rendering in RGraphSpace.
}
\details{
\code{GraphSpace} objects are designed to bridge the gap between network
analysis (via \code{igraph}) and high-quality visualization (via \code{ggplot2}).
The constructor ensures that all necessary aesthetics for
\code{\link{geom_graphspace}} are pre-processed and validated.

\strong{Coordinate System and Normalization:}
By default, the constructor expects coordinates in the \code{x} and \code{y}
vertex attributes, along with unique IDs in the \code{name} vertex
attribute. If these are not provided, the constructor will generate
sequential IDs and assign a layout using the
\code{\link[igraph]{layout_nicely}} function. These coordinates define the
relative positioning of nodes. For optimal rendering, it is recommended
to pass the object through \code{\link{normalizeGraphSpace}} after
construction. This converts vertex positions to Normalized Parent Coordinates
(NPC), ensuring the graph remains centered and scaled relative to the
plotting area.

\strong{Data Structure:}
The resulting object stores nodes and edges in separate internal slots,
preserving metadata such as \code{nodeSize} and \code{edgeLineColor}.
If an \code{igraph} object is provided without specific styling attributes,
\code{GraphSpace} will assign the default values defined in the
\code{\link{geom_graphspace}} aesthetics. Users can also specify custom
variables in the input graph to be used as aesthetics within the
\code{ggplot2} grammar.

\strong{Arrowhead Mapping:}
The \code{arrowType} attribute (see \emph{Arrowhead types} section)
allows for a mapping between symbolic aliases (such as \code{"-->"})
and internal integer codes. This is useful for assigning interaction
types in directed or undirected graphs (e.g., activation vs. inhibition).
}
\section{Vertex attributes}{

The following attributes in \code{g} are evaluated by the constructor:

\tabular{ll}{
\code{nodeSize} \tab Numeric \code{[0, 100]}, representing \% of the plotting space. \cr
\code{nodeShape} \tab Integer code \code{[0-25]}; see \link[graphics]{points}. \cr
\code{nodeColor} \tab A valid color name or hexadecimal code. \cr
\code{nodeLineWidth} \tab Border thickness; see \link[grid]{gpar}. \cr
\code{nodeLineColor} \tab A valid color name or hexadecimal code. \cr
\code{nodeLabel} \tab Character string (\code{NA} will omit labels). \cr
\code{nodeLabelSize} \tab Font size in \code{pts}; see \link[grid]{gpar}. \cr
\code{nodeLabelColor} \tab A valid color name or hexadecimal code.
}
}

\section{Edge attributes}{

The following attributes in \code{g} are evaluated by the constructor:

\tabular{ll}{
\code{edgeLineWidth} \tab Edge thickness; see \code{\link[grid]{gpar}}. \cr
\code{edgeLineColor} \tab A valid color name or hexadecimal code. \cr
\code{edgeLineType}  \tab Line style (e.g., "solid", "dashed"); see \code{\link[grid]{gpar}}. \cr
\code{arrowType}     \tab Arrowhead style (see \emph{Arrowhead types} section).
}
}

\section{Arrowhead types}{


Arrowheads are controlled via the \code{arrowType} attribute using
integer or character codes (see examples in the \emph{RGraphSpace} vignette).

In directed graphs, arrows follow the edge list orientation by default,
representing forward directions (\emph{e.g.}, \code{A -> B}).
While undirected graphs do not show arrows by default, specific styles
can be manually assigned for detailed visualization, including forward,
backward, or bidirectional arrowheads.

\subsection{Directed graphs (A -> B):}{
\tabular{lll}{
\strong{Code} \tab \strong{Alias} \tab \strong{Description} \cr
\code{0} \tab \code{"---"} \tab No arrow \cr
\code{1} \tab \code{"-->"} \tab Forward arrow \cr
\code{-1} \tab \code{"--|"} \tab Forward bar
}
}

\subsection{Undirected graphs (A -- B):}{
\tabular{lll}{
\strong{Code} \tab \strong{Alias} \tab \strong{Description} \cr
\code{0} \tab \code{"---"} \tab No arrow \cr
\code{1} \tab \code{"-->"} \tab Forward arrow \cr
\code{2} \tab \code{"<--"} \tab Backward arrow \cr
\code{3} \tab \code{"<->"} \tab Bidirectional arrow \cr
\code{4} \tab \code{"|->"} \tab Forward arrow / backward bar \cr
\code{-1} \tab \code{"--|"} \tab Forward bar \cr
\code{-2} \tab \code{"|--"} \tab Backward bar \cr
\code{-3} \tab \code{"|-|"} \tab Bidirectional bar \cr
\code{-4} \tab \code{"<-|"} \tab Backward arrow / forward bar
}
}
}

\examples{
library(RGraphSpace)
library(igraph)

# Create a star graph
gtoy1 <- make_full_graph(15)

# Custom attributes
V(gtoy1)$nodeSize <- 5
E(gtoy1)$edgeLineColor <- "red"
E(gtoy1)$arrowType <- "-->"

# Create a GraphSpace
gs <- GraphSpace(gtoy1)

}
\seealso{
\code{\link{geom_graphspace}}, \code{\link{plotGraphSpace}}
}
\author{
Sysbiolab.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-methods.R
\name{GraphSpace,igraph-method}
\alias{GraphSpace,igraph-method}
\alias{GraphSpace}
\alias{GraphSpace,data.frame-method}
\title{Create a GraphSpace object}
\usage{
\S4method{GraphSpace}{igraph}(
  g,
  layout = NULL,
  verbose = TRUE,
  mar = deprecated(),
  image = deprecated(),
  ...
)

\S4method{GraphSpace}{data.frame}(g, verbose = TRUE, ...)
}
\arguments{
\item{g}{An \link[igraph]{igraph} object or a \code{data.frame} used to
initialize a \code{GraphSpace} object. If an \code{igraph} is provided,
it should include vertex coordinates in \code{x} and \code{y} attributes,
and vertex labels in the \code{name} attribute. If a \code{data.frame} is
provided, it must contain at least \code{x} and \code{y} columns
representing the node coordinates; additional columns will be treated
as vertex attributes. For graphs requiring edge definitions, use the
\code{igraph} initialization.}

\item{layout}{An optional numeric matrix with two columns for \code{x} and
\code{y} vertex coordinates. If provided, it overrides coordinates in \code{g}.}

\item{verbose}{A logical value. If \code{TRUE}, displays detailed messages.}

\item{mar}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Deprecated since
RGraphSpace 1.1.1; use \link{normalizeGraphSpace} instead.}

\item{image}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Deprecated since
RGraphSpace 1.1.1; use \link{normalizeGraphSpace} instead.}

\item{...}{Additional arguments passed to the \code{GraphSpace} constructor.}
}
\value{
A \linkS4class{GraphSpace} class object.
}
\description{
\code{GraphSpace} is the main constructor for
\linkS4class{GraphSpace} objects, designed to store graph data and
metadata for optimized rendering in RGraphSpace.
}
\details{
\code{GraphSpace} objects are designed to bridge the gap between network
analysis (via \code{igraph}) and high-quality visualization (via \code{ggplot2}).
The constructor ensures that all necessary aesthetics for
\code{\link{geom_graphspace}} are pre-processed and validated.

\strong{Coordinate System and Normalization:}
By default, the constructor expects coordinates in the \code{x} and \code{y}
vertex attributes, along with unique IDs in the \code{name} vertex
attribute. If these are not provided, the constructor will generate
sequential IDs and assign a layout using the
\code{\link[igraph]{layout_nicely}} function. These coordinates define the
relative positioning of nodes. For optimal rendering, it is recommended
to pass the object through \code{\link{normalizeGraphSpace}} after
construction. This converts vertex positions to Normalized Parent Coordinates
(NPC), ensuring the graph remains centered and scaled relative to the
plotting area.

\strong{Data Structure:}
The resulting object stores nodes and edges in separate internal slots,
preserving metadata such as \code{nodeSize} and \code{edgeLineColor}.
If an \code{igraph} object is provided without specific styling attributes,
\code{GraphSpace} will assign the default values defined in the
\code{\link{geom_graphspace}} aesthetics. Users can also specify custom
variables in the input graph to be used as aesthetics within the
\code{ggplot2} grammar.

\strong{Arrowhead Mapping:}
The \code{arrowType} attribute (see \emph{Arrowhead types} section)
allows for a mapping between symbolic aliases (such as \code{"-->"})
and internal integer codes. This is useful for assigning interaction
types in directed or undirected graphs (e.g., activation vs. inhibition).
}
\section{Vertex attributes}{

The following attributes in \code{g} are evaluated by the constructor:

\tabular{ll}{
\code{nodeSize} \tab Numeric \code{[0, 100]}, representing \% of the plotting space. \cr
\code{nodeShape} \tab Integer code \code{[0-25]}; see \link[graphics]{points}. \cr
\code{nodeColor} \tab A valid color name or hexadecimal code. \cr
\code{nodeLineWidth} \tab Border thickness; see \link[grid]{gpar}. \cr
\code{nodeLineColor} \tab A valid color name or hexadecimal code. \cr
\code{nodeLabel} \tab Character string (\code{NA} will omit labels). \cr
\code{nodeLabelSize} \tab Font size in \code{pts}; see \link[grid]{gpar}. \cr
\code{nodeLabelColor} \tab A valid color name or hexadecimal code.
}
}

\section{Edge attributes}{

The following attributes in \code{g} are evaluated by the constructor:

\tabular{ll}{
\code{edgeLineWidth} \tab Edge thickness; see \code{\link[grid]{gpar}}. \cr
\code{edgeLineColor} \tab A valid color name or hexadecimal code. \cr
\code{edgeLineType}  \tab Line style (e.g., "solid", "dashed"); see \code{\link[grid]{gpar}}. \cr
\code{arrowType}     \tab Arrowhead style (see \emph{Arrowhead types} section).
}
}

\section{Arrowhead types}{


Arrowheads are controlled via the \code{arrowType} attribute using
integer or character codes (see examples in the \emph{RGraphSpace} vignette).

In directed graphs, arrows follow the edge list orientation by default,
representing forward directions (\emph{e.g.}, \code{A -> B}).
While undirected graphs do not show arrows by default, specific styles
can be manually assigned for detailed visualization, including forward,
backward, or bidirectional arrowheads.

\subsection{Directed graphs (A -> B):}{
\tabular{lll}{
\strong{Code} \tab \strong{Alias} \tab \strong{Description} \cr
\code{0} \tab \code{"---"} \tab No arrow \cr
\code{1} \tab \code{"-->"} \tab Forward arrow \cr
\code{-1} \tab \code{"--|"} \tab Forward bar
}
}

\subsection{Undirected graphs (A -- B):}{
\tabular{lll}{
\strong{Code} \tab \strong{Alias} \tab \strong{Description} \cr
\code{0} \tab \code{"---"} \tab No arrow \cr
\code{1} \tab \code{"-->"} \tab Forward arrow \cr
\code{2} \tab \code{"<--"} \tab Backward arrow \cr
\code{3} \tab \code{"<->"} \tab Bidirectional arrow \cr
\code{4} \tab \code{"|->"} \tab Forward arrow / backward bar \cr
\code{-1} \tab \code{"--|"} \tab Forward bar \cr
\code{-2} \tab \code{"|--"} \tab Backward bar \cr
\code{-3} \tab \code{"|-|"} \tab Bidirectional bar \cr
\code{-4} \tab \code{"<-|"} \tab Backward arrow / forward bar
}
}
}

\examples{
library(igraph)

# Create a star graph
gtoy1 <- make_full_graph(15)

# Custom attributes
V(gtoy1)$nodeSize <- 5
E(gtoy1)$edgeLineColor <- "red"
E(gtoy1)$arrowType <- "-->"

# Create a GraphSpace
gs <- GraphSpace(gtoy1)

}
\seealso{
\code{\link{geom_graphspace}}, \code{\link{plotGraphSpace}}
}
\author{
Sysbiolab.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-edgespace.R
\name{StatEdgeSpace}
\alias{StatEdgeSpace}
\title{Attribute Processing for GeomEdgeSpace}
\format{
A \code{ggproto} object.
}
\usage{
StatEdgeSpace
}
\description{
Manage visual attribute precedence (color, size, shape) for \code{GeomEdgeSpace}
objects.
}
\section{Attribute Priority}{

\enumerate{
\item Explicit \code{aes()} mappings.
\item Fixed \code{geom_edgespace()} arguments.
\item Original graph attributes (via \code{optional_aes}).
}

During the \code{setup_data} stage, the Stat invokes internal functions
to resolve value priority:
\enumerate{
\item \strong{Explicit Mapping}: Values defined by the user inside \code{aes()}.
\item \strong{Fixed Parameters}: Constant values passed as arguments in the \code{geom_edgespace()} call.
\item \strong{Graph Attributes}: Original attributes stored within the GraphSpace
object, retrieved from the data columns.
}
}

\seealso{
\code{\link{geom_edgespace}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-methods.R
\name{gs_nodes,GraphSpace-method}
\alias{gs_nodes,GraphSpace-method}
\alias{names}
\alias{names<-}
\alias{gs_nodes}
\alias{gs_edges}
\alias{gs_vcount}
\alias{gs_ecount}
\alias{gs_vertex_attr}
\alias{gs_edge_attr}
\alias{gs_vertex_attr<-}
\alias{gs_edge_attr<-}
\alias{gs_edges,GraphSpace-method}
\alias{gs_vcount,GraphSpace-method}
\alias{gs_ecount,GraphSpace-method}
\alias{names,GraphSpace-method}
\alias{names<-,GraphSpace-method}
\alias{gs_vertex_attr,GraphSpace-method}
\alias{gs_vertex_attr<-,GraphSpace-method}
\alias{gs_edge_attr,GraphSpace-method}
\alias{gs_edge_attr<-,GraphSpace-method}
\title{Accessors for applying essential igraph methods to modify
attributes of GraphSpace objects.}
\usage{
\S4method{gs_nodes}{GraphSpace}(x)

\S4method{gs_edges}{GraphSpace}(x)

\S4method{gs_vcount}{GraphSpace}(x)

\S4method{gs_ecount}{GraphSpace}(x)

\S4method{names}{GraphSpace}(x)

\S4method{names}{GraphSpace}(x) <- value

\S4method{gs_vertex_attr}{GraphSpace}(x, name, ...)

\S4method{gs_vertex_attr}{GraphSpace}(x, name, ...) <- value

\S4method{gs_edge_attr}{GraphSpace}(x, name, ...)

\S4method{gs_edge_attr}{GraphSpace}(x, name, ...) <- value
}
\arguments{
\item{x}{A \linkS4class{GraphSpace} class object}

\item{value}{The new value of the attribute.}

\item{name}{Name of the attribute.}

\item{...}{Additional arguments passed to igraph methods.}
}
\value{
Updated \linkS4class{GraphSpace} object.
}
\description{
Access and modify individual slots of a GraphSpace
object. Selected 'igraph' methods are applied to the 'graph' slot and
propagated to downstream components.
}
\examples{
# Load a demo igraph
data('gtoy1', package = 'RGraphSpace')

# Create a new GraphSpace object
gs <- GraphSpace(gtoy1)

#--- Usage of GraphSpace attribute accessors:

# Get a data frame with nodes for plotting methods
gs_nodes(gs)

# Get a data frame with edges for plotting methods
gs_edges(gs)

# Get vertex count
gs_vcount(gs)

# Get edge count
gs_ecount(gs)

# Get vertex names
names(gs)

# Access all vertex attributes
gs_vertex_attr(gs)

# Access a specific vertex attribute
gs_vertex_attr(gs, "nodeLabel")

# Modify a single value within a vertex attribute
gs_vertex_attr(gs, "nodeSize")["n1"] <- 10

# Replace an entire vertex attribute
gs_vertex_attr(gs, "nodeSize") <- 10

# Alternative syntax using `$` accessor
gs_vertex_attr(gs)$nodeSize <- 10

# Access a specific edge attribute
gs_edge_attr(gs, "edgeLineColor")

# Replace an entire edge attribute
gs_edge_attr(gs, "edgeLineWidth") <- 1

 # Alternative syntax using `$` for edge attributes
gs_edge_attr(gs)$edgeLineWidth <- 3

}
\seealso{
\code{\link[igraph]{vertex_attr}}, \code{\link[igraph]{edge_attr}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annotation-gspace.R
\name{annotation_gspace}
\alias{annotation_gspace}
\alias{annotation_gspace_image}
\title{Annotate a GraphSpace Plot}
\usage{
annotation_gspace(..., type = "image")

annotation_gspace_image(
  raster,
  interpolate = FALSE,
  opacity = 1,
  flip.v = FALSE,
  flip.h = FALSE
)
}
\arguments{
\item{...}{Arguments forwarded to the annotation handler selected by
\code{type}.}

\item{type}{A string specifying the annotation type. Currently available:
\code{"image"} (default), dispatched to \code{annotation_gspace_image()}.}

\item{raster}{An image to be displayed. Accepted types:
\itemize{
\item A \code{\link{GraphSpace}} object — the image is extracted via
\code{\link{gs_image}}.
\item A \code{raster} object
(see \code{\link[grDevices]{as.raster}}).
\item A \code{matrix} or 3D array (RGB/RGBA), coerced to
\code{raster} automatically.
}}

\item{interpolate}{A logical value indicating whether to apply linear
interpolation when the image is rendered at a different resolution than
its native size. Defaults to \code{FALSE}.}

\item{opacity}{A numeric value in \code{[0, 1]} controlling the
transparency of the image. \code{1} is fully opaque (default);
\code{0} is fully transparent.}

\item{flip.v}{A logical value; if \code{TRUE}, the image is flipped
vertically (top-to-bottom). Defaults to \code{FALSE}.}

\item{flip.h}{A logical value; if \code{TRUE}, the image is flipped
horizontally (left-to-right). Defaults to \code{FALSE}.}
}
\value{
A ggplot2 layer object that can be added to a \code{ggplot()}
call with \code{+}, or \code{invisible(NULL)} with a warning if the
image could not be resolved.
}
\description{
\code{annotation_gspace()} is a generic dispatcher for adding annotation
layers to a \code{ggplot}-based \code{GraphSpace} plot. The \code{type}
argument selects the annotation type; additional arguments are forwarded
to the corresponding handler.
}
\examples{

# Assuming 'gs' is a GraphSpace object with 
# an image stored in gs_image(gs)

\dontrun{
# Pass a GraphSpace object directly
ggplot(gs) +
  annotation_gspace(gs) +
  geom_edgespace() +
  geom_nodespace()

# Extract the image explicitly
ggplot(gs) +
  annotation_gspace(gs_image(gs)) +
  geom_edgespace() +
  geom_nodespace()

# Dim the background and flip vertically
ggplot(gs) +
  annotation_gspace(gs, opacity = 0.5, flip.v = TRUE) +
  geom_edgespace() +
  geom_nodespace()
  
}
}
\seealso{
\code{\link[ggplot2]{annotation_raster}},
\code{\link{gs_image}},
\code{\link{geom_nodespace}},
\code{\link{geom_edgespace}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-themes.R
\name{theme_gspace}
\alias{theme_gspace}
\alias{theme_gspace_th0}
\alias{theme_gspace_th1}
\alias{theme_gspace_th2}
\alias{theme_gspace_th3}
\alias{theme_gspace_coords}
\alias{theme_gspace_legend}
\title{RGraphSpace ggplot2 themes}
\usage{
theme_gspace_th0(
  txt_size = 1,
  leg_size = 1,
  bg_color = "grey95",
  key_fill = FALSE,
  key_colour = FALSE,
  ...
)

theme_gspace_th1(
  txt_size = 1,
  leg_size = 1,
  bg_color = "grey95",
  key_fill = FALSE,
  key_colour = FALSE,
  ...
)

theme_gspace_th2(
  txt_size = 1,
  leg_size = 1,
  bg_color = "grey95",
  key_fill = FALSE,
  key_colour = FALSE,
  ...
)

theme_gspace_th3(
  txt_size = 1,
  leg_size = 1,
  bg_color = "grey95",
  key_fill = FALSE,
  key_colour = FALSE,
  ...
)

theme_gspace_coords(
  theme = "th0",
  is_norm = FALSE,
  xlab = "Graph coordinates 1",
  ylab = "Graph coordinates 2",
  expand = NULL,
  ...
)

theme_gspace_legend(leg_size = 1, key_fill = FALSE, key_colour = FALSE, ...)
}
\arguments{
\item{txt_size}{Numeric value to scale plot- and axis-related text elements.}

\item{leg_size}{Numeric value to scale legend-related elements.}

\item{bg_color}{A color name or hex code specifying the panel background.}

\item{key_fill}{Logical; if TRUE, treats the fill legend as discrete
to adjust key size.}

\item{key_colour}{Logical; if TRUE, treats the colour legend as discrete
to adjust key size.}

\item{...}{Additional arguments passed to \code{theme_gspace_th*} and
\link[ggplot2]{ggtheme}.}

\item{theme}{Character string specifying the GraphSpace theme variant.
Options: \code{th0}, \code{th1}, \code{th2}, and \code{th3}.}

\item{is_norm}{Logical; if TRUE, assumes plot coordinates are
already normalized in \code{[0, 1]}.}

\item{xlab}{The title for the 'x' axis.}

\item{ylab}{The title for the 'y' axis.}

\item{expand}{A range expansion factor applied to both the lower and
upper limits of the 'x' and 'y' scales.}
}
\value{
\code{theme_gspace_th*()} return a \code{ggplot2} theme object.

\code{theme_gspace_coords()} returns a list containing scale and theme
components that can be added to a \pkg{ggplot2} plot.

\code{theme_gspace_legend()} returns a list of theme and guide components.
}
\description{
A set of \pkg{ggplot2} themes used by \pkg{RGraphSpace} plots.
}
\details{
\code{theme_gspace_th0()} is a minimal wrapper around
\link[ggplot2]{theme_gray} that simplifies axis and legend scaling.
The \code{txt_size} and \code{leg_size} arguments aggregate related
\link[ggplot2]{theme} parameters for quick thematic overrides.

\code{theme_gspace_th1()} builds on \code{theme_gspace_th0()} and
modifies grid lines, axis appearance, and panel borders.

\code{theme_gspace_th2()} is similar to \code{theme_gspace_th1()} with
simplified grid elements and a customizable panel background.

\code{theme_gspace_th3()} is similar to \code{theme_gspace_th2()} but with
slightly adjusted margins, tick appearance, and legend formatting.

The \code{theme_gspace_coords()} is a helper function that also adds axes
scales for normalized coordinates. It configures axis breaks, limits,
and expansion for graph layouts. Plot coordinates are ideally normalized
to the interval \code{[0, 1]}.

\code{theme_gspace_legend()} is helper function that adjusts legend text,
title, and key sizes by a single scaling factor.
}
\examples{
library(ggplot2)

ggplot(mtcars, aes(wt, mpg)) +
  geom_point() +
  theme_gspace_th0()
  
ggplot(mtcars, 
  aes(scales::rescale(wt), 
      scales::rescale(mpg))) +
  geom_point() +
  theme_gspace_coords("th2", is_norm = TRUE)
  
# Small scale legends
ggplot(mtcars, aes(wt, mpg, fill = factor(cyl))) + 
  geom_point(shape = 21) + 
  theme_gspace_legend(0.8)

}
\seealso{
\link[ggplot2]{ggtheme}, \link[ggplot2]{theme}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-methods.R
\name{names,GraphSpace-method}
\alias{names,GraphSpace-method}
\alias{names}
\alias{gs_names}
\alias{gs_nodes}
\alias{gs_edges}
\alias{gs_graph}
\alias{gs_image}
\alias{gs_image<-}
\alias{gs_fdata}
\alias{gs_fdata<-}
\alias{gs_features}
\alias{gs_nfeatures}
\alias{gs_vcount}
\alias{gs_ecount}
\alias{gs_vertex_attr}
\alias{gs_vertex_attr<-}
\alias{gs_edge_attr}
\alias{gs_edge_attr<-}
\alias{gs_names,GraphSpace-method}
\alias{gs_nodes,GraphSpace-method}
\alias{gs_edges,GraphSpace-method}
\alias{gs_image,GraphSpace-method}
\alias{gs_image<-,GraphSpace-method}
\alias{gs_graph,GraphSpace-method}
\alias{gs_fdata,GraphSpace-method}
\alias{gs_fdata<-,GraphSpace-method}
\alias{gs_nfeatures,GraphSpace-method}
\alias{gs_features,GraphSpace-method}
\alias{gs_vcount,GraphSpace-method}
\alias{gs_ecount,GraphSpace-method}
\alias{gs_vertex_attr,GraphSpace-method}
\alias{gs_vertex_attr<-,GraphSpace-method}
\alias{gs_edge_attr,GraphSpace-method}
\alias{gs_edge_attr<-,GraphSpace-method}
\alias{$,GraphSpace-method}
\alias{$<-,GraphSpace-method}
\alias{[[<-,GraphSpace-method}
\alias{as.igraph.GraphSpace}
\title{Accessors and attribute utilities for GraphSpace objects}
\usage{
\S4method{names}{GraphSpace}(x)

\S4method{gs_names}{GraphSpace}(x)

\S4method{gs_nodes}{GraphSpace}(x, ...)

\S4method{gs_edges}{GraphSpace}(x)

\S4method{gs_image}{GraphSpace}(x)

\S4method{gs_image}{GraphSpace}(x) <- value

\S4method{gs_graph}{GraphSpace}(x)

\S4method{gs_fdata}{GraphSpace}(x)

\S4method{gs_fdata}{GraphSpace}(x) <- value

\S4method{gs_nfeatures}{GraphSpace}(x)

\S4method{gs_features}{GraphSpace}(x)

\S4method{gs_vcount}{GraphSpace}(x)

\S4method{gs_ecount}{GraphSpace}(x)

\S4method{gs_vertex_attr}{GraphSpace}(x, name, ...)

\S4method{gs_vertex_attr}{GraphSpace}(x, name, ...) <- value

\S4method{gs_edge_attr}{GraphSpace}(x, name, ...)

\S4method{gs_edge_attr}{GraphSpace}(x, name, ...) <- value

\S4method{$}{GraphSpace}(x, name)

\S4method{$}{GraphSpace}(x, name) <- value

\method{as.igraph}{GraphSpace}(x, ...)
}
\arguments{
\item{x}{A \linkS4class{GraphSpace} class object}

\item{...}{Additional arguments passed to extraction methods.}

\item{value}{Replacement value for the selected slot or attribute.}

\item{name}{Name of the attribute.}
}
\value{
Updated \linkS4class{GraphSpace} object.
}
\description{
Access and modify individual components of a
\linkS4class{GraphSpace} object. Selected \pkg{igraph} methods are
applied to the internal graph representation and propagated to
downstream node and edge components.
}
\details{
For \code{gs_nodes()}, the optional \code{vars} argument specifies
node-associated features retrieved from the \code{fdata}
container. See also \code{\link{gs_fetch_features}}.
}
\examples{
library(RGraphSpace)
library(igraph)

# Load a demo igraph
data('gtoy1', package = 'RGraphSpace')

# Create a new GraphSpace object
gs <- GraphSpace(gtoy1)

#--- Usage of GraphSpace attribute accessors:

# Vertex names
names(gs)

# Vertex attribute names
gs_names(gs)

# Get a data frame with nodes
gs_nodes(gs)

# Get a data frame with edges
gs_edges(gs)

# Get vertex count
gs_vcount(gs)

# Get edge count
gs_ecount(gs)

# Access all vertex attributes
gs_vertex_attr(gs)

# Access a specific vertex attribute
gs_vertex_attr(gs, "nodeLabel")

# Modify a single value within a vertex attribute
gs_vertex_attr(gs, "nodeSize")["n1"] <- 10

# Replace an entire vertex attribute
gs_vertex_attr(gs, "nodeSize") <- 10

# Alternative syntax using `$` accessor
gs_vertex_attr(gs)$nodeSize <- 10

# Access a specific edge attribute
gs_edge_attr(gs, "edgeLineColor")

# Replace an entire edge attribute
gs_edge_attr(gs, "edgeLineWidth") <- 1

# Alternative syntax using `$` for edge attributes
gs_edge_attr(gs)$edgeLineWidth <- 3

# Add an image and rescale graph coordinates to image space
# Images may be provided as a raster or numeric matrix
gs_image(gs) <- as_colorraster(volcano)
gs <- normalizeGraphSpace(gs)

}
\seealso{
\code{\link[igraph]{vertex_attr}}, \code{\link[igraph]{edge_attr}},
\code{\link{gs_fetch_features}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-classes.R
\docType{class}
\name{GraphSpace-class}
\alias{GraphSpace-class}
\title{GraphSpace: An S4 class for igraph objects}
\value{
An S4 class object.
}
\description{
GraphSpace: An S4 class for igraph objects
}
\section{Slots}{

\describe{
\item{\code{nodes}}{A data frame with xy-vertex coordinates.}

\item{\code{edges}}{A data frame with edges.}

\item{\code{graph}}{An igraph object.}

\item{\code{image}}{A raster background image matrix.}

\item{\code{pars}}{A list with parameters.}

\item{\code{misc}}{A list with intermediate objects for downstream methods.}
}}

\section{Constructor}{

see \code{\link{GraphSpace}} constructor.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inject-nodespace.R
\name{inject_nodespace}
\alias{inject_nodespace}
\title{Dynamic Scale Injection for Edge Clipping}
\usage{
inject_nodespace(...)
}
\arguments{
\item{...}{Additional parameters passed to other methods (currently ignored).}
}
\value{
An object of class \code{inject_nodespace}, which interacts with the
ggplot2 \code{+} operator.
}
\description{
Utility function for \strong{RGraphSpace} that enables edge layers to scan
adjacent nodes and determine their dimensions. This information is used
to compute arrow clipping offsets, preventing edge geometry from
overlapping node symbols.
}
\details{
This function operates in two stages within the \code{ggplot2} workflow:
\enumerate{
\item \strong{Capture:} It scans the \code{plot} layers for a \link{GeomNodeSpace}
to extract both mapping variables (from \code{aes()}) and static parameters
(specifically \code{size} and \code{stroke}).
\item \strong{Injection:} It locates \link{GeomEdgeSpace} layers and injects
scale rules, captured mappings, and fixed parameters into the geometry
parameters.
}

This "lazy injection" calculates edge clipping based on the actual scales
used by the nodes, even if scales are defined after the layers.

\strong{Note:} \code{inject_nodespace()} must be called last in the \code{ggplot} chain to
allow the function to correctly scan all previously added layers and scales.
}
\examples{

library(igraph)
library(ggplot2)

# Generate a toy star graph
gtoy1 <- make_star(15, mode="out")

# Set node and edge attributes
V(gtoy1)$my_node_var <- runif(vcount(gtoy1), 1, 20)
E(gtoy1)$my_edge_var <-  runif(ecount(gtoy1), 1, 20)

# Create a GraphSpace object with a circular layout
gs <- GraphSpace(gtoy1, layout = layout_in_circle(gtoy1))

\dontrun{
# Build the plot
# Note that inject_nodespace() is called at the end to
# synchronize node sizes with edge clipping.
ggplot() +
  geom_edgespace(aes(colour = my_edge_var), data = gs) +
  geom_nodespace(aes(size = my_node_var), data = gs) +
  scale_size(range = c(2, 15)) +
  inject_nodespace()
}

}
\seealso{
\link{geom_edgespace}, \link{geom_nodespace}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-themes.R
\name{theme_gspace}
\alias{theme_gspace}
\alias{theme_gspace_th0}
\alias{theme_gspace_th1}
\alias{theme_gspace_th2}
\alias{theme_gspace_th3}
\alias{theme_gspace_coords}
\alias{theme_gspace_legend}
\title{RGraphSpace ggplot2 themes}
\usage{
theme_gspace_th0(
  txt_size = 1,
  leg_size = 1,
  bg_colour = "grey95",
  discrete_fill = FALSE,
  discrete_colour = FALSE,
  ...
)

theme_gspace_th1(
  txt_size = 1,
  leg_size = 1,
  bg_colour = "grey95",
  discrete_fill = FALSE,
  discrete_colour = FALSE,
  ...
)

theme_gspace_th2(
  txt_size = 1,
  leg_size = 1,
  bg_colour = "grey95",
  discrete_fill = FALSE,
  discrete_colour = FALSE,
  ...
)

theme_gspace_th3(
  txt_size = 1,
  leg_size = 1,
  bg_colour = "grey95",
  discrete_fill = FALSE,
  discrete_colour = FALSE,
  ...
)

theme_gspace_coords(
  theme = "th0",
  is_norm = FALSE,
  xlab = "Graph coordinates 1",
  ylab = "Graph coordinates 2",
  expand = NULL,
  ...
)

theme_gspace_legend(
  leg_size = 1,
  discrete_fill = FALSE,
  discrete_colour = FALSE,
  ...
)
}
\arguments{
\item{txt_size}{Numeric value to scale plot- and axis-related text elements.}

\item{leg_size}{Numeric value to scale legend-related elements.}

\item{bg_colour}{A color name or hex code specifying the panel background.}

\item{discrete_fill}{Logical; if TRUE, treats the fill legend as discrete
to adjust key size.}

\item{discrete_colour}{Logical; if TRUE, treats the colour legend as discrete
to adjust key size.}

\item{...}{Additional arguments passed to \code{theme_gspace_th*} and
\link[ggplot2]{ggtheme}.}

\item{theme}{Character string specifying the GraphSpace theme variant.
Options: \code{th0}, \code{th1}, \code{th2}, and \code{th3}.}

\item{is_norm}{Logical; if TRUE, assumes plot coordinates are
already normalized in \code{[0, 1]}.}

\item{xlab}{The title for the 'x' axis.}

\item{ylab}{The title for the 'y' axis.}

\item{expand}{A range expansion factor applied to both the lower and
upper limits of the 'x' and 'y' scales.}
}
\value{
\code{theme_gspace_th*()} return a \code{ggplot2} theme object.

\code{theme_gspace_coords()} returns a list containing scale and theme
components that can be added to a \pkg{ggplot2} plot.

\code{theme_gspace_legend()} returns a list of theme and guide components.
}
\description{
A set of \pkg{ggplot2} themes used by \pkg{RGraphSpace} plots.
}
\details{
\code{theme_gspace_th0()} is a minimal wrapper around
\link[ggplot2]{theme_gray} that simplifies axis and legend scaling.
The \code{txt_size} and \code{leg_size} arguments aggregate related
\link[ggplot2]{theme} parameters for quick thematic overrides.

\code{theme_gspace_th1()} builds on \code{theme_gspace_th0()} and
modifies grid lines, axis appearance, and panel borders.

\code{theme_gspace_th2()} is similar to \code{theme_gspace_th1()} with
simplified grid elements and a customizable panel background.

\code{theme_gspace_th3()} is similar to \code{theme_gspace_th2()} but with
slightly adjusted margins, tick appearance, and legend formatting.

The \code{theme_gspace_coords()} is a helper function that also adds axes
scales for normalized coordinates. It configures axis breaks, limits,
and expansion for graph layouts. Plot coordinates are ideally normalized
to the interval \code{[0, 1]}.

\code{theme_gspace_legend()} is helper function that adjusts legend text,
title, and key sizes by a single scaling factor.
}
\examples{
library(RGraphSpace)
library(ggplot2)

ggplot(mtcars, aes(wt, mpg)) +
  geom_point() +
  theme_gspace_th0()
  
ggplot(mtcars, 
  aes(scales::rescale(wt), 
      scales::rescale(mpg))) +
  geom_point() +
  theme_gspace_coords("th2", is_norm = TRUE)
  
# Reduce legend element sizes
ggplot(mtcars, aes(wt, mpg, fill = factor(cyl))) + 
  geom_point(shape = 21) + 
  theme_gspace_legend(0.8)

}
\seealso{
\link[ggplot2]{ggtheme}, \link[ggplot2]{theme}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-classes.R
\docType{class}
\name{GraphSpace-class}
\alias{GraphSpace-class}
\title{GraphSpace: An S4 class for igraph objects}
\value{
An S4 class object.
}
\description{
GraphSpace: An S4 class for igraph objects
}
\section{Slots}{

\describe{
\item{\code{nodes}}{A data frame containing node coordinates, attributes, and metadata.}

\item{\code{edges}}{A data frame containing edge relationships and attributes.}

\item{\code{graph}}{An \code{\link[igraph]{igraph}} object representing the graph
structure.}

\item{\code{image}}{A \code{raster} object (see \code{\link[grDevices]{as.raster}}) used
as background image.}

\item{\code{fdata}}{A \code{\link[Matrix]{Matrix}} object storing high-dimensional
feature data associated with graph nodes.}

\item{\code{pars}}{A list with parameters.}

\item{\code{misc}}{A list with intermediate objects for downstream methods.}

\item{\code{uuid}}{A Universally Unique Identifier (UUID) for the object instance.}
}}

\section{Constructor}{

see \code{\link{GraphSpace}} constructor.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-ggplot-constructor.R
\name{inject_nodespace}
\alias{inject_nodespace}
\title{Dynamic Scale Injection for Edge Clipping}
\usage{
inject_nodespace(...)
}
\arguments{
\item{...}{Additional parameters passed to other methods (currently ignored).}
}
\value{
An object of class \code{inject_nodespace}, which interacts with the
ggplot2 \code{+} operator.
}
\description{
Utility function for \strong{RGraphSpace} that enables edge layers to scan
adjacent nodes and determine their dimensions. This information is used
to compute arrow clipping offsets, preventing edge geometry from
overlapping node symbols.
}
\details{
This function operates in two stages within the \code{ggplot2} workflow:
\enumerate{
\item \strong{Capture:} It scans the \code{plot} layers for a \link{GeomNodeSpace}
to extract both mapping variables (from \code{aes()}) and static parameters
(specifically \code{size} and \code{stroke}).
\item \strong{Injection:} It locates \link{GeomEdgeSpace} layers and injects
scale rules, captured mappings, and fixed parameters into the geometry
parameters.
}

This "lazy injection" calculates edge clipping based on the actual scales
used by the nodes, even if scales are defined after the layers.

\strong{Note:} \code{inject_nodespace()} must be called last in the \code{ggplot} chain to
allow the function to correctly scan all previously added layers and scales.
}
\examples{
library(RGraphSpace)
library(igraph)
library(ggplot2)

# Generate a toy star graph
gtoy1 <- make_star(15, mode="out")

# Set node and edge attributes
V(gtoy1)$my_node_var <- runif(vcount(gtoy1), 1, 20)
E(gtoy1)$my_edge_var <-  runif(ecount(gtoy1), 1, 20)

# Create a GraphSpace object with a circular layout
gs <- GraphSpace(gtoy1, layout = layout_in_circle(gtoy1))

\dontrun{
# Build the plot
# Note that inject_nodespace() is called at the end to
# synchronize node sizes with edge clipping.
ggplot() +
  geom_edgespace(aes(colour = my_edge_var), data = gs) +
  geom_nodespace(aes(size = my_node_var), data = gs) +
  scale_size(range = c(2, 15)) +
  inject_nodespace()
}

}
\seealso{
\link{geom_edgespace}, \link{geom_nodespace}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-methods.R
\docType{methods}
\name{plotGraphSpace,GraphSpace-method}
\alias{plotGraphSpace,GraphSpace-method}
\alias{plotGraphSpace}
\alias{plotGraphSpace,igraph-method}
\title{Wrapper function to plot GraphSpace objects in ggplot2}
\usage{
\S4method{plotGraphSpace}{GraphSpace}(
  gs,
  theme = "th0",
  xlab = "Graph coordinates 1",
  ylab = "Graph coordinates 2",
  font.size = 1,
  bg.color = "grey95",
  add.labels = FALSE,
  node.labels = NULL,
  label.size = 3,
  label.color = "grey20",
  add.image = FALSE,
  raster = FALSE,
  dpi = 300,
  dev = "cairo_png"
)

\S4method{plotGraphSpace}{igraph}(gs, ...)
}
\arguments{
\item{gs}{Either an \code{igraph} or \linkS4class{GraphSpace} class object.
If \code{gs} is an \code{igraph}, then it must include \code{x}, \code{y},
and \code{name}  vertex attributes (see \code{\link{GraphSpace}}).}

\item{theme}{Name of a custom RGraphSpace theme. These themes
(from 'th0' to 'th3') consist of preconfigured ggplot settings,
which can subsequently refine using \code{\link[ggplot2]{ggplot2}}.}

\item{xlab}{The title for the 'x' axis of a 2D-image space.}

\item{ylab}{The title for the 'y' axis of a 2D-image space.}

\item{font.size}{A single numeric value passed to ggplot themes.}

\item{bg.color}{A single color for background.}

\item{add.labels}{A logical value indicating whether to plot vertex labels.}

\item{node.labels}{A vector of vertex names to be highlighted in the graph
space. This argument overrides 'add.labels'.}

\item{label.size}{A size argument passed to \code{\link[ggplot2]{geom_text}}.}

\item{label.color}{A color passed to \code{\link[ggplot2]{geom_text}}.}

\item{add.image}{A logical value indicating whether to add a background
image, when one is available (see \code{\link{GraphSpace}}).}

\item{raster}{A logical value indicating whether to rasterize the main plot.
See \code{\link[ggrastr]{rasterise}} for further specifications.}

\item{dpi}{Raster resolution, in dots per inch.}

\item{dev}{Device used in the \code{\link[ggrastr]{rasterise}} call.}

\item{...}{Additional arguments passed to the
\code{\link{plotGraphSpace}} function.}
}
\value{
A ggplot-class object.
}
\description{
\code{plotGraphSpace()} is a High-level plotting interface that translates
\code{igraph} and \code{GraphSpace} data objects into \code{ggplot2} layers.
}
\examples{
library(igraph)

# Load a demo igraph
data('gtoy1', package = 'RGraphSpace')

# Generate a ggplot for gtoy1
plotGraphSpace(gtoy1)

# Create a star graph
gtoy_star <- make_full_graph(15)

# Example of setting node and edge attributes
V(gtoy_star)$nodeSize <- 5
E(gtoy_star)$edgeLineColor <- "red"
E(gtoy_star)$arrowType <- "<->"

# Create a GraphSpace object
gs_star <- GraphSpace(gtoy_star)

# Normalize graph coordinates
gs_star <- normalizeGraphSpace(gs_star)

# Generate a ggplot for gs_star
plotGraphSpace(gs_star)

}
\seealso{
\code{\link{GraphSpace}}
}
\author{
Sysbiolab.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-nodespace.R
\name{StatNodeSpace}
\alias{StatNodeSpace}
\title{Attribute Processing for GeomNodeSpace}
\format{
A \code{ggproto} object.
}
\usage{
StatNodeSpace
}
\description{
Manage visual attribute precedence (color, size, shape) for \code{GeomNodeSpace}
objects.
}
\section{Attribute Priority}{

\enumerate{
\item Explicit \code{aes()} mappings.
\item Fixed \code{geom_nodespace()} arguments.
\item Original graph attributes (via \code{optional_aes}).
}

During the \code{setup_data} stage, the Stat invokes internal functions
to resolve value priority:
\enumerate{
\item \strong{Explicit Mapping}: Values defined by the user inside \code{aes()}.
\item \strong{Fixed Parameters}: Constant values passed as arguments in the \code{geom_nodespace()} call.
\item \strong{Graph Attributes}: Original attributes stored within the GraphSpace
object, retrieved from the data columns.
}
}

\seealso{
\code{\link{geom_nodespace}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-graphspace.R
\name{geom_graphspace}
\alias{geom_graphspace}
\title{Draw node and edge elements in a 2D graph layout}
\usage{
geom_graphspace(
  mapping = NULL,
  data = NULL,
  stat = "identity",
  position = "identity",
  ...,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = FALSE,
  arrow_size = 1,
  arrow_offset = 0.01
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[ggplot2:aes]{ggplot2::aes()}}.
These mappings override global aesthetics and are not inherited
from the top-level plot.}

\item{data}{A \link{GraphSpace} object.}

\item{stat}{The statistical transformation to use on the data.
Defaults to \code{identity}.}

\item{position}{Position adjustment, either as a string or
the result of a call to a position adjustment function.}

\item{...}{Additional parameters passed to the underlying
drawing function in \link{GeomGraphSpace}.}

\item{na.rm}{Logical. Should missing values be removed?
Defaults to \code{FALSE}.}

\item{show.legend}{Logical or a named logical vector indicating
whether this layer should be included in legends.}

\item{inherit.aes}{Logical. If \code{FALSE} (default), the layer will use
aesthetics defined in \code{mapping}.}

\item{arrow_size}{Numeric scaling factor controlling arrowhead
geometry (see 'drawing' section).}

\item{arrow_offset}{Numeric value controlling the base offset of arrows
at edge endpoints (see 'drawing' section).}
}
\value{
A ggplot2 layer that renders node glyphs defined by
\link{GeomGraphSpace}.
}
\description{
Constructor for \link{GeomGraphSpace} ggproto objects.

A wrapper around \link[ggplot2]{geom_point} that enables direct use of
node attributes stored in \link{GraphSpace} objects as aesthetics.

This geom is designed to map node-level attributes (e.g., \code{fill},
\code{size}) or any aesthetics supported by \link[ggplot2]{GeomPoint}.
}
\section{Aesthetics for node drawing}{


Nodes are drawn in the main layer of \code{geom_graphspace()}, which
understands \link[ggplot2]{geom_point} aesthetics.

If these aesthetics are not explicitly provided in \code{aes()}, they
are automatically retrieved from the \link{GraphSpace} object.

\tabular{ll}{
\strong{\code{x}, \code{y}, \code{vertex}} \tab
Required (automatically supplied).\cr
\code{fill} \tab Node interior colour (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
\code{colour} \tab Node border colour (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
\code{alpha} \tab Transparency (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
\code{shape} \tab Node shape (see \link{points} and \link[ggplot2]{aes_linetype_size_shape}).\cr
\code{size} \tab Node size (see \emph{drawing} section and \link[ggplot2]{aes_linetype_size_shape}).\cr
\code{stroke} \tab Node line width (see \link[ggplot2]{gg_par} and \link[ggplot2]{aes_linetype_size_shape}).
}

Required aesthetics \code{x}, \code{y}, and \code{vertex} are supplied from
the \link{GraphSpace} object and do not need to be manually mapped.

Additional parameters can be passed to control fixed values for the layer.
For example: \code{fill = "red"}, \code{stroke = 3}, \code{alpha = 0.5}, or \code{shape = 21}.

The interpretation of \strong{\code{size}} depends on how it is provided:
\itemize{
\item \strong{As an aesthetic}: When mapped within \code{aes()}, \code{size}
follows the behavior of \link[ggplot2]{geom_point}, using absolute
units to ensure consistency with the plot legends.
\item \strong{As a parameter}: When set outside \code{aes()}, \code{size} is
treated as a percentage of the viewport (\code{[0, 100]}), scaling
in \code{npc} units. This allows nodes to resize dynamically with
viewport changes.
}
}

\section{Edge context-aware parameters}{


These parameters control the edge appearance. If not explicitly provided,
they are automatically retrieved from the \link{GraphSpace} object.
They can be a single value or a vector matching the number of edges:

\tabular{ll}{
\code{edge_colour} \tab Node border colour.\cr
\code{edge_linetype} \tab Edge line type.\cr
\code{edge_linewidth} \tab Edge line width.\cr
\code{edge_alpha} \tab Edge transparency.
}
}

\section{Edge global parameters}{


These parameters apply globally to all edges in the layer:

\tabular{ll}{
\code{arrow_size} \tab Arrow scaling factor (default = 1).\cr
\code{arrow_offset} \tab Arrow offset from nodes (default = 0.01).\cr
\code{arrow_lineend} \tab Line end style (see \link[grid]{gpar}).\cr
\code{arrow_linejoin} \tab Line join style (see \link[grid]{gpar}).
}

\strong{arrow_size} is a numeric scaling factor controlling arrowhead geometry.
The value is interpreted in the same numeric space as line width (\code{lwd}),
ensuring consistent scaling between edge strokes and arrowheads.

\strong{arrow_offset} is an additive term that offsets arrow endpoints
uniformly in graph space and is bounded by the edge length, in NPC units.

Arrowhead types are specified in the \link{GraphSpace} constructor.
}

\examples{

# Make a demo igraph
library(igraph)
gtoy1 <- make_star(15, mode="out")

# Set some node attributes
V(gtoy1)$nodeSize <- runif(vcount(gtoy1), 1, 20)
V(gtoy1)$nodeColor <- rainbow(vcount(gtoy1))

# Set some variables
V(gtoy1)$user_var1 <- runif(vcount(gtoy1), 1, 3)^3
V(gtoy1)$user_var2 <-  rep(c(1, 2, 3), each = 5)

# Create a GraphSpace object
gs <- GraphSpace(gtoy1, layout = layout_in_circle(gtoy1))

\dontrun{

# Example 1: Nodes scaling with the legend
# When 'size' is mapped inside aes(), it follows
# ggplot2 default behavior: sizes are translated 
# to absolute units (mm) via 'scale_size()'.

ggplot() + 
  geom_graphspace(
  mapping = aes(size = nodeSize, fill = user_var2), 
  data = gs, arrow_offset = 0.01) + 
  scale_size(range = c(1, 12)) + 
  theme(aspect.ratio = 1)
  
# Example 2: Nodes scaling with the viewport
# When 'size' is passed as a node attribute, 
# as inherited from the igraph object, it is
# interpreted as a percentage of the plot area
# and translated to NPC units.

ggplot() +
  geom_graphspace(mapping = aes(fill = user_var2), 
  data = gs, arrow_offset = 0.01) + 
  theme(aspect.ratio = 1)
  
}

}
\seealso{
\link{GraphSpace}, \link{geom_nodespace}, \link{geom_edgespace},
\link[ggplot2]{geom_point}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-methods.R
\docType{methods}
\name{plotGraphSpace,GraphSpace-method}
\alias{plotGraphSpace,GraphSpace-method}
\alias{plotGraphSpace}
\alias{plotGraphSpace,igraph-method}
\alias{plotGraphSpace,tbl_graph-method}
\alias{plotGraphSpace,gs_graph-method}
\title{Wrapper function to plot GraphSpace objects in ggplot2}
\usage{
\S4method{plotGraphSpace}{GraphSpace}(
  gs,
  theme = "th0",
  xlab = "Graph coordinates 1",
  ylab = "Graph coordinates 2",
  font.size = 1,
  bg.color = "grey95",
  add.labels = FALSE,
  node.labels = NULL,
  label.size = 3,
  label.color = "grey20",
  add.image = TRUE,
  raster = FALSE,
  dpi = 300,
  dev = "cairo_png"
)

\S4method{plotGraphSpace}{igraph}(gs, ...)

\S4method{plotGraphSpace}{tbl_graph}(gs, ...)

\S4method{plotGraphSpace}{gs_graph}(gs, ...)
}
\arguments{
\item{gs}{Either an \code{igraph} or \linkS4class{GraphSpace} class object.
If \code{gs} is an \code{igraph}, then it must include \code{x}, \code{y},
and \code{name}  vertex attributes (see \code{\link{GraphSpace}}).}

\item{theme}{Name of a custom RGraphSpace theme. These themes
(from 'th0' to 'th3') consist of preconfigured ggplot settings,
which can subsequently refine using \code{\link[ggplot2]{ggplot2}}.}

\item{xlab}{The title for the 'x' axis of a 2D-image space.}

\item{ylab}{The title for the 'y' axis of a 2D-image space.}

\item{font.size}{A single numeric value passed to ggplot themes.}

\item{bg.color}{A single color for background.}

\item{add.labels}{A logical value indicating whether to plot vertex labels.}

\item{node.labels}{A vector of vertex names to be highlighted in the graph
space. This argument overrides 'add.labels'.}

\item{label.size}{A size argument passed to \code{\link[ggplot2]{geom_text}}.}

\item{label.color}{A color passed to \code{\link[ggplot2]{geom_text}}.}

\item{add.image}{A logical value indicating whether to add a background
image, when one is available (see \code{\link{GraphSpace}}).}

\item{raster}{A logical value indicating whether to rasterize the main plot.
See \code{\link[ggrastr]{rasterise}} for further specifications.}

\item{dpi}{Raster resolution, in dots per inch.}

\item{dev}{Device used in the \code{\link[ggrastr]{rasterise}} call.}

\item{...}{Additional arguments passed to the
\code{\link{plotGraphSpace}} function.}
}
\value{
A ggplot-class object.
}
\description{
\code{plotGraphSpace()} is a High-level plotting interface that translates
\code{igraph} and \code{GraphSpace} data objects into \code{ggplot2} layers.
}
\examples{
library(RGraphSpace)
library(igraph)

# Load a demo igraph
data('gtoy1', package = 'RGraphSpace')

# Generate a ggplot for gtoy1
plotGraphSpace(gtoy1)

# Create a star graph
gtoy_star <- make_full_graph(15)

# Example of setting node and edge attributes
V(gtoy_star)$nodeSize <- 5
E(gtoy_star)$edgeLineColor <- "red"
E(gtoy_star)$arrowType <- "<->"

# Create a GraphSpace object
gs_star <- GraphSpace(gtoy_star)

# Normalize graph coordinates
gs_star <- normalizeGraphSpace(gs_star)

# Generate a ggplot for gs_star
plotGraphSpace(gs_star)

}
\seealso{
\code{\link{GraphSpace}}
}
\author{
Sysbiolab.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-misc.R
\name{as_colorraster}
\alias{as_colorraster}
\title{Map numeric values to a color raster}
\usage{
as_colorraster(x, palette = hcl.colors(30), na.color = "white")
}
\arguments{
\item{x}{A numeric vector or matrix containing values to be mapped to colors.}

\item{palette}{A vector of colors used as the palette. By default,
\code{hcl.colors(30)} is used.}

\item{na.color}{Color used for \code{NA} values. Defaults to \code{NA}.}
}
\value{
A raster object as produced by \code{as.raster()}.
}
\description{
Helper function that converts numeric values to colors and
returns a raster image. Useful for visualizing numeric matrices
as color backgrounds.
}
\details{
Values in \code{x} are rescaled to the range of the palette using
\code{scales::rescale()}, and each value is mapped to a corresponding
color. If \code{x} is a matrix, the resulting raster preserves the same
dimensions.
}
\examples{
# Convert the volcano matrix to a color raster
img <- as_colorraster(volcano)
plot(img)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-features.R
\name{gs_features-utils}
\alias{gs_features-utils}
\alias{gs_fetch_features}
\alias{gs_add_features}
\title{Manipulate node features in a GraphSpace object}
\usage{
gs_fetch_features(x, vars = NULL, as_df = FALSE)

gs_add_features(x, data)
}
\arguments{
\item{x}{A \code{GraphSpace} object.}

\item{vars}{Character vector specifying feature names to extract.
If \code{NULL}, all features are returned.}

\item{as_df}{Logical. If \code{TRUE}, returns a \code{data.frame}.
Otherwise returns the original backend representation.}

\item{data}{A matrix-like or \code{data.frame} object containing node
features. Rows must correspond to node identifiers.}
}
\value{
\itemize{
\item \code{gs_fetch_features()} returns a matrix-like object or
\code{data.frame} containing the selected node features.

\item \code{gs_add_features()} returns a modified
\code{GraphSpace} object.
}
}
\description{
Utilities for extracting and adding node-associated features stored in the
\code{fdata} container of a \code{GraphSpace} object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-graphspace.R
\name{geom_graphspace}
\alias{geom_graphspace}
\title{Draw node and edge elements in a 2D graph layout}
\usage{
geom_graphspace(
  mapping = NULL,
  data,
  stat = "identity",
  position = "identity",
  ...,
  na.rm = FALSE,
  show.legend = NA,
  inherit.aes = FALSE,
  arrow_size = 1,
  arrow_offset = 0.01,
  raster = FALSE,
  dpi = NULL,
  dev = "cairo",
  scale = 1
)
}
\arguments{
\item{mapping}{Set of aesthetic mappings created by \code{\link[ggplot2:aes]{ggplot2::aes()}}.
These mappings override global aesthetics and are not inherited
from the top-level plot.}

\item{data}{A \link{GraphSpace} object.}

\item{stat}{The statistical transformation to use on the data.
Defaults to \code{identity}.}

\item{position}{Position adjustment, either as a string or
the result of a call to a position adjustment function.}

\item{...}{Additional parameters passed to the underlying
drawing function in \link{GeomGraphSpace}.}

\item{na.rm}{Logical. Should missing values be removed?
Defaults to \code{FALSE}.}

\item{show.legend}{Logical or a named logical vector indicating
whether this layer should be included in legends.}

\item{inherit.aes}{Logical. If \code{FALSE} (default), the layer will use
aesthetics defined in \code{mapping}.}

\item{arrow_size}{Numeric scaling factor controlling arrowhead
geometry (see 'drawing' section).}

\item{arrow_offset}{Numeric value controlling the base offset of arrows
at edge endpoints (see 'drawing' section).}

\item{raster}{Logical. Should node glyphs be rasterized?
Rasterization support is based on \code{\link[ggrastr]{rasterise}}.}

\item{dpi}{Numeric. Rasterization resolution.}

\item{dev}{Character. Rasterization backend. One of \code{"cairo"},
\code{"ragg"}, \code{"ragg_png"}, or \code{"cairo_png"}.}

\item{scale}{Numeric. Rasterization scaling factor
(see \code{\link[ggrastr]{rasterise}}).}
}
\value{
A ggplot2 layer that renders node glyphs defined by
\link{GeomGraphSpace}.
}
\description{
Constructor for \link{GeomGraphSpace} ggproto objects.

A wrapper around \link[ggplot2]{geom_point} that enables direct use of
node attributes stored in \link{GraphSpace} objects as aesthetics.

This geom is designed to map node-level attributes (e.g., \code{fill},
\code{size}) or any aesthetics supported by \link[ggplot2]{GeomPoint}.
}
\section{Aesthetics for node drawing}{


Nodes are drawn in the main layer of \code{geom_graphspace()}, which
understands \link[ggplot2]{geom_point} aesthetics.

If these aesthetics are not explicitly provided in \code{aes()}, they
are automatically retrieved from the \link{GraphSpace} object.

\tabular{ll}{
\strong{\code{x}, \code{y}, \code{vertex}} \tab
Required (automatically supplied).\cr
\code{fill} \tab Node interior colour (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
\code{colour} \tab Node border colour (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
\code{alpha} \tab Transparency (see \link[ggplot2]{aes_colour_fill_alpha}).\cr
\code{shape} \tab Node shape (see \link{points} and \link[ggplot2]{aes_linetype_size_shape}).\cr
\code{size} \tab Node size (see \emph{drawing} section and \link[ggplot2]{aes_linetype_size_shape}).\cr
\code{stroke} \tab Node line width (see \link[ggplot2]{gg_par} and \link[ggplot2]{aes_linetype_size_shape}).
}

Required aesthetics \code{x}, \code{y}, and \code{vertex} are supplied from
the \link{GraphSpace} object and do not need to be manually mapped.

Additional parameters can be passed to control fixed values for the layer.
For example: \code{fill = "red"}, \code{stroke = 3}, \code{alpha = 0.5}, or \code{shape = 21}.

The interpretation of \strong{\code{size}} depends on how it is provided:
\itemize{
\item \strong{As an aesthetic}: When mapped within \code{aes()}, \code{size}
follows the behavior of \link[ggplot2]{geom_point}, using absolute
units to ensure consistency with the plot legends.
\item \strong{As a parameter}: When set outside \code{aes()}, \code{size} is
treated as a percentage of the viewport (\code{[0, 100]}), scaling
in \code{npc} units. This allows nodes to resize dynamically with
viewport changes.
}
}

\section{Edge context-aware parameters}{


These parameters control the edge appearance. If not explicitly provided,
they are automatically retrieved from the \link{GraphSpace} object.
They can be a single value or a vector matching the number of edges:

\tabular{ll}{
\code{edge_colour} \tab Node border colour.\cr
\code{edge_linetype} \tab Edge line type.\cr
\code{edge_linewidth} \tab Edge line width.\cr
\code{edge_alpha} \tab Edge transparency.
}
}

\section{Edge global parameters}{


These parameters apply globally to all edges in the layer:

\tabular{ll}{
\code{arrow_size} \tab Arrow scaling factor (default = 1).\cr
\code{arrow_offset} \tab Arrow offset from nodes (default = 0.01).\cr
\code{arrow_lineend} \tab Line end style (see \link[grid]{gpar}).\cr
\code{arrow_linejoin} \tab Line join style (see \link[grid]{gpar}).
}

\strong{arrow_size} is a numeric scaling factor controlling arrowhead geometry.
The value is interpreted in the same numeric space as line width (\code{lwd}),
ensuring consistent scaling between edge strokes and arrowheads.

\strong{arrow_offset} is an additive term that offsets arrow endpoints
uniformly in graph space and is bounded by the edge length, in NPC units.

Arrowhead types are specified in the \link{GraphSpace} constructor.
}

\examples{
library(RGraphSpace)
library(igraph)
library(ggplot2)

# Make a demo igraph
gtoy1 <- make_star(15, mode="out")

# Set some node attributes
V(gtoy1)$nodeSize <- runif(vcount(gtoy1), 1, 20)
V(gtoy1)$nodeColor <- rainbow(vcount(gtoy1))

# Set some variables
V(gtoy1)$user_var1 <- runif(vcount(gtoy1), 1, 3)^3
V(gtoy1)$user_var2 <-  rep(c(1, 2, 3), each = 5)

# Create a GraphSpace object
gs <- GraphSpace(gtoy1, layout = layout_in_circle(gtoy1))

\dontrun{

# Example 1: Nodes scaling with the legend
# When 'size' is mapped inside aes(), it follows
# ggplot2 default behavior: size is translated 
# to absolute units (mm) via 'scale_size()'.

ggplot() + 
  geom_graphspace(
  mapping = aes(size = nodeSize, fill = user_var2), 
  data = gs, arrow_offset = 0.01) + 
  scale_size(range = c(1, 12)) + 
  theme(aspect.ratio = 1)
  
# Example 2: Nodes scaling with the viewport
# When 'size' is passed as a node attribute, 
# inherited from the igraph object, it is
# interpreted as a percentage of the plotting 
# area and translated to NPC units.

ggplot() +
  geom_graphspace(mapping = aes(fill = user_var2), 
  data = gs, arrow_offset = 0.01) + 
  theme(aspect.ratio = 1)
  
}

}
\seealso{
\link{GraphSpace}, \link{geom_nodespace}, \link{geom_edgespace},
\link[ggplot2]{geom_point}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-edgespace.R
\docType{data}
\name{GeomEdgeSpace}
\alias{GeomEdgeSpace}
\title{GeomEdgeSpace: a ggplot2 prototype for GraphSpace-class methods}
\format{
An object of class \code{GeomEdgeSpace} (inherits from \code{GeomSegment}, \code{Geom}, \code{ggproto}, \code{gg}) of length 7.
}
\usage{
GeomEdgeSpace
}
\description{
\code{GeomEdgeSpace} is the underlying \link[ggplot2]{ggproto} object
used by \link{geom_edgespace} to draw edge elements in a graph layout.

This geom is designed for network diagrams, where graph attributes
are often already in their final form (e.g., hex colors).
}
\section{Aesthetics}{


\code{GeomEdgeSpace} understands ggplot2's conventions for segment-like geoms.
}

\seealso{
\link{geom_edgespace}, \link[ggplot2]{geom_segment}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-misc.R
\docType{data}
\name{gtoys}
\alias{gtoys}
\alias{gtoy1}
\alias{gtoy2}
\title{Toy 'igraph' objects}
\format{
igraph
}
\source{
This package.
}
\usage{
data(gtoy1)
}
\value{
A pre-processed igraph object.
}
\description{
Small 'igraph' objects used for workflow demonstrations.
All graphs include 'x', 'y', and 'name' vertex attributes.
}
\examples{
data(gtoy1)
data(gtoy2)
}
\keyword{gtoys}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-classes.R
\name{.generate_gs_uuid}
\alias{.generate_gs_uuid}
\title{Generate a unique identifier for GraphSpace objects}
\usage{
.generate_gs_uuid()
}
\value{
A character string containing a unique alphanumeric ID.
}
\description{
This helper function creates a unique ID without relying on the R
Random Number Generator (RNG), making it immune to \code{set.seed()}.
It combines the Process ID (PID), high-precision system time, and
a system-level temporary identifier to ensure uniqueness across
parallel processes and rapid sequential calls.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-ggplot-constructor.R
\name{ggplot-GraphSpace}
\alias{ggplot-GraphSpace}
\alias{ggplot.GraphSpace}
\title{Using ggplot2 with GraphSpace objects}
\usage{
\method{ggplot}{GraphSpace}(data, mapping = NULL, ...)
}
\arguments{
\item{data}{A \link{GraphSpace} object.}

\item{mapping}{Set of aesthetic mappings created by \link[ggplot2]{aes}.
Passed to \link[ggplot2]{ggplot}.}

\item{...}{Additional arguments passed to \link[ggplot2]{ggplot}.}
}
\value{
A \code{gspace_plot} object extending \link[ggplot2]{ggplot}.
}
\description{
GraphSpace objects can be used directly with \pkg{ggplot2}, allowing node
attributes and high-dimensional feature data to be mapped through standard
aesthetic mappings without manual data extraction.
This integration enables:
\itemize{
\item Lazy evaluation of node attributes and feature data.
\item Automatic synchronization of node metadata for standard ggplot2
geoms such as \link[ggplot2]{geom_point}.
\item Automatic propagation of node metadata required for edge clipping
and arrow placement in GraphSpace-native geoms.
}
}
\details{
When a GraphSpace object is supplied to \code{ggplot()}, RGraphSpace
extends the standard ggplot2 build process to automatically resolve
GraphSpace variables and synchronize node metadata required for edge
rendering.

When using \code{ggplot()}, neither \link{nodespace_handler} nor
\link{inject_nodespace} need to be called explicitly.
}
\examples{
library(RGraphSpace)
library(igraph)
library(ggplot2)

# Generate a toy star graph
gtoy1 <- make_star(15, mode = "out")
V(gtoy1)$my_node_var <- runif(vcount(gtoy1), 1, 20)

# Create a GraphSpace object
gs <- GraphSpace(gtoy1, layout = layout_in_circle(gtoy1))

\dontrun{

# Example 1: Using RGraphSpace-native geoms
# Edge clipping metadata are injected automatically
ggplot(gs) +
  geom_edgespace(colour = "red") +
  geom_nodespace(aes(size = my_node_var), 
  fill = "steelblue", stroke = 2) +
  scale_size(range = c(2, 15))

# Example 2: Mixing native and general geoms
# Note possible clipping mismatch when combining
# geom_edgespace() with generic ggplot2 node geoms.
# Since geom_point() does not expose the final rendered
# node radius to RGraphSpace, edge clipping is estimated
# from layer parameters and may not exactly match the
# displayed node geometry.
ggplot(gs) +
  geom_edgespace(colour = "red") +
  geom_point(aes(x, y, size = my_node_var), 
  fill = "steelblue", stroke = 2, shape = 21) +
  scale_size(range = c(2, 15))

}

}
\seealso{
\link{GraphSpace}, \link{geom_nodespace}, \link{geom_edgespace},
\link{inject_nodespace}, \link{nodespace_handler}, \link{edgespace_handler}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-normalize.R
\name{normalizeGraphSpace,GraphSpace,missing-method}
\alias{normalizeGraphSpace,GraphSpace,missing-method}
\alias{normalizeGraphSpace}
\alias{normalizeGraphSpace,GraphSpace,ANY-method}
\alias{cropGraphSpace,GraphSpace-method}
\alias{cropGraphSpace}
\title{Normalize node coordinates to graph and image spaces}
\usage{
normalizeGraphSpace(gs, image = NULL, ...)

\S4method{normalizeGraphSpace}{GraphSpace,ANY}(
  gs,
  image,
  ...,
  mar = 0.1,
  flip.x = FALSE,
  flip.y = FALSE,
  rotate.xy = FALSE,
  flip.v = FALSE,
  flip.h = FALSE,
  verbose = TRUE
)

\S4method{cropGraphSpace}{GraphSpace}(gs, crop.coord = c(0, 1, 0, 1), verbose = TRUE)
}
\arguments{
\item{gs}{A \code{GraphSpace} object to be normalized.}

\item{image}{An optional background image. When provided, \code{x} and
\code{y} coordinates must represent pixel positions in the image matrix.}

\item{...}{Additional arguments passed to specific normalization workflows.}

\item{mar}{A single numeric value in \code{[0, 0.5]} controlling the size of
the outer margins around the graph. Without an image, \code{mar} specifies
symmetric margins as a fraction of the graph space. With an image,
\code{mar} is interpreted as a fraction of the available image margins
surrounding the graph.}

\item{flip.x}{Logical; whether to flip the node coordinates along the x-axis.}

\item{flip.y}{Logical; whether to flip the node coordinates along the y-axis.
Useful for aligning nodes with image backgrounds, which often use an
inverted coordinate system.}

\item{rotate.xy}{Logical; whether to rotate x-y coordinates.}

\item{verbose}{A single logical value specifying to display detailed
messages (when \code{verbose=TRUE}) or not (when \code{verbose=FALSE}).}

\item{flip.v}{Logical; whether to vertically flip the background image
matrix (top-to-bottom) to align with the graph coordinate system.}

\item{flip.h}{Logical; whether to horizontally flip the background image
matrix (left-to-right) to align with the graph coordinate system.}

\item{crop.coord}{An optional numeric vector of length four specifying a
cropping region (xmin, xmax, ymin, ymax), with values in normalized
coordinates \code{[0,1]}.}
}
\value{
A \code{GraphSpace} object with updated \code{nodes}
and \code{image} slots.
}
\description{
Accessory functions to normalize node coordinates in GraphSpace,
either by centering them within the graph boundaries
or by mapping them to pixel coordinates of a background image.
}
\details{
These functions provide different strategies for coordinate transformation:
\itemize{
\item \bold{normalizeGraphSpace}: Re-scales node coordinates to a
\code{[0, 1]} unit square based on the graph's bounding box (when
\code{image} is missing) or maps them to pixel coordinates (when
\code{image} is provided). It handles image-to-graph alignment via
\code{flip.} and \code{rotate.} arguments, used to adjust the graph
origin with the image matrix layout. Users should be aware of the
potential discrepancy between image matrix orientation (top-down) and
graph coordinates (bottom-up). The function attempts to automatically
adjust the y-axis to align the graph's bottom-up coordinates with the
image's top-down layout, but further manual adjustments might be required.
\item \bold{cropGraphSpace}: Subsets the normalized graph space into a
specific region defined by \code{crop.coord}. It recalculates node positions
and background image boundaries to maintain spatial consistency after cropping.
This function requires a previously normalized \code{GraphSpace} object.
}
}
\note{
This is an accessory function typically called during
the preprocessing of \code{GraphSpace} objects before rendering.
}
\examples{
library(igraph)

# Create a star graph
gtoy1 <- make_full_graph(15)

# Create a GraphSpace
gs <- GraphSpace(gtoy1)

gs <- normalizeGraphSpace(gs)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-misc.R
\name{as_colorraster}
\alias{as_colorraster}
\title{Map numeric values to a color raster}
\usage{
as_colorraster(x, palette = hcl.colors(30), na.color = "white")
}
\arguments{
\item{x}{A numeric vector or matrix containing values to be mapped to colors.}

\item{palette}{A vector of colors used as the palette. By default,
\code{hcl.colors(30)} is used.}

\item{na.color}{Color used for \code{NA} values. Defaults to \code{NA}.}
}
\value{
A raster object as produced by \code{as.raster()}.
}
\description{
Helper function that converts numeric values to colors and
returns a raster image. Useful for visualizing numeric matrices
as color backgrounds.
}
\details{
Values in \code{x} are rescaled to the range of the palette using
\code{scales::rescale()}, and each value is mapped to a corresponding
color. If \code{x} is a matrix, the resulting raster preserves the same
dimensions.
}
\examples{
library(RGraphSpace)

# Convert the volcano matrix to a color raster
img <- as_colorraster(volcano)
plot(img)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-graphspace.R
\name{GeomGraphSpace}
\alias{GeomGraphSpace}
\title{GeomGraphSpace: a ggplot2 prototype for GraphSpace-class methods}
\usage{
GeomGraphSpace
}
\description{
\code{GeomGraphSpace} is the underlying \link[ggplot2]{ggproto} object
used by \link{geom_graphspace} to draw node and edge elements in a
graph layout.

This geom is designed for network diagrams, where graph attributes
are often already in their final form (e.g., hex colors).
}
\section{Aesthetics}{


\code{GeomGraphSpace} understands ggplot2's conventions for point-like geoms.
}

\seealso{
\link{geom_graphspace}, \link[ggplot2]{geom_point}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-coercion.R
\name{as.GraphSpace}
\alias{as.GraphSpace}
\alias{as.GraphSpace.default}
\alias{as.GraphSpace.igraph}
\alias{as.GraphSpace.tbl_graph}
\alias{as.GraphSpace.data.frame}
\alias{as.GraphSpace.Seurat}
\title{Convert objects to GraphSpace}
\usage{
as.GraphSpace(x, ...)

\method{as.GraphSpace}{default}(x, ...)

\method{as.GraphSpace}{igraph}(x, ...)

\method{as.GraphSpace}{tbl_graph}(x, ...)

\method{as.GraphSpace}{data.frame}(x, ...)

\method{as.GraphSpace}{Seurat}(x, space = c("embedding", "spatial"), layer = NULL, ...)
}
\arguments{
\item{x}{An object to be converted.}

\item{...}{Additional arguments passed to methods associated with the
selected \code{space}.}

\item{space}{Character specifying the coordinate space used for node
geometry. Either \code{"embedding"} or \code{"spatial"}. See details.}

\item{layer}{Name of the layer from which node data should be retrieved
(see \code{\link[SeuratObject]{LayerData}}).}
}
\value{
A \code{GraphSpace} object.
}
\description{
S3 generic function for coercing objects into a \code{GraphSpace} object.
}
\details{
Unified entry point for converting graph, spatial, and high-dimensional
data into a \code{GraphSpace} object.

Graph objects are imported either through native methods or via
\link[tidygraph]{as_tbl_graph} when available.

For \strong{Seurat} objects, coordinate extraction depends on the
selected \code{space}:
\itemize{
\item \code{space = "embedding"} uses the first two dimensions returned by
\code{\link[SeuratObject]{Embeddings}}.

\item \code{space = "spatial"} uses tissue coordinates returned by
\code{\link[SeuratObject]{GetTissueCoordinates}}.
}

Assay data are stored in the \code{data} slot of the resulting
\code{GraphSpace} object. Node metadata from \code{x@meta.data} are
appended to the node table.
}
\seealso{
\code{\linkS4class{GraphSpace}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geom-edgespace.R
\name{GeomEdgeSpace}
\alias{GeomEdgeSpace}
\title{GeomEdgeSpace: a ggplot2 prototype for GraphSpace-class methods}
\usage{
GeomEdgeSpace
}
\description{
\code{GeomEdgeSpace} is the underlying \link[ggplot2]{ggproto} object
used by \link{geom_edgespace} to draw edge elements in a graph layout.

This geom is designed for network diagrams, where graph attributes
are often already in their final form (e.g., hex colors).
}
\section{Aesthetics}{


\code{GeomEdgeSpace} understands ggplot2's conventions for segment-like geoms.
}

\seealso{
\link{geom_edgespace}, \link[ggplot2]{geom_segment}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-misc.R
\docType{data}
\name{gtoys}
\alias{gtoys}
\alias{gtoy1}
\alias{gtoy2}
\title{Toy 'igraph' objects}
\format{
igraph
}
\source{
This package.
}
\usage{
data(gtoy1)
}
\value{
A pre-processed igraph object.
}
\description{
Small 'igraph' objects used for workflow demonstrations.
All graphs include 'x', 'y', and 'name' vertex attributes.
}
\examples{
library(RGraphSpace)
data(gtoy1)
data(gtoy2)
}
\keyword{gtoys}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-normalize.R
\name{normalizeGraphSpace,GraphSpace-method}
\alias{normalizeGraphSpace,GraphSpace-method}
\alias{normalizeGraphSpace}
\alias{cropGraphSpace,GraphSpace-method}
\alias{cropGraphSpace}
\title{Normalize node coordinates to graph and image spaces}
\usage{
\S4method{normalizeGraphSpace}{GraphSpace}(
  gs,
  mar = 0.1,
  use_image = FALSE,
  flip.x = FALSE,
  flip.y = FALSE,
  rotate.xy = FALSE,
  flip.v = FALSE,
  flip.h = FALSE,
  verbose = TRUE,
  image = deprecated()
)

\S4method{cropGraphSpace}{GraphSpace}(gs, crop.coord = c(0, 1, 0, 1), verbose = TRUE)
}
\arguments{
\item{gs}{A \code{GraphSpace} object to be normalized.}

\item{mar}{A single numeric value in \code{[0, 0.5]} controlling the size of
the outer margins around the graph. Without an image, \code{mar} specifies
symmetric margins as a fraction of the graph space. With an image,
\code{mar} is interpreted as a fraction of the available image margins
surrounding the graph.}

\item{use_image}{Logical; if an image is available, whether to use it as
a background reference map. When enabled, \code{x} and \code{y} graph
coordinates are interpreted as pixel coordinates in the image matrix.
Images can be inspected and assigned with \code{\link{gs_image}}.}

\item{flip.x}{Logical; whether to flip the node coordinates along the x-axis.}

\item{flip.y}{Logical; whether to flip the node coordinates along the y-axis.
Useful for aligning nodes with image backgrounds, which often use an
inverted coordinate system.}

\item{rotate.xy}{Logical; whether to rotate x-y coordinates.}

\item{flip.v}{Logical; whether to vertically flip the background image
matrix (top-to-bottom) to align with the graph coordinate system.}

\item{flip.h}{Logical; whether to horizontally flip the background image
matrix (left-to-right) to align with the graph coordinate system.}

\item{verbose}{A single logical value specifying to display detailed
messages (when \code{verbose=TRUE}) or not (when \code{verbose=FALSE}).}

\item{image}{Deprecated from RGraphSpace 1.3.0; use \link{gs_image} instead.}

\item{crop.coord}{An optional numeric vector of length four specifying a
cropping region (xmin, xmax, ymin, ymax), with values in normalized
coordinates \code{[0,1]}.}
}
\value{
A \code{GraphSpace} object with updated \code{nodes}
and \code{image} slots.
}
\description{
Accessory functions to normalize node coordinates in GraphSpace,
either by centering them within the graph boundaries
or by mapping them to pixel coordinates of a background image.
}
\details{
These functions provide different strategies for coordinate transformation:
\itemize{
\item \bold{normalizeGraphSpace}: Re-scales node coordinates to a
\code{[0, 1]} unit square based on the graph's bounding box (when
\code{use_image = FALSE}) or maps them to pixel coordinates (when
\code{use_image = TRUE} and an image is provided; see \link{gs_image}).
It handles image-to-graph alignment via \code{flip.\*} and
\code{rotate.\*} arguments, used to adjust the graph origin with the
image matrix layout. Users should be aware of the potential discrepancy
between image matrix orientation (top-down) and graph coordinates
(bottom-up). The function attempts to automatically adjust the y-axis to
align the graph's bottom-up coordinates with the image's top-down layout,
but further manual adjustments might be required.
\item \bold{cropGraphSpace}: Subsets the normalized graph space into a
specific region defined by \code{crop.coord}.
It recalculates node positions and background image boundaries to maintain
spatial consistency after cropping. This function requires a previously
normalized \code{GraphSpace} object.
}
}
\note{
This is an accessory function typically called during
the preprocessing of \code{GraphSpace} objects before rendering.
}
\examples{
library(RGraphSpace)
library(igraph)

# Create a star graph
gtoy1 <- make_full_graph(30)

# Create a GraphSpace
gs <- GraphSpace(gtoy1)

gs <- normalizeGraphSpace(gs)

gs_crop <- cropGraphSpace(gs, 
         crop.coord = c(0, 0.75, 0, 0.75))

}
\seealso{
\code{\link{gs_image}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gspace-methods.R
\name{GraphSpace-internal}
\alias{GraphSpace-internal}
\alias{.DollarNames.GraphSpace}
\title{Internal methods for GraphSpace}
\usage{
\method{.DollarNames}{GraphSpace}(x, pattern = "")
}
\arguments{
\item{x, pattern}{Internal arguments.}
}
\description{
Exported solely to enable RStudio auto-completion
and should not be called directly by the user.
}
\keyword{internal}
